!*--------------------------------------------------------------------------

!**
!* pchim ::: Piecewise Cubic Hermite Interpolation to Monotone data.
!*
!*   n     (i) : number of data points.  (Error return if n<2 .)
!*               If n=2, simply does linear interpolation.
!*   x     (i) : Real array of independent variable values.  The elements
!*               of x must be strictly increasing:
!*                   x(i-1) < x(i),  i = 2(1)n.
!*              (Error return if not.)
!*   f     (i) : real array of dependent variable values to be interpola-
!*               ted. f(1+(i-1)*incfd) is the value corresponding to x(i).
!*   d     (o) : real array of derivative values at the x-values.
!*               If the data are monotonic, these values will determine
!*               a monotone cubic Hermite function.
!*               The value corresponding to x(i) is stored in
!*                   d(1+(i-1)*incfd),  i=1(1)n.
!*               No other entries in d are changed.
!*   incfd (i) : increment between successive values in f and d.
!*               This argument is provided primarily for 2-D applications.
!*               (Error return if  incfd<1 .)
!*   ierr  (o) : error flag.  (See Return Value, below.)
!*
!* pchim computes derivatives needed to determine a monotone piecewise
!* cubic Hermite interpolant the data given in X and F.
!*
!* Default boundary conditions are provided which are compatible with
!* monotonicity.  (Use pchic if user control of boundary conditions is
!* desired.)
!*
!* If the data are only piecewise monotonic, the interpolant will have an
!* extremum at each point where monotonicity switches direction.  (Use
!* pchic if control is desired over such cases.)
!*
!* To facilitate two-dimensional applications, pchim allows an increment
!* between successive values of the f- and d-arrays.  For normal usage,
!* with incfd=1, pchim may be called with one-dimensional arrays f and d.
!*
!* The resulting piecewise cubic Hermite function may be evaluated by
!* pchfe or pchfd.
!*
!* Return Value: ierr = 0 indicates a normal return (no errors).
!*   Warning error:
!*     ierr < 0   means that ierr switches in the direction of
!*                monotonicity were detected.
!*   Fatal errors:
!*     ierr = -1  if n < 2 .
!*     ierr = -2  if incfd < 1 .
!*     ierr = -3  if the x-array is not strictly increasing.
!*     The d-array has not been changed in any of these cases.
!*     NOTE:  The above errors are checked in the order listed, and
!*            following arguments have **NOT** been validated.
!*
!**
!* Implementation notes:
!*
!* References:  1. F. N. Fritsch and J. Butland, A method for construc-
!*                ting local monotone piecewise cubic interpolants, SIAM
!*                Journal on Scientific and Statistical Computing 5, 2
!*                (June 1984), pp. 300-304.
!*              2. F. N. Fritsch and R. E. Carlson, Monotone piecewise
!*                cubic interpolation, SIAM Journal on Numerical Ana-
!*                lysis 17, 2 (April 1980), pp. 238-246.
!*
!* Notes :
!*  1. The function  pchst(arg1,arg2)  is assumed to return zero if either
!*     argument is zero, +1 if they are of the same sign, and -1 if they
!*     are of opposite sign.
!*
!*  2. Argument incfd is used only to cause the compiler to generate
!*     efficient code for the subscript expressions (1+(i-1)*incfd) .
!*     The normal usage, in which pchim is called with one-dimensional
!*     arrays f and d, is probably non-Fortran 77, in the strict sense,
!*     but it works on all systems in which pchim has been tested.
!*
!* Revision History  (yymmdd means mm/dd/19yy)
!*  811103  Date written by Fritsch, F. N., (LLNL).
!*  820201  1. Introduced  PCHST  to reduce possible over/under-
!*             flow problems.
!*          2. Rearranged derivative formula for same reason.
!*  820602  1. Modified end conditions to be continuous functions
!*             of data when monotonicity switches in next interval.
!*          2. Modified formulas so end conditions are less prone
!*             of over/underflow problems.
!*  980202  To reduce single/double differences, removed all DATA-loaded
!*          constants except ZERO by the use of mixed-mode arithmetic.
!*/

      subroutine pchim (n, x, f, d, incfd, ierr)
      integer  n, incfd
      real  x(*), f(incfd,*), d(incfd,*)
      integer  ierr

      Other declarations
      Executable code

      end

!*--------------------------------------------------------------------------
