#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
int
request_locator(int *wsid, int *x, int *y, int *win_width, int *win_height)

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
  Routine:  request_locator
  Author:  Jeff Rowe,
           Lawrence Livermore National Lab,
           Livermore Computing Center,
           Information Management & Graphics Group
           Tel: (510) 422-4042, email: jrowe@llnl.gov
  Modifications:
    Created: Dec 97

  Description: Get mouse position in NCAR X window.

  Parameters:
	int *wsid  < workstation identifier
	int *x  < pointer to x value.
	int *y  < pointer to y value.
	int *win_width  < pointer to x window width.
	int *win_height  < pointer to x window height.
	int *x  > int x value relative to window. 0,0 at lower left.
	int *y  > int y value relative to window. 0,0 at lower left.
	request_locator   > 0 if no error, 1 if X display not initialized.
			    2 if window not initiailized/ user destroyed.
			    5 expose event recieve. Caller should
			      redraw picture and call request_locator.
			    7 if window doesn't match wsid.
  Procedure: Start a while loop to receive X Button events.
  If a destroy/close window event received, perform the requested action.
  App should use llnlgetxwin to determine window status.

=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
{
/* A basic X program to read the mouse position from an X window
   Assumes X window open.
 */
#define debug 0
    Display *llnlgetxdpy();
    Window llnlgetxwin();

    Display *the_display = NULL;        /* the display in use. this has been set by the
                                           NCAR X driver */
    Window the_window;          /* the window that is opened */
    Window cwindow;             /* client message window id of exited window  */
    XWindowAttributes xwa;      /* Get window attributes        */

    Drawable drawable_root;
    int xwind, ywind;
    unsigned int width, height, border_width, drawable_depth;

    int the_screen;             /* the screen that is in use */
    XEvent the_event;           /* the structure for input events */
    XButtonEvent *button_event; /* to avoid casts */
    XClientMessageEvent *client_event;  /* to avoid casts */
    int expose = 0;             /* Dec 98   */

    *x = 0;
    *y = 0;

/* initialize */

    the_display = llnlgetxdpy(*wsid);
    if (the_display == NULL)
        return 1;

    the_window = llnlgetxwin(*wsid);
    if (the_window == 0)
        return 2;

/* Get the current size of the window */

    XGetGeometry(the_display, the_window, &drawable_root,
                 &xwind, &ywind, &width, &height, &border_width,
                 &drawable_depth);
    *win_width = width;
    *win_height = height;
    if (debug) {
        printf
            ("request_locator:WSID: %d Window %x geometry:  xloc %d yloc %d width %d height %d\n",
             *wsid, the_window, xwind, ywind, width, height);
    }
/*The event types to receive are requested in gxwind.c  */
/* Enter the main input event loop. */

    while (True) {
/* Get the next event. Note that the first event will be an exposure event
   -but this should already have been taken care of by the NCAR
   X Driver

      XNextEvent(the_display, &the_event);
 */
        /* Get only the_window specific events. This ignores window delete protocols
         */
        XWindowEvent(the_display, the_window,
                     (ExposureMask | ButtonPressMask | KeyPressMask |
                      ButtonReleaseMask | StructureNotifyMask),
                     &the_event);

        if (debug)
            printf(" request_locator:WSID: %d got event %d\n", *wsid,
                   the_event.type);

/* the outer clauss check for the type of event.
   The inner processing for a button down mouse event is dependent
   on the window in which the event occurred.
 */

        switch (the_event.type) {
        case ClientMessage:

            /* Get window and continue if not current window  */
            client_event = (XClientMessageEvent *) & the_event;
            cwindow = client_event->window;
            if (debug)
                printf("client message %d recieved for window %x\n",
                       the_event.type, cwindow);
            /* Destroy window as requested by user. 
               Should ask user if sure window should be destroy then
               take appropriate action.
               if(cwindow != the_window) break;
             */
            {
                llnldelxwin(cwindow);
                *x = 0;
                *y = 0;
                return 2;
            }

        case DestroyNotify:
            break;
            /*
               {*x =0; *y=0; return 7;}
             */

        case Expose:           /* Dec 98 */

            expose = 1;
            /* 
             * find the last expose event on the event queue.
             */
            /* if count non zero, more expose events follow  */
            if (the_event.xexpose.count == 0) {
                expose += 1;
                /*
                 * Remove any other pending Expose events from 
                 * the queue to avoid multiple repaints. */
                /*SUPPRESS570 */
                while (XCheckTypedEvent(the_display, Expose, &the_event)) ;

            }
            return 5;
        case ConfigureNotify:
            /* Window has been resizxed. Call X11_ClearWorkstation to
               change width and height and to notify GKS of new limits.
               Assumes application will redraw image
             */
            expose = 1;
            /* X11_ClearWorkstation (); */
            /* User should do a clear workstation with gclear_ws (X_WIN_WKID, 1)
               It calls X11_ClearWorkstation () with the correct GKSC context and determines
               if new X window limits are needed.
             */
            if (debug) {
                XGetWindowAttributes(the_display, the_window, &xwa);
                printf("Window x %d, y%d width %d height %d\n", xwa.x,
                       xwa.y, xwa.width, xwa.height);
            }
            return 5;           /* end Dec 98 */
        case ButtonPress:
            break;
        case ButtonRelease:
            button_event = (XButtonEvent *) & the_event;
            /* See if button in wsid's window */
            if (debug) {
                printf
                    ("  request_locator: window is 0x%1x, state is %d, button is %d, x is %d, y is %d\n",
                     button_event->window, button_event->state,
                     button_event->button, button_event->x,
                     button_event->y);
            }
            /* See if window matches wsid's window. */
            if (button_event->window == the_window) {
/* Return x, y to caller */

                *x = (button_event->x);
                *y = (button_event->y);

                return 0;
            } else {
                *x = 0;
                *y = 0;
                return 7;
            }
        default:
            break;
        }
    }                           /* of while */
    /* should not get here. */
/* compiler says cannot get here
	  {*x =0; *y=0; return 6;}
*/
}
