c==============================================================================
c
c  ezclrm
c
c  Purpose: convert an array of real8's into an array of indices into a
c           colormap for EZC Fillmesh plotting using a linear mapping,
c           with optional user-supplied limits.
c
c  Arguments:
c     ncol  Number of colors in colormap.
c     z     The array to be mapped.
c     ireg  The associated region map.
c     kmax,lmax  The dimensions of z and ireg.
c     ndx   The output array of color indices, in range [1,ncol].
c     zz    Auxiliary output array.  This routine returns zz(1)=zmin,
c           zz(2)=zmax, zz(5)=0 (to signal linear mapping).
c           If the input value of zz(5) is negative, reverse the mapping.
c     zlim  If the input values zlim(1) and zlim(2) are different, these
c           are treated as user-supplied limits.  (No order assumed.)
c
c  Created: Fred N. Fritsch, 980325
c
c  Revision History:
c     - Developed from ezclr8 by adding arguments ireg,kmax,lmax and code
c       to ignore elements of z for which ireg=0.  (FNF 980325)
c     - Moved range calculation to ezclrlim.  (FNF 980326)
c     - Implemented user-supplied limits.  (FNF 990108)
c     - Added zlim as a separate argument, to be consistent with the
c       ezcmp8 family of routines.  (FNF 000427)
c     - Made zz(5) an input, to determine order of color map. (FNF 010629)
c
c==============================================================================


      subroutine ezclrm(ncol, z, ireg, kmax, lmax, ndx, zz, zlim)
c
      integer        ncol, ireg(kmax,lmax), kmax, lmax, ndx(kmax,lmax)
      real(Size8)    z(kmax,lmax), zz(5), zlim(2)
c
      integer        k, l, ncm1
      external       ezclrlim

c   Define function to map f (in [0.,1.]) to color index (in [0,nind]),
c   mapping any negative f's to 0, f > 1 to nind.  (FNF)
      integer      nind, colmap
      real(Size8)  f
      colmap(f,nind) = max(0, min( int(f*nind), nind) )

c   Initialize.
c      Points not otherwise set will have color index 1.
      do l = 1, lmax
         do k = 1, kmax
            ndx(k,l) = 1
         enddo
      enddo

c   Find range of data.
      if (zlim(1) <> zlim(2)) then
         zz(1) = min( zlim(1), zlim(2) )
         zz(2) = max( zlim(1), zlim(2) )
      else
         call ezclrlim(z, ireg, kmax, lmax, zz(1), zz(2))
      endif

c   Check for zz(1) .eq. zz(2) -- constant data or ireg==0.
      if ( zz(1) < zz(2) ) then
c        Convert data to color indices.
         ncm1 = ncol - 1
         do l = 1, lmax
            do k = 1, kmax
               if (ireg(k,l) > 0)  then
                 ndx(k,l) = \
                  colmap( (z(k,l) - zz(1))/(zz(2) - zz(1)), ncm1) + 1
c               Check for reverse order desired.
                 if (zz(5) < 1) ndx(k,l) = ncol+1 - ndx(k,l)
               endif
            enddo
         enddo
      endif
c
      zz(3) = 0
      zz(4) = 0
      zz(5) = 0     #linear mapping
      return
      end
