c==============================================================================
c
c  ezclrmln
c
c  Purpose: convert an array of real8's into an array of indices into a
c           colormap for EZC Fillmesh plotting using a logarithmic mapping,
c           with optional user-supplied limits.
c
c  Arguments:
c     ncol  Number of colors in colormap.
c     z     The array to be mapped.
c     ireg  The associated region map.
c     kmax,lmax  The dimensions of z and ireg.
c     ndx   The output array of color indices, in range [1,ncol].
c     zz    Auxiliary output array.  This routine returns zz(1)=zmin,
c           zz(2)=zmax, zz(5)=1 (to signal logarithmic mapping).
c           If the input value of zz(5) is negative, reverse the mapping.
c     zlim  If the input values zlim(1) and zlim(2) are different, these
c           are treated as user-supplied limits.  (No order assumed.)
c
c  Created: Fred N. Fritsch, 980325
c
c  Revision History:
c     - Developed from ezclr8ln by adding arguments ireg,kmax,lmax and
c       code to ignore elements of z for which ireg=0.  (FNF 980325)
c     - Moved range calculation to ezclrlim.  (FNF 980326)
c     - Implemented user-supplied limits.  (FNF 990108)
c     - Added code to map data to user-supplied limits.  (FNF 000420)
c     - Added zlim as a separate argument, to be consistent with the
c       ezcmp8 family of routines.  (FNF 000427)
c     - Made zz(5) an input, to determine order of color map. (FNF 010629)
c
c==============================================================================


      subroutine ezclrmln(ncol, z, ireg, kmax, lmax, ndx, zz, zlim)
c
      integer        ncol, ireg(kmax,lmax), kmax, lmax, ndx(kmax,lmax)
      real(Size8)    z(kmax,lmax), zz(5), zlim(2)
c
      integer        k, l, ncm1
      real(Size8)    zmin, zmax, zmap
      external       ezclrlim
c
      integer utsizeof, utstrcod
      external utsizeof, utstrcod
      integer numdata, type8
      Dynamic(ztmp,real(Size8),1)

c   Define function to map f (in [0.,1.]) to color index (in [0,nind]),
c   mapping any negative f's to 0, f > 1 to nind.  (FNF)
      integer      nind, colmap
      real(Size8)  f
      colmap(f,nind) = max(0, min( int(f*nind), nind) )

c   Initialize.
c
c      Points not otherwise set will have color index 1.
      do l = 1, lmax
         do k = 1, kmax
            ndx(k,l) = 1
         enddo
      enddo

c   Find range of data.
      if (zlim(1) <> zlim(2)) then
         zz(1) = min( zlim(1), zlim(2) )
         zz(2) = max( zlim(1), zlim(2) )
      else
         call ezclrlim(z, ireg, kmax, lmax, zz(1), zz(2))
      endif

c   Check for valid data.
      if (zz(1) <= 0 | zz(2) <= 0) then
         call ezcremrk ("ezclrmln: nonpositive data or limits")
         go to 30
      endif
      zmin = log(zz(1))
      zmax = log(zz(2))

c   Check for zmin .eq. zmax -- constant data.
      if ( zmin < zmax ) then

c       Allocate work array.
         type8 = utstrcod(Quote(real(Size8)))
         numdata = utsizeof(kmax*lmax,type8)
         call osallot(Point(ztmp),numdata,"EZCCLRMLN:ztmp ")

c       Take logs of data (safe, because of previous tests).
         do l = 1, lmax
            do k = 1, kmax
               zmap = min( max(z(k,l),zz(1)), zz(2) )  # Map to [zz(1),zz(2)].
               if (ireg(k,l) > 0)  ztmp(k+(l-1)*kmax) = log(zmap)
            enddo
         enddo
c        (Recall that zmin,zmax are already log-mapped.)

c       Convert data to color indices.
         ncm1 = ncol - 1
         do l = 1, lmax
            do k = 1, kmax
               if (ireg(k,l) > 0)  then
                 ndx(k,l) = \
                colmap( (ztmp(k+(l-1)*kmax) - zmin)/(zmax - zmin), ncm1) + 1
c               Check for reverse order desired.
                 if (zz(5) < 1) ndx(k,l) = ncol+1 - ndx(k,l)
               endif
            enddo
         enddo
c
         call osfree(Point(ztmp),"ezn.ztmp")
      endif
c
   30 continue
      zz(3) = 0
      zz(4) = 0
      zz(5) = 1     #logarithmic mapping
      return
      end
