c==============================================================================
c
c  ezclrmnml
c
c  Purpose: convert an array of real8's into an array of indices into a
c           colormap for EZC Fillmesh plotting using a normal mapping,
c           with optional user-supplied limits.
c
c  Arguments:
c     ncol  Number of colors in colormap.
c     z     The array to be mapped.
c     ireg  The associated region map.
c     kmax,lmax  The dimensions of z and ireg.
c     ndx   The output array of color indices, in range [1,ncol].
c     zz    Auxiliary output array.  This routine returns zz(1)=zmin,
c           zz(2)=zmax, zz(3)=zbar, zz(4)=zsigma, zz(5)=zmap=2 (to signal
c           normal mapping).
c           If the input value of zz(5) is negative, reverse the mapping.
c     zlim  If the input values zlim(1) and zlim(2) are different, these
c           are treated as user-supplied limits.  (No order assumed.)
c
c  Created: Fred N. Fritsch, 980325
c
c  Revision History:
c     - Developed from ezclr8nml by adding arguments ireg,kmax,lmax and
c       code to ignore elements of z for which ireg=0.  (FNF 980325)
c     - Moved range calculation to ezclrlim.  (FNF 980326)
c     - Implemented user-supplied limits.  (FNF 990108)
c     - Added code to map data to user-supplied limits and to avoid
c       dividing by zero in the case of constant data.  (FNF 000420)
c     - Added zlim as a separate argument, to be consistent with the
c       ezcmp8 family of routines.  (FNF 000427)
c     - Made zz(5) an input, to determine order of color map. (FNF 010628)
c
c==============================================================================


      subroutine ezclrmnml(ncol, z, ireg, kmax, lmax, ndx, zz, zlim)
c
      integer        ncol, ireg(kmax,lmax), kmax, lmax, ndx(kmax,lmax)
      real(Size8)    z(kmax,lmax), zz(5), zlim(2)
c
      integer        i, k, l, n
      real(Size8)    zsum, zsumsq, factor
      external       ezclrlim
c
      integer utsizeof, utstrcod
      external utsizeof, utstrcod
      integer numdata, type8
      Dynamic(ztmp,real(Size8),1)

c   Initialize.

c      Points not otherwise set will have color index 1.
      do l = 1, lmax
         do k = 1, kmax
            ndx(k,l) = 1
         enddo
      enddo

c   Find range of data.
      if (zlim(1) <> zlim(2)) then
         zz(1) = min( zlim(1), zlim(2) )
         zz(2) = max( zlim(1), zlim(2) )
      else
         call ezclrlim(z, ireg, kmax, lmax, zz(1), zz(2))
      endif

c   Check for zz(1) .eq. zz(2) -- constant data or ireg==0.
      if ( zz(1) < zz(2) ) then

c       Allocate work array.
         type8 = utstrcod(Quote(real(Size8)))
         numdata = utsizeof(kmax*lmax,type8)
         call osallot(Point(ztmp),numdata,"EZCCLRMNML:ztmp ")

c       Set the values for working variables.
c       When this loop finishes, n will be the number of points for
c       which ireg(k,l)>0, and ztmp=0 where this is not true, so those
c       points do not contribute the the sums computed later.
         n = 0
         do l = 1, lmax
            do k = 1, kmax
               if (ireg(k,l) > 0) then
c                  Pick up value, mapped to [zz(1),zz(2)].
                  ztmp(k+(l-1)*kmax) = min( max(z(k,l),zz(1)), zz(2) )
                  n = n + 1
               else
                  ztmp(k+(l-1)*kmax) = 0
               endif
            enddo
         enddo

c       Find the standard deviation.
         zsum = 0.
         do i = 1, kmax*lmax
            zsum = zsum + ztmp(i)
         enddo
         zz(3) = zsum/n
      
         zsumsq = 0.
         do i = 1, kmax*lmax
            zsumsq = zsumsq + ztmp(i)**2
         enddo
         zz(4) = sqrt((zsumsq-(zsum**2/n))/(n-1))

c       Convert data to color indices by normalizing the data
c       between -2*factor and 2*factor.
c       The zsigma factor is determined by 4*factor=ncol.
         factor = ncol/4
         do l = 1, lmax
            do k = 1, kmax
               i = k + (l-1)*kmax
               if (ireg(k,l) > 0)  then
c               Special check to avoid dividing by zero (constant data).
                 if ( zz(4) <= 0.) then
                  ndx(k,l) = ncol/2
                 else
                  ztmp(i) = (ztmp(i) - zz(3))*factor / zz(4)
                  if (ztmp(i) > 2*factor) ztmp(i) = 2*factor
                  if (ztmp(i) < -2*factor) ztmp(i) = -2*factor
                  ndx(k,l) = int(ztmp(i)) + int(ncol/2)
                  ndx(k,l) = max(1, min(ndx(k,l), ncol) )
c                  Check for reverse order desired.
                  if (zz(5) < 1) ndx(k,l) = ncol+1 - ndx(k,l)
                 endif
               endif
            enddo
         enddo
c
         call osfree(Point(ztmp),"ezn.ztmp")
      endif
c
      zz(5) = 2     #normal mapping
      return
      end
