c==============================================================================
c
c  ezcmp8ln
c
c  Purpose: Convert an array of real8's into an array of indices into a
c     colormap using a logarithmic mapping, with optional user-supplied
c     limits.
c
c  Arguments:
c     ncol  Number of colors in colormap.
c     n     Number of elements in array z.
c     z     The array to be mapped.
c     ndx   The output array of color indices, in range [1,ncol].
c     zz    Auxiliary output array.  This routine returns zz(1)=zmin,
c           zz(2)=zmax, zz(5)=zmap=1 (to signal logarithmic mapping).
c           If the input value of zz(5) is negative, reverse the mapping.
c     zlim  If the input values zlim(1), zlim(2) are not equal, they are
c           treated as user-supplied limits.  (Ordering not assumed.)
c
c  Routines called: eznminmx, osallot, osfree, utsizeof, utstrcod.
c
c  Created: Fred N. Fritsch, 26 April 2000
c
c  Revision History:
c     - Created from 000414 version of ezclr8ln, as modified to call
c       eznmimmx to compute array limits.  Note that ANY two equal
c       values cause computation of the actual array limits.  Also
c       modified to explicitly map to [zz(1),zz(2)], as in ezclrmln.
c     - Made zz(5) an input, to determine order of color map. (FNF 010702)
c==============================================================================

      subroutine ezcmp8ln (ncol, n, z, ndx, zz, zlim)

c Declare arguments.
      integer        ncol, n, ndx(n)
      real(Size8)    z(n), zz(5), zlim(2)

c Declare externals.
      integer utsizeof, utstrcod
      external utsizeof, utstrcod

c Declare local variables.
      integer i, imax, imin, ncm1
      real(Size8)  zmap, zmax, zmin
      integer numdata, type8
      Dynamic(ztmp,real(Size8),1)

c Define function to map f (in [0.,1.]) to color index (in [0,nind]),
c mapping any negative f's to 0, f > 1 to nind.  (FNF)
      integer      nind, colmap
      real(Size8)  f
      colmap(f,nind) = max(0, min( int(f*nind), nind) )

c Find range of data.
      if (zlim(1) <> zlim(2)) then
         zz(1) = min( zlim(1), zlim(2) )
         zz(2) = max( zlim(1), zlim(2) )
      else
         call eznminmx (z, 1, n, 1, zz(1), zz(2), imin, imax)
      endif

c Check for valid data.
      if (zz(1) <= 0 | zz(2) <= 0) then
         call ezcremrk ("ezccmapln: nonpositive data or limits")
         go to 30
      endif
      zmin = log(zz(1))
      zmax = log(zz(2))

c Check for the case of equal limits.
      if (zmax == zmin) then
         do i=1,n
            ndx(i) = 1
         enddo
c
         go to 30
      endif

c Allocate work array.
      type8 = utstrcod(Quote(real(Size8)))
      numdata = utsizeof(n,type8)
      call osallot(Point(ztmp),numdata)

c Take logs of data (safe, because of previous tests).
      do i=1, n
         zmap = min( max(z(i),zz(1)), zz(2))
         ztmp(i) = log(zmap)
      enddo
c     (Recall that zmin,zmax are already log-mapped.)

c Convert data to color indices.
      ncm1 = ncol - 1
      do i=1, n
         ndx(i) = colmap( (ztmp(i) - zmin)/(zmax - zmin), ncm1) + 1
c         Check for reverse order desired.
         if (zz(5) < 1) ndx(i) = ncol+1 - ndx(i)
      enddo

      call osfree(Point(ztmp))
   30 continue
      zz(3) = 0
      zz(4) = 0
      zz(5) = 1    #logarithmic mapping
      return
      end
