c==============================================================================
c
c  ezcmp8nml
c
c  Purpose: Convert an array of real8's into an array of indices into a
c     colormap using a normal mapping, with optional user-supplied limits.
c
c  Arguments:
c     ncol  Number of colors in colormap.
c     n     Number of elements in array z.
c     z     The array to be mapped.
c     ndx   The output array of color indices, in range [1,ncol].
c     zz    Auxiliary output array.  This routine returns zz(1)=zmin,
c           zz(2)=zmax, zz(3)=zbar, zz(4)=zsigma, zz(5)=zmap=2 (to signal
c           normal mapping).
c           If the input value of zz(5) is negative, reverse the mapping.
c     zlim  If the input values zlim(1), zlim(2) are not equal, they are
c           treated as user-supplied limits.  (Ordering not assumed.)
c
c  Routines called: eznminmx, osallot, osfree, utsizeof, utstrcod.
c
c  Created: Fred N. Fritsch, 26 April 2000
c
c  Revision History:
c     - Created from 000420 version of ezclr8nml, as modified to call
c       eznmimmx to compute array limits.  Note that ANY two equal
c       values cause computation of the actual array limits.
c     - Made zz(5) an input, to determine order of color map. (FNF 010702)
c==============================================================================

      subroutine ezcmp8nml (ncol, n, z, ndx, zz, zlim)

c Declare arguments.
      integer        ncol, n, ndx(n)
      real(Size8)    z(n), zz(5), zlim(2)

c Declare externals.
      integer utsizeof, utstrcod
      external utsizeof, utstrcod

c Declare local variables.
      integer      i, imax, imin
      real(Size8)  zsum, zsumsq, factor
      integer numdata, type8
      Dynamic(ztmp,real(Size8),1)

c Initialize zz.

c Find range of data.
      if (zlim(1) <> zlim(2)) then
         zz(1) = min( zlim(1), zlim(2) )
         zz(2) = max( zlim(1), zlim(2) )
      else
         call eznminmx (z, 1, n, 1, zz(1), zz(2), imin, imax)
      endif

c Check for constant data.
      if (zz(1) == zz(2)) then
         do i=1,n
            ndx(i) = 1
         enddo

         go to 930
      endif

c Allocate work array.
      type8 = utstrcod(Quote(real(Size8)))
      numdata = utsizeof(n,type8)
      call osallot(Point(ztmp),numdata)

c Set the values for working variables.
      do i=1, n
	 ztmp(i) = z(i)
      enddo

c Compute the mean and standard deviation.
      zsum = 0.
      do i=1, n
         zsum = zsum + ztmp(i)
      enddo
c      The following, of course, assumes n>0.
      zz(3) = zsum/n
      
      zsumsq = 0.
      do i=1, n
         zsumsq = zsumsq + ztmp(i)**2
      enddo
c      The following assumes n>1.
      zz(4) = sqrt((zsumsq-(zsum**2/n))/(n-1))

c Handle the case of constant data, to avoid division by zero.
c  (This is needed, despite earlier test, in case user limits
c   are distinct but data is actually constant.)
      if (zz(4) <= 0.) then
         do i=1, n
            ndx(i) = ncol/2
         enddo
         go to 900
      endif

c Convert data to color indices by normalizing the data between
c -2*factor and 2*factor.
c The zsigma factor is determined by 4*factor=ncol.
      factor = ncol/4
      do i=1, n
         ztmp(i) = (ztmp(i) - zz(3))*factor / zz(4)
         if (ztmp(i) > 2*factor) ztmp(i) = 2*factor
         if (ztmp(i) < -2*factor) ztmp(i) = -2*factor
         ndx(i) = int(ztmp(i))+int(ncol/2)
         if (ndx(i) > ncol) ndx(i) = ncol
         if (ndx(i) <= 0) ndx(i) = 1
c         Check for reverse order desired.
         if (zz(5) < 1) ndx(i) = ncol+1 - ndx(i)
      enddo

 900  continue
      call osfree(Point(ztmp))
 930  continue
      zz(5) = 2    #normal mapping
      return
      end
