c=======================================================================
c
c  surface plotting z(i,j) = F(x(i), y(j)).
c
c  Notes:
c   o this routine is not designed to stand alone.  it will only work
c     with the basis system.
c
c   o the scratch array should be at least 2*nx*ny words long, ie.
c     at least twice the size of z(); we get the scratch array
c     dynamically.
c
c   o it is recommended that the viewpoint be outside the plotted
c     box by a factor of 2-5 on the horizontal and 5-10 on the vertical
c     and the point should be well outside the box on the z-coordinate.
c     since i am using perspective viewing, we don't want the object
c     to be too small, so we might have to use smaller multipliers.
c     the point we will look towards is the origin: by design, the
c     origin will be the center of the plotted surface.
c
c   o we accept any size arrays on input now, and provide data thinning
c     as necessary to produce a good plot.  the number of data points
c     used in the plot is user-controllable via basis.  the actual arrays
c     used in plotting will be allocated and freed within this routine.
c
c   o CAUTION: Input array view is changed internally by srfplot.
c
c  Created:   Bert Still [chs], 930302
c  Revision history:
c     moved surface plotting pckage into EZN graphics, 931101 [chs]
c     integrate changes into the repository version of the code. 930719 [chs]
c      -  move loading, scaling, and viewpoint calculations into subroutines
c           to simplify the main routine, 930519 [chs]
c      -  add array thinning to prevent overcrowding on the plot, 930519 [chs]
c     allow for fixed scaling and label on/off. 930608 [chs]
c     allow the title size to dynamically change with the length of
c          the title. 930325 [chs]
c     fake restricted plotting by adding PICSCALE.  by telling the
c          SETR call that the image is larger than it really is,
c          we can force SRFACE to shrink the image a little but still
c          fill the plotting area nicely. 930325 [chs]
c     restricted plotting quadrants won't work because SRFACE redefines
c          the standard transformation.   added code removed, 930323 [chs]
c     add restricted plotting quadrants, 930323 [chs]
c     add legends in top right corner, 930322 [chs]
c     add perspective viewing by using SETR(), 930317 [chs]
c     change where titles get placed based on eyeball, 930311 [chs]
c     add titling, 930308 [chs]
c     add data scaling and eyeball calculation, 930304 [chs]
c     add dynamic allocation of the scratch array, 930303 [chs]
c     Implemented new variables srfiskrt,srfhskrt, 970428 (FNF)
c     Made xlab,ylab local variables, so srfxtle,srfytle will not be
c          modified by srfl2u, 970430 (FNF)
c
c  Known problems:
c     The axis label positioning isn't quite right for certain values
c     of view(2).  This is especially noticeable when srfiskrt=1.
c     [FNF 970430]
c
c=======================================================================

      subroutine srfplot(x,y,z,in_x,in_y,view)

Use(Surface_Plot)
Use(Srf_Internal_Parms)
Use(NCAR_Surface_Int)

      real(Size4)    VIEWSCALE, PICSCALE
      PARAMETER     (VIEWSCALE=12.5_Size4)
      PARAMETER     (PICSCALE=1.1_Size4)
      INTEGER        in_x, in_y, nx, ny, utgetcl, ichht
      INTEGER        ixflip, iyflip, ixswap, iyswap
      CHARACTER*80   xlab, ylab
      INTEGER        xlablen, ylablen, zlablen
      real(Size4)    x(in_x), y(in_y), z(in_x,in_y)
      real(Size4)    low(3), high(3)
      real(Size4)    view(2), sight(6), r0
      INTEGER        i
      CHARACTER*128  outmsg
      INTEGER        srfscal

c  initialize sight to zeros.
      do i=1,6
         sight(i) = 0.0_Size4
      enddo

c  set some common block values (like FRAME off)
      IFR_N = 0
      if (srfiskrt == 1) then
         ISKRT_N = 1
#	  The following assumes z-data scaled to [-1,1].
         HSKRT_N = -(1.0_Size4 + srfhskrt)
      else
         ISKRT_N = 0
      endif

c  put data into plotting arrays, and scale it to prevent pancake plots
      nx = min(in_x, srfnpx)
      ny = min(in_y, srfnpy)
      call allot("srfxplot", nx)
      call allot("srfyplot", ny)
      call allot("srfzplot", nx*ny)
      call srfload(x, y, z, in_x, in_y, srfxplot, srfyplot, srfzplot, nx, ny)
      if (srfscal(srfxplot, srfyplot, srfzplot, nx, ny, low, high) <> OK)
     &    goto 900

c  set up viewport information
      call srfvpt(view, ixflip, iyflip, ixswap, iyswap)

c  other initialization
      sight(1) = VIEWSCALE * cos(view(2)) * cos(view(1))
      sight(2) = VIEWSCALE * cos(view(2)) * sin(view(1))
      sight(3) = VIEWSCALE * (sin(view(2)) + 1.0_Size4)
      r0 = sqrt(sight(1)*sight(1) + sight(2)*sight(2) + sight(3)*sight(3))

c  allocate the work array
      call allot("srfscr", 2*nx*ny)

c  check to see that a graphics device exists and fire up a new frame
      call pltstart
      call ezcnf

c  plot the title: center string, set height (30pt), and write text
      call gselnt(0)
      ichht = min0(800 / utgetcl(srftitle), 30)
      call agpwrt(0.5_Size4, 0.95_Size4, srftitle, utgetcl(srftitle), ichht, 0, 0)

c  get true label lengths, and put legends into the plotting
c      window using:  AGPWRT(x, y, msg, len, ichsize, angle, center)
      xlablen = utgetcl(srfxtle)
      ylablen = utgetcl(srfytle)
      zlablen = utgetcl(srfztle)
      if (srftopln == 1) then
         write(outmsg,101) srfxtle(1:xlablen)//' in ', low(1), ' to ', high(1)
         call agpwrt(1.0_Size4, 0.98_Size4, outmsg, utgetcl(outmsg), 2, 0, 1)
         write(outmsg,101) srfytle(1:ylablen)//' in ', low(2), ' to ', high(2)
         call agpwrt(1.0_Size4, 0.96_Size4, outmsg, utgetcl(outmsg), 2, 0, 1)
         write(outmsg,101) srfztle(1:zlablen)//' in ', low(3), ' to ', high(3)
         call agpwrt(1.0_Size4, 0.94_Size4, outmsg, utgetcl(outmsg), 2, 0, 1)
      else
         write(outmsg,101) srfxtle(1:xlablen)//' in ', low(1), ' to ', high(1)
         call agpwrt(0.98_Size4, 0.09_Size4, outmsg, utgetcl(outmsg), 2, 0, 1)
         write(outmsg,101) srfytle(1:ylablen)//' in ', low(2), ' to ', high(2)
         call agpwrt(0.98_Size4, 0.07_Size4, outmsg, utgetcl(outmsg), 2, 0, 1)
         write(outmsg,101) srfztle(1:zlablen)//' in ', low(3), ' to ', high(3)
         call agpwrt(0.98_Size4, 0.05_Size4, outmsg, utgetcl(outmsg), 2, 0, 1)
      endif

c  set the parameters: SETR (XMIN, XMAX, YMIN, YMAX, ZMIN, ZMAX, R0)
      call setr (-PICSCALE, PICSCALE, -PICSCALE, PICSCALE,
                 -PICSCALE, PICSCALE, r0)

c  do the plot: SRFACE(X, Y, Z, WORK, MX, NX, NY, SIGHT, STEREO) 
      call srface(srfxplot, srfyplot, srfzplot, srfscr, nx, nx, ny, sight, 0._4)

c  convert lowercase to uppercase in labels because PWRZS can't
c       process lowercase
      xlab = srfxtle
      ylab = srfytle
      call srfl2u(xlab)
      call srfl2u(ylab)

c  label the axes
      if (srflabel == 1) then
         ichht = min0(630 / max0(xlablen,ylablen), 30)
         call pwrzs(0.0_Size4, 1.3_Size4*ixswap, 0.0_Size4, xlab, xlablen, ichht, 
                    -1*ixflip, -2*ixflip, 0)
         call pwrzs(-1.3_Size4*iyswap, 0.0_Size4, 0.0_Size4, ylab, ylablen, ichht,
                    -2*iyflip, 1*iyflip, 0)
      endif

      call sflush       #????
c  call the external notification routine
      if (srfxn == 1) then
         call ezcsetxn
      endif

c  cleanup and exit
 900  continue
      call basfree("srfscr")
      call basfree("srfzplot")
      call basfree("srfyplot")
      call basfree("srfxplot")

      return
 101  format(A, G10.3, A, G10.3)
      end

