/*
 * EPGRA.C - PGS interface for EZN
 *
 */

#include <pgs.h>

#define NPTR    32
#define NLINES  64
#define LSIZE  120

#define FillMesh     10
#define PolyFill     13
#define PolyFunc     14

typedef struct s_display_db display_db;

struct s_display_db {
    int maxdata;
    int maxnc;                  /* current maximum size of arrays dimensioned by nc */
    int nc;                     /* number of elements currently in the display list */
    int ndata;

    void **data;

    REAL *dminc;                /* (0:maxdata) _real8 */
    REAL *dmaxc;
    REAL *dminac;
    REAL *dmaxac;

    REAL *xlim;                 /* frame limits in X direction */
    REAL *ylim;                 /* frame limits in Y direction */

    FIXNUM *dnoec;              /* (0:maxdata) _logical */
    FIXNUM *n1;                 /* first dimension of data */
    FIXNUM *n2;                 /* second dimension of data */

    FIXNUM *zi;                 /* index into data for z */
    FIXNUM *xi;                 /* index into data for x */
    FIXNUM *yi;                 /* index into data for y */

/* fillmesh specific */

    FIXNUM *zexti;              /* index into data for z extrema */

/* mesh plot specific */

    FIXNUM *iregi;              /* index into data for ireg */
    FIXNUM *kpregi;             /* index into data for kept regions */
    FIXNUM *kllmti;             /* index into data for k,l limits for mesh plot */
    FIXNUM *bndflagi;           /* index into data for boundary plot flag for mesh plot */

    char *lstylec;              /* l line style for mesh plot */
    char *kstylec;              /* k line style for mesh plot */
    char *lcolorc;              /* l line color for mesh plot */
    char *kcolorc;              /* k line color for mesh plot */

/* vector specific */

    FIXNUM *vyi;
    FIXNUM *vxi;

/* ray specific */

    FIXNUM *pwi;                /* index into data for power */
    FIXNUM *lasi;               /* index to laser number */
    FIXNUM *nrayi;              /* index to the number of rays in each laser */
    FIXNUM *raylabi;            /* index to the labels for rays  */
    FIXNUM *dispmarki;          /* index to the dispmarks for rays */
    FIXNUM *nraypti;            /* index to the number of plot points in each ray */

/* curve, markers and general  */

    FIXNUM *labeli;             /* index to label to use for curve */
    char *stylec;               /* curve style */
    FIXNUM *namec;              /* curve type */
    char *fontc;                /* font to use */
    char *markc;                /* mark to use for curve */
    REAL *marksc;               /* mark size pararmeter for curve marks */
    REAL *thickc;               /* line thickness to use */
    char *colorc;               /* color to use for curve */
    char *legendc;              /* legend to use for curve */
    char *preleg;               /* prefix to legend */
    char *dshp;                 /* dash-pattern holder */

/* miscellaneous */

    int ixprev;
    int ixpreva;
};                              /* default stack elements for plotv, plotm, plotc, plotf */


/* graphics display list data */

static display_db grdl;

static REAL _vxmn = 0.17, _vxmx = 0.82, _vymn = 0.27, _vymx = 0.90;

/*--------------------------------------------------------------------------*/

/*                              STRING HELPERS                              */

/*--------------------------------------------------------------------------*/

/* MK_FNAME - filename assignment for a given filetype
 *          - FNAMEI is the input filename so far
 *          - FTYPE  is the input filetype, it should be of the form:
 *          -      "xxx"    (alias "XXX", "xxxfile", "XXXFILE" allowed)
 *          -      "xxxlog" (alias "XXXLOG", "xxxlogfile", "XXXLOGFILE" allowed)
 *          - FNAMEO is the output filename with extension added
 *          - the calling program is expected to check the result
 *          - take appropriate action
 */

static void mk_fname(char *fnamei, char *ftype, char *fnameo, int log)
{
    int nt;
    char type[MAXLINE];

    strcpy(type, ftype);
    SC_str_lower(type);

/* whack off "file" from type if it exists */
    nt = strlen(type);
    if ((nt > 4) && (strcmp(type + nt - 4, "file") == 0))
        type[nt - 4] = '\0';

/* if it's a log file type */
    if (log)
        sprintf(fnameo, "%s.%slog", fnamei, type);

/* otherwise just use the name */
    else
        strcpy(fnameo, fnamei);

    SC_str_lower(fnameo);

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/*  EZCGFNAM - generate a unique graphics filename with proper extension
 *           - the returned fname will have the form
 *           -
 *           -           NAMER_NAMESFX.EXT
 *           -
 *           - where
 *           -   NAMER    is the input "root" of the filename,
 *           -   NAMESFX  is a unique 3-digit file suffix, and
 *           -   EXT      is an extension based on ftype
 *           - in addition to NAMER, FTYPE, and FNAME, the arguments
 *           - to ezcgfnam are:
 *           -   FNSFX    is the trial value for namesfx on input
 *           -            if succ = TRUE, the returned value is one
 *           -            greater than
 *           -            the actual suffix in the returned fname;
 *           -            i.e., the value
 *           -            appropriate for the nextto ezcgfnam
 *           -            if fnsfx(in) was acceptable, fnsfx(out)=fnsfx(in)+1
 *           -   suff     is the output logical success/fail flag
 *           -            if suff=.false., then fname will be blank
 *           -            and fnsfx will remain unchanged
 *           - the ftype-to-ext mapping is done in subroutine mk_fname
 *           - NOTE: If fnsfx(in) >= 1000 (so not representable as
 *           - a 3-digit number) the generated name will have form
 *           -
 *           -       NAMER_1_NAMESFX.EXT
 *           -
 *           - this means that a total of 1999 filenames can be generated for a
 *           - given namer and ftype  (Anything > 1999 will give error)
 */

FIXNUM F77_ID(ezcgfnam_, ezcgfnam, EZCGFNAM)
    (FIXNUM *pnc, F77_string pnamer,
     FIXNUM *pnt, F77_string pftype,
     FIXNUM *pnn, F77_string pfname,
     FIXNUM *pfnsfx, FIXNUM *plog)
{
    int nc, nt;
    int fnsfx, log;
    int namesfx, savesfx;
    char nameo[MAXLINE], namei[MAXLINE], namew[MAXLINE];
    char *namer, *ftype, *fname;
    struct stat sbf;

    nc = *pnc;
    nt = *pnt;
    fnsfx = *pfnsfx;
    log = *plog;

    namer = SC_F77_C_STRING(pnamer);
    ftype = SC_F77_C_STRING(pftype);
    fname = SC_F77_C_STRING(pfname);

    namer[nc] = '\0';
    ftype[nt] = '\0';
    fname[0] = '\0';

    namesfx = fnsfx;
    do {
        strcpy(namei, namer);
        savesfx = namesfx;
        if (namesfx >= 1000) {
            namesfx -= 1000;
            if (namesfx >= 1000) {
                printf
                    ("ezcgfnam: Could not build unique suffix for %s file.",
                     ftype);
                strcpy(fname, " ");
                return ((FIXNUM) FALSE);
            }

            sprintf(namei, "%s.1", namer);
        }

        sprintf(namew, "%s.%03d", namei, namesfx + 1);

/* append proper extension, based on ftype */
        mk_fname(namew, ftype, nameo, log);

        namesfx++;
        if (savesfx >= 1000)
            namesfx += 1000;
    }

    while (stat(nameo, &sbf) == 0);

/* namesfx is one greater than the suffix used in nameo
 * or 1000 greater than that if suffix has rolled over)
 */
    strcpy(fname, nameo);
    *pnn = strlen(fname);
    *pfnsfx = namesfx;

    return ((FIXNUM) TRUE);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* _EP_GETCL - return index of last non-blank in w or 1 */

static int _EP_getcl(char *w, int n)
{
    int i, c;

    for (i = 0; i < n; i++) {
        c = w[i];
        if (c == '\0')
            break;
    }

    for (; i > 0; i--) {
        c = w[i - 1];
        if (c != ' ')
            break;
    }

    return (i);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* _EP_FIX_STRING - make a Fortran string OK */

static void _EP_fix_string(char *s, int n)
{
    int nc;

    nc = _EP_getcl(s, n);
    s[nc] = '\0';

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* EZCGETCL - return index of last non-blank in w or 1 */

FIXNUM F77_ID(ezcgetcl_, ezcgetcl, EZCGETCL)(F77_string pw, FIXNUM n) {
    char *w;

    w = SC_F77_C_STRING(pw);

    return (_EP_getcl(w, (int) n));
}

/*--------------------------------------------------------------------------*/

/*                             RENDER HELPER                                */

/*--------------------------------------------------------------------------*/

/* _EP_SET_VIEWPORT - setup rendering and adornment */

static pcons *_EP_set_viewport(PG_device * dev, pcons * info)
{
    REAL *pv;

    pv = FMAKE_N(REAL, 4, "_EP_SET_VIEWPORT:pv");
    pv[0] = _vxmn;
    pv[1] = _vxmx;
    pv[2] = _vymn;
    pv[3] = _vymx;

    info = SC_change_alist(info, "VIEW-PORT", SC_REAL_P_S, pv);

    PG_set_viewport(dev, _vxmn, _vxmx, _vymn, _vymx);

    return (info);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* _EP_SET_WC - find and set the WC for the picture */

static void _EP_set_wc(PG_device * dev)
{
    int i, m, nc, ipl, npl, id, inx, iny;
    int n1x, n2x;
    FIXNUM *xi, *yi, *n1, *n2, *namec;
    REAL xmn, xmx, ymn, ymx, xc, yc;
    REAL *x, *y, **data;

    nc = grdl.nc;
    data = (REAL **) grdl.data;
    namec = grdl.namec;
    xi = grdl.xi;
    yi = grdl.yi;
    n1 = grdl.n1;
    n2 = grdl.n2;

    xmn = HUGE;
    xmx = -HUGE;
    ymn = HUGE;
    ymx = -HUGE;

    npl = 0;

    for (ipl = 0; ipl < nc; ipl++) {
        id = namec[ipl];
        if ((id < 12) && (id != 8)) {
            inx = xi[ipl] - 1;
            iny = yi[ipl] - 1;

            n1x = n1[inx];
            n2x = n2[inx];

            x = data[inx];
            y = data[iny];

            m = n1x * n2x;
            for (i = 0; i < m; i++) {
                xc = x[i];
                yc = y[i];

                xmn = min(xmn, xc);
                xmx = max(xmx, xc);
                ymn = min(ymn, yc);
                ymx = max(ymx, yc);
            }

            npl++;
        }
    }

    if (npl > 0)
        PG_set_window(dev, xmn, xmx, ymn, ymx);

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* _EP_RENDER_INFO - setup rendering and adornment */

static void _EP_render_info(PG_device *dev, PG_graph *data,
                            int render, int dd, FIXNUM *irg,
                            int clr, int sty, REAL wid,
                            int scat, int mark, REAL marksc)
{
    int nde, nre;
    REAL ms;
    REAL *dpex, *rpex, *rdex, *prx;
    pcons *info;

    PG_get_marker_scale(dev, ms);
    marksc = (marksc == 0.0) ? ms : marksc;

/* set attributes on the mapping */
    info = (pcons *) data->f->map;

    if (irg != NULL)
        info = SC_add_alist(info, "EXISTENCE", SC_INTEGER_P_S, irg);

    data->f->map = (byte *) info;

/* set attributes on the graph */
    info = (pcons *) data->info;

    if (dd == 2)
        info = PG_set_tds_info(info, render, CARTESIAN,
                               sty, clr, 10, 1.0, wid,
                               0.0, 0.0, 0.0, HUGE);
    else
        info = PG_set_line_info(info, render, CARTESIAN,
                                sty, scat, mark, clr, CENTER, wid);

    SC_CHANGE_VALUE_ALIST(info, int, SC_REAL_P_S, "MARKER-SCALE", marksc);

    info = _EP_set_viewport(dev, info);

    data->info = (byte *) info;
    data->rendering = render;

    _PG_axis_tick_type = LEFT_OF_AXIS;
    _PG_palette_orientation = VERTICAL;

    PG_turn_data_id(dev, TRUE);
    PG_set_marker_scale(dev, marksc);

/* register the range extrema */
    PG_find_extrema(data, 0.0, &dpex, &rpex, &nde, NULL, &nre, &rdex);

    prx = ((dev->autorange == TRUE) || (rpex == NULL)) ? rdex : rpex;
    PG_register_range_extrema(dev, nre, prx);

    SFREE(rdex);

/* set the domain limits */
    PG_set_window(dev, grdl.xlim[0], grdl.xlim[1],
		       grdl.ylim[0], grdl.ylim[1]);

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* _EP_DEF_ROUND_MARKER - define a marker which is a circle of radius R */

static int _EP_def_round_marker(REAL r)
{
    int i, id, n;
    REAL da, ac;
    REAL x1[NPTR], x2[NPTR], y1[NPTR], y2[NPTR];

    n = NPTR;
    da = DEG_RAD * 360.0 / ((REAL) n);
    for (i = 0; i < n; i++) {
        ac = i * da;
        x1[i] = r * cos(ac);
        y1[i] = r * sin(ac);
    }

    x2[0] = x1[n - 1];
    y2[0] = y1[n - 1];
    for (i = 1; i < n; i++) {
        x2[i] = x1[i - 1];
        y2[i] = y1[i - 1];
    }

    id = PG_def_marker(n, x1, y1, x2, y2);

    return (id);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* _EP_PLOT_ATTRS - fetch plot attribute for plot IPL */

static void _EP_plot_attrs(int ipl,
                           int *pclr, int *psty, REAL * pwid,
                           int *pmrk, REAL * pmrs)
{
    int clr, sty, mrk;
    int crossid, circleid, dotid;
    REAL wid, mrs;
    REAL x1[10], x2[10], y1[10], y2[10];
    char *styn, *clrn, *mrkn;

/* get color */
    if (pclr != NULL) {
        clrn = grdl.colorc + ipl * 16;
        clr = 1;
        if (strncmp(clrn, "red", 3) == 0)
            clr = 13;
        else if (strncmp(clrn, "green", 5) == 0)
            clr = 11;
        else if (strncmp(clrn, "blue", 4) == 0)
            clr = 10;
        else if (strncmp(clrn, "cyan", 4) == 0)
            clr = 12;
        else if (strncmp(clrn, "magenta", 7) == 0)
            clr = 15;
        else if (strncmp(clrn, "yellowg", 7) == 0)
            clr = 5;
        else if (strncmp(clrn, "yellow", 6) == 0)
            clr = 14;
        else if (strncmp(clrn, "coral", 5) == 0)
            clr = 7;
        else if (strncmp(clrn, "springg", 7) == 0)
            clr = 5;
        else if (strncmp(clrn, "slatebl", 7) == 0)
            clr = 10;
        else if (strncmp(clrn, "skyblu", 6) == 0)
            clr = 4;
        else if (strncmp(clrn, "orange", 6) == 0)
            clr = 9;
        else if (strncmp(clrn, "gray33", 6) == 0)
            clr = 2;
        else if (strncmp(clrn, "lavend", 6) == 0)
            clr = 8;
        else if (strncmp(clrn, "orchid", 6) == 0)
            clr = 6;
        else if (strncmp(clrn, "gray70", 6) == 0)
            clr = 3;

        *pclr = clr;
    }

/* get line style */
    if (psty != NULL) {
        styn = grdl.stylec + ipl * 8;
        sty = LINE_SOLID;
        if (strncmp(styn, "soli", 4) == 0)
            sty = LINE_SOLID;
        else if (strncmp(styn, "dash", 4) == 0)
            sty = LINE_DASHED;
        else if (strncmp(styn, "dott", 4) == 0)
            sty = LINE_DOTTED;
        else if (strncmp(styn, "dotd", 4) == 0)
            sty = LINE_DOTDASHED;

        *psty = sty;
    }

/* get line width */
    if (pwid != NULL) {
        wid = grdl.thickc[ipl];
        *pwid = 0.5 * wid;
    }

/* get the marker */
    if (pmrk != NULL)

/* define 'x' marker */
    {
        x1[0] = -0.71;
        y1[0] = -0.71;

        x2[0] = 0.71;
        y2[0] = 0.71;

        x1[1] = 0.71;
        y1[1] = -0.71;

        x2[1] = -0.71;
        y2[1] = 0.71;

        crossid = PG_def_marker(2, x1, y1, x2, y2);

/* define circle and dot markers */
        circleid = _EP_def_round_marker(1.0);
        dotid = _EP_def_round_marker(0.05);

        mrkn = grdl.markc + ipl * 8;
        mrk = 0;
        if (strncmp(mrkn, "plus", 4) == 0)
            mrk = 0;
        else if (strncmp(mrkn, "aste", 4) == 0)
            mrk = 1;
        else if (strncmp(mrkn, "cros", 4) == 0)
            mrk = crossid;
        else if (strncmp(mrkn, "circ", 4) == 0)
            mrk = circleid;
        else if (strncmp(mrkn, "dot", 3) == 0)
            mrk = dotid;

        *pmrk = mrk;
    }

/* get the marker scale */
    if (pmrs != NULL) {
        mrs = grdl.marksc[ipl];
        *pmrs = 0.005 * mrs;
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* _EP_REND_11 - render 1D 1D datasets */

static void _EP_rend_11(FIXNUM *devid, FIXNUM *pipl, FIXNUM *pdid,
                        int render, char *title)
{
    int id, ipl;
    int inx, iny, n1x, n1y, n2x, n2y, strx, stry;
    int lc, ls, olc, ols, di;
    int mrk, sct, disjoint;
    REAL lw, olw, mrs;
    REAL *x, *y;
    PM_set *domain, *range;
    PG_device *dev;
    PG_graph *data;
    PG_palette *pal;

    dev = SC_GET_POINTER(PG_device, *devid);
    pal = dev->current_palette;

    ipl = *pipl - 1;
    di  = *pdid;

    _EP_plot_attrs(ipl, &lc, &ls, &lw, &mrk, &mrs);

    inx = grdl.xi[ipl] - 1;
    iny = grdl.yi[ipl] - 1;
    id  = grdl.namec[ipl];
   
    x = grdl.data[inx];
    y = grdl.data[iny];

    n2x = grdl.n2[inx];
    n2y = grdl.n2[iny];

/* disjoint polyline */
    if ((n2x != 1) || (n2y != 1))
       {disjoint = TRUE;
        strx     = grdl.n1[inx];
	stry     = grdl.n1[iny];
        n1x      = n2x;
	n1y      = n2y;

        PG_set_viewport(dev, _vxmn, _vxmx, _vymn, _vymx);

	PG_get_line_color(dev, &olc);
	PG_get_line_style(dev, &ols);
	PG_get_line_width(dev, &olw);

	PG_set_line_color(dev, lc);
	PG_set_line_style(dev, ls);
	PG_set_line_width(dev, lw);

        PG_draw_disjoint_polyline_2(dev, x, y, n1x, FALSE, TRUE);

	PG_set_line_color(dev, olc);
	PG_set_line_style(dev, ols);
	PG_set_line_width(dev, olw);

       }

/* contigous curve */
    else
       {disjoint = FALSE;
        n1x      = grdl.n1[inx];
	n1y      = grdl.n1[iny];

	sct = (id == 1);            /* Mark in EZN.V */

	if (n1x == n1y) {
	   domain = PM_make_set(" ", SC_REAL_S, FALSE, 1, n1x, 1, x);

	   range = PM_make_set(" ", SC_REAL_S, FALSE, 1, n1y, 1, y);

	   data = PG_make_graph_from_sets("  ", domain, range,
					  N_CENT,
					  SC_PCONS_P_S, NULL, 'A', NULL);

/* set attributes on the graph */
	   _EP_render_info(dev, data, render, 1, NULL,
			   lc, ls, lw, sct, mrk, mrs);

	   PG_set_clear_mode(FALSE);
	   PG_set_palette(dev, "standard");
	   PG_turn_data_id(dev, di);

	   PG_draw_graph(dev, data);

	   PG_rl_graph(data, TRUE, TRUE);

	   dev->current_palette = pal;
	   };
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* _EP_REND_21 - render 2D 1D datasets */

static void _EP_rend_21(FIXNUM *devid, FIXNUM *pipl, FIXNUM **preg,
                        FIXNUM **pkreg, FIXNUM **pklmt, FIXNUM **pbndfl,
                        int render, char *title)
{
    int ipl, ls, lc, mrk;
    int inx, iny, inf, n1x, n2x, n1y, n2y;
    FIXNUM *irg;
    REAL lw, mrs;
    REAL *x, *y, *f;
    PM_set *domain, *range;
    PG_device *dev;
    PG_graph *data;

    dev = SC_GET_POINTER(PG_device, *devid);

    ipl = *pipl - 1;
    irg = *preg;
/*
    krg = *pkreg;
    klm = *pklmt;
    bnd = *pbndfl;
*/
    inx = grdl.xi[ipl] - 1;
    iny = grdl.yi[ipl] - 1;
    inf = grdl.zi[ipl] - 1;
    n1x = grdl.n1[inx];
    n2x = grdl.n2[inx];
    n1y = grdl.n1[iny];
    n2y = grdl.n2[iny];

    x = grdl.data[inx];
    y = grdl.data[iny];
    f = grdl.data[inf];

    _EP_plot_attrs(ipl, &lc, &ls, &lw, &mrk, &mrs);

    if ((n1x == n1y) && (n2x == n2y)) {
        domain = PM_make_set(" ", SC_REAL_S, FALSE, 2, n1x, n2x, 2, x, y);

        range = PM_make_set("f", SC_REAL_S, FALSE,
                            2, n1x - 1, n2x - 1, 1, f);

        data = PG_make_graph_from_sets("  ", domain, range,
                                       Z_CENT,
                                       SC_PCONS_P_S, NULL, 'A', NULL);

/* set attributes on the graph */
        _EP_render_info(dev, data, render, 2, irg,
                        lc, ls, lw, 0, mrk, mrs);

        PG_draw_graph(dev, data);

        PG_rl_graph(data, TRUE, TRUE);
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* _EP_REND_21_AC - render 2D 1D datasets of polygons */

static void _EP_rend_21_ac(FIXNUM *devid, FIXNUM *pipl, FIXNUM **preg,
			   FIXNUM **pkreg, FIXNUM **pklmt, FIXNUM **pbndfl,
			   int render, char *title)
{
    int i, j;
    int ipl, ls, lc, mrk;
    int inx, iny, inf, n1y, n2y;
    int npp, nply, ne, nc;
    int *bnc, *bnp;
    long **bnd, *edge, *cell;
    FIXNUM *irg;
    REAL lw, mrs;
    REAL *x, *y, *f;
    PM_set *domain, *range;
    PG_device *dev;
    PG_graph *data;
    PM_mesh_topology *mt;

    dev = SC_GET_POINTER(PG_device, *devid);

    ipl = *pipl - 1;
    irg = *preg;

    inx  = grdl.xi[ipl] - 1;
    iny  = grdl.yi[ipl] - 1;
    inf  = grdl.zi[ipl] - 1;
    npp  = grdl.n1[inx];
    nply = grdl.n2[inx];
    n1y  = grdl.n1[iny];
    n2y  = grdl.n2[iny];

    x = grdl.data[inx];
    y = grdl.data[iny];
    f = grdl.data[inf];

    _EP_plot_attrs(ipl, &lc, &ls, &lw, &mrk, &mrs);

    if ((npp == n1y) && (nply == n2y)) {
        bnp = FMAKE_N(int, 3, "_EP_REND_21_AC:bnp");
        bnc = FMAKE_N(int, 3, "_EP_REND_21_AC:bnc");
        bnd = FMAKE_N(long *, 3, "_EP_REND_21_AC:bnd");
        mt  = PM_make_topology(2, bnp, bnc, bnd);

	bnp[0] = 1;
	bnp[1] = 1;
	bnp[2] = 1;

	bnc[0] = npp*nply;
	bnc[1] = npp*nply;
	bnc[2] = nply;

	bnd[0] = NULL;
	bnd[1] = FMAKE_N(long, bnc[1]+1, "_EP_REND_21_AC:bnd[1]");
	bnd[2] = FMAKE_N(long, bnc[2]+1, "_EP_REND_21_AC:bnd[2]");

	edge = bnd[1];
	cell = bnd[2];

	ne = 0;
	nc = 0;
	for (i = 0; i < nply; i++)
	    {cell[nc++] = ne;
	     for (j = 0; j < npp; j++)
	         edge[ne++] = i*npp + j;};
	cell[nc++] = ne;

        domain = PM_make_ac_set(" ", SC_REAL_S, FALSE, mt, 2, x, y);

        range = PM_make_set("f", SC_REAL_S, FALSE,
                            1, nply, 1, f);

        data = PG_make_graph_from_sets("  ", domain, range,
                                       Z_CENT,
                                       SC_PCONS_P_S, NULL, 'A', NULL);

/* set attributes on the graph */
        _EP_render_info(dev, data, render, 2, irg,
                        lc, ls, lw, 0, mrk, mrs);

        PG_draw_graph(dev, data);

        PG_rl_graph(data, TRUE, TRUE);
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* _EP_DRAW_POLY - render 2D 1D datasets of polygons */

static void _EP_draw_poly(FIXNUM *devid, FIXNUM *pipl, FIXNUM **preg,
			  FIXNUM **pkreg, FIXNUM **pklmt, FIXNUM **pbndfl)
{
    int i, j, m;
    int ipl, ls, lc, mrk;
    int inx, iny, inf;
    int npp, nply, npt;
    REAL lw, mrs;
    REAL *x, *y, *xpt, *ypt;
    PG_device *dev;
    PG_palette *pl;

    dev = SC_GET_POINTER(PG_device, *devid);

    pl = dev->current_palette;

    PG_set_palette(dev, "standard");

    ipl = *pipl - 1;

    inx  = grdl.xi[ipl] - 1;
    iny  = grdl.yi[ipl] - 1;
    inf  = grdl.zi[ipl] - 1;
    npp  = grdl.n1[inx];
    nply = grdl.n2[inx];

    x = grdl.data[inx];
    y = grdl.data[iny];

    _EP_plot_attrs(ipl, &lc, &ls, &lw, &mrk, &mrs);

    PG_set_color_fill(dev, lc, TRUE);

    xpt = FMAKE_N(REAL, npp, "_EP_DRAW_POLY:xpt");
    ypt = FMAKE_N(REAL, npp, "_EP_DRAW_POLY:ypt");

    m = 0;
    for (i = 0; i < nply; i++) {
        npt = 0;
	for (j = 0; j < npp; j++) {
	    xpt[npt] = x[m];
	    ypt[npt] = y[m];
	    m++;

	    npt++;
	};

	PG_shade_poly(dev, xpt, ypt, npt);
    };

    SFREE(xpt);
    SFREE(ypt);

    dev->current_palette = pl;

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* _EP_REND_22 - render 2D 2D datasets */

static void _EP_rend_22(FIXNUM *devid, FIXNUM *pipl, FIXNUM **preg,
                        FIXNUM **pkreg, FIXNUM **pklmt, FIXNUM **pbndfl,
                        int render, char *title)
{
    int ipl, lc, ls, mrk;
    int inx, iny, inu, inv, n1x, n2x, n1y, n2y;
    FIXNUM *irg;
    REAL lw, mrs;
    REAL *x, *y, *u, *v;
    PM_set *domain, *range;
    PG_device *dev;
    PG_graph *data;

    dev = SC_GET_POINTER(PG_device, *devid);

    ipl = *pipl - 1;
    irg = *preg;
/*
    krg = *pkreg;
    klm = *pklmt;
    bnd = *pbndfl;
*/
    inx = grdl.xi[ipl] - 1;
    iny = grdl.yi[ipl] - 1;
    inu = grdl.vxi[ipl] - 1;
    inv = grdl.vyi[ipl] - 1;
    n1x = grdl.n1[inx];
    n2x = grdl.n2[inx];
    n1y = grdl.n1[iny];
    n2y = grdl.n2[iny];

    x = grdl.data[inx];
    y = grdl.data[iny];
    u = grdl.data[inu];
    v = grdl.data[inv];

    _EP_plot_attrs(ipl, &lc, &ls, &lw, &mrk, &mrs);

    if ((n1x == n1y) && (n2x == n2y)) {
        domain = PM_make_set(" ", SC_REAL_S, FALSE, 2, n1x, n2x, 2, x, y);

        range = PM_make_set("f", SC_REAL_S, FALSE,
                            2, n1x - 1, n2x - 1, 2, u, v);

        data = PG_make_graph_from_sets("  ", domain, range,
                                       N_CENT,
                                       SC_PCONS_P_S, NULL, 'A', NULL);

/* set attributes on the graph */
        _EP_render_info(dev, data, render, 2, irg,
                        lc, ls, lw, 0, mrk, mrs);

        PG_draw_graph(dev, data);

        PG_rl_graph(data, TRUE, TRUE);
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* PEDAT - let C side know about the data set
 *       - this is only here because GDB won't show
 *       - FORTRAN common block contents
 */

void F77_ID(pedat_, pedat, PEDAT)(FIXNUM *maxdata, FIXNUM *maxnc, FIXNUM *nc,
				  FIXNUM *ndata, void **data,
				  REAL *dminc, REAL *dmaxc,
				  REAL *dminac, REAL *dmaxac,
				  FIXNUM *dnoec, FIXNUM *n1, FIXNUM *n2,
				  FIXNUM *zi, FIXNUM *xi, FIXNUM *yi,
				  FIXNUM *zexti, FIXNUM *iregi, FIXNUM *kpregi,
				  FIXNUM *kllmti, FIXNUM *bndflagi,
				  char **lstylec, char **kstylec,
				  char **lcolorc, char **kcolorc,
				  FIXNUM *vyi, FIXNUM *vxi, FIXNUM *pwi,
				  FIXNUM *lasi, FIXNUM *nrayi, FIXNUM *raylabi,
				  FIXNUM *dispmarki, FIXNUM *nraypti,
				  FIXNUM *labeli, char **stylec,
				  FIXNUM *namec, char **fontc,
				  char **markc, REAL *marksc,
				  REAL *thickc, char **colorc,
				  char **legendc, char **preleg,
				  char *dshp, FIXNUM *ixprev, FIXNUM *ixpreva,
				  REAL *xfrlim, REAL *yfrlim)
{
    grdl.maxdata   = *maxdata;
    grdl.maxnc     = *maxnc;
    grdl.nc        = *nc;
    grdl.ndata     = *ndata;
    grdl.data      = data;
    grdl.dminc     = dminc;
    grdl.dmaxc     = dmaxc;
    grdl.dminac    = dminac;
    grdl.dmaxac    = dmaxac;
    grdl.dnoec     = dnoec;
    grdl.n1        = n1;
    grdl.n2        = n2;
    grdl.zi        = zi;
    grdl.xi        = xi;
    grdl.yi        = yi;
    grdl.zexti     = zexti;
    grdl.iregi     = iregi;
    grdl.kpregi    = kpregi;
    grdl.kllmti    = kllmti;
    grdl.bndflagi  = bndflagi;
    grdl.lstylec   = *lstylec;
    grdl.kstylec   = *kstylec;
    grdl.lcolorc   = *lcolorc;
    grdl.kcolorc   = *kcolorc;
    grdl.vyi       = vyi;
    grdl.vxi       = vxi;
    grdl.pwi       = pwi;
    grdl.lasi      = lasi;
    grdl.nrayi     = nrayi;
    grdl.raylabi   = raylabi;
    grdl.dispmarki = dispmarki;
    grdl.nraypti   = nraypti;
    grdl.labeli    = labeli;
    grdl.stylec    = *stylec;
    grdl.namec     = namec;
    grdl.fontc     = *fontc;
    grdl.markc     = *markc;
    grdl.marksc    = marksc;
    grdl.thickc    = thickc;
    grdl.colorc    = *colorc;
    grdl.legendc   = *legendc;
    grdl.preleg    = *preleg;
    grdl.dshp      = dshp;
    grdl.ixprev    = *ixprev;
    grdl.ixpreva   = *ixpreva;
    grdl.xlim      = xfrlim;
    grdl.ylim      = yfrlim;

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* EZCWCLAB - write msg at x, y angle, center into a box which has
 *          - dimensions ht x wd of the viewport
 */

void ezcwclab(PG_device * dev, REAL y, char *msg)
{
    PG_palette *pal;

    pal = dev->current_palette;

    PG_set_palette(dev, "standard");

    PG_set_text_color(dev, dev->WHITE);

    PG_center_label(dev, y, msg);

    dev->current_palette = pal;

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* _EP_WTEXT - write msg at x, y angle, center into a box which has
 *           - dimensions ht x wd of the viewport
 */

void _EP_wtext(PG_device * dev, REAL x, REAL y, char *msg, REAL ht,
               REAL wd, REAL angle, REAL center, int clr)
{
    REAL dx, dy, ca, sa;
    PG_palette *pal;

    sa = sin(angle);
    ca = cos(angle);

    pal = dev->current_palette;

    PG_set_palette(dev, "standard");

    PG_set_text_color(dev, clr);

    PG_set_char_path(dev, ca, sa);
    PG_get_text_ext_NDC(dev, msg, &dx, &dy);

    x -= 0.5 * (ca*dx + sa*dy);
    y -= 0.5 * (-sa*dx + ca*dy);

    PG_write_NDC(dev, x, y, "%s", msg);

    PG_set_char_path(dev, 1.0, 0.0);

    dev->current_palette = pal;

#if 0
       real(Size8) x, y, ht, wd, angle, center
       character*(*) msg
       real(Size8) v1, v2, v3, v4, size1, size2, hw, u1, u2, u3, u4, a
       integer isize, n, ll

/* get viewport: u1-u4, ll are ignored */
       call getset(v1, v2, v3, v4, u1, u2, u3, u4, ll)

/* character aspect ratio hw (1.75 default) */
       call pcgetr('HW', hw) 
       n = len(msg)
       a = angle*acos(-1._Size8)/180._Size8

/* height the text box will take at character width isize is: */
/*       hsize=isize*hw*(cos(a)+sin(a)*n) */
/*       wsize=isize*(sin(a)+cos(a)*n) */
/* The constraints on isize are thus */
/*       ht*(v4-v3) >= isize*(hw*cos(a)+sin(a)*n) */
/*       wd*(v2-v1) >= isize*(hw*sin(a)+cos(a)*n) */
        size1=ht*(v4-v3)/(hw*cos(a)+sin(a)*n)
        size2=wd*(v2-v1)/(hw*sin(a)+cos(a)*n)

c plotter units 1024/screen
c agpwrt misinterprets sizes < 5
       isize = 1024*min(size1, size2)
       isize = max(5, isize)
       call agpwrt(x, y, msg, n, isize, int(angle), int(center))

       call ezcsetxn
#endif

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* EZCWTEXT - write msg at x, y angle, center into a box which has
 *          - dimensions ht x wd of the viewport
 */

void F77_ID(ezcwtext_, ezcwtext, EZCWTEXT)(REAL * px, REAL * py,
                                           F77_string msg,
                                           REAL * pht, REAL * pwd,
                                           REAL * pa, REAL * pc)
{

/*
    REAL xc, yc, ht, wd, an, al;
    char *t;
    PG_device *dev;

    dev = SC_GET_POINTER(PG_device, *devid);
    t   = SC_F77_C_STRING(msg);

    xc  = *px;
    yc  = *py;
    ht  = *pht;
    wd  = *pwd;
    an  = *pa;
    al  = *pc;

    _EP_wtext(dev, xc, yc, t, ht, wd, an, al, dev->WHITE);
*/
    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* PEADT - let C do the memory management hence the ADD
 *       - function on the display list
 */

FIXNUM F77_ID(peadt_, peadt, PEADT)(REAL * x, FIXNUM *di, FIXNUM *dj,
                                    FIXNUM *pfl)
{
    FIXNUM j;

    j = 0;

    return (j);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* PESTP - setup of globals for EZN on PGS */

void F77_ID(pestp_, pestp, PESTP)()
{

/*

c ... set up dash patterns
        pattern(1) = "$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$"
        pattern(2) = "$$$'$$$'$$$'$$$'$$$'$$$'$$$'$$$'"
        pattern(3) = "$'$'$'$'$'$'$'$'$'$'$'$'$'$'$'$'"
        pattern(4) = "$'$'$$$'$'$'$$$'$'$'$$$'$'$'$$$'"
        pattern(5) = "$$$$$$$>$$$$$$$>$$$$$$$>$$$$$$$>"
        pattern(6) = "$$$$$$$<$$$$$$$<$$$$$$$<$$$$$$$<"
        pattern(7) = "''''''''''''''''''''''''''''''''"

c ... set up defaults
c ... autograph
	call agseti('DASH/SELECTOR.', 1)
	call agseti('DASH/LENGTH.', LDASH)
	null = char(0)

	call agsetf('WINDOW.', 1._Size8)
	call ezcminsz(12)

c ... conpack
	do i = -3, -2
	   call cpseti('PAI - parameter array index', i)
	   call cpseti('CLU - contour level usage', 1)
	   call cpsetc('CLD - contour level dash pattern', pattern(3))
	enddo

	call cpseti('SET -- calling of set by conpack', 0)
	call cpsetr('SPV -- special value', NCARSPV)
	call cpsetr('ORV -- out of range value', NCARSPV)
	call cpsetc('ILT -- info level text', ' ')

c ... these are matters of taste
        call cpsetr('CWM -- character width multiplier', 1.2_Size8)
        call cpsetc('HLT -- hi/lo text labels', 'H''L')

c ... small wedges need labels on them too...
        call cpsetr('RC1 -- contour label positioning from start', .05_Size8)
*/

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* PEADN - adorn plot for EZN on PGS */

void F77_ID(peadn_, peadn, PEADN)(FIXNUM *devid, F77_string psc,
                                  F77_string pgx, F77_string pgy)
{
    int xls, yls, grd;
    char *scalea, *gridxa, *gridya;
    PG_device *dev;

    dev = SC_GET_POINTER(PG_device, *devid);

    gridxa = SC_F77_C_STRING(pgx);
    gridya = SC_F77_C_STRING(pgy);
    scalea = SC_F77_C_STRING(psc);

    _EP_fix_string(scalea, 8);

    PG_set_finish_state(dev, FALSE);
    PG_set_clear_mode(CLEAR_FRAME);

    PG_clear_window(dev);

    dev->supress_setup = TRUE;

/* set up the log scales */
    xls = FALSE;
    yls = FALSE;
    if (strcmp(scalea, "linlog") == 0)
        yls = TRUE;

    else if (strcmp(scalea, "loglin") == 0)
        xls = TRUE;

    else if (strcmp(scalea, "loglog") == 0) {
        xls = TRUE;
        yls = TRUE;
    }

    PG_set_axis_log_scale(dev, xls, yls);

/* set up grid drawing */
    grd = FALSE;
    grd = ((strncmp(gridxa, "on", 2) == 0) ||
           (strncmp(gridya, "on", 2) == 0));

    PG_turn_grid(dev, grd);

/* set the world coordinate system */
    _EP_set_wc(dev);

#if 0
c ... variable sizes
c ... merge spaces where possible
      if (~ezcfixed) then
         if (titler .eq. " ") titlert = 0._Size8
         if (titlel .eq. " ") titlelf = 0._Size8
         if (ezctitle .eq. " ") suptitle = 0._Size8
         if (titlet  .eq.  " ") titletp = 0._Size8
         if (titleb  .eq.  " ") titlebt = 0._Size8 
	 if (~conplot & ~fmplot & ~raypw) conkey = 0._Size8
         xdivider = leginfoa*xdivider
      endif

c ... compute proper margin spaces
      view(1) = quad(1)+titlelf
      view(2) = quad(2)-titlert-conkey
      if (ezcsuper) then
         view(3) = quad(3)+xdivider+titlebt
         view(4) = quad(4)-suptitle-titletp
      else
         view(3) = quad(3)+xdivider+titlebt+suptitle
         view(4) = quad(4)-titletp
      endif

      call agsetp('GRAPH.', view, 4)

c ... grid
c ... attempt to use the type size to make things fit
      grid(1) = jminsz*10/(1024*(view(2)-view(1)))
      grid(1) = max(.1_Size8, min(grid(1), .25_Size8))
      grid(2) = .98_Size8
      grid(3) = jminsz*10/(1024*(view(2)-view(1)))
      grid(3) = max(.1_Size8, min(grid(3), .25_Size8))
      grid(4) = .98_Size8   
      grid(5) = 0._Size8

      if (scalea = "linlog") then
      elseif (scalea = "loglin") then
      elseif (scalea = "loglog") then
      elseif (scalea = "equal") then
         grid(5) = 10._Size8  /* value for GRID/SHAPE. */
      endif

      call agsetp('X.', xspec, 3)
      call agsetp('Y.', yspec, 3)
      call agsetp('GRID.', grid, 5)

      if (textonly = NO) then
         call agseti('BAC.', 1)
         call agseti('LABEL/CONTROL.', 0)
         if (gridxa = "on") then
            call agsetf('AXIS/BOTTOM/TICKS/MAJOR/LENGTH/INWARD.', 1.0_Size8)
            call agsetf('AXIS/TOP/CONTROL.', -1.0_Size8)
         else
            call agsetf('AXIS/BOTTOM/TICKS/MAJOR/LENGTH/INWARD.', 0.015_Size8)
            call agsetf('AXIS/TOP/CONTROL.', 4.0_Size8)
         endif
         if (gridya = "on") then
            call agsetf('AXIS/LEFT/TICKS/MAJOR/LENGTH/INWARD.', 1.0_Size8)
            call agsetf('AXIS/RIGHT/CONTROL.', -1.0_Size8)
         else
            call agsetf('AXIS/LEFT/TICKS/MAJOR/LENGTH/INWARD.', 0.015_Size8)
            call agsetf('AXIS/RIGHT/CONTROL.', 4.0_Size8)
         endif
      else
         call agseti('BAC.', 4)
      endif

c ... none of args to agstup effective; all set above
      call agstup(0._Size8, 0, 0, 0, 0, 0._Size8, 0, 0, 0, 0) 
      call agback
      call sflush	/* Flush buffer for complete grid output. */

c ... Do the log       
      call ezcfrlog	/* set up frame count and filename */
      if (suptitle >0._Size8) then
  	  logmsg = "   "//ezctitle(1:utgetcl(ezctitle))
	  call ezcwrtlog(logmsg)
      endif
      if (titletp >0._Size8) then
	  logmsg = "   "//titlet(1:utgetcl(titlet))
	  call ezcwrtlog(logmsg)
      endif
#endif

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* PECRV - curve plotter for EZN on PGS */

void F77_ID(pecrv_, pecrv, PECRV)(FIXNUM *devid, FIXNUM *pipl,
                                  FIXNUM *pdid)
{
    _EP_rend_11(devid, pipl, pdid, CARTESIAN, "Curve");

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* PEMSH - mesh plotter for EZN on PGS */

void F77_ID(pemsh_, pemsh, PEMSH)(FIXNUM *devid, FIXNUM *pipl,
                                  FIXNUM **preg, FIXNUM **pkreg,
                                  FIXNUM **pklmt, FIXNUM **pbndfl,
                                  FIXNUM *pdid)
{
    _EP_rend_21(devid, pipl, preg, pkreg, pklmt, pbndfl,
                PLOT_MESH, "Mesh");

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* PEFPY - fill-poly plotter for EZN on PGS */

void F77_ID(pefpy_, pefpy, PEFPY)(FIXNUM *devid, FIXNUM *pipl, FIXNUM *pityp,
                                  FIXNUM **preg, FIXNUM **pkreg,
                                  FIXNUM **pklmt, FIXNUM **pbndfl)
{

    switch (*pityp)
       {case FillMesh :
	     _EP_rend_21(devid, pipl, preg, pkreg, pklmt, pbndfl,
			 PLOT_FILL_POLY, "Fill Poly");
             break;

        case PolyFunc :
	     _EP_rend_21_ac(devid, pipl, preg, pkreg, pklmt, pbndfl,
			    PLOT_FILL_POLY, "Fill Poly AC");
             break;

        case PolyFill :
	     _EP_draw_poly(devid, pipl, preg, pkreg, pklmt, pbndfl);
             break;};

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* PECNT - contour plotter for EZN on PGS */

void F77_ID(pecnt_, pecnt, PECNT)(FIXNUM *devid, FIXNUM *pipl,
                                  FIXNUM **preg, FIXNUM **pkreg,
                                  FIXNUM **pklmt, FIXNUM **pbndfl)
{
    _EP_rend_21(devid, pipl, preg, pkreg, pklmt, pbndfl,
                PLOT_CONTOUR, "Mesh Contour");

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* PEIMG - image plotter for EZN on PGS */

void F77_ID(peimg_, peimg, PEIMG)(FIXNUM *devid, FIXNUM *pipl,
                                  FIXNUM **preg, FIXNUM **pkreg,
                                  FIXNUM **pklmt, FIXNUM **pbndfl)
{
    int i, m;
    int ipl, inx, iny, inf, n1, n2;
    int maxes[2];
    int *f;
    REAL xmn, xmx, ymn, ymx, fmn, fmx, fc;
    REAL *x, *y;
    PG_device *dev;
    pcons *info;

    dev = SC_GET_POINTER(PG_device, *devid);

    ipl = *pipl - 1;

    inx = grdl.xi[ipl] - 1;
    iny = grdl.yi[ipl] - 1;
    inf = grdl.zi[ipl] - 1;
    n1 = grdl.n1[inf];
    n2 = grdl.n2[inf];

    x = grdl.data[inx];
    y = grdl.data[iny];
    f = grdl.data[inf];

    xmn = x[0];
    xmx = x[1];
    ymn = y[0];
    ymx = y[1];

    maxes[0] = n1;
    maxes[1] = n2;

    m = n1 * n2;
    fmn = HUGE;
    fmx = -HUGE;
    for (i = 0; i < m; i++) {
        fc = f[i];
        fmn = min(fmn, fc);
        fmx = max(fmx, fc);
    }

    info = _EP_set_viewport(dev, NULL);

    dev->supress_setup = 0;

    PG_draw_image_nc_lr(dev, " ", SC_INTEGER_S, f,
                        xmn, xmx, ymn, ymx, fmn, fmx, maxes, info);

    SC_free_alist(info, 3);

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* PEVCT - vector plotter for EZN on PGS */

void F77_ID(pevct_, pevct, PEVCT)(FIXNUM *devid, FIXNUM *pipl,
                                  FIXNUM **preg, FIXNUM **pkreg,
                                  FIXNUM **pklmt, FIXNUM **pbndfl,
                                  REAL * pvs)
{
    REAL vscalea, vhs;
    PG_device *dev;

    dev = SC_GET_POINTER(PG_device, *devid);

    vscalea = 20.0 * (*pvs);
    vhs = 0.008;

/* setup the vector attributes */
    PG_set_vec_attr(dev,
                    VEC_SCALE, vscalea,
                    VEC_FIXHEAD, TRUE, VEC_HEADSIZE, vhs,
/*
		    VEC_MAXISZE, 0.05,
		    VEC_LINETHICK, wid,
		    VEC_COLOR, clr,
*/
                    0);

    _EP_rend_22(devid, pipl, preg, pkreg, pklmt, pbndfl,
                PLOT_VECTOR, "Vector");

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* PERAY - ray plotter for EZN on PGS */

void F77_ID(peray_, peray, PERAY)(devid, x, y, reg,
                                  kpreg, kllmt, bndflag, cthick, ccolor,
                                  imark)
{

#if 0
           cthick = thickc(iplot)
	   numcol = ibasis("ezd.numcol")	/* number of color indices */
           indbase = ibasis("ezd.indbase")	/* base color index */

	   Point(laslist) = data(lasi(iplot))
           Point(nray) = data(nrayi(iplot))
           Point(raylab) = data(raylabi(iplot))
           Point(dispmark) = data(dispmarki(iplot))
           Point(nraypt) = data(nraypti(iplot))
           Point(y) = data(yi(iplot))
           Point(x) = data(xi(iplot))
           Point(pw) = data(pwi(iplot))
           laserct = n1(lasi(iplot))
           rayct = n1(raylabi(iplot))
           rayptct = n1(pwi(iplot))		/* total ray points count */

           if (raypw) then	/* ray color to show power */
	      call osallot(Point(rpwidx), rayptct)
	      call osallot(Point(rpwmax), rayct)
              pltrayct = 0	/* plot ray count */
              rpwsum = 0.	/* sum of max power  */
              bptidx = 1
              do i = 1, rayct
                 ptct = nraypt(i)  /* number of points in the ray */
                 if (ptct > 0) then       /* skip the ray with no points */
                    pltrayct = pltrayct + 1
                    pwmin = pw(bptidx)
                    pwmax = pw(bptidx)

c ... Per JAH (4/2/99): Since ray power never increases as it 
c ... propagates, min and max power calculation is trivial.
                    pwmin = pw(bptidx)
                    pwmax = pw(bptidx+ptct-1)
                    rpwmax(i) = pwmax	/* store the max power of the ray */
                    rpwsum = rpwsum + rpwmax(i)	/* total max power */
                    if (colorc(iplot) .eq. "relpow") then

c ... normalize pw over each ray
                       if (pwmin  .eq.  pwmax) then
			  do k = 1, ptct
                             pw(bptidx+k-1) = 1.
                          enddo
		       else
                          do k = 1, ptct
			     pw(bptidx+k-1)=(pw(bptidx+k-1)-pwmin)/(pwmax-pwmin)
                          enddo
                       endif
                    endif

c ... reset bptidx for the next ray
                    bptidx = bptidx + ptct 
                 endif
              enddo  /* End loop over rays. */

              ccolor = ezcctoi("fgcolor")
              rpwbar = rpwsum/pltrayct	/* average max power */

c ... Changes requested by JAH 4/2/99:  (FNF 11/02/99)
              pwmax = pw(1)          /* max power of ray #1 */
              pwmin = pw(nraypt(1))  /* min power of ray #1 */

c ... Just loop over rays, not every point along every ray.
              kk = nraypt(1)  /* number of points on ray #1 */
              do k = 2, rayptct
                 kk = kk+1
                 if (pwmax < pw(kk)) pwmax = pw(kk)
                 kk = kk + nraypt(i)-1
                 if (pwmin > pw(kk)) pwmin = pw(kk)
              enddo

c ... End of JAH changes.
c ... convert to color idx linearly
              call ezclr4(numcol, rayptct, pw, rpwidx)
           else
	      ccolor = ezcctoi(colorc(iplot))
           endif

           bptidx = 1
           do i = 1, rayct

c ... number of points in the ray
              ptct = nraypt(i)
              if (ptct > 0) then	/* skip the ray with no points */

c ... set raylabel
                 if (ezcraylab  .eq.  "on") then
                    idum = imkstr(raylab(i), raylabel)
                    call ezcdshl(raylabel(1:idum), mod((i-1)*ezclbshft, 4))
                    call ezcdshs
                 endif

c ... mark the beginning point with marker
c ... Turned off per JAH to avoid ATC-GKS bug, 11/07/96 (FNF)
c ... call points(x(bptidx), y(bptidx), 1, -4, 0)
c ... plot the ray
                 if (raypw) then	/* color the ray with power values */
		    if (ezcthickray  .eq.  "on") then	/* thick ray option */
                       cthick = int(rpwmax(i)/rpwbar + 1.5)
                    endif
                    do j = 1, ptct-1
                       ccolor = rpwidx(bptidx+j)+indbase-1
                       call agcurv(x(bptidx+j-1), 1, y(bptidx+j-1), 1, 2, 1)
                    enddo
                 else      
                    call agcurv(x(bptidx), 1, y(bptidx), 1, ptct, 1)
                 endif

c ... add arrows to indicate the ray direction
                 if (raypw) ccolor = ezcctoi("fgcolor")  /* for raypower arrows */
                 call ezcarrow(x(bptidx), y(bptidx), ptct)

c ... reset bptidx for the next ray
                 bptidx = bptidx + ptct 

c ... plot dispmark if assigned (at the end of the ray)
                 if (dispmark(i) .ne. 0) then
                    call points(x(bptidx-1), y(bptidx-1), 1, dispmark(i), 0)
	         endif
              endif
           enddo

           if (raypw) call osfree(Point(rpwidx))
           if (raypw) call osfree(Point(rpwmax))
	      
#endif

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* PETXT - text plotter for EZN on PGS */

void F77_ID(petxt_, petxt, PETXT)(FIXNUM *devid, FIXNUM *pipl)
{
    int ipl, inx, iny, inz, inu;
    int ang, alg, clr;
    char *t;
    REAL xc, yc;
    REAL *x, *y, *u;
    PG_device *dev;

    dev = SC_GET_POINTER(PG_device, *devid);

    ipl = *pipl - 1;

    inx = grdl.xi[ipl] - 1;
    iny = grdl.yi[ipl] - 1;
    inz = grdl.zi[ipl] - 1;
    inu = grdl.vxi[ipl] - 1;

    x = grdl.data[inx];
    y = grdl.data[iny];
    u = grdl.data[inu];
    t = grdl.data[inz];

    ang = u[2];
    alg = u[3];

    _EP_plot_attrs(ipl, &clr, NULL, NULL, NULL, NULL);

    xc = x[0];
    yc = y[0];
    WtoS(dev, xc, yc);

    _EP_wtext(dev, xc, yc, t, 0.0, 0.0, ang, alg, clr);

/*

               Point(tdata) = data(vxi(iplot))
               Point(sdata) = data(zi(iplot))

           n = tdata(1)
           k = tdata(2)
	   ccolor = ezcctoi(colorc(iplot))
	   call gsplci(ccolor)
           if (k>4) then
              k = k*(pquad(2)-pquad(1))*(pquad(4)-pquad(3))
              k = max(k, jminsz)
           endif

           call agpwrt(x, y, sdata(1:n), n, k, int(tdata(3)), int(tdata(4)))

	   ccolor = ezcctoi(defcolor)
	   if (ccolor .ne. -1) then
	      call gsplci(ccolor)
           else
	      call gsplci(1)
	   endif
*/

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* PEFTX - text plotter for EZN on PGS */

void F77_ID(peftx_, peftx, PEFTX)(FIXNUM *devid, FIXNUM *pipl)
{
    F77_ID(petxt_, petxt, PETXT) (devid, pipl);

/*
               Point(tdata) = data(vxi(iplot))
               Point(sdata) = data(zi(iplot))

           n = tdata(1)
           k = tdata(2)
	   ccolor = ezcctoi(colorc(iplot))
	   call gsplci(ccolor)
           if (k>4) then
              k = k*(pquad(2)-pquad(1))*(pquad(4)-pquad(3))
              k = max(k, jminsz)
           endif

	   tmpx = cfux(x(1))
	   tmpy = cfuy(y(1))
	   call gsclip(0)
           call agpwrt(tmpx, tmpy, sdata(1:n), n, k, int(tdata(3)), int(tdata(4)))
	   call gsclip(1)

	   ccolor = ezcctoi(defcolor)
	   if (ccolor .ne. -1) then
	      call gsplci(ccolor)
           else
	      call gsplci(1)
	   endif
*/

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* PELEG - legend for contour plots */

void F77_ID(peleg_, peleg, PELEG)(devid, x, y, tdata, sdata)
{

#if 0

c ... write to the log
         j = min(52, ezctctn)
            do i=1, j, 5
               write(logmsg, 80) (ezctcl(k), ezctctlev(k), k=i, min(i+4, j))
               call ezcwrtlog(logmsg)
            enddo
         if (j < ezctctn) then
            do i=j+1, ezctctn, 5
               write(logmsg, 80) (" ", ezctctlev(k), k=i, min(i+4, ezctctn))
               call ezcwrtlog(logmsg)
            enddo
         endif
         if (conerr) then  /* write error message */
            write(logmsg, '(a)') "**** A 'recoverable' NCAR error has occurred in EZCCONPL."
            call ezcwrtlog(logmsg)
            write(logmsg, '(a)') "     Too many filled contour levels?"
            call ezcwrtlog(logmsg)
         endif

c ... write to the graphic display
         if (conkey > 0 & ezcconkey  .eq.  "on") then
            call set(
                  quad(2)-conkey, quad(2), quad(3), quad(4), 
                  0._Size8, 1._Size8, 0._Size8, 1._Size8, 
                  1)

c ... this makes sure it isnt too big for the type size
            mlines2 = (quad(4)-quad(3))*1024/(hw*jminsz)
            mlines = min(ezctctn, 52, mlines2)
            if (mlines > 0) then

c ... make sure the lines dont spread out too much
               mlines1 = (quad(4)-quad(3))*nlines*0.75_Size8
               j = max(mlines1, mlines)

c ... set polygon vertices for label box
	       clbpx(1) = 0.0_Size8
	       clbpx(2) = 1.0_Size8
	       clbpx(3) = clbpx(2)
	       clbpx(4) = clbpx(1)
               call gsfais(1)	/* set up polygon fill solid */
               if (ezccksfill .ne. "solid") call gsfais(0)
               do i=1, mlines
	          clbpy(1) = 1._Size8-(i*(1._Size8/j))
	          clbpy(2) = clbpy(1)
	          clbpy(3) = clbpy(1)+(1._Size8/j)
	          clbpy(4) = clbpy(3)

c ... determine the color index
	          cfidx = ezctctci(i)
	          if (cfidx < 0) then	/* conflict color assignement */
	             call gsfais(0)	/* use hollow fill to distingush it */
	             cfidx = 1
	             call gsfaci(cfidx) 	
	             call gfa(4, clbpx, clbpy)
	             if (ezccksfill  .eq.  "solid") call gsfais(1)	
						/* set back to solid fill */
                  else
	             if (cfidx > 1) then	/* avoid use fgcolor for fill */
	                call gsfaci(cfidx) 	/* color index */
	                call gfa(4, clbpx, clbpy)
                     endif
	          endif

c ... write the contour label
	          if (ezclabel  .eq.  "off" | ezclabel  .eq.  "on") then
                     write(msg(1), 90) ezctctlev(i)
	          else
                     write(msg(1), 80) ezctcl(i), ezctctlev(i)
	          endif
                  call _EP_wtext(
                      0.0_Size8, 
                      1._Size8-(i-0.5_Size8)*(1._Size8/j), 
                      msg(1)(1:13), 
                      1._Size8/j, 
                      1.0_Size8, 
                      0._Size8, 
                     -1._Size8)
               enddo
	    endif
	 endif
90    format("  ", 1Pe11.3)
80    format(5(a, " ", 1Pe11.3, "  "))

      endif
#endif

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* PEPAL - palette for fill-poly plots */

void F77_ID(pepal_, pepal, PEPAL)(devid, x, y, tdata, sdata)
{

#if 0
         call set(
              quad(2)-conkey, quad(2), quad(3), quad(4), 
              0._Size8, 1._Size8, 0._Size8, 1._Size8, 
              1)

c ... find base index and max index of color map
         indbase = ibasis("ezd.indbase")
         indmax = ibasis("ezd.indmax")
         numcol = ibasis("ezd.numcol")

c ... list 16 colors evenly selected from the color spectrum
         j = min(numcol, 16)
         nn = 40	/* set 40 lines of annotation as the size */

c ... set polygon vertices for label box
         clbpx(1) = 0.0_Size8
         clbpx(2) = 1.0_Size8
         clbpx(3) = clbpx(2)
         clbpx(4) = clbpx(1)
         call gsfais(1)   /* set up polygon fill solid */

c ... Add switch for solid/hollow like for contour levels (FNF 5/22/97)
         if ((fmplot & ezcfmfill .ne. "solid") |
              (raypw  & ezcpwfill .ne. "solid")) call gsfais(0)
         do i=1, j+1
            clbpy(1) = 1._Size8-(i*(1._Size8/nn))
            clbpy(2) = clbpy(1)
            clbpy(3) = clbpy(1)+(1._Size8/nn)
            clbpy(4) = clbpy(3)

c ... determine the color index
            cfidx = (i-1)*((indmax-indbase)/j)+indbase
            call gsfaci(cfidx)      /* color index */
            call gfa(4, clbpx, clbpy)
         enddo
         call gsfais(0)

         call gsplci(1)
         if (fmplot & zmap .eq. 1) zrange = (log(zmax)-log(zmin))/j
         do i=1, j+1 
            if (raypw) then
               write(msg(1), 90) pwmin+(i-1)*((pwmax-pwmin)/j)
            else
               /* write the fillmesh label */
               if (zmap  .eq.  0) then
                  write(msg(1), 90) zmin+(i-1)*((zmax-zmin)/j)
               elseif (zmap  .eq.  1) then
	     	  write(msg(1), 90) zmin * exp((i-1)*zrange)
               else
	          write(msg(1), 90) zbar+ (((i-1)*numcol/j)-numcol/2)*
						zsigma*4/numcol
               endif
            endif
            call _EP_wtext(
                        0._Size8, 
                        1._Size8-(i-0.5_Size8)*(1._Size8/nn), 
                        msg(1)(1:13), 
                        1._Size8/nn, 
                        1._Size8, 
                        0._Size8, 
                       -1._Size8)
         enddo
      endif
#endif

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* _EP_TITLES - plot titles for EZN on PGS */

static void _EP_titles(PG_device * dev,
                       REAL titlelf, char *titlel,
                       REAL titlert, char *titler,
                       REAL titlebt, char *titleb,
                       REAL titletp, char *titlet,
                       REAL suptitle, char *ezctitle, int ezcsuper)
{
    REAL ang;

    ang = 90.0 * DEG_RAD;

/* left */
    if (titlel[0] != ' ') {
        _EP_fix_string(titlel, LSIZE);
        _EP_wtext(dev, _vxmn - 0.15, 0.50, titlel, 0.90, 1.00, ang, 0.0,
                  dev->WHITE);
    }

/* right */
    if (titler[0] != ' ') {
        _EP_fix_string(titler, LSIZE);
        _EP_wtext(dev, _vxmx + 0.05, 0.50, titler, 0.90, 1.00, ang, 0.0,
                  dev->WHITE);
    }

/* bottom */
    if (titleb[0] != ' ') {
        _EP_fix_string(titleb, LSIZE);
        _EP_wtext(dev, 0.5, _vymn - 0.06, titleb, 1.00, 0.880, 0.0, 0.0,
                  dev->WHITE);
    }

/* top */
    if (titlet[0] != ' ') {
        _EP_fix_string(titlet, LSIZE);
        _EP_wtext(dev, 0.5, _vymx + 0.03, titlet, 1.00, 0.880, 0.0, 0.0,
                  dev->WHITE);
    }

/* write out ezctitle */
    suptitle = 0.8;             /* GOTCHA: remove this when suptitle is correct */
    if (suptitle > 0.0) {
        _EP_fix_string(ezctitle, LSIZE);

        if (ezcsuper)
            ezcwclab(dev, 0.98, ezctitle);
        else
            ezcwclab(dev, 0.02, ezctitle);
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* PEANN - annotations for frame to finish the picture
 *       - NOTE: the definition of legendc and preleg is:
 *       -      legendc(maxnc)          _character*(LSIZE)  
 *       -      preleg(maxnc)           _character*16
 */

void F77_ID(peann_, peann, PEANN)(FIXNUM *devid, FIXNUM *pli, REAL * pxd,
                                  FIXNUM *pnf,
                                  REAL * ptlf, F77_string ptl,
                                  REAL * ptrt, F77_string ptr,
                                  REAL * ptbt, F77_string ptb,
                                  REAL * pttp, F77_string ptt,
                                  REAL * ptsp, F77_string pte,
                                  FIXNUM *pezcsup)
{
    int i, iplot, nc, sz;
    int ezcsuper, leginfoa;
    int mlines, lastc, lastcc, ileg, jleg;
    char *titlel, *titler, *titleb, *titlet, *ezctitle;
    char msg[NLINES][LSIZE];
    char *legendc, *preleg, *s, *t;
    REAL titlelf, titlert, titlebt, titletp, suptitle;
    PG_device *dev;

    dev = SC_GET_POINTER(PG_device, *devid);

    PG_set_axis_log_scale(dev, FALSE, FALSE);

    titlelf = *ptlf;
    titlert = *ptrt;
    titlebt = *ptbt;
    titletp = *pttp;
    suptitle = *ptsp;
    ezcsuper = *pezcsup;

    titlel = SC_F77_C_STRING(ptl);
    titler = SC_F77_C_STRING(ptr);
    titleb = SC_F77_C_STRING(ptb);
    titlet = SC_F77_C_STRING(ptt);
    ezctitle = SC_F77_C_STRING(pte);

    leginfoa = *pli;

    nc = grdl.nc;
    legendc = grdl.legendc;
    preleg = grdl.preleg;

    sz = dev->type_size;
    PG_set_font(dev, "courier", "medium", 14);

/* titles */
    _EP_titles(dev,
               titlelf, titlel, titlert, titler,
               titlebt, titleb, titletp, titlet,
               suptitle, ezctitle, ezcsuper);

#if 0

    c...Do the contour levels annotation if (conplot)
        then call peleg(devid, Point(x), Point(y), Point(tdata),
                        Point(sdata))
            endif c...Do the fillmesh or ray power annotation clrann =.
            false.if (fmplot)
            then c if (zmap.ne.3)
                then
/* input is a array of physics quantities */
                    c...write to the graphic display
                    if (conkey.gt.0.and.ezcfmkey.eq."on")
                    clrann =.true.c endif endif if (raypw)
                        then
/* ray power plot invoked */
                            if (conkey.gt.0.and.ezcpwkey.eq."on")
                            clrann =.true.
                                endif
                                c...control of color annotation if (clrann)
                                then call pepal(devid, Point(x), Point(y),
                                                Point(tdata), Point(sdata))
                                    endif
#endif
/* do the legend */
                                    if (leginfoa != 0)

/* decide how many lines max for the legend */
                                {
                                    mlines = 10;
                                    lastc = -2;
                                    ileg = 0;
                                    for (iplot = 0; iplot < nc; iplot++) {
                                        s = legendc + iplot * LSIZE;
                                        t = preleg + iplot * 16;

                                        _EP_fix_string(s, LSIZE);
                                        _EP_fix_string(t, LSIZE);

                                        if (*s != '\0') {
                                            jleg = (ileg % mlines);
                                            ileg++;
                                            if (jleg == 0) {
                                                lastcc = lastc + 3;
                                                if (lastcc > LSIZE - 10)
                                                    break;
                                            }

                                            if (*t != '\0')
                                                sprintf(msg[jleg],
                                                        "%s: %s", t, s);
                                            else
                                                strcpy(msg[jleg], s);

                                            lastc =
                                                max(strlen(msg[jleg]),
                                                    lastc);
                                        }
                                    }

                                    ileg = min(mlines, ileg);
                                    if (ileg > 0) {
                                        REAL xmn, xmx, ymn, ymx;
                                        REAL xoffset, yoffset, dy;

                                        PG_get_frame(dev, &xmn, &xmx, &ymn,
                                                     &ymx);
                                        xoffset = xmn + 0.05;

                                        yoffset = _vymn - 0.1;
                                        dy = 0.1 * yoffset;

                                        for (i = 0; i < ileg; i++) {
                                            PG_write_NDC(dev, xoffset,
                                                         yoffset, msg[i]);
                                            yoffset -= dy;
                                        }
                                    }
                                }

    PG_set_font(dev, "helvetica", "medium", sz);

    PG_finish_plot(dev);

    dev->supress_setup = FALSE;

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* PECLR - clear device for EZN on PGS */

void F77_ID(peclr_, peclr, PECLR)(FIXNUM *devid)
{
    PG_device *dev;

    dev = SC_GET_POINTER(PG_device, *devid);

    if (dev != NULL) {
        if (!HARDCOPY_DEVICE(dev))
            PG_clear_screen(dev);
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* PEEND - end EZN on PGS */

void F77_ID(peend_, peend, PEEND)(void)
{
    PG_rl_all();

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
