/*
 * access.c
 */

#include "rt.h"

#define bassubl(i,j) bassubs[SSWIDTH *(j - 1) + i - 1]

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Parxerr(char *msg, int i)
{
    char errmsg[80];

    sprintf(errmsg, "...Error in computation... %s.\n", msg);
    BasError(errmsg, strlen(errmsg));
    sprintf(errmsg,
            "The error occurred in the %d component of the expression.\n",
            i);
    BasError(errmsg, strlen(errmsg));
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Integer binary operation: pu and pv point to integer arrays,
 * n gives the length of each, and op is one of the operations.
 * store * pu op * pv into * ru one element at a time.
 * n1, n2, and m are used for matrix multiply and divide.
 */

Integer Binopi(Address ru, Address pu, Address pv, Integer n, Integer op,
               Integer n1, Integer n2, Integer m)
{
    static Integer nine = 9;
    int i, j, k, jj, jm, li, ri;
    Integer *t, *s1, *s2;

    t = (Integer *) ru;
    s1 = (Integer *) pu;
    s2 = (Integer *) pv;
    switch (op) {
    case ADD:
        for (i = 0; i < n; i++)
            t[i] = s1[i] + s2[i];
        return (OK);
    case SUBTRACT:
        for (i = 0; i < n; i++)
            t[i] = s1[i] - s2[i];
        return (OK);
    case MULTIPLY:
        for (i = 0; i < n; i++)
            t[i] = s1[i] * s2[i];
        return (OK);
    case DIVIDE:
        for (i = 0; i < n; i++)
            if (s2[i] == 0)
                t[i] = I1mach(&nine);
            else
                t[i] = s1[i] / s2[i];
        return (OK);
    case POWER:
        for (i = 0; i < n; i++)
            if (s1[i] == 0 && s2[i] == 0) {
                Parxerr("0 ** 0", i);
                return ERR;
            } else
                t[i] = (Integer) pow((double) s1[i], (double) s2[i]);
        return (OK);
    case DOTPRODUCT:
        *t = 0;
        for (i = 0; i < n; i++)
            *t += s1[i] * s2[i];
        return (OK);
    case MATRIXMULTIPLY:
        /* See below for explanation of indexing... */
        for (j = 0, jm = 0, k = 0; j < n2; j++, jm += m)
            for (i = 0; i < n1; i++, k++) {
                t[k] = 0;
                for (jj = 0, li = i, ri = jm; jj < m; jj++, li += n1, ri++)
                    t[k] += s1[li] * s2[ri];
            }
        return (OK);
    }
    return (OK);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Indexing for matrix multiply:
 * (Recall that FORTRAN indexing is in column major order)
 * t is a 2D FORTRAN array whose dimensions are n1 by n2.
 * The offset of the i, j element of t is k = i + j * n1.
 * Because the inner loop is on i, k actually runs consecutively
 * from 0 through n1 * n2 - 1.
 * s1 is a 2D FORTRAN ARRAY whose dimensions are n1 by m.
 * The offset of the i, jj element of s1 is li = i + jj * n1.
 * Thus, in the inner loop on jj, li starts out at i and
 * increments by n1.
 * s2 is a 2D FORTRAN ARRAY whose dimensions are m by n2.
 * The offset of the jj, j element of s2 is ri = jj + j * m.
 * We use jm to represent j * m. Because the innermost loop is
 * on jj, ri runs consecutively starting with jm.
 */

Integer Binopr(Address ru, Address pu, Address pv, Integer n, Integer op,
               Integer n1, Integer n2, Integer m)
{
    /* Real binary operation: pu and pv point to real arrays,
     * n gives the length of each, and op is one of the operations.
     * store * pu op * pv into * ru one element at a time.
     * n1, n2, and m are used for matrix multiply and divide.
     */
    int i, j, k, jj, jm, li, ri;
    float *t, *s1, *s2;
    Address ipiv;
    Integer info;
    static char msg1[] = "sgesv failed, matrix is singular.";

    t = (float *) ru;
    s1 = (float *) pu;
    s2 = (float *) pv;
    switch (op) {
    case ADD:
        for (i = 0; i < n; i++)
            t[i] = s1[i] + s2[i];
        return (OK);
    case SUBTRACT:
        for (i = 0; i < n; i++)
            t[i] = s1[i] - s2[i];
        return (OK);
    case MULTIPLY:
        for (i = 0; i < n; i++)
            t[i] = s1[i] * s2[i];
        return (OK);
    case DIVIDE:
        for (i = 0; i < n; i++)
            if (s2[i] == 0)
                t[i] = (float) infinity;
            else
                t[i] = s1[i] / s2[i];
        return (OK);
    case POWER:
        for (i = 0; i < n; i++)
            if (s1[i] == 0. && s2[i] == 0.) {
                Parxerr("0. ** 0.", i);
                return ERR;
            } else
                t[i] = (float) pow((double) s1[i], (double) s2[i]);
        return (OK);
    case DOTPRODUCT:
        *t = 0;
        for (i = 0; i < n; i++)
            *t += s1[i] * s2[i];
        return (OK);
    case MATRIXMULTIPLY:
        /* See above for explanation of indexing... */
        for (j = 0, jm = 0, k = 0; j < n2; j++, jm += m)
            for (i = 0; i < n1; i++, k++) {
                t[k] = 0;
                for (jj = 0, li = i, ri = jm; jj < m; jj++, li += n1, ri++)
                    t[k] += s1[li] * s2[ri];
            }
        return (OK);
    case MATRIXDIVIDE:
        Osallotn(&ipiv, n1, "BINOPR:ipiv");
        for (i = 0; i < n1 * n2; i++)
            t[i] = s1[i];
        Sgesv(&n1, &n2, s2, &n1, (Integer *) ipiv, t, &n1, &info);
        Osfreen(&ipiv, "ipiv in MATRIXDIVIDE");
        if (info != 0) {
            Remark(msg1, strlen(msg1));
            return (ERR);
        }
        return (OK);
    }
    return (OK);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Double binary operation: pu and pv point to double arrays,
 * n gives the length of each, and op is one of the operations.
 * store * pu op * pv into * ru one element at a time.
 * n1, n2, and m are used for matrix multiply and divide.
 */

Integer Binopd(Address ru, Address pu, Address pv, Integer n, Integer op,
               Integer n1, Integer n2, Integer m)
{
    int i, j, k, jj, jm, li, ri;
    real8 *t, *s1, *s2;
    Address ipiv;
    Integer info;
    static char msg1[] = "sgesv failed, matrix is singular.";

    t = (real8 *) ru;
    s1 = (real8 *) pu;
    s2 = (real8 *) pv;
    switch (op) {
    case ADD:
        for (i = 0; i < n; i++)
            t[i] = s1[i] + s2[i];
        return (OK);
    case SUBTRACT:
        for (i = 0; i < n; i++)
            t[i] = s1[i] - s2[i];
        return (OK);
    case MULTIPLY:
        for (i = 0; i < n; i++)
            t[i] = s1[i] * s2[i];
        return (OK);
    case DIVIDE:
        for (i = 0; i < n; i++)
            if (s2[i] == 0.)
                t[i] = (real8) infinity;
            else
                t[i] = s1[i] / s2[i];
        return (OK);
    case POWER:
        for (i = 0; i < n; i++)
            if (s1[i] == 0. && s2[i] == 0.) {
                Parxerr("0. ** 0.", i);
                return ERR;
            } else
                t[i] = (real8) pow((double) s1[i], (double) s2[i]);
        return (OK);
    case DOTPRODUCT:
        *t = 0;
        for (i = 0; i < n; i++)
            *t += s1[i] * s2[i];
        return (OK);
    case MATRIXMULTIPLY:
        /* See above for explanation of indexing... */
        for (j = 0, jm = 0, k = 0; j < n2; j++, jm += m)
            for (i = 0; i < n1; i++, k++) {
                t[k] = 0;
                for (jj = 0, li = i, ri = jm; jj < m; jj++, li += n1, ri++)
                    t[k] += s1[li] * s2[ri];
            }
        return (OK);
    case MATRIXDIVIDE:
        Osallotn(&ipiv, n1, "BINOPD:ipiv");
        for (i = 0; i < n1 * n2; i++)
            t[i] = s1[i];
        Sgesv8(&n1, &n2, s2, &n1, (Integer *) ipiv, t, &n1, &info);
        Osfreen(&ipiv, "ipiv in MATRIXDIVIDE");
        if (info != 0) {
            Remark(msg1, strlen(msg1));
            return (ERR);
        }
        return (OK);
    }
    return (OK);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Complex binary operation: pu and pv point to
 * complex arrays, n gives the length of each, and op is one of
 * the operations. store * pu op * pv into * ru one element at
 * a time. Assume a complex is two consecutive floats.
 * n1, n2, and m are used for matrix multiply and divide.
 */
Integer Binopc(Address ru, Address pu, Address pv, Integer n, Integer op,
               Integer n1, Integer n2, Integer m)
{
    int i, j, k, jj, jm, li, ri;
    float *t, *s1, *s2;
    float r2, s2r, s2i;

    t = (float *) ru;
    s1 = (float *) pu;
    s2 = (float *) pv;
    switch (op) {
    case ADD:
        for (i = 0; i < 2 * n; i++)
            t[i] = s1[i] + s2[i];
        return (OK);
    case SUBTRACT:
        for (i = 0; i < 2 * n; i++)
            t[i] = s1[i] - s2[i];
        return (OK);
    case MULTIPLY:
        for (i = 0; i < n; i++) {
            t[2 * i] =
                s1[2 * i] * s2[2 * i] - s1[2 * i + 1] * s2[2 * i + 1];
            t[2 * i + 1] =
                s1[2 * i] * s2[2 * i + 1] + s1[2 * i + 1] * s2[2 * i];
        }
        return (OK);
    case DIVIDE:
        for (i = 0; i < n; i++) {
            r2 = max(fabs(s2[2 * i]), fabs(s2[2 * i + 1]));
            if (r2 == 0.) {     /* Zero denominator */
                t[2 * i] = (float) infinity;
                t[2 * i + 1] = 0.;
            } else {
                s2r = s2[2 * i] / r2;
                s2i = s2[2 * i + 1] / r2;
                r2 = s2[2 * i] * s2r + s2[2 * i + 1] * s2i;
                t[2 * i] = (s1[2 * i] * s2r + s1[2 * i + 1] * s2i) / r2;
                t[2 * i + 1] =
                    (s1[2 * i + 1] * s2r - s1[2 * i] * s2i) / r2;
            }
        }
        return (OK);
    case POWER:
        return (Cpow(t, s1, s2, &n));
    case DOTPRODUCT:
        t[0] = 0;
        t[1] = 0;
        for (i = 0; i < n; i++) {
            t[0] += s1[2 * i] * s2[2 * i] - s1[2 * i + 1] * s2[2 * i + 1];
            t[1] += s1[2 * i] * s2[2 * i + 1] + s1[2 * i + 1] * s2[2 * i];
        }
        return (OK);
    case MATRIXMULTIPLY:
        /* See above for explanation of indexing... */
        for (j = 0, jm = 0, k = 0; j < n2; j++, jm += m)
            for (i = 0; i < n1; i++, k++) {
                t[2 * k] = 0;
                t[2 * k + 1] = 0;
                for (jj = 0, li = i, ri = jm; jj < m; jj++, li += n1, ri++) {
                    t[2 * k] += s1[2 * li] * s2[2 * ri] -
                        s1[2 * li + 1] * s2[2 * ri + 1];
                    t[2 * k + 1] += s1[2 * li] * s2[2 * ri + 1] +
                        s1[2 * li + 1] * s2[2 * ri];
                }
            }
        return (OK);
    }
    return (OK);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

#ifdef HAS_DOUBLE_COMPLEX

/* Double complex binary operation: pu and pv point to
 * double complex arrays, n gives the length of each, and op is one of
 * the operations. store * pu op * pv into * ru one element at
 * a time. Assume a double complex is two consecutive floats.
 * n1, n2, and m are used for matrix multiply and divide.
 */

Integer Binopdc(Address ru, Address pu, Address pv, Integer n, Integer op,
                Integer n1, Integer n2, Integer m)
{
    int i, j, k, jj, jm, li, ri;
    real8 *t, *s1, *s2;
    real8 r2, s2r, s2i;

    t = (real8 *) ru;
    s1 = (real8 *) pu;
    s2 = (real8 *) pv;
    switch (op) {
    case ADD:
        for (i = 0; i < 2 * n; i++)
            t[i] = s1[i] + s2[i];
        return (OK);
    case SUBTRACT:
        for (i = 0; i < 2 * n; i++)
            t[i] = s1[i] - s2[i];
        return (OK);
    case MULTIPLY:
        for (i = 0; i < n; i++) {
            t[2 * i] =
                s1[2 * i] * s2[2 * i] - s1[2 * i + 1] * s2[2 * i + 1];
            t[2 * i + 1] =
                s1[2 * i] * s2[2 * i + 1] + s1[2 * i + 1] * s2[2 * i];
        }
        return (OK);
    case DIVIDE:
        for (i = 0; i < n; i++) {
            r2 = max(fabs(s2[2 * i]), fabs(s2[2 * i + 1]));
            if (r2 == 0.) {     /* Zero denominator */
                t[2 * i] = (real8) infinity;
                t[2 * i + 1] = 0.;
            } else {
                s2r = s2[2 * i] / r2;
                s2i = s2[2 * i + 1] / r2;
                r2 = s2[2 * i] * s2r + s2[2 * i + 1] * s2i;
                t[2 * i] = (s1[2 * i] * s2r + s1[2 * i + 1] * s2i) / r2;
                t[2 * i + 1] =
                    (s1[2 * i + 1] * s2r - s1[2 * i] * s2i) / r2;
            }
        }
        return (OK);
    case POWER:
        return (Dcpow(t, s1, s2, &n));
    case DOTPRODUCT:
        t[0] = 0;
        t[1] = 0;
        for (i = 0; i < n; i++) {
            t[0] += s1[2 * i] * s2[2 * i] - s1[2 * i + 1] * s2[2 * i + 1];
            t[1] += s1[2 * i] * s2[2 * i + 1] + s1[2 * i + 1] * s2[2 * i];
        }
        return (OK);
    case MATRIXMULTIPLY:
        /* See above for explanation of indexing... */
        for (j = 0, jm = 0, k = 0; j < n2; j++, jm += m)
            for (i = 0; i < n1; i++, k++) {
                t[2 * k] = 0;
                t[2 * k + 1] = 0;
                for (jj = 0, li = i, ri = jm; jj < m; jj++, li += n1, ri++) {
                    t[2 * k] += s1[2 * li] * s2[2 * ri] -
                        s1[2 * li + 1] * s2[2 * ri + 1];
                    t[2 * k + 1] += s1[2 * li] * s2[2 * ri + 1] +
                        s1[2 * li + 1] * s2[2 * ri];
                }
            }
        return (OK);
    }
    return (OK);
}

#endif

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

Integer Binoprg(Address ru, Address pu, Address pv, Integer n, Integer op)
{
    Integer *t, *s1, *s2;
    int i;

    t = (Integer *) ru;
    s1 = (Integer *) pu;
    s2 = (Integer *) pv;
    if (op == ADD) {
        for (i = 0; i < n; i++) {
            t[3 * i] = s1[3 * i] + s2[3 * i];
            t[3 * i + 1] = s1[3 * i + 1] + s2[3 * i + 1];
            if (s1[3 * i + 2] == 1)
                t[3 * i + 2] = s2[3 * i + 2];
            else if (s2[3 * i + 2] == 1)
                t[3 * i + 2] = s1[3 * i + 2];
            else if (s1[3 * i + 2] != s2[3 * i + 2]) {
                Parxerr("mismatched RANGE strides", i);
                return (ERR);
            }
        }
    } else if (op == SUBTRACT) {
        for (i = 0; i < n; i++) {
            t[3 * i] = s1[3 * i] - s2[3 * i];
            t[3 * i + 1] = s1[3 * i + 1] - s2[3 * i + 1];
            if (s1[3 * i + 2] == 1)
                t[3 * i + 2] = s2[3 * i + 2];
            else if (s2[3 * i + 2] == 1)
                t[3 * i + 2] = s1[3 * i + 2];
            else if (s1[3 * i + 2] != s2[3 * i + 2]) {
                Parxerr("mismatched RANGE strides", i);
                return (ERR);
            }
        }
        for (i = 0; i < n; i++) {
        }
    }
    return OK;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

Integer Binopl(Address ru, Address pu, Address pv, Integer n, Integer op,
               Integer type)
{
    logical *t, *s1, *s2;
    Address *a1, *a2;
    Integer *i1, *i2;
    float *f1, *f2;
    double *d1, *d2;
    int i;
    static char
        msg1[] = "pardobin: impossible binary operator for type Address.",
        msg2[] = "pardobin: impossible binary operator.",
        msg3[] = "parxbin: weird operator got by me.";

    t = (logical *) ru;
    switch (op) {
    case OR:
        s1 = (logical *) pu;
        s2 = (logical *) pv;
        for (i = 0; i < n; i++)
            t[i] = s1[i] || s2[i];
        return (OK);
    case AND:
        s1 = (logical *) pu;
        s2 = (logical *) pv;
        for (i = 0; i < n; i++)
            t[i] = s1[i] && s2[i];
        return (OK);
    default:
        switch (type) {
        case LOGICAL:
            Remark(msg2, strlen(msg2));
            return (ERR);
        case FWA:
            a1 = (Address *) pu;
            a2 = (Address *) pv;
            if (op == EQU)
                for (i = 0; i < n; i++)
                    t[i] = a1[i] == a2[i];
            else if (op == NE)
                for (i = 0; i < n; i++)
                    t[i] = a1[i] != a2[i];
            else {
                Remark(msg1, strlen(msg1));
                return (ERR);
            }
            return (OK);
        case INTEGER:
            i1 = (Integer *) pu;
            i2 = (Integer *) pv;
            switch (op) {
            case LT:
                for (i = 0; i < n; i++)
                    t[i] = i1[i] < i2[i];
                return (OK);
            case LE:
                for (i = 0; i < n; i++)
                    t[i] = i1[i] <= i2[i];
                return (OK);
            case GT:
                for (i = 0; i < n; i++)
                    t[i] = i1[i] > i2[i];
                return (OK);
            case GE:
                for (i = 0; i < n; i++)
                    t[i] = i1[i] >= i2[i];
                return (OK);
            case EQU:
                for (i = 0; i < n; i++)
                    t[i] = i1[i] == i2[i];
                return (OK);
            case NE:
                for (i = 0; i < n; i++)
                    t[i] = i1[i] != i2[i];
                return (OK);
            default:
                Remark(msg2, strlen(msg2));
                return (ERR);
            }
        case REAL:
            f1 = (float *) pu;
            f2 = (float *) pv;
            switch (op) {
            case LT:
                for (i = 0; i < n; i++)
                    t[i] = f1[i] < f2[i];
                return (OK);
            case LE:
                for (i = 0; i < n; i++)
                    t[i] = f1[i] <= f2[i];
                return (OK);
            case GT:
                for (i = 0; i < n; i++)
                    t[i] = f1[i] > f2[i];
                return (OK);
            case GE:
                for (i = 0; i < n; i++)
                    t[i] = f1[i] >= f2[i];
                return (OK);
            case EQU:
                for (i = 0; i < n; i++)
                    t[i] = f1[i] == f2[i];
                return (OK);
            case NE:
                for (i = 0; i < n; i++)
                    t[i] = f1[i] != f2[i];
                return (OK);
            default:
                Remark(msg2, strlen(msg2));
                return (ERR);
            }
        case DOUBLE:
            d1 = (double *) pu;
            d2 = (double *) pv;
            switch (op) {
            case LT:
                for (i = 0; i < n; i++)
                    t[i] = d1[i] < d2[i];
                return (OK);
            case LE:
                for (i = 0; i < n; i++)
                    t[i] = d1[i] <= d2[i];
                return (OK);
            case GT:
                for (i = 0; i < n; i++)
                    t[i] = d1[i] > d2[i];
                return (OK);
            case GE:
                for (i = 0; i < n; i++)
                    t[i] = d1[i] >= d2[i];
                return (OK);
            case EQU:
                for (i = 0; i < n; i++)
                    t[i] = d1[i] == d2[i];
                return (OK);
            case NE:
                for (i = 0; i < n; i++)
                    t[i] = d1[i] != d2[i];
                return (OK);
            default:
                Remark(msg2, strlen(msg2));
                return (ERR);
            }
        case COMPLEX:
#   ifndef HAS_DOUBLE_COMPLEX
        case DOUBLE_C:
#   endif
            f1 = (float *) pu;
            f2 = (float *) pv;
            switch (op) {
            case EQU:
                for (i = 0; i < n; i++) {
                    t[i] = f1[2 * i] == f2[2 * i] &&
                        f1[2 * i + 1] == f2[2 * i + 1];
                }
                return (OK);
            case NE:
                for (i = 0; i < n; i++) {
                    t[i] = f1[2 * i] != f2[2 * i] ||
                        f1[2 * i + 1] != f2[2 * i + 1];
                }
                return (OK);
            default:
                Remark(msg3, strlen(msg3));
                return (ERR);
            }
#   ifdef HAS_DOUBLE_COMPLEX
        case DOUBLE_C:
            d1 = (double *) pu;
            d2 = (double *) pv;
            switch (op) {
            case EQU:
                for (i = 0; i < n; i++) {
                    t[i] = d1[2 * i] == d2[2 * i] &&
                        d1[2 * i + 1] == d2[2 * i + 1];
                }
                return (OK);
            case NE:
                for (i = 0; i < n; i++) {
                    t[i] = d1[2 * i] != d2[2 * i] ||
                        d1[2 * i + 1] != d2[2 * i + 1];
                }
                return (OK);
            default:
                Remark(msg3, strlen(msg3));
                return (ERR);
            }
#   endif
        case RANGE:
            i1 = (Integer *) pu;
            i2 = (Integer *) pv;
            switch (op) {
            case EQU:
                for (i = 0; i < n; i++) {
                    t[i] = i1[3 * i] == i2[3 * i] &&
                        i1[3 * i + 1] == i2[3 * i + 1] &&
                        i1[3 * i + 2] == i2[3 * i + 2];
                }
                return (OK);
            case NE:
                for (i = 0; i < n; i++) {
                    t[i] = i1[3 * i] != i2[3 * i] ||
                        i1[3 * i + 1] != i2[3 * i + 1] ||
                        i1[3 * i + 2] != i2[3 * i + 2];
                }
                return (OK);
            default:
                Remark(msg3, strlen(msg3));
                return (ERR);
            }
        }
    }
    return (OK);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

Integer Binoplch(Address lt, Address u, Address v, Integer lenl,
                 Integer lens1, Integer lens2, Integer op)
{
    /* Comparison of two character arrays */
    logical *t;
    char *s1, *s2;
    integer s1base, s2base;
    int i;
    char msg[80];

    t = (logical *) lt;
    s1 = (char *) u;
    s2 = (char *) v;

    switch (op) {
    case EQU:
        for (i = 0, s1base = 0, s2base = 0; i < lenl;
             i++, s1base += lens1, s2base += lens2) {
            t[i] = (logical) StrEQ(s1, s1base, lens1, s2, s2base, lens2);
        }
        return (OK);
    case NE:
        for (i = 0, s1base = 0, s2base = 0; i < lenl;
             i++, s1base += lens1, s2base += lens2) {
            t[i] =
                (logical) (!StrEQ(s1, s1base, lens1, s2, s2base, lens2));
        }
        return (OK);
    default:
        sprintf(msg, "parxchop: improper operation for characters: %ld",
                (long) op);
        Remark(msg, strlen(msg));
        return (ERR);
    }
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Perform the unary NOT on a logical variable.
 */

Integer Unopnot(Address lt, Integer n)
{
    logical *t;
    int i;

    t = (logical *) lt;
    for (i = 0; i < n; i++)
        t[i] = !t[i];
    return (OK);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Increment the integer variable at at with FORTRAN offset offs.
 */

void Incr(Address at, Integer offs)
{
    Integer *it;

    it = (Integer *) at;
    it[offs - 1]++;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Perform the unary negation of an arithmetic variable.
 */

Integer Unopneg(Address at, Integer n, Integer type)
{
    Integer *ti;
    float *tr;
    double *td;
    int i;
    static char
        msg[] = "pardoneg: illegal type of operand for unary minus.";

    switch (type) {
    case INTEGER:
        ti = (Integer *) at;
        for (i = 0; i < n; i++)
            ti[i] = -ti[i];
        return (OK);
    case REAL:
        tr = (float *) at;
        for (i = 0; i < n; i++)
            tr[i] = -tr[i];
        return (OK);
    case DOUBLE:
        td = (double *) at;
        for (i = 0; i < n; i++)
            td[i] = -td[i];
        return (OK);
    case COMPLEX:
# ifndef HAS_DOUBLE_COMPLEX
    case DOUBLE_C:
# endif
        tr = (float *) at;
        for (i = 0; i < n; i++) {
            tr[2 * i] = -tr[2 * i];
            tr[2 * i + 1] = -tr[2 * i + 1];
        }
        return (OK);
# ifdef HAS_DOUBLE_COMPLEX
    case DOUBLE_C:
        td = (double *) at;
        for (i = 0; i < n; i++) {
            td[2 * i] = -td[2 * i];
            td[2 * i + 1] = -td[2 * i + 1];
        }
        return (OK);
# endif
    default:
        Remark(msg, strlen(msg));
        return (ERR);
    }
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

Integer Maxop(Address ru, Address pu, Integer n, Integer inc, Integer type)
{
    Integer *ti, *si;
    float *tr, *sr;
    double *td, *sd;
    int i, j;
    static char msg[] = "max: type code bad, something wrong";

    switch (type) {
    case INTEGER:
        ti = (Integer *) ru;
        si = (Integer *) pu;
        for (i = 0, j = 0; i < n; i++, j += inc)
            ti[i] = ti[i] > si[j] ? ti[i] : si[j];
        return (OK);
    case REAL:
        tr = (float *) ru;
        sr = (float *) pu;
        for (i = 0, j = 0; i < n; i++, j += inc)
            tr[i] = tr[i] > sr[j] ? tr[i] : sr[j];
        return (OK);
    case DOUBLE:
        td = (double *) ru;
        sd = (double *) pu;
        for (i = n - 1, j = inc * (n - 1); i >= 0; i--, j -= inc)
            td[i] = td[i] > sd[j] ? td[i] : sd[j];
        return (OK);
    default:
        Baderr(msg, strlen(msg));
    }
    return (OK);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

Integer Minop(Address ru, Address pu, Integer n, Integer inc, Integer type)
{
    Integer *ti, *si;
    float *tr, *sr;
    double *td, *sd;
    int i, j;
    static char msg[] = "min: type code bad, something wrong";

    switch (type) {
    case INTEGER:
        ti = (Integer *) ru;
        si = (Integer *) pu;
        for (i = 0, j = 0; i < n; i++, j += inc)
            ti[i] = ti[i] < si[j] ? ti[i] : si[j];
        return (OK);
    case REAL:
        tr = (float *) ru;
        sr = (float *) pu;
        for (i = 0, j = 0; i < n; i++, j += inc)
            tr[i] = tr[i] < sr[j] ? tr[i] : sr[j];
        return (OK);
    case DOUBLE:
        td = (double *) ru;
        sd = (double *) pu;
        for (i = n - 1, j = inc * (n - 1); i >= 0; i--, j -= inc)
            td[i] = td[i] < sd[j] ? td[i] : sd[j];
        return (OK);
    default:
        Baderr(msg, strlen(msg));
    }
    return (OK);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Rangei(Address ru, Integer ilow, Integer ihigh, Integer iinc)
{
    Integer *t;

    t = (integer *) ru;
    t[0] = ilow;
    t[1] = ihigh;
    t[2] = iinc;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Ranger(Address ru, Integer itot, float rlow, float rhi, float rdelta)
{
    float *t;
    int i;

    t = (float *) ru;
    t[0] = rlow;
    for (i = 1; i < itot - 1; i++)
        t[i] = t[i - 1] + rdelta;
    t[itot - 1] = rhi;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Ranged(Address ru, Integer itot, double rlow, double rhi,
            double rdelta)
{
    double *t;
    int i;

    t = (double *) ru;
    t[0] = rlow;
    for (i = 1; i < itot - 1; i++)
        t[i] = t[i - 1] + rdelta;
    t[itot - 1] = rhi;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Concatenate two byte strings s1 and s2(of lengths len1 and len2)
 * and put the result into t, which is presumed to be of len1 + len2.
 * Since these are FORTRAN character strings, or numerical values, do not
 * put a null at the end.
 */
void Bincat(Address t, Address s1, Address s2, Integer len1, Integer len2)
{
    bcopy(s1, t, len1);
    bcopy(s2, (Address) ((char *) t + len1), len2);
}

/*--------------------------------------------------------------------------*/

/*                          DATA MOVERS                                     */

/*--------------------------------------------------------------------------*/

/* MVALUE - Move n items from s(offset o) to t */

void Mvalue(Address s, integer o, integer ityp, Address *t, integer n)
{
    integer ne;
    int bpi, bo, nb;
    char *ps;

    if (ityp < 0)
        bpi = -ityp;
    else {
        ne = 1;
        bpi = Utsizeof(&ne, &ityp) * BYTES_PER_WORD;
    }

    bo = bpi * (o - 1);
    nb = bpi * n;
    ps = (char *) s + bo;

    memcpy(t, ps, nb);

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MVALUER - Move n items from s(offset o) to t
 *         - these elements are across a column of matrix s
 */

void Mvaluer(Address s, integer o, integer ityp, Address t,
             integer n, Integer row)
{
    integer ne;
    int i, bpi, bo, offs;
    char *ps, *pt;

    if (ityp < 0)
        bpi = -ityp;
    else {
        ne = 1;
        bpi = Utsizeof(&ne, &ityp) * BYTES_PER_WORD;
    }

    for (i = 0, offs = o - 1; i < n; i++, offs += row) {
        bo = bpi * offs;
        ps = (char *) s + bo;
        pt = (char *) t + i * bpi;

        memcpy(pt, ps, bpi);
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MCOMPARE - compare n items from s and t */

integer Mcompare(Address s, Address t, integer ityp, integer n)
{
    integer ne, ok;
    int bpi, nb;

    ne = 1;
    bpi = Utsizeof(&ne, &ityp) * BYTES_PER_WORD;
    nb = bpi * n;

    ok = memcmp(t, s, nb);

    return (ok);
}

/*--------------------------------------------------------------------------*/

/* NOTE: these appear only to have been used by the old printing routines
 *       and are not longer needed for that purpose
 *       in any case MVALUE and MVALUER can be used in their place
 *       if nobody needs them by 01/2001 then remove them altogether
 */

#if 0

/*--------------------------------------------------------------------------*/

void Rngvalue(Address prng, Integer *rng)
{
    /* prng points to a range. Assign its components to rng. */
    memcpy((Address) rng, prng, 3 * sizeof(Integer));
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void I2value(Address pi, Integer *i)
{
    /* pi points to something two words long. Move it to i. */
    memcpy((Address) i, pi, 2 * sizeof(Integer));
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void I4value(Address pi, Integer *i)
{
    /* pi points to something four words long. Move it to i. */
    memcpy((Address) i, pi, 4 * sizeof(Integer));
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Invalue(Address s, integer o, Integer *t, integer n)
{
    /* Move n integer values from Address(offset o) to t */
    memcpy((Address) t,
           (Address) ((Integer *) s + o - 1), n * sizeof(Integer));
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Lnvalue(Address s, integer o, logical *t, integer n)
{
    /* Move n logical values from Address(offset o) to t */
    memcpy((Address) t,
           (Address) ((logical *) s + o - 1), n * sizeof(logical));
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Rnvalue(Address s, integer o, float *t, integer n)
{
    /* Move n float values from Address(offset o) to t */
    memcpy((Address) t,
           (Address) ((float *) s + o - 1), n * sizeof(float));
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Dnvalue(Address s, integer o, real8 *t, integer n)
{
    /* Move n real8 values from s(offset o) to t */
    memcpy((Address) t,
           (Address) ((real8 *) s + o - 1), n * sizeof(real8));
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Cnvalue(Address s, integer o, float *t, integer n)
{
    /* Move n complex values from s(offset o) to t */
    memcpy((Address) t,
           (Address) ((float *) s + 2 * (o - 1)), 2 * n * sizeof(float));
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Dcnvalue(Address s, integer o, real8 *t, integer n)
{
    /* Move n double complex values from s(offset o) to t */
    memcpy((Address) t,
           (Address) ((real8 *) s + 2 * (o - 1)), 2 * n * sizeof(real8));
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Invaluer(Address s, integer o, integer *t, integer n, Integer row)
{
    /* Move n integer values from s(offset o) to t   */
    /* These elements are across a column of matrix s */
    int i, offs;

    for (i = 0, offs = o - 1; i < n; i++, offs += row) {
        t[i] = *((integer *) s + offs);
    }
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void I3valuer(Address s, integer o, integer *t, integer n, Integer row)
{
    /* Move n range values from s(offset o) to t       */
    /* These elements are across a column of matrix s   */
    int i, offs;

    for (i = 0, offs = 3 * (o - 1); i < n; i++, offs += 3 * row) {
        t[3 * i] = *((integer *) s + offs);
        t[3 * i + 1] = *((integer *) s + offs + 1);
        t[3 * i + 2] = *((integer *) s + offs + 2);
    }
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Lnvaluer(Address s, integer o, logical *t, integer n, Integer row)
{
    /* Move n logical values from Address(offset o) to t */
    /* These elements are across a column of matrix s     */
    int i, offs;

    for (i = 0, offs = o - 1; i < n; i++, offs += row) {
        t[i] = *((logical *) s + offs);
    }
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Rnvaluer(Address s, integer o, float *t, integer n, Integer row)
{
    /* Move n float values from Address(offset o) to t */
    /* These elements are across a column of matrix s   */
    int i, offs;

    for (i = 0, offs = o - 1; i < n; i++, offs += row) {
        t[i] = *((float *) s + offs);
    }
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Dnvaluer(Address s, integer o, real8 *t, integer n, Integer row)
{
    /* Move n real8 values from s(offset o) to t       */
    /* These elements are across a column of matrix s   */
    int i, offs;

    for (i = 0, offs = o - 1; i < n; i++, offs += row) {
        t[i] = *((real8 *) s + offs);
    }
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Cnvaluer(Address s, integer o, float *t, integer n, Integer row)
{
    /* Move n complex values from s(offset o) to t   */
    /* These elements are across a column of matrix s */
    int i, offs;

    for (i = 0, offs = 2 * (o - 1); i < n; i++, offs += 2 * row) {
        t[2 * i] = *((float *) s + offs);
        t[2 * i + 1] = *((float *) s + offs + 1);
    }
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Move n double complex values from s(offset o) to t
 * These elements are across a column of matrix s
 */

void Dcnvaluer(Address s, integer o, real8 *t, integer n, Integer row)
{
    int i, offs;

    for (i = 0, offs = 2 * (o - 1); i < n; i++, offs += 2 * row) {
        t[2 * i] = *((real8 *) s + offs);
        t[2 * i + 1] = *((real8 *) s + offs + 1);
    }
}

/*--------------------------------------------------------------------------*/

#endif

/*--------------------------------------------------------------------------*/

/* ACCESS an Address value whose address is known.
 */
Address Avalue(Address t)
{
    return (*((Address *) t));
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* t points to an array of Addresses, extract the oth one.
 *(like FORTRAN, o starts with 1).
 */
Address Avalueo(Address t, Integer o)
{
    return (((Address *) t)[o - 1]);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* t points to an array of Addresses, assign aval to the oth one.
 */

void Savalueo(Address t, Integer o, Address aval)
{
    ((Address *) t)[o - 1] = aval;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* ACCESS a double precision value whose address is known.
 */

double Dvalue(Address t)
{
    return (*((double *) t));
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* t points to an array of doubles. extract the oth One.
 * (like FORTRAN, o starts with 1).
 */

double Dvalueo(Address t, Integer o)
{
    return ((double *) t)[o - 1];
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* ACCESS an integer value whose address is known.
 */

integer Ivalue(Address t)
{
    return (*((integer *) t));
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* t points to an array of integers. extract the oth One.
 * (like FORTRAN, o starts with 1).
 */

integer Ivalueo(Address t, Integer o)
{
    return (((Integer *) t)[o - 1]);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* t points to an array of integers. assign ival to the oth one.
 */

void Sivalueo(Address t, Integer o, Integer ival)
{
    ((Integer *) t)[o - 1] = ival;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* ACCESS a logical value whose address is known.
 */
logical Lvalue(Address t)
{
    return (*((logical *) t));
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* ACCESS a logical value whose address is known.
 */

logical Lvalueo(Address t, Integer o)
{
    return (((logical *) t)[o - 1]);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* t points at an array of logicals. set the oth one to lval.
 */

void Slvalueo(Address t, Integer o, logical lval)
{
    ((logical *) t)[o - 1] = lval;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* set n values starting at t, offset o, to value s.
 */

void Lvaluea(logical s, Address t, Integer o, Integer n)
{
    int i;

    for (i = 0; i < n; i++) {
        ((logical *) t + o)[i] = s;
    }
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* ACCESS a real value whose address is known.
 */

float Rvalue(Address t)
{
    return (*((float *) t));
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* t points to an array of reals. extract the oth One.
 * (like FORTRAN, o starts with 1).
 */

float Rvalueo(Address t, Integer o)
{
    return (((float *) t)[o - 1]);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* t points to an array of real8's. extract the oth One.
 */

real8 R8valueo(Address t, Integer o)
{
    return (((real8 *) t)[o - 1]);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RETURN a complex value whose address is known.
 */

void Cvalue(Address s, float *t)
{
    t[0] = ((float *) s)[0];
    t[1] = ((float *) s)[1];
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* t points to an array of complexes. extract the oth One.
 */

void Cvalueo(Address s, Integer o, float *t)
{
    t[0] = ((float *) s)[2 * (o - 1)];
    t[1] = ((float *) s)[2 * (o - 1) + 1];
}

/*--------------------------------------------------------------------------*/

#ifdef HAS_DOUBLE_COMPLEX

/*--------------------------------------------------------------------------*/

/* RETURN a double complex value whose address is known.
 */

void Dcvalue(Address s, real8 *t)
{
    t[0] = ((real8 *) s)[0];
    t[1] = ((real8 *) s)[1];
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* t points to an array of double complexes. extract the oth One.
 */

void Dcvalueo(Address s, Integer o, real8 *t)
{
    t[0] = ((real8 *) s)[2 * (o - 1)];
    t[1] = ((real8 *) s)[2 * (o - 1) + 1];
}

/*--------------------------------------------------------------------------*/

#endif

/*--------------------------------------------------------------------------*/

/* Broadcast the item src into the array given by tar which is
 * tarsize items long; the type code is tc.
 */

void Brdcast(Address src, Address tar, Integer tarsize, Integer tc)
{
    int i;
    Integer itemsize;

    itemsize = Utbytesin(tc);
    for (i = 0; i < itemsize * tarsize; i += itemsize)
        bcopy(src, (Address) ((char *) tar + i), itemsize);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Assign nwrds words from src to tar, offset by toffs words.
 */

void Assgn(Address src, Address tar, Integer toffs, Integer nwrds)
{
    bcopy(src, (void *) ((char *) tar + toffs * sizeof(Integer)),
          nwrds * sizeof(Integer));
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Assign nwrds words from src offset by soffs words, to tar,
 * offset by toffs words.
 */

void Assgn2(Address src, Address tar, Integer soffs, Integer toffs,
            Integer nwrds)
{
    bcopy((void *) ((char *) src + soffs * sizeof(Integer)),
          (void *) ((char *) tar + toffs * sizeof(Integer)),
          nwrds * sizeof(Integer));
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Store the transpose of n X m matrix src into m X n matrix tar,
 * both being of type tc.
 */

void Transp(Address src, Address tar, Integer n, Integer m, Integer tc)
{
    int i, j, nbytes, lencol;
    char *s, *s1, *t;

    nbytes = Utbytesin(tc);
    lencol = n * nbytes;
    for (i = 0, t = (char *) tar, s1 = (char *) src; i < n;
         i++, s1 += nbytes)
        for (j = 0, s = s1; j < m; j++, t += nbytes, s += lencol)
            bcopy((void *) s, (void *) t, nbytes);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* CCOERCE - coerce numerical types: nitems of type t1 in src to
 *         - nitems of type t2 in tar
 */

Integer Ccoerce(Address src, Address tar, Integer nitems, Integer t1,
                Integer t2)
{
    Integer *isrc, *itar;
    Address *atar;
    float *rsrc, *rtar;
    real8 *dsrc, *dtar;
    logical *lsrc, *ltar;
    int i;

/* switch on type of source */
    switch (t1) {
    case INTEGER:
        isrc = (Integer *) src;

/* switch on type of target */
        switch (t2) {
        case FWA:
            atar = (Address *) tar;
            for (i = 0; i < nitems; i++)
                atar[i] = (Address) isrc[i];
            return (OK);
        case LOGICAL:
            ltar = (logical *) tar;
            for (i = 0; i < nitems; i++)
                ltar[i] = (logical) (isrc[i] != 0);
            return (OK);
        case REAL:
            rtar = (float *) tar;
            for (i = 0; i < nitems; i++)
                rtar[i] = (float) isrc[i];
            return (OK);
        case DOUBLE:
            dtar = (real8 *) tar;
            for (i = 0; i < nitems; i++)
                dtar[i] = (real8) isrc[i];
            return (OK);
        case COMPLEX:
            rtar = (float *) tar;
            for (i = 0; i < nitems; i++) {
                rtar[2 * i] = (float) isrc[i];
                rtar[2 * i + 1] = 0.0;
            }
            return (OK);
        case DOUBLE_C:
            dtar = (real8 *) tar;
            for (i = 0; i < nitems; i++) {
                dtar[2 * i] = (real8) isrc[i];
                dtar[2 * i + 1] = 0.0;
            }
            return (OK);
        case RANGE:
            itar = (Integer *) tar;
            for (i = 0; i < nitems; i++) {
                itar[3 * i] = isrc[i];
                itar[3 * i + 1] = isrc[i];
                itar[3 * i + 2] = 1;
            }
            return (OK);
        default:
            return (ERR);
        }

    case REAL:
        rsrc = (float *) src;

/* switch on type of target */
        switch (t2) {
        case INTEGER:
            itar = (Integer *) tar;
            for (i = 0; i < nitems; i++)
                itar[i] = (Integer) (rsrc[i]);
            return (OK);
        case DOUBLE:
            dtar = (real8 *) tar;
            for (i = 0; i < nitems; i++)
                dtar[i] = (real8) (rsrc[i]);
            return (OK);
        case COMPLEX:
            rtar = (float *) tar;
            for (i = 0; i < nitems; i++) {
                rtar[2 * i] = rsrc[i];
                rtar[2 * i + 1] = 0.0;
            }
            return (OK);
        case DOUBLE_C:
            dtar = (real8 *) tar;
            for (i = 0; i < nitems; i++) {
                dtar[2 * i] = (real8) rsrc[i];
                dtar[2 * i + 1] = 0.0;
            }
            return (OK);
        case RANGE:
            itar = (Integer *) tar;
            for (i = 0; i < nitems; i++) {
                itar[3 * i] = (Integer) rsrc[i];
                itar[3 * i + 1] = (Integer) rsrc[i];
                itar[3 * i + 2] = 1;
            }
            return (OK);
        default:
            return (ERR);
        }

    case DOUBLE:
        dsrc = (real8 *) src;

/* switch on type of target */
        switch (t2) {
        case INTEGER:
            itar = (Integer *) tar;
            for (i = 0; i < nitems; i++)
                itar[i] = (Integer) dsrc[i];
            return (OK);
        case REAL:
            rtar = (float *) tar;
            for (i = 0; i < nitems; i++)
                rtar[i] = (float) dsrc[i];
            return (OK);
        case COMPLEX:
            rtar = (float *) tar;
            for (i = 0; i < nitems; i++) {
                rtar[2 * i] = (float) dsrc[i];
                rtar[2 * i + 1] = 0.0;
            }
            return (OK);
        case DOUBLE_C:
            dtar = (real8 *) tar;
            for (i = 0; i < nitems; i++) {
                dtar[2 * i] = dsrc[i];
                dtar[2 * i + 1] = 0.0;
            }
            return (OK);
        case RANGE:
            itar = (Integer *) tar;
            for (i = 0; i < nitems; i++) {
                itar[3 * i] = (Integer) dsrc[i];
                itar[3 * i + 1] = (Integer) dsrc[i];
                itar[3 * i + 2] = 1;
            }
            return (OK);
        default:
            return (ERR);
        }

    case COMPLEX:
        rsrc = (float *) src;

/* switch on type of target */
        switch (t2) {
        case DOUBLE_C:
            dtar = (real8 *) tar;
            for (i = 0; i < nitems; i++) {
                dtar[2 * i] = (real8) rsrc[2 * i];
                dtar[2 * i + 1] = (real8) rsrc[2 * i + 1];
            }
            return (OK);
        case DOUBLE:
            dtar = (real8 *) tar;
            for (i = 0; i < nitems; i++)
                dtar[i] = (real8) rsrc[2 * i];
            return (OK);
        case REAL:
            rtar = (float *) tar;
            for (i = 0; i < nitems; i++)
                rtar[i] = rsrc[2 * i];
            return (OK);
        case INTEGER:
            itar = (Integer *) tar;
            for (i = 0; i < nitems; i++)
                itar[i] = (Integer) rsrc[2 * i];
            return (OK);
        default:
            return (ERR);
        }

    case DOUBLE_C:
        dsrc = (real8 *) src;

/* switch on type of target */
        switch (t2) {
        case COMPLEX:
            rtar = (float *) tar;
            for (i = 0; i < nitems; i++) {
                rtar[2 * i] = (float) dsrc[2 * i];
                rtar[2 * i + 1] = (float) dsrc[2 * i + 1];
            }
            return (OK);
        case DOUBLE:
            dtar = (real8 *) tar;
            for (i = 0; i < nitems; i++)
                dtar[i] = dsrc[2 * i];
            return (OK);
        case REAL:
            rtar = (float *) tar;
            for (i = 0; i < nitems; i++)
                rtar[i] = (float) dsrc[2 * i];
            return (OK);
        case INTEGER:
            itar = (Integer *) tar;
            for (i = 0; i < nitems; i++)
                itar[i] = (Integer) dsrc[2 * i];
            return (OK);
        default:
            return (ERR);
        }

    case LOGICAL:
        lsrc = (logical *) src;
        if (t2 == INTEGER) {
            itar = (Integer *) tar;
            for (i = 0; i < nitems; i++)
                itar[i] = (Integer) lsrc[i];
            return (OK);
        } else
            return (ERR);
    default:
        return (ERR);
    }
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* COBJCPY - make a copy of the data associated with stack element SRC
 *         - for the stack element DST
 */

void Cobjcpy(Integer *dst, Integer *src)
{
    Integer n, nb;
    void *dp, *sp;

    Parget(dst);

    n = Parlen(dst);
    nb = Utsizeof(&n, &src[TC - 1]) * CHAR_PER_WORD;

    dp = Paraddr(dst);
    sp = Paraddr(src);

    memcpy(dp, sp, nb);

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Iset(Integer *tar, Integer src)
{
    *tar = src;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Rset(float *tar, float src)
{
    *tar = src;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Dset(real8 *tar, real8 src)
{
    *tar = src;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Lset(logical *tar, logical src)
{
    *tar = src;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Cset(float *tar, float *src)
{
    tar[0] = src[0];
    tar[1] = src[1];
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void C8set(real8 *tar, real8 *src)
{
    tar[0] = src[0];
    tar[1] = src[1];
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

#ifdef HAS_DOUBLE_COMPLEX
void Dcset(real8 *tar, real8 *src)
{
    tar[0] = src[0];
    tar[1] = src[1];
}
#else
void Dcset(float *tar, float *src)
{
    tar[0] = src[0];
    tar[1] = src[1];
}
#endif

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Iseto(Integer *tar, Integer i, Integer src)
{
    tar[i - 1] = src;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Rseto(float *tar, Integer i, float src)
{
    tar[i - 1] = src;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Dseto(real8 *tar, Integer i, real8 src)
{
    tar[i - 1] = src;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/*
 * add the object described by s to structure described by sstru,
 * including a structure piece code
 *(FNF 7/31/97) Changed "call remark" statements to "call baserror".
 *  struadd adds the parser stack element described by s to the structure
 *  sstru
 */
Integer Struadd(Integer *sstru, Integer *s, Integer strcod)
{
    static char
        msg1[] = "struadd: tried to add null component",
        msg2[] = "struadd: first argument not a structure";
    Integer *ips, *ipso, *sc, sx[SSWIDTH];
    int n, ns, nold;
    Integer len;

    if (s[TC - 1] == 0) {
        BasError(msg1, strlen(msg1));
        return (ERR);
    }
    ips = (Integer *) Paraddr(sstru);
    sc = (Integer *) Paraddr(s);
    if (ips[0] != STRUCTURE) {
        BasError(msg2, strlen(msg2));
        return (ERR);
    }
    nold = ips[1];
    if (s[TC - 1] != STRUCTURE) {
        len = Parlen(s);
        ns = Utsizeof(&len, &(s[TC - 1]));
        n = nold + 18 + ns;
    } else {
        ns = sc[1];
        n = nold + ns;
    }
    Parcopy(sstru, sx);
    ipso = ips;
    sstru[N1 - 1] = n;
    Parget(sstru);
    ips = (Integer *) Paraddr(sstru);
    memcpy((void *) ips, (void *) ipso, nold * sizeof(Integer));
    Parkill(sx);
    ips[1] = n;
    ips[2] += 1;
    if (s[TC - 1] != STRUCTURE) {
        ips[nold] = strcod;
        ips[nold + 1] = ns + 18;
        ips[nold + 2] = s[TC - 1];
        ips[nold + 3] = s[N - 1];
        memcpy((void *) (&(ips[nold + 4])), (void *) (&(s[N1 - 1])),
               14 * sizeof(Integer));
        memcpy((void *) (&(ips[nold + 18])), (void *) sc,
               ns * sizeof(Integer));
    } else {
        memcpy((void *) (&(ips[nold])), (void *) sc, ns * sizeof(Integer));
    }
    return (OK);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* STRUCTG - extracts the ipiece-th piece from the structure bstruct(note that
 *         - bstruct is the first word address of the structure's content, not
 *         - the address of its stack element)
 */

Integer Structg(Address bstruct, Integer ipiece, Integer *sx,
                logical addrflag)
{
    static char msg[] = "Improper subscript on structure";
    int npieces, locpiece, ipc, sxn, numwords;
    Integer *istr, *lstruct, len;

/* check that this piece exists */
    lstruct = (Integer *) bstruct;
    npieces = lstruct[2];
    if (npieces < ipiece) {
        BasError(msg, strlen(msg));
        return (ERR);
    }

/* locate fwa of piece */
    for (locpiece = 4, ipc = 1; ipc < ipiece; ipc++)
        locpiece += lstruct[locpiece];

/* act on whether it's a structure or not
 * differences also occur depending on whether value is stored by copy
 * or address
 */
    if (addrflag) {
        sx[PTR - 1] =
            locpiece + (lstruct[locpiece - 1] == STRUCTURE ? 0 : 18);
        return (OK);
    }

    if (lstruct[locpiece - 1] == STRUCTURE) {
        sx[N - 1] = 1;
        sx[TC - 1] = STRUCTURE;
        sx[N1 - 1] = lstruct[locpiece];
        sx[M1 - 1] = 1;
        Parget(sx);
        istr = (Integer *) Paraddr(sx);
        memcpy(istr, lstruct + locpiece - 1, sx[N1 - 1] * sizeof(Integer));
        return OK;
    } else {
        sx[N - 1] = lstruct[locpiece + 2];
        sx[TC - 1] = lstruct[locpiece + 1];
        sxn = sx[N - 1];

        memcpy(sx + N1 - 1, lstruct + locpiece + 3, sxn * sizeof(Integer));
        memcpy(sx + M1 - 1, lstruct + locpiece + 10,
               sxn * sizeof(Integer));

        Parget(sx);

        istr = (Integer *) Paraddr(sx);
        len = Parlen(sx);
        numwords = Utsizeof(&len, &(sx[TC - 1]));
        memcpy(istr, lstruct + locpiece + 17, numwords * sizeof(Integer));
    }
    return OK;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/*
 *                              PARSUBL
 * Programmed by : Zane C. Motteler
 * Date         : August 1989
 * Purpose      : Handle the reduction <lhs> ::= <lhs> <args> when <lhs> is
 *                a variable, possibly having been previously subscripted.
 *                The intention is to allow multiply-subscripted variables
 *                to appear on the left side of an assignment statement, as
 *                well as to support the postponement of the(par)fetching
 *                of variables as long as possible.
 * Details      : <lhs> hasn't been looked up, so none of the sets of subscripts
 *                except for the very last can be checked for legality for
 *                whatever object it is. However, one set of subscripts after
 *                another set of subscripts can be checked for conformity with
 *                the earlier one.
 *                Here's what subscripted subscripts mean:
 *                If a:b:c is subscripted by d:e:f, then g:h:i results, where
 *                 1) d must be in the range a:b:c(meaning it must be of the
 *                      form a+n*c and must be between a and b) and is taken to be
 *                      the new first range element g;
 *                 2)(e - d)/f + 1 is taken to be the number of elements
 *                      wanted(the second subscript may not be reachable,
 *                      e. g.,(4:13:2));
 *                 3) the new stride i is taken to be f*c, so the new
 *                      second range element(last subscript h) will be
 *                      d +(e - d)/f *(f*c).
 *                 this treatment is consistent with the way multiple sub-
 *                 scripting with positive strides worked in the past, but
 *                 is completely general in that it allows arbitrary strides.
 *                 detailed error checking for consistency within the above
 *                 is carried out and informative comments given if an error
 *                 is detected. Only the final set of subscripts is checked by
 *                 'parfetch'. Thus for instance if x is declared
 *                      integer x(100)
 *                 then referring to
 *                      x(-100:100:2)(6:30:2)
 *                 will give a correct subsequence of x, namely(6:54:4),
 *                 because the final set of subscripts falls within the correct
 *                 range, even though the first does not.
 *               (ZCM 8/26/1999)
 *                Added the capability of multiple subscripting where one
 *                or more of the subscripts is an arbitrary 1d array.
 *                 How it works is similar to what was described above.
 *                 The minimum and maximum of the first subscript are known.
 *                (The N1 element is the minimum, and the M1 element is the
 *                 maximum, if it is a RANGE; the N1 element is DEFAULT + 1
 *                 if it was an array, and the N1 element indexes into an
 *                 array of stack elements bassubl. This array's minimum
 *                 and maximum must be found.) Then, whether the next subscript
 *                 is an integer scalar or array, or a RANGE, each of its
 *                 values must be between the minimum and maximum of the
 *                 first subscript. A RANGE consisting of a single value
 *                 results if the second subscript is a scalar, and otherwise
 *                 a new 1d array of subscripts results.
 * Future work  : It ought to be possible to generalize the single subscripting
 *                with a 1d array fairly simply to multiple subscripting. This
 *                would be tedious but not hard, and probably should not be
 *                undertaken unless there is demand from users.
 * Parameters   :
 *                 ns -- the number of subscripts specified by <args>
 * Revised: October 1989(ZCM)
 *      improved error printouts when debug is on
 * Revised: January 1990(ZCM)
 *      to support lazy evaluation, parfetch must be called for subscripts
 *      which have type LHS.
 * Revised: July 1990(ZCM)
 *      add the 'if (symbolic)' block so that subscripts of displayed variables
 *      will be printed.
 *(FNF 7/29/97) Changed "write(STDERR,..." statements to "call baserror".
 *(ZCM 8/10/1999) Eliminated all arguments but ns; use global runtime stack
 * and link pointers to access the stack; use the new stack frame structure.
 *(ZCM 8/26/1999) If all subscripts in a subscript after the first are
 *      defaulted, then it is treated as a NOP.
 * Reprogrammed: February 2000(ZCM) redone in C.
 */

void Parsubl(Integer ns)
{
    static Integer cpy = COPY;
    static Integer Inn = INTEGER;
    static logical false = (logical) 0;
    static char
        msg1[] =
        "parsubl: you have too many variables subscripted by arrays!",
        msg2[] = "parsubl: trouble with variable named '", msg3[] =
        "has more subscripts in a later specification than earlier.",
        msg4[] = "The variable in question is currently UNKNOWN.";
    char errmsg[121];
    Integer nns, ival, ilow, ihi, istr, nelts, nlast, i, j, k, lcomma,
        /* Declarations for the 1d case: */
     
        
        
        isub,
        isublow,
        isubhi,
        isubcnt,
        si[SSWIDTH],
        imax, imin, istep, *irange, *isubarr1, *isubarr2, *isubarr3;

    if (symbolic) {
        if (ns >= 1) {
            nxp[rlinkp - 2][lxp[rlinkp - 2]] = '(';
            lxp[rlinkp - 2] += 1;
            lcomma = 0;
            for (j = rlinkp; j < rlinkp + ns; j++) {
                if (lcomma == 0)
                    lcomma = 1;
                else {
                    nxp[rlinkp - 2][lxp[rlinkp - 2]] = ',';
                    lxp[rlinkp - 2] += 1;
                }
                strncpy(&nxp[rlinkp - 2][lxp[rlinkp - 2]], nxp[j], lxp[j]);
                lxp[rlinkp - 2] += lxp[j];
            }
            nxp[rlinkp - 2][lxp[rlinkp - 2]] = ')';
            lxp[rlinkp - 2] += 1;
        }
    }

/*
 *      Handle the case of a stack element which has been subscripted with one
 *      subscript and is about to be subscripted again with one subscript
 *      separately here:
 */
    if (ssl(NS, rlinkp - 1) == 1 && ns == 1) {
        /*
         * A singly subscripted array is singly subscripted so we
         * have lots of cases to worry about. If we drop out of this
         * loop, any error will be handled by older code.
         */
        if (ssl(N1, rlinkp - 1) == DEFAULT + 1) {
            /* The earlier set of subscripts was a 1d array; get its info */
            isub = ssl(M1, rlinkp - 1); /* indexes into array of stack elements */
            isublow = bassubl(M1, isub);
            isubcnt = bassubl(N1, isub);
            isubhi = isublow + isubcnt - 1;
            isubarr1 = (Integer *) Paraddr(&(bassubl(1, isub)));
            /* Fetch the new set of subscripts */
            Parfetch(&(ssl(1, rlinkp + 1)), nxp[rlinkp], &lxp[rlinkp],
                     &cpy, &false, COMMENTSIZE);
            if (ssl(N, rlinkp + 1) == 0 && ssl(TC, rlinkp + 1) != RANGE &&
                ssl(TC, rlinkp + 1) != NULL_TYPE &&
                Parint(&(ssl(1, rlinkp + 1)), &ival) != ERR) {
                /* Second subscript is a scalar */
                if (ival >= isublow && ival <= isubhi) {
                    /* It's in range */
                    ival = isubarr1[ival - isublow];
                    bassubl(N1 + 1, isub) -= 1;
                    if (bassubl(N1 + 1, isub) == 0) {
                        /* reference count reached 0, so release */
                        Parrel(&(bassubl(1, isub)));
                        bassubl(TC, isub) = NULL_TYPE;
                    }
                    ssl(N1, rlinkp - 1) = ival;
                    ssl(M1, rlinkp - 1) = ival;
                    ssl(I1, rlinkp - 1) = 1;
                    return;
                } else {
                    /* It's out of range */
                    sprintf(errmsg,
                            "parsubl: last subscript %ld  is not between %ld and %ld.",
                            (long) ival, (long) isublow, (long) isubhi);
                    BasError(errmsg, strlen(errmsg));
                    Kaboom(0);
                }
            } else if (ssl(TC, rlinkp + 1) == RANGE) {
                /* Second subscript is a RANGE */
                irange = (Integer *) Paraddr(&(ssl(1, rlinkp + 1)));
                if (irange[0] >= isublow && irange[0] <= isubhi &&
                    irange[1] >= isublow && irange[1] <= isubhi) {
                    /* The range is OK to get elements out of the subscript array */
                    nelts = (irange[1] - irange[0]) / irange[2] + 1;
                    /* Build a new stack element si for the new subscript array
                     * Note that when you take a subrange of something, the lower
                     * subscript is always the beginning value of the range.
                     */
                    si[N - 1] = 1;
                    si[TC - 1] = INTEGER;
                    si[N1 - 1] = nelts;
                    si[M1 - 1] = irange[0];
                    si[I1 - 1] = 1;
                    Parget(si);
                    isubarr3 = (Integer *) Paraddr(si);
                    for (j = 0, i = irange[0]; i <= irange[1];
                         i += irange[2], j++)
                        isubarr3[j] = isubarr1[i - isublow];
                    bassubl(N1 + 1, isub) -= 1;
                    if (bassubl(N1 + 1, isub) == 0) {
                        /* reference count reached 0, so release */
                        Parrel(&(bassubl(N1 + 1, isub)));
                        bassubl(TC, isub) = NULL_TYPE;
                    }
                    /* Now we need to build a new entry for bassubl */
                    for (k = 1; k <= SUBSLENGTH; k++) {
                        if (bassubl(TC, k) == NULL_TYPE) {
                            /* found! save subscript stack element */
                            Parcopy(si, &(bassubl(1, k)));
                            /* put signal in operand stack element */
                            ssl(N1, rlinkp - 1) = DEFAULT + 1;
                            ssl(M1, rlinkp - 1) = k;
                            ssl(I1, rlinkp - 1) = 1;    /* dummy */
                            break;
                        }
                        if (k == SUBSLENGTH) {
                            BasError(msg1, strlen(msg1));
                            Kaboom(0);
                        }
                    }
                    return;
                } else {
                    sprintf(errmsg,
                            "parsubl: last subscript range %ld:%ld is not between %ld and %ld.",
                            (long) irange[0], (long) irange[1],
                            (long) isublow, (long) isubhi);
                    BasError(errmsg, strlen(errmsg));
                    Kaboom(0);
                }
            } else if (ssl(TC, rlinkp + 1) == INTEGER) {
                /* Second subscript is also an integer array */
                isubarr2 = (Integer *) Paraddr(&(ssl(1, rlinkp + 1)));
                nelts = ssl(N1, rlinkp + 1);
                imax = isubarr2[0];
                imin = isubarr2[0];
                for (i = 1; i < nelts; i++) {
                    if (imax < isubarr2[i])
                        imax = isubarr2[i];
                    if (imin > isubarr2[i])
                        imin = isubarr2[i];
                }
                if (imin >= isublow && imin <= isubhi &&
                    imax >= isublow && imax <= isubhi) {
                    /* All subscripts are in range--create new array */
                    si[N1 - 1] = 1;
                    si[TC - 1] = INTEGER;
                    si[N1 - 1] = nelts;
                    si[M1 - 1] = imin;
                    si[I1 - 1] = 1;
                    Parget(si);
                    isubarr3 = (Integer *) Paraddr(si);
                    for (i = 1; i <= nelts; i++)
                        isubarr3[i - 1] =
                            isubarr1[isubarr2[i - 1] - isublow];
                    bassubl(N1 + 1, isub) -= 1;
                    if (bassubl(N1 + 1, isub) == 0) {
                        /* reference count reached 0, so release */
                        Parrel(&(bassubl(1, isub)));
                        bassubl(TC, isub) = NULL_TYPE;
                    }
                    /* Now we need to build a new entry for bassubl */
                    for (k = 1; k <= SUBSLENGTH; k++) {
                        if (bassubl(TC, k) == NULL_TYPE) {
                            /* found! save subscript stack element */
                            Parcopy(si, &(bassubl(1, k)));
                            bassubl(N1 + 1, k) = 1;     /* can be parrel'ed */
                            /* put signal in operand stack element */
                            ssl(N1, rlinkp - 1) = DEFAULT + 1;
                            ssl(M1, rlinkp - 1) = k;
                            ssl(I1, rlinkp - 1) = 1;    /* dummy */
                            break;
                        }
                        if (k == SUBSLENGTH) {
                            BasError(msg1, strlen(msg1));
                            Kaboom(0);
                        }
                    }
                    return;
                } else {
                    /* Some are out of range */
                    sprintf(errmsg,
                            "parsubl: array with min and max subscripts %ld and %ld",
                            (long) imin, (long) imax);
                    BasError(errmsg, strlen(errmsg));
                    sprintf(errmsg,
                            "         can't index array with min and max %ld and %ld.",
                            (long) isublow, (long) isubhi);
                    BasError(errmsg, strlen(errmsg));
                    Kaboom(0);
                }
            }
            /* If it's something else, we'll just drop out and catch it later. */
        } else {
            /* The earlier set of subscripts must be a RANGE and the 
             * second set is an integer array. All other cases can be
             * handled later. */
            ilow = ssl(N1, rlinkp - 1);
            ihi = ssl(M1, rlinkp - 1);
            if (ilow == DEFAULT && ihi == DEFAULT) {
                /* With both subscripts defaulted, use the array untouched.
                 * Now we need to build a new entry for bassubl */
                for (k = 1; k <= SUBSLENGTH; k++) {
                    if (bassubl(TC, k) == NULL_TYPE) {
                        /* found! save subscript stack element */
                        Parcopy(&(ssl(1, rlinkp + 1)), &(bassubl(1, k)));
                        bassubl(N1 + 1, k) = 1; /* can be parrel'ed */
                        /* put signal in operand stack element */
                        ssl(N1, rlinkp - 1) = DEFAULT + 1;
                        ssl(M1, rlinkp - 1) = k;
                        ssl(I1, rlinkp - 1) = 1;        /* dummy */
                        break;
                    }
                    if (k == SUBSLENGTH) {
                        BasError(msg1, strlen(msg1));
                        Kaboom(0);
                    }
                }
                return;
            }
            if (ilow == ihi)
                /* Range only has one element, degenerate case */
                istep = 0;
            else
                istep = ssl(I1, rlinkp - 1);
            isubarr2 = (Integer *) Paraddr(&(ssl(1, rlinkp + 1)));
            nelts = ssl(N1, rlinkp + 1);
            imax = isubarr2[0];
            imin = isubarr2[0];
            for (i = 1; i < nelts; i++) {
                if (imax < isubarr2[i])
                    imax = isubarr2[i];
                if (imin > isubarr2[i])
                    imin = isubarr2[i];
            }
            if ((ilow == DEFAULT && imin <= ihi && imax <= ihi) ||
                (ihi == DEFAULT && imin >= ilow && imax >= ilow) ||
                (imin >= ilow && imin <= ihi && imax >= ilow &&
                 imax <= ihi)) {
                /* All subscripts are in range--create new array */
                si[N - 1] = 1;
                si[TC - 1] = INTEGER;
                si[N1 - 1] = nelts;
                si[M1 - 1] = imin;
                si[I1 - 1] = 1; /* dummy */
                Parget(si);
                isubarr3 = (Integer *) Paraddr(si);
                for (i = 0; i < nelts; i++)
                    isubarr3[i] = ilow + (isubarr2[i] - 1) * istep;
                /* Now we need to build a new entry for bassubl */
                for (k = 1; k <= SUBSLENGTH; k++) {
                    if (bassubl(TC, k) == NULL_TYPE) {
                        /* found! save subscript stack element */
                        Parcopy(si, &(bassubl(1, k)));
                        bassubl(N1 + 1, k) = 1; /*can be parrel'ed */
                        /* put signal in operand stack element */
                        ssl(N1, rlinkp - 1) = DEFAULT + 1;
                        ssl(M1, rlinkp - 1) = k;
                        ssl(I1, rlinkp - 1) = 1;        /* dummy */
                        break;
                    }
                    if (k == SUBSLENGTH) {
                        BasError(msg1, strlen(msg1));
                        Kaboom(0);
                    }
                }
                return;
            } else {
                /* Some are out of range */
                sprintf(errmsg,
                        "parsubl: array with min and max subscripts %ld and %ld",
                        (long) imin, (long) imax);
                BasError(errmsg, strlen(errmsg));
                sprintf(errmsg,
                        "         can't index into range %ld:%ld:%ld.",
                        (long) ilow, (long) ihi, (long) istep);
                BasError(errmsg, strlen(errmsg));
                Kaboom(0);
            }
        }
    }
    if (ssl(NS, rlinkp - 1) == 0 && ns > 0) {
        /* this is the first set of subscripts, so uncritically accept it */
        Parsubc(&ns);
        if (ns == 0)
            return;
    } else if (ssl(NS, rlinkp - 1) > 0 && ns > 0) {
        if (ns > ssl(NS, rlinkp - 1)) {
            if (lxp[rlinkp - 2] > 72)
                lxp[rlinkp - 2] = 72;
            nxp[rlinkp - 2][lxp[rlinkp - 2]] = '\0';
            sprintf(errmsg, "%s%s'.", msg2, nxp[rlinkp - 2]);
            BasError(errmsg, strlen(errmsg));
            BasError(msg3, strlen(msg3));
            if (debug == YES)
                Kaboom(1);
            else
                Kaboom(0);
        }
        nns = ssl(NS, rlinkp - 1);
        if (nns == 1 && ssl(TC, rlinkp + 1) == NULL_TYPE) {
            if (ssl(NS, rlinkp - 1) > 0)
                /* Treat subsequent all-default as NOP */
                return;
            ssl(NS, rlinkp - 1) = 0;
            return;
        }
        for (i = 1; i <= nns; i++) {
            if (i > ns) {
                /* later spec has fewer subscripts so the old minimum will be the
                 * new default: */
            } else {
                j = rlinkp + i;
              l10:
                Parfetch(&(ssl(1, j)), nxp[j - 1], &(lxp[j - 1]), &cpy,
                         &false, lxp[j - 1]);
                switch (ssl(TC, j)) {
                case INTEGER:
                    if (Parint(&(ssl(1, j)), &ival) == ERR) {
                        if (lxp[rlinkp - 2] > 71)
                            lxp[rlinkp - 2] = 71;
                        nxp[rlinkp - 2][lxp[rlinkp - 2]] = '\0';
                        sprintf(errmsg,
                                "parsubl: error is in subscript %ld of '%s'.",
                                (long) i, nxp[rlinkp - 2]);
                        BasError(errmsg, strlen(errmsg));
                        Kaboom(0);
                    }
                    ilow = ival;
                    ihi = ival;
                    istr = 1;
                    break;
                case RANGE:
                    irange = (Integer *) Paraddr(&(ssl(1, j)));
                    ilow = irange[0];
                    ihi = irange[1];
                    istr = irange[2];
                    break;
                case NULL_TYPE:
                    if (ssl(I1 - 1 + i, rlinkp - 1) > 0) {
                        ilow = ssl(N1 - 1 + i, rlinkp - 1);
                        ihi = ssl(N1 - 1 + i, rlinkp - 1);
                    } else {
                        ilow = ssl(M1 - 1 + i, rlinkp - 1);
                        ihi = ssl(M1 - 1 + i, rlinkp - 1);
                    }
                    break;
                default:
                    if (lxp[rlinkp - 2] > 71)
                        lxp[rlinkp - 2] = 71;
                    nxp[rlinkp - 2][lxp[rlinkp - 2]] = '\0';
                    sprintf(errmsg,
                            "parsubl: possible trouble with variable named '%s'.",
                            nxp[rlinkp - 2]);
                    BasError(errmsg, strlen(errmsg));
                    if (ssl(N, rlinkp - 1) == UNKNOWN)
                        BasError(msg4, strlen(msg4));
                    if (Parcoerc(&(ssl(1, j)), &Inn) == ERR) {
                        sprintf(errmsg,
                                "Subscript %ld can't be coerced to integer.",
                                (long) i);
                        BasError(errmsg, strlen(errmsg));
                        Kaboom(1);
                    } else
                        goto l10;
                }
                if (istr == 0) {
                    if (lxp[rlinkp - 2] > 71)
                        lxp[rlinkp - 2] = 71;
                    nxp[rlinkp - 2][lxp[rlinkp - 2]] = '\0';
                    sprintf(errmsg,
                            "parsubl: trouble with variable named '%s'.",
                            nxp[rlinkp - 2]);
                    BasError(errmsg, strlen(errmsg));
                    sprintf(errmsg, "Stride on subscript %ld is zero.",
                            (long) i);
                    BasError(errmsg, strlen(errmsg));
                    if (debug == YES)
                        Kaboom(1);
                    else
                        Kaboom(0);
                }
                nelts = (ihi - ilow) / istr;
                nlast = ilow + nelts * istr * ssl(I1 - 1 + i, rlinkp - 1);
                if (ssl(N1 - 1 + i, rlinkp - 1) == DEFAULT &&
                    ssl(M1 - 1 + i, rlinkp - 1) == DEFAULT)
                    goto l20;
            }
            if ((ilow - ssl(N1 - 1 + i, rlinkp - 1)) % ssl(I1 - 1 + i,
                                                           rlinkp - 1) != 0
                || (ssl(M1 - 1 + i, rlinkp - 1) != DEFAULT &&
                    ilow / ssl(I1 - 1 + i, rlinkp - 1) > ssl(M1 - 1 + i,
                                                             rlinkp -
                                                             1) / ssl(I1 -
                                                                      1 +
                                                                      i,
                                                                      rlinkp
                                                                      -
                                                                      1)))
            {
                if (lxp[rlinkp - 2] > 71)
                    lxp[rlinkp - 2] = 71;
                nxp[rlinkp - 2][lxp[rlinkp - 2]] = '\0';
                sprintf(errmsg,
                        "parsubl: trouble with variable named '%s'.",
                        nxp[rlinkp - 2]);
                BasError(errmsg, strlen(errmsg));
                sprintf(errmsg,
                        "First value in subrange(%ld:%ld:%ld) of subscript %ld",
                        (long) ilow, (long) ihi, (long) istr, (long) i);
                BasError(errmsg, strlen(errmsg));
                sprintf(errmsg,
                        "         does not lie in earlier range(%ld:%ld:%ld).",
                        (long) ssl(N1 - 1 + i, rlinkp - 1),
                        (long) ssl(M1 - 1 + i, rlinkp - 1),
                        (long) ssl(I1 - 1 + i, rlinkp - 1));
                BasError(errmsg, strlen(errmsg));
                if (debug == YES)
                    Kaboom(1);
                else
                    Kaboom(0);
            } else if (nelts < 0) {
                if (lxp[rlinkp - 2] > 71)
                    lxp[rlinkp - 2] = 71;
                nxp[rlinkp - 2][lxp[rlinkp - 2]] = '\0';
                sprintf(errmsg,
                        "parsubl: trouble with variable named '%s'.",
                        nxp[rlinkp - 2]);
                BasError(errmsg, strlen(errmsg));
                sprintf(errmsg,
                        "Incorrect subrange(%ld:%ld:%ld) specified for subscript %ld.",
                        (long) ilow, (long) ihi, (long) istr, (long) i);
                BasError(errmsg, strlen(errmsg));
                if (debug == YES)
                    Kaboom(1);
                else
                    Kaboom(0);
            } else if (ssl(I1 - 1 + i, rlinkp - 1) < 0 &&
                       ssl(N1 - 1 + i, rlinkp - 1) != DEFAULT &&
                       nlast > ssl(N1 - 1 + i, rlinkp - 1)) {
                if (lxp[rlinkp - 2] > 71)
                    lxp[rlinkp - 2] = 71;
                nxp[rlinkp - 2][lxp[rlinkp - 2]] = '\0';
                sprintf(errmsg,
                        "parsubl: trouble with variable named '%s'.",
                        nxp[rlinkp - 2]);
                BasError(errmsg, strlen(errmsg));
                sprintf(errmsg,
                        "True last subscript %ld exceeds first value of old subrange.",
                        (long) nlast);
                BasError(errmsg, strlen(errmsg));
                sprintf(errmsg,
                        "         old subrange:(%ld:%ld:%ld).",
                        (long) ssl(N1 - 1 + i, rlinkp - 1),
                        (long) ssl(M1 - 1 + i, rlinkp - 1),
                        (long) ssl(I1 - 1 + i, rlinkp - 1));
                BasError(errmsg, strlen(errmsg));
                sprintf(errmsg,
                        "         new subrange:(%ld:%ld:%ld).",
                        (long) ilow, (long) ihi, (long) istr);
                BasError(errmsg, strlen(errmsg));
                if (debug == YES)
                    Kaboom(1);
                else
                    Kaboom(0);
            } else if (ssl(I1 - 1 + i, rlinkp - 1) > 0 &&
                       ssl(M1 - 1 + i, rlinkp - 1) != DEFAULT &&
                       nlast > ssl(M1 - 1 + i, rlinkp - 1)) {
                if (lxp[rlinkp - 2] > 71)
                    lxp[rlinkp - 2] = 71;
                nxp[rlinkp - 2][lxp[rlinkp - 2]] = '\0';
                sprintf(errmsg,
                        "parsubl: trouble with variable named '%s'.",
                        nxp[rlinkp - 2]);
                BasError(errmsg, strlen(errmsg));
                sprintf(errmsg,
                        "True last subscript %ld exceeds last value of old subrange.",
                        (long) nlast);
                BasError(errmsg, strlen(errmsg));
                sprintf(errmsg,
                        "         old subrange:(%ld:%ld:%ld).",
                        (long) ssl(N1 - 1 + i, rlinkp - 1),
                        (long) ssl(M1 - 1 + i, rlinkp - 1),
                        (long) ssl(I1 - 1 + i, rlinkp - 1));
                BasError(errmsg, strlen(errmsg));
                sprintf(errmsg,
                        "         new subrange:(%ld:%ld:%ld).",
                        (long) ilow, (long) ihi, (long) istr);
                BasError(errmsg, strlen(errmsg));
                if (debug == YES)
                    Kaboom(1);
                else
                    Kaboom(0);
            }
          l20:
            ssl(N1 - 1 + i, rlinkp - 1) = ilow;
            ssl(M1 - 1 + i, rlinkp - 1) = ilow +
                nelts * istr * ssl(I1 - 1 + i, rlinkp - 1);
            ssl(I1 - 1 + i, rlinkp - 1) *= istr;
        }
        /* note this routine does nothing at all if ns == 0 */
    }
    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Calculate the max and min of an integer array given by adata and len.
 */

void Cminmax(Address adata, Integer len, Integer *max, Integer *min)
{
    int i;
    Integer *idata;

    idata = (Integer *) adata;
    *max = idata[0];
    *min = idata[0];
    for (i = 1; i < len; i++) {
        if (*max < idata[i])
            *max = idata[i];
        if (*min > idata[i])
            *min = idata[i];
    }
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* same as above, except it calculates the three max and min for range.
 */
void Cminmaxr(Address adata, Integer len, Integer *max1, Integer *min1,
              Integer *max2, Integer *min2, Integer *max3, Integer *min3)
{
    int i;
    Integer *idata;

    idata = (Integer *) adata;
    *max1 = idata[0];
    *max2 = idata[1];
    *max3 = idata[2];
    *min1 = idata[0];
    *min2 = idata[1];
    *min3 = idata[2];
    for (i = 1; i < len; i++) {
        if (*max1 < idata[3 * i])
            *max1 = idata[3 * i];
        if (*max2 < idata[3 * i + 1])
            *max2 = idata[3 * i + 1];
        if (*max3 < idata[3 * i + 2])
            *max3 = idata[3 * i + 2];
        if (*min1 > idata[3 * i])
            *min1 = idata[3 * i];
        if (*min2 > idata[3 * i + 1])
            *min2 = idata[3 * i + 1];
        if (*min3 > idata[3 * i + 2])
            *min3 = idata[3 * i + 2];
    }
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
