/*
 * cfcn10.c
 *
 */

#include "cfcn.h"

#if defined(AXP) || defined (IRIX64) || defined (LINUXA)
#   define INTEGER_SIZE WORD_BIT
#else
#   define INTEGER_SIZE LONG_BIT
#endif

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* CLEGENDRE - handler for the legendre builtin 
 *           - Computes the first n Legendre polynomials on array y. stores them in
 #           - a new array shaped like y but with an additional dimension whose
 #           - size is n and whose value selects the appropriate polynomial
 */

Integer Clegendre(Integer *sxi, f_entry *thisf, Integer nargs, Integer *syi)
{
    float *ry, *rx, rmul1, rdiv;
    real8 *dy, *dx, dmul1, ddiv;
    int i, j, i0, i1, i2;
    Integer n, len, szi[SSWIDTH], ndim;

    static char
        errmsg1[] = "legendre: argument has more than 6 dimensions.",
        errmsg2[] = "legendre: argument must be < 1 in absolute value.",
        errmsg3[] = "legendre: second argument must be an integer.",
        errmsg4[] = "legendre: second argument must be non-negative.";

    ndim = syi[N - 1];
    if (ndim > 6) {
        BasError(errmsg1, strlen(errmsg1));
        return ERR;
    }
    Parcopy(syi, sxi);
    len = Parlen(sxi);
    if (sxi[TC - 1] == REAL) {
        ry = (float *) Paraddr(syi);
        for (i = 0; i < len; i++)
            if (abs(ry[i] > 1.)) {
                BasError(errmsg2, strlen(errmsg2));
                return ERR;
            }
    } else {                    /* DOUBLE */
        dy = (real8 *) Paraddr(syi);
        for (i = 0; i < len; i++)
            if (abs(dy[i] > 1.)) {
                BasError(errmsg2, strlen(errmsg2));
                return ERR;
            }
    }
    Fcnargb(2, szi);
    if (Parint(szi, &n) == ERR) {
        BasError(errmsg3, strlen(errmsg3));
        return ERR;
    }
    Parrel(szi);
    sxi[N - 1] += 1;
    sxi[N1 + ndim - 1] = n + 1;
    sxi[M1 + ndim - 1] = 0;
    if (n < 0) {
        BasError(errmsg4, strlen(errmsg4));
        return ERR;
    }
    Parget(sxi);
    if (sxi[TC - 1] == REAL) {
        rx = (float *) Paraddr(sxi);
        if (n == 0) {
            for (i = 0; i < len; i++)
                rx[i] = 1.0;    /* First polynomial is 1.0 */
        } else if (n >= 1) {
            for (i = 0; i < len; i++) {
                rx[i] = 1.0;    /* First polynomial is 1.0 */
                if (n > 0)
                    rx[i + len] = ry[i];        /* Second polynomial is x  */
            }
            for (j = 1, i0 = 0, i1 = len, i2 = 2 * len, rmul1 = 3.;
                 j < n; j++, rmul1 += 2.) {
                rdiv = (float) j + 1.;
                for (i = 0; i < len; i++, i0++, i1++, i2++)
                    rx[i2] =
                        (rmul1 * ry[i] * rx[i1] -
                         (float) j * rx[i0]) / rdiv;
            }
        }
    } else {                    /* DOUBLE */
        dx = (real8 *) Paraddr(sxi);
        if (n == 0) {
            for (i = 0; i < len; i++)
                dx[i] = 1.0;    /* First polynomial is 1.0 */
        } else if (n >= 1) {
            for (i = 0; i < len; i++) {
                dx[i] = 1.0;    /* First polynomial is 1.0 */
                dx[i + len] = dy[i];    /* Second polynomial is x  */
            }
            for (j = 1, i0 = 0, i1 = len, i2 = 2 * len, dmul1 = 3.;
                 j < n; j++, dmul1 += 2.) {
                ddiv = (real8) j + 1.;
                for (i = 0; i < len; i++, i0++, i1++, i2++)
                    dx[i2] =
                        (dmul1 * dy[i] * dx[i1] -
                         (real8) j * dx[i0]) / ddiv;
            }
        }
    }
    Parrel(syi);
    return OK;
}

/*--------------------------------------------------------------------------*/

/*                               BIT HANDLERS                               */

/*--------------------------------------------------------------------------*/

/* GET_BITNO - get the value of the second argument to be used as bitno */

static Integer get_bitno(void)
{
    Integer szi[SSWIDTH], bitno;

    Fcnargb(2, szi);
    if (szi[TC - 1] != INTEGER) {
        BasError("bitwise operation: second argument must be integer.",
                 51);
        return (-100);
    }

    if ((Parlen(szi) != 1) || (Parint(szi, (Integer *) &bitno) == ERR)) {
        BasError
            ("bitwise operation: second argument must be integer scalar.",
             58);
        return (-100);
    }

    Parrel(szi);

    return (bitno);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* GET_BITSIZE - get the value of the third argument to be used as bitsize */

static Integer get_bitsize(void)
{
    Integer szi[SSWIDTH], bitsize;
    char errmsg[OUTLINE];

    Fcnargb(3, szi);

    if ((Parlen(szi) != 1) || (Parint(szi, (Integer *) &bitsize) == ERR)) {
        BasError
            ("bitwise operation: third argument must be integer scalar.",
             57);
        return (-1000);
    }

    Parrel(szi);

    if (bitsize <= 0 || bitsize > INTEGER_SIZE) {
        sprintf(errmsg, "ishftc: size of %ld is not between 1 and %ld.",
                (long) bitsize, (long) INTEGER_SIZE);
        BasError(errmsg, strlen(errmsg));
        return (-1000);
    }

    return (bitsize);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* GET_LX - get the logical array being operated on by bit operations */

static void get_lx(Integer *plenx, logical **pix, Integer *sxi)
{
    Integer lenx;
    logical *lx;

    sxi[TC - 1] = LOGICAL;

    Parget(sxi);
    lx = (logical *) Paraddr(sxi);
    lenx = Parlen(sxi);

    *plenx = lenx;
    *pix = lx;

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* GET_IX - get the integer array being operated on by bit operations */

static void get_ix(Integer *plenx, unsigned Integer **pix, Integer *sxi)
{
    Integer lenx;
    unsigned Integer *ix;

    ix = (unsigned Integer *) Paraddr(sxi);
    lenx = Parlen(sxi);

    *plenx = lenx;
    *pix = ix;

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* GET_IZ - get the integer array being operated on by bit operations */

static void get_iz(Integer *plenz, unsigned Integer **piz, Integer *szi)
{
    Integer lenz;
    unsigned Integer *iz;

    Fcnargb(2, szi);
    if (szi[TC - 1] != INTEGER) {
        BasError("bitwise operation: second argument must be integer.",
                 51);
        lenz = -1;
        iz = NULL;

    } else {
        lenz = Parlen(szi);
        iz = (unsigned Integer *) Paraddr(szi);
    }

    *plenz = lenz;
    *piz = iz;

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* CBTEST - do the btest builtin */

Integer Cbtest(Integer *sxi, f_entry *thisf, Integer nargs, Integer *syi)
{
    Integer lenx, bitno;
    unsigned Integer *iy, bitmask;
    logical *lx;
    int i;
    char errmsg[MAXOUT];

    Parcopy(syi, sxi);

    get_lx(&lenx, &lx, sxi);

    iy = Paraddr(syi);

    bitno = get_bitno();
    if ((bitno < 0) || (bitno > INTEGER_SIZE - 1)) {
        sprintf(errmsg, "single bit operations: bit number < 0 or > %ld.",
                (long) (WORD_BIT - 1));
        BasError(errmsg, strlen(errmsg));
        return (ERR);
    }

    bitmask = 1 << bitno;

    for (i = 0; i < lenx; i++)
        lx[i] = (logical) (iy[i] & bitmask ? 1 : 0);

    Parrel(syi);

    return (OK);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* CIBSET - do the ibset builtin */

Integer Cibset(Integer *sxi, f_entry *thisf, Integer nargs, Integer *syi)
{
    Integer lenx, bitno;
    unsigned Integer *ix, bitmask;
    int i;
    char errmsg[MAXOUT];

    Parcopy(syi, sxi);

    get_ix(&lenx, &ix, sxi);

    bitno = get_bitno();
    if ((bitno < 0) || (bitno > INTEGER_SIZE - 1)) {
        sprintf(errmsg, "single bit operations: bit number < 0 or > %ld.",
                (long) (WORD_BIT - 1));
        BasError(errmsg, strlen(errmsg));
        return (ERR);
    }

    bitmask = 1 << bitno;

    for (i = 0; i < lenx; i++)
        ix[i] = ix[i] | bitmask;

    return (OK);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* CIBCLR - do the ibclr builtin */

Integer Cibclr(Integer *sxi, f_entry *thisf, Integer nargs, Integer *syi)
{
    Integer lenx, bitno;
    unsigned Integer *ix, bitmask;
    int i;
    char errmsg[MAXOUT];

    Parcopy(syi, sxi);

    get_ix(&lenx, &ix, sxi);

    bitno = get_bitno();
    if ((bitno < 0) || (bitno > INTEGER_SIZE - 1)) {
        sprintf(errmsg, "single bit operations: bit number < 0 or > %ld.",
                (long) (WORD_BIT - 1));
        BasError(errmsg, strlen(errmsg));
        return (ERR);
    }

    bitmask = 1 << bitno;
    bitmask = ~bitmask;

    for (i = 0; i < lenx; i++)
        ix[i] = ix[i] & bitmask;

    return (OK);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* CISHFT - do ishft builtin */

Integer Cishft(Integer *sxi, f_entry *thisf, Integer nargs, Integer *syi)
{
    Integer lenx, bitno;
    unsigned Integer *ix;
    int i;
    char errmsg[MAXOUT];

    Parcopy(syi, sxi);

    get_ix(&lenx, &ix, sxi);

    bitno = get_bitno();
    if (abs(bitno) > INTEGER_SIZE) {
        sprintf(errmsg, "ishft: shift of %ld is too large (> %ld).",
                (long) bitno, (long) INTEGER_SIZE);
        BasError(errmsg, strlen(errmsg));
        return (ERR);
    }

    if (bitno > 0)
        for (i = 0; i < lenx; i++)
            ix[i] = ix[i] << bitno;

    else if (bitno < 0) {
        bitno = -bitno;
        for (i = 0; i < lenx; i++)
            ix[i] = ix[i] >> bitno;
    }

    return (OK);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* CISHFTC - do the ishftc builtin */

Integer Cishftc(Integer *sxi, f_entry *thisf, Integer nargs, Integer *syi)
{
    Integer lenx, bitno;
    unsigned Integer *ix, itmp, bitmask, cbitmask;
    unsigned Integer ishftpart, bitsize, ibit, isone;
    int i;
    char errmsg[MAXOUT];

    Parcopy(syi, sxi);

    get_ix(&lenx, &ix, sxi);

    bitno = get_bitno();
    bitsize = get_bitsize();

    bitmask = (~((unsigned Integer) 0)) >> (INTEGER_SIZE - bitsize);
    cbitmask = ~bitmask;

    if (abs(bitno) > bitsize) {
        sprintf(errmsg,
                "ishftc: shift of %ld is greater than size of %ld.",
                (long) bitno, (long) bitsize);
        BasError(errmsg, strlen(errmsg));
        return (ERR);
    }

    ibit = ((unsigned Integer) 1) << max(bitsize - 1, 0);

/* change to equivalent right shift */

/* left shift */
    if (bitno > 0)
        bitno = bitsize - bitno;

/* right shift */
    else
        bitno = abs(bitno);

    for (i = 0; i < lenx; i++) {
        ishftpart = ix[i] & bitmask;

/* Note on below: nothing happens if bitno == 0
 * or if |bitno| == bitsize
 */
        if (bitno != 0 && bitno != bitsize) {
            itmp = bitno;
            while (itmp > 0) {
                isone = ishftpart & 1;
                ishftpart = ishftpart >> 1;
                if (isone)
                    ishftpart = ishftpart | ibit;
                itmp -= 1;
            }
        }

        ix[i] = (ix[i] & cbitmask) | ishftpart;
    }

    return (OK);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* CIBITS - do ibits builtin */

Integer Cibits(Integer *sxi, f_entry *thisf, Integer nargs, Integer *syi)
{
    Integer lenx, bitno;
    unsigned Integer *ix, bitmask, cbitmask;
    unsigned Integer bitsize;
    int i;
    char errmsg[MAXOUT];

    Parcopy(syi, sxi);

    get_ix(&lenx, &ix, sxi);

    bitno = get_bitno();
    bitsize = get_bitsize();

    bitmask = (~((unsigned Integer) 0)) >> (INTEGER_SIZE - bitsize);
    cbitmask = ~bitmask;

    if (bitno < 0 || bitno < bitsize - 1) {
        sprintf(errmsg,
                "ibits: pos = %ld, len = %ld must be >= 0 and pos >= len - 1.",
                (long) bitno, (long) bitsize);
        BasError(errmsg, strlen(errmsg));
        return (ERR);
    }

    bitmask = bitmask << (bitno - bitsize + 1);

    for (i = 0; i < lenx; i++)
        ix[i] = (ix[i] & bitmask) >> (bitno - bitsize + 1);

    return (OK);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* CIAND - do the iand builtin */

Integer Ciand(Integer *sxi, f_entry *thisf, Integer nargs, Integer *syi)
{
    Integer szi[SSWIDTH], lenx, lenz;
    unsigned Integer *ix, *iz, itmp;
    int i;

    Parcopy(syi, sxi);

    get_ix(&lenx, &ix, sxi);
    get_iz(&lenz, &iz, szi);

    if ((lenz != lenx) && (lenx != 1) && (lenz != 1)) {
        BasError("bitwise operation: argument size mismatch.", 42);
        return (ERR);
    }

/* broadcast left to right */
    if ((lenx == 1) && (lenz > 1)) {
        itmp = ix[0];

        Parrel(sxi);
        Parcopy(szi, sxi);
        Parget(sxi);

        ix = (unsigned Integer *) Paraddr(sxi);
    }

    if ((lenx == 1) && (lenz == 1))
        ix[0] = ix[0] & iz[0];

    else if (lenx == 1)
        for (i = 0; i < lenz; i++)
            ix[i] = itmp & iz[i];

    else if (lenz == 1)
        for (i = 0; i < lenx; i++)
            ix[i] = ix[i] & iz[0];

    else
        for (i = 0; i < lenx; i++)
            ix[i] = ix[i] & iz[i];

    Parrel(szi);

    return (OK);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* CIOR - do the ior builtin */

Integer Cior(Integer *sxi, f_entry *thisf, Integer nargs, Integer *syi)
{
    Integer szi[SSWIDTH], lenx, lenz;
    unsigned Integer *ix, *iz, itmp;
    int i;

    Parcopy(syi, sxi);

    get_ix(&lenx, &ix, sxi);
    get_iz(&lenz, &iz, szi);

    if ((lenz != lenx) && (lenx != 1) && (lenz != 1)) {
        BasError("bitwise operation: argument size mismatch.", 42);
        return (ERR);
    }

/* broadcast left to right */
    if ((lenx == 1) && (lenz > 1)) {
        itmp = ix[0];

        Parrel(sxi);
        Parcopy(szi, sxi);
        Parget(sxi);

        ix = (unsigned Integer *) Paraddr(sxi);
    }

    if ((lenx == 1) && (lenz == 1))
        ix[0] = ix[0] | iz[0];

    else if (lenx == 1)
        for (i = 0; i < lenz; i++)
            ix[i] = itmp | iz[i];

    else if (lenz == 1)
        for (i = 0; i < lenx; i++)
            ix[i] = ix[i] | iz[0];

    else
        for (i = 0; i < lenx; i++)
            ix[i] = ix[i] | iz[i];

    Parrel(szi);

    return (OK);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* CIEOR - do the ieor builtin */

Integer Cieor(Integer *sxi, f_entry *thisf, Integer nargs, Integer *syi)
{
    Integer szi[SSWIDTH], lenx, lenz;
    unsigned Integer *ix, *iz, itmp;
    int i;

    Parcopy(syi, sxi);

    get_ix(&lenx, &ix, sxi);
    get_iz(&lenz, &iz, szi);

    if ((lenz != lenx) && (lenx != 1) && (lenz != 1)) {
        BasError("bitwise operation: argument size mismatch.", 42);
        return (ERR);
    }

/* broadcast left to right */
    if ((lenx == 1) && (lenz > 1)) {
        itmp = ix[0];

        Parrel(sxi);
        Parcopy(szi, sxi);
        Parget(sxi);

        ix = (unsigned Integer *) Paraddr(sxi);
    }

    if ((lenx == 1) && (lenz == 1))
        ix[0] = ix[0] ^ iz[0];

    else if (lenx == 1)
        for (i = 0; i < lenz; i++)
            ix[i] = itmp ^ iz[i];

    else if (lenz == 1)
        for (i = 0; i < lenx; i++)
            ix[i] = ix[i] ^ iz[0];

    else
        for (i = 0; i < lenx; i++)
            ix[i] = ix[i] ^ iz[i];

    Parrel(szi);

    return (OK);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* CUNOT - do the unot builtin */

Integer Cunot(Integer *sxi, f_entry *thisf, Integer nargs, Integer *syi)
{
    Integer lenx;
    unsigned Integer *ix;
    int i;

    Parcopy(syi, sxi);

    get_ix(&lenx, &ix, sxi);

    for (i = 0; i < lenx; i++)
        ix[i] = ~ix[i];

    return (OK);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
