/*
  So how does this set of signal handlers work, anyway?  There are 3
  signal handling routines defined in the Basis kernel:

    yuck(): Handles BUS, FPE, ILL, SEGV, SYS, USR1, and USR2
    ruthere1(): Handles INT
    basterm(): Handles TERM
  
  Starting from the bottom up, basterm() is actually a Fortran routine,
  intended to be user-replaceable.  For example, some batch job systems
  use SIGTERM to tell a process to exit gracefully.  Codes running
  under such a system might catch SIGTERM and make a restart file
  before exiting.  The default routine (in term.m) just causes Basis to
  exit immediately, without a core dump on most operating systems.

  Ruthere1 handles the interrupt signal, usually generated by a user at
  a keyboard with a Control-C sequence.  It merely sets a flag, for
  later query by the ruthere() routine.  Ruthere may be called at any
  time from user code.  If an interrupt has occurred since it was last
  called, it enters the debugger, or not, depending on whether the
  no_recovery flag has been set by a call to errortrp(), aka
  Error_trap().

  Error trapping for SIGINT and the other signals is turned on during
  Basis initialization.  A later call to errortrp("batch") or
  errortrp("off"), etc. (errortrp can be called from the interpreter)
  leaves all handlers installed for their signals as shown above, but
  sets the no_recovery flag to 1 (TRUE).

  If a signal in the group BUS, FPE, ILL, SEGV, SYS, USR1, or USR2 is
  delivered, yuck() prints a message per the signal, sets handling to
  SIG_IGN, and calls Kaboom.  Kaboom() prints a traceback with ERRPRT.
  If no_recovery == 1, Kaboom sets the handler to SIG_DFL.  It then
  calls Rtfin and BASISEND, which should never return (although an
  unwanted core dump may occur if additional signals are generated
  during the exit sequence).  Alternatively, if the no_recovery flag is
  0, Kaboom longjumps back into Parctl(), which PAROOPS a message,
  unblocks the signal (if any), restores yuck() as the signal handler,
  and restarts PAREXE1, the command processing loop.

  SIGFPE gets special handling, due to the various architecture
  specific interactions of floating point exceptions and this signal.
  Note that HPUX simply uses the reliable_signal() function to set up
  the handler, whereas Sun (both SunOS and Solaris) uses another
  function ieee_handler() to do this task.  The fpu_reset() function
  sets the floating point unit to throw an exception for invalid
  operations, overflow, and division by zero, and sets the signal
  handler for SIGFPE as per its argument.  HPUX is complicated by the
  fact that its (Fortran library) floating point exception code
  *requires* the user to replace matherr() in order to get control of
  certain library function exceptions, such as sqrt(-1.0), and so
  forth.  The HPUX Floating Point Guide makes a point of saying that
  matherr() is deprecated and will not be supported by future releases,
  but they give no alternative that I am aware of at this time
  (December, 1995).

  Disposition of Basis core dumps is normally determined by the
  "coredump" global variable, which is 0 (FALSE) by default.  During an
  exit, the thread of control goes from BASISEND to basexit, to basfin,
  to Finish, which raises SIGABRT if coredump is TRUE.

  Unwanted core dumps could be produced if no_recovery is TRUE, and one
  of SIGFPE, SEGV, etc. is generated during the exit sequence (Rtfin
  or BASISEND).  Schemes other than setting the handler to SIG_DFL (see
  Kaboom() below) are much more complicated, however, and SIG_IGN would
  risk an infinite loop in the case above.
*/

#include "rt.h"

/* Declare Fortran routines we need to call */
#define Parexe1  FF_ID(parexe1,  PAREXE1)
#define Paroops  FF_ID(paroops,  PAROOPS)
#define Basisend FF_ID(basisend, BASISEND)
#define Basterm  FF_ID(basterm,  BASTERM)
#define Debugger FF_ID(debugger, DEBUGGER)

extern void Paroops(void *);
extern void Parexe1(void);
extern void Basisend(void *);
extern void Basterm(Integer);
extern void Debugger(void);

static void yuck(int);
static void ruthere1(int);

#ifdef SIGSYS
int recover_from[] = { SIGBUS, SIGSYS, SIGILL, SIGSEGV, SIGUSR1, SIGUSR2 };
#else
int recover_from[] = { SIGBUS, SIGILL, SIGSEGV, SIGUSR1, SIGUSR2 };
#endif
jmp_buf jmp_pt;
int iKaboom = 0;
int no_recovery = 1;
int last_signal = 0;

#ifdef __STDC__
typedef void (*Sigfunc) (int);
static void fpu_reset(Sigfunc handler);
#else
typedef void (*Sigfunc) ();
static void fpu_reset();
#endif

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/*
 *   Following implementation of signal was lifted from ``Advanced Programming
 *  in the UNIX Environment'', by W. Richard Stevens, 1992, Addison-Wesley,
 *  pp. 298.  It has the advantage over signal() that additional occurences
 *  of the trapped signal are blocked while in the signal handler, and
 *  disposition of the signal handler never reverts to SIG_DFL unless
 *  explicitly set by the programmer.
 */
Sigfunc reliable_signal(signo, func)
int signo;
Sigfunc func;
{
#if defined(HP700)
    /* I had trouble with HPUX and sigaction.  Interrupts (SIGINT) caused
     * an immediate SIGILL upon return from ruthere1.  I suspect something
     * in sigaction is not "Fortran-clean", in terms of saving and restoring
     * registers.
     */
    signal(signo, func);
    return ((Sigfunc) 0);
#else
    struct sigaction act, oact;

    act.sa_handler = func;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
    if (signo == SIGALRM) {
#ifdef SA_INTERRUPT
        act.sa_flags |= SA_INTERRUPT;   /* SunOS */
#endif
    } else {
#ifdef SA_RESTART
        act.sa_flags |= SA_RESTART;     /* SVR4, 4.3+BSD (But not HPUX!) */
#endif
    }
    if (sigaction(signo, &act, &oact) < 0)
        return (SIG_ERR);
    return (oact.sa_handler);
#endif
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void sigunblock(sig)
int sig;
{
/* Sigunblock shouldn't really be necessary.  However, I found that on
   the DEC Alpha, SIGINT started out blocked, and couldn't be caught
   unless they were explicitly unblocked, once. */
    sigset_t newmask;

    sigemptyset(&newmask);
    sigaddset(&newmask, sig);
    if (sigprocmask(SIG_UNBLOCK, &newmask, NULL) < 0) {
        char errmsg[80];

        sprintf(errmsg, "Sigunblock: Failed to unblock signal %d", sig);
        BasError(errmsg, strlen(errmsg));
    }
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Kaboom(idum)
Integer idum;
{
    int i;
    int table_size = sizeof(recover_from) / sizeof(recover_from[0]);
    static Integer one = 1;

    Errprt();
    fflush(stdout);
    if (no_recovery == 1) {
        char errmsg[80];

        for (i = 0; i < table_size; i++) {
            reliable_signal(recover_from[i], SIG_DFL);
        }
        sprintf(errmsg, "Error recovery disabled, exiting.");
        BasError(errmsg, strlen(errmsg));
        fpu_reset(SIG_DFL);
        Rtfin(-1);
        Basisend(&one);         /* C calls Fortran, so must pass arg by reference */
    } else {
        fpu_reset(yuck);
        longjmp(jmp_pt, 1);
    }
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/*
 * Dosigterm is called by the default BASTERM routine.  BASTERM doesn't
 * call raise() directly from Fortran because I can't see any easy way
 * to get the literal integer corresponding to SIGTERM inside Fortran.
 */
void Dosigterm()
{
    reliable_signal(SIGTERM, SIG_DFL);
    raise(SIGTERM);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Parctl()
{
    if (setjmp(jmp_pt) != 0) {
/*
  setjmp sets jmp_pt and returns with 0 the very first time.            
  Kaboom returns here with a nonzero value after an irretrievable error.
*/
        Paroops((Address) &iKaboom);
        if (last_signal) {      /* Then call to kaboom came via signal */
            sigunblock(last_signal);
            reliable_signal(last_signal, yuck);
            last_signal = 0;
        }
    }
    reliable_signal(SIGINT, ruthere1);  /* Redundant, unless you're coming
                                           from a  DEBUG> abort */
    Parexe1();
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

static int imhere = 0;
static void ruthere1(sig)
int sig;
{
    imhere = 1;
    reliable_signal(SIGINT, ruthere1);
    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Ruthere0()
{
    reliable_signal(SIGINT, ruthere1);
#if defined(AXP)
    sigunblock(SIGINT);
#endif
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Ruthere()
{
    if (imhere > 0) {
        imhere = 0;
        if (no_recovery == 0) {
            reliable_signal(SIGINT, SIG_IGN);
            Debugger();
            reliable_signal(SIGINT, ruthere1);
        } else {
            Kaboom(0);
        }
    }
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

static void yuck(sig)
int sig;
{
    char *msg;
    char errmsg[120];

    switch (sig) {
    case SIGBUS:
        msg = "Bus error (often misaligned memory reference).";
        break;
    case SIGFPE:
        msg = "Floating point error.";
        break;
    case SIGILL:
        msg = "Illegal instruction.";
        break;
    case SIGSEGV:
        msg = "Segmentation violation.";
        break;
#ifdef SIGSYS
    case SIGSYS:
        msg = "Bad arguments to system call.";
        break;
#endif
    case SIGUSR1:
    case SIGUSR2:
        msg = "EZD error or other user-software-generated signal.";
        break;
    default:
        msg = "? Unknown signal code.";
    }
    sprintf(errmsg, "Basis routine 'yuck' caught signal %d: %s", sig, msg);
    BasError(errmsg, strlen(errmsg));
    last_signal = sig;
    reliable_signal(sig, SIG_IGN);
    Kaboom((Integer) 0);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

Integer Basis_stat = 0;

Integer Basisexe(s)
char *s;
{
    int system();

    Basis_stat = (Integer) (system(s) >> 8);
    return (Basis_stat);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

Integer Basisstat()
{
    return (Basis_stat);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Code for fpu_reset() is quite architecture-specific.  Basically, I'm
 * trying to set up things so that overflow, division by zero, and
 * invalid operations cause SIGFPE to be sent to the process.  Seems
 * easy enough, doesn't it?
 */

#if defined(HP700)

static void fpu_reset(handler)
Sigfunc handler;
{
    fpsetdefaults();
    reliable_signal(SIGFPE, handler);
    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

int matherr(x)
struct exception *x;
{
    char errmsg[80];

    switch (x->type) {
    case DOMAIN:
        sprintf(errmsg, "Domain error in %s(%g)", x->name, x->arg1);
        BasError(errmsg, strlen(errmsg));
        raise(SIGFPE);

    case SING:
        sprintf(errmsg, "Singularity in %s(%g)", x->name, x->arg1);
        BasError(errmsg, strlen(errmsg));
        raise(SIGFPE);

    case OVERFLOW:
        sprintf(errmsg, "Overflow in %s(%g)", x->name, x->arg1);
        BasError(errmsg, strlen(errmsg));
        raise(SIGFPE);

    case UNDERFLOW:
        sprintf(errmsg, "Underflow in %s(%g)", x->name, x->arg1);
        BasError(errmsg, strlen(errmsg));
        raise(SIGFPE);

    case TLOSS:
    case PLOSS:
        sprintf(errmsg, "Loss of significance in %s(%g)", x->name,
                x->arg1);
        BasError(errmsg, strlen(errmsg));
        raise(SIGFPE);

    default:
        raise(SIGFPE);
    }
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

#elif defined(SUN4)

#include <floatingpoint.h>
static void fpu_reset(handler)
sigfpe_handler_type handler;
{
    char errmsg[80];
    char *mode = "exception", *in = "all", *out;
    void abrupt_underflow_();

    if (ieee_flags("clearall", mode, in, &out) != 0) {
        sprintf(errmsg, "Fpu_reset: Ieee_flags returned non-zero!");
        BasError(errmsg, strlen(errmsg));
    }
    if (ieee_handler("set", "common", handler) != 0) {
        sprintf(errmsg, "Fpu_reset: Ieee_handler returned non-zero!");
        BasError(errmsg, strlen(errmsg));
    }
    abrupt_underflow_();        /* (This doesn't really seem to work.) */
    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

#elif defined(SOL)

#include <ieeefp.h>
static void fpu_reset(handler)
Sigfunc handler;
{
    fpsetmask(FP_X_INV | FP_X_DZ | FP_X_OFL);
    reliable_signal(SIGFPE, handler);
    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

#elif defined(SGI) || defined(IRIX64)

#include <sigfpe.h>

static void haha(unsigned int **x)
{
    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

static void fpu_reset(handler)
Sigfunc handler;
{
    handle_sigfpes(_ON,
                   _EN_OVERFL | _EN_DIVZERO | _EN_INVALID,
                   NULL, _ABORT_ON_ERROR, haha);

    reliable_signal(SIGFPE, handler);
    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

#elif defined(AXP)

#include <machine/fpu.h>
static void fpu_reset(handler)
Sigfunc handler;
{
    unsigned long fp_control =
        IEEE_TRAP_ENABLE_INV | IEEE_TRAP_ENABLE_DZE | IEEE_TRAP_ENABLE_OVF;

    ieee_set_fp_control(fp_control);
    reliable_signal(SIGFPE, handler);
    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

#elif defined(RS6000)

/* Untested as of 11Dec95... */
#include <fptrap.h>
static void fpu_reset(handler)
Sigfunc handler;
{
    fp_trap(FP_TRAP_FASTMODE);
    fp_enable(TRP_INVALID | TRP_DIV_BY_ZERO | TRP_OVERFLOW);
    reliable_signal(SIGFPE, handler);
    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

#elif defined(LINUX) || defined(LINUXA)
/* GOTCHA: on a LINUXA system, LINUX is also defined by pact,
 *         so we must also test for __i386__ and __alpha__
 */

/* see also: fpu_control.h or i386/fpu_control.h, __setfpucw function */

#ifdef __i386__
static void fpu_reset(Sigfunc handler)
{
    unsigned int cw = 0x1372;

    __asm__("fldcw %0": :"m"(cw));
    reliable_signal(SIGFPE, handler);
    return;
}
#endif

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

#ifdef __alpha__
#include <asm/fpu.h>

static void fpu_reset(Sigfunc handler)
{
    /* Digital Unix, True64 unix compatability */
#if 1
    unsigned long fp_control =
        IEEE_TRAP_ENABLE_INV | IEEE_TRAP_ENABLE_DZE | IEEE_TRAP_ENABLE_OVF;

    ieee_set_fp_control(fp_control);
#else
    /* Intel Linux compatibility */
#include <fpu_control.h>
    /*
     * IM: Invalid operation mask
     * DM: Denormalized operand mask
     * ZM: Zero-divide mask
     * OM: Overflow mask
     * UM: Underflow mask
     * PM: Precision (inexact result) mask
     */
    __setfpucw(_FPU_MASK_IM | _FPU_MASK_DM | _FPU_MASK_ZM | _FPU_MASK_OM);
#endif
    reliable_signal(SIGFPE, handler);
    return;
}
#endif

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

#else

static void fpu_reset(handler)
Sigfunc handler;
{
    reliable_signal(SIGFPE, handler);
    return;
}

#endif

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Error_trap(flag)
char *flag;
{
    int i;
    int table_size = sizeof(recover_from) / sizeof(recover_from[0]);

    reliable_signal(SIGFPE, yuck);
    fpu_reset(yuck);
    if (strcmp(flag, "on") == 0) {
        no_recovery = 0;
        for (i = 0; i < table_size; i++) {
            reliable_signal(recover_from[i], yuck);
        }
    } else {
        no_recovery = 1;
    }
    /* SIGTERM is always caught by the (user-replaceable) BASTERM. */
    reliable_signal(SIGTERM, Basterm);
    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Scaffolding code, was useful at least twice so far.  Prints out
 * blocked signals.
 */
#if 0

void pr_mask(str)
char *str;
{
    sigset_t sigset;
    int errno_save;
    char errmsg[80];

    errno_save = errno;
    if (sigprocmask(0, NULL, &sigset) < 0) {
        sprintf(errmsg, "sigprocmask returned non-zero");
        BasError(errmsg, strlen(errmsg));
    }
    sprintf(errmsg, "%s", str);
    BasError(errmsg, strlen(errmsg));
    if (sigismember(&sigset, SIGINT)) {
        sprintf(errmsg, "SIGINT ");
        BasError(errmsg, strlen(errmsg));
    }
    if (sigismember(&sigset, SIGFPE)) {
        sprintf(errmsg, "SIGFPE ");
        BasError(errmsg, strlen(errmsg));
    }
    BasError(" ", 1)
        errno = errno_save;
}
#endif

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
