/*
 * HAND.C - handlers for efficient executino of builtin functions
 *
 */

#include "cfcn.h"

typedef double (*PFD_AN) (void);
typedef double (*PFD_AD) (double x);

typedef void (*PFDC_AN) (dcomplex *x, int n);
typedef void (*PFDC_APDC) (dcomplex *x, dcomplex *y, int n);
typedef void (*PFD_APDC) (double *x, dcomplex *y, int n);

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* PAR_ZARG - zero argument functions */

Integer Par_zarg(Integer *sxi, f_entry *thisf, Integer nargs, Integer *syi)
{
    int i;
    Integer n;
    PFVoid *info;

    info = (PFVoid *) thisf->info;

/* set up syi */
    if (nargs == 0) {
        syi[N - 1] = 0;
        syi[TC - 1] = (sizeof(int) == 8) ? REAL : DOUBLE;
        Parget(syi);
    }

    Parcopy(syi, sxi);
    if (sxi[TC - 1] == INTEGER)
        n = Parcoerc(sxi, &myreal);

    n = Parlen(sxi);

    switch (sxi[TC - 1]) {
    case REAL:
    {
        float *rx;
        PFD_AN df;

        rx = (float *) Paraddr(sxi);

        df = (PFD_AN) info[0];
        for (i = 0; i < n; i++)
            rx[i] = (*df) ();
    }
        break;

    case DOUBLE:
    {
        real8 *dx;
        PFD_AN df;

        dx = (real8 *) Paraddr(sxi);

        df = (PFD_AN) info[0];
        for (i = 0; i < n; i++)
            dx[i] = (*df) ();
    }
        break;

/* NOTE: do all complex math in C functions and in order to
 *       make this efficient as possible have them do an array
 *       at a time
 */
    case COMPLEX:
    {
        dcomplex *dcx;
        Integer nt;
        PFDC_AN dcf;

        nt = DOUBLE_C;
        Parcoerc(sxi, &nt);

        dcx = (dcomplex *) Paraddr(sxi);
        dcf = (PFDC_AN) info[1];
        (*dcf) (dcx, (int) n);

        nt = COMPLEX;
        Parcoerc(sxi, &nt);
    }

        break;

    case DOUBLE_C:
    {
        dcomplex *dcx;
        PFDC_AN dcf;

        dcx = (dcomplex *) Paraddr(sxi);
        dcf = (PFDC_AN) info[1];
        (*dcf) (dcx, (int) n);
    }

        break;
    }

    return (OK);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* PAR_SARG - single argument functions */

Integer Par_sarg(Integer *sxi, f_entry *thisf, Integer nargs, Integer *syi)
{
    int i;
    Integer n;
    PFVoid *info;

    info = (PFVoid *) thisf->info;

    Parcopy(syi, sxi);
    if (sxi[TC - 1] == INTEGER)
        n = Parcoerc(sxi, &myreal);

    n = Parlen(sxi);

    switch (sxi[TC - 1]) {
    case REAL:
    {
        float *rx;
        double d;
        PFD_AD df;

        rx = (float *) Paraddr(sxi);

        df = (PFD_AD) info[0];
        for (i = 0; i < n; i++) {
            d = (*df) (rx[i]);
            if (d > 0.0) {
                d = min(d, FLT_MAX);
                d = max(d, FLT_MIN);
            } else if (d < 0.0) {
                d = max(d, -FLT_MAX);
                d = min(d, -FLT_MIN);
            }
            rx[i] = d;
        }
    }
        break;

    case DOUBLE:
    {
        real8 *dx;
        PFD_AD df;

        dx = (real8 *) Paraddr(sxi);

        df = (PFD_AD) info[0];
        for (i = 0; i < n; i++)
            dx[i] = (*df) (dx[i]);
    }
        break;

/* NOTE: do all complex math in C functions and in order to
 *       make this efficient as possible have them do an array
 *       at a time
 */
    case COMPLEX:
    {
        dcomplex *dcx;
        Integer nt;
        PFDC_APDC dcf;

        nt = DOUBLE_C;
        Parcoerc(sxi, &nt);

        dcx = (dcomplex *) Paraddr(sxi);
        dcf = (PFDC_APDC) info[1];
        (*dcf) (dcx, dcx, (int) n);

        nt = COMPLEX;
        Parcoerc(sxi, &nt);
    }

        break;

    case DOUBLE_C:
    {
        dcomplex *dcx;
        PFDC_APDC dcf;

        dcx = (dcomplex *) Paraddr(sxi);
        dcf = (PFDC_APDC) info[1];
        (*dcf) (dcx, dcx, (int) n);
    }

        break;
    }

    return (OK);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* PAR_SRARG - single argument functions returning a double */

Integer Par_srarg(Integer *sxi, f_entry *thisf,
                  Integer nargs, Integer *syi)
{
    int i;
    Integer n;
    PFVoid *info;

    info = (PFVoid *) thisf->info;

    Parcopy(syi, sxi);
    if (sxi[TC - 1] == INTEGER)
        n = Parcoerc(sxi, &myreal);

    n = Parlen(sxi);

    switch (sxi[TC - 1]) {
    case REAL:
    {
        float *rx;
        PFD_AD df;

        rx = (float *) Paraddr(sxi);

        df = (PFD_AD) info[0];
        for (i = 0; i < n; i++)
            rx[i] = (*df) (rx[i]);
    }
        break;

    case DOUBLE:
    {
        real8 *dx;
        PFD_AD df;

        dx = (real8 *) Paraddr(sxi);

        df = (PFD_AD) info[0];
        for (i = 0; i < n; i++)
            dx[i] = (*df) (dx[i]);
    }
        break;

/* NOTE: do all complex math in C functions and in order to
 *       make this efficient as possible have them do an array
 *       at a time
 */
    case COMPLEX:
    {
        Integer nt;
        real8 *rx;
        dcomplex *dcx;
        PFD_APDC dcf;

        sxi[TC - 1] = DOUBLE;
        Parget(sxi);

        rx = (real8 *) Paraddr(sxi);

        nt = DOUBLE_C;
        Parcoerc(syi, &nt);

        dcx = (dcomplex *) Paraddr(syi);
        dcf = (PFD_APDC) info[1];
        (*dcf) (rx, dcx, (int) n);

        nt = REAL;
        Parcoerc(sxi, &nt);

        Parrel(syi);
    }
        break;

    case DOUBLE_C:
    {
        real8 *rx;
        dcomplex *dcx;
        PFD_APDC dcf;

        sxi[TC - 1] = DOUBLE;
        Parget(sxi);

        rx = (real8 *) Paraddr(sxi);
        dcx = (dcomplex *) Paraddr(syi);

        dcf = (PFD_APDC) info[1];
        (*dcf) (rx, dcx, (int) n);

        Parrel(syi);
    }
        break;
    }

    return (OK);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
