/*
 * ihash.c - integer hash tables
 *
 *  Fortran entries are:
 *  itable =  ihashnew(isize) -- create a Ihash table, return index
 *  ientry =  ihashfind(itable,key) -- does itable have entry with given key?
 *            ihashfind returns -1 if no such key, 0 iff key == 0
 *                      returns entry index
 *  ientry =  ihashput(itable,key,ivalue) -- put key,value into table
 *                  or replaces value at key; returns entry index
 *  ivalue = ihashget(itable, ientry) -- return value associated with entry 
 *
 */

#include "rt.h"
#include "ihash.h"

#define NIhash 2999
#define NTABLE 25

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

integer IhashFunction(key)
integer key;
{
    integer hv;

    hv = key % NIhash;

    return (hv);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

typedef struct node {
    integer key;
    integer value;
    integer next;
} Node;

typedef struct {
    integer h[NIhash];
    integer capacity, size;
    Node *nodes;
} IhashTable;

static IhashTable *Tables[NTABLE];
static unsigned integer ntable = 0;

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* create a Ihash table holding around asize elements
 * returns index into set of Ihash tables, zero-based
 */
integer IhashNew(integer asize)
{
    int i;
    integer size;
    IhashTable *newTable;

    size = ((asize < 50) ? 50 : asize);
    if (ntable >= NTABLE) {
        char errmsg[80];

        sprintf(errmsg, "Ihashnew: no more table space.");
        BasError(errmsg, strlen(errmsg));
        exit(1);
    }
#ifdef PCK_SCORE
    newTable = FMAKE(IhashTable, "PERM|IHASHNEW:newTable");
#else
    newTable = (IhashTable *) malloc(sizeof(IhashTable));
#endif

    for (i = 0; i < NIhash; i++) {
        newTable->h[i] = -1;
    }
    newTable->capacity = size;

#ifdef PCK_SCORE
    newTable->nodes =
        FMAKE_N(Node, newTable->capacity, "PERM|IHASHNEW:nodes");
#else
    newTable->nodes = (Node *) malloc(sizeof(Node) * (newTable->capacity));
#endif

    newTable->size = 0;
    Tables[ntable] = newTable;
    IhashPut(ntable, (integer) 0, (integer) 0);
    return (ntable++);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

static integer IhashNewNode(integer jtable, integer k, integer v)
{
    Node *oldNodes;
    int old_capacity, i;

    if (Tables[jtable]->size >= Tables[jtable]->capacity) {
        old_capacity = Tables[jtable]->capacity;
        oldNodes = Tables[jtable]->nodes;
        Tables[jtable]->capacity *= 2;

#ifdef PCK_SCORE
        Tables[jtable]->nodes =
            FMAKE_N(Node, Tables[jtable]->capacity, "IHASHNEWNODE:nodes");
#else
        Tables[jtable]->nodes =
            (Node *) malloc(sizeof(Node) * (Tables[jtable]->capacity));
#endif

        for (i = 0; i < old_capacity; i++) {
            Tables[jtable]->nodes[i] = oldNodes[i];
        }

#ifdef PCK_SCORE
        SFREE(oldNodes);
#else
        free(oldNodes);
#endif

    }
    Tables[jtable]->nodes[Tables[jtable]->size].key = k;
    Tables[jtable]->nodes[Tables[jtable]->size].value = v;
    Tables[jtable]->nodes[Tables[jtable]->size].next = -1;
    return (Tables[jtable]->size++);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void FreeNodes()
{
    int i;

    for (i = 0; i < ntable; i++) {
#   ifdef PCK_SCORE
        SFREE(Tables[i]->nodes);
#   else
        free(Tables[i]->nodes);
#   endif
    }
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

integer IhashFind(integer jtable, integer k)
{
    integer hvalue, kentry;
    Node *current;
    IhashTable *t;

    assert(jtable >= 0);
    assert(jtable < NTABLE);
    t = Tables[jtable];

    hvalue = IhashFunction(k);
    kentry = t->h[hvalue];
    while (kentry != -1) {
        current = &(t->nodes[kentry]);
        if (k == current->key) {
            return (kentry);
        }
        kentry = current->next;
    }
    return (kentry);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* add a new key, value pair, or change the value
 */

integer IhashPut(integer jtable, integer k, integer v)
{
    integer hvalue, kentry, kprevious;
    Node *current;

    assert(jtable >= 0);
    assert(jtable < NTABLE);

    hvalue = IhashFunction(k);
    kentry = Tables[jtable]->h[hvalue];
    if (kentry == -1) {
        kentry = IhashNewNode(jtable, k, v);
        Tables[jtable]->h[hvalue] = kentry;
        return (kentry);
    }
    while (kentry != -1) {
        current = &(Tables[jtable]->nodes[kentry]);
        if (k == current->key) {
            current->value = v;
            return (kentry);
        }
        kprevious = kentry;
        kentry = current->next;
    }
    kentry = IhashNewNode(jtable, k, v);
    Tables[jtable]->nodes[kprevious].next = kentry;
    return (kentry);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

integer IhashGet(integer jtable, integer jentry)
{
    assert(jtable >= 0);
    assert(jtable < NTABLE);
    assert(jentry >= 0);
    assert(jentry < Tables[jtable]->size);
    return ((Tables[jtable]->nodes[jentry]).value);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
