/*==================================================================*/
/*                                                                  */
/* basis.h                                                          */
/*                                                                  */
/* Header file containing useful routines, structures, and          */
/* constants for interfacing C programs with the Basis system.      */
/*                                                                  */
/* Written by: Scott W. Haney and James A. Crotinger                */
/* Created:    04/06/89                                             */
/* Expanded by Peter A. Willmann      8/3/89                        */
/*                                                                  */
/*------------------------------------------------------------------*/
/*                                                                  */
/* Revision History                                                 */
/*                                                                  */
/*   04/06/89  - File created.                                      */
/*   04/15/89  - Plotting glue added.                               */
/*   07/06/89  - Routines made static so header can be used in more */
/*               than one source file.  Also, new fcd routines      */
/*               used.                                              */
/*   07/07/89  - Made some GRAFL-III routines return longs in       */
/*               accordance with the BASIS manual.                  */
/*   07/09/89  - Added glue for plabel.                             */
/*   09/05/89  - Added additional data definitions; merged with     */
/*               fcd.h.                                             */
/*                                                                  */
/*==================================================================*/
 
/********************************************************************/
/*   Conditional Includes                                           */
/********************************************************************/
 
#ifndef __BASIS_H__
 
   #define __BASIS_H__
 
 
/*==================================================================*/
/*                  F C D  D E F I N I T I O N S                    */
/*==================================================================*/
 
/*------------------------------------------------------------------*/
/*                                                                  */
/*  Description of data types:                                      */
/*                                                                  */
/*    fcd        - Fortran Character Descriptor structure           */
/*    chptr      - C character pointer structure                    */
/*                                                                  */
/*  Description of conversion routines:                             */
/*                                                                  */
/*    fcd CtoFStr(char *)  - takes a pointer to a NULL terminated   */
/*                           C string, and returns an fcd pointing  */
/*                           to the same data location.             */
/*                                                                  */
/*    char * F0toCStr(fcd) - takes an fcd pointing to a NULL ter-   */
/*                           minated Fortran string, and returns    */
/*                           a pointer to the same data location.   */
/*                                                                  */
/*    char * FtoCStr(fcd)  - takes an fcd and returns a pointer     */
/*                           to a NULL terminated C string which    */
/*                           contains a COPY of the data pointed    */
/*                           to by the fcd.                         */
/*                                                                  */
/*    NOTE: The pointer returned by FtoCStr() should be considered  */
/*          to be a constant pointer, and it should be free()'d     */
/*          when you are finished with it.                          */
/*                                                                  */
/*------------------------------------------------------------------*/
 
   #define __FCD_H__
 
   #define _PROTOTYPES
 
   #ifdef PACT
   # include <score.h>
   #else
   # include <assert.h>
   # include <string.h>
   # include <stdlib.h>
   #endif

   #if defined(CRAY2)
       struct __fcd {
           unsigned long offset :  6;    /* in bits! */
           unsigned long length : 26;    /* in bits! */
           unsigned long waddr  : 32;
       };
   #elif defined(CRAY1) || defined(CRAYXMP)
       struct __fcd {
           unsigned long unused : 17;
           unsigned long length : 17;    /* in bits! */
           unsigned long offset :  6;    /* in bits! */
           unsigned long waddr  : 24;
       };
   #else
       #error "Host machine not specified: use -DCRAY1, etc!"
   #endif
 
   struct __charptr {
       unsigned long byteoffset :  3;
       unsigned long unused     : 29;
       unsigned long waddr      : 32;
       };
 
   typedef union {
       char *contents;
       struct __fcd fields;
       } fcd;
 
   typedef union {
       char *contents;
       struct __charptr fields;
       } chptr;
 
 
   static fcd CtoFStr(register char *cstr)
       {
       register fcd retFCD;
       register chptr temp;
 
       temp.contents = cstr;
 
       retFCD.fields.waddr  = temp.fields.waddr;
       retFCD.fields.offset = (temp.fields.byteoffset << 3);
       retFCD.fields.length = (strlen(cstr) << 3);
 
       return (retFCD);
       }
 
   static char *F0toCStr(register fcd fstr)
       {
       register chptr retChptr;
 
       retChptr.fields.waddr      = fstr.fields.waddr;
       retChptr.fields.byteoffset = (fstr.fields.offset >> 3);
 
   /*--------------------------------------------------------------*/
   /* Test to make sure that the fstr really was NULL terminated.  */
   /* Unfortunately, if main() was a Fortran program, printf()     */
   /* won't work, and this just causes the program to die with an  */
   /* abort(). To disable the test, compile with -DNDEBUG          */
   /*--------------------------------------------------------------*/
 
       assert(retChptr.contents[fstr.fields.length >> 3] == '\0');
 
       return (retChptr.contents);
       }
 
   static char *FtoCStr(register fcd fstr)
       {
       register char *retChar;
       register chptr temp;
       register long len;
 
       temp.fields.waddr      = fstr.fields.waddr;
       temp.fields.byteoffset = (fstr.fields.offset >> 3);
       len                    = (fstr.fields.length >> 3);

#ifdef PCK_SCORE
       retChar = FMAKE_N(char, len + 1, "FTOCSTR:retChar");
#else
       retChar = (char *) malloc(len + 1);
#endif

       retChar[len] = '\0';
       strncpy(retChar, temp.contents, len);

       return (retChar);
       }
 
   static char *FtoCPtr(register fcd fstr)
       {
       register chptr retChptr;
 
       retChptr.fields.waddr      = fstr.fields.waddr;
       retChptr.fields.byteoffset = (fstr.fields.offset >> 3);
 
       return (retChptr.contents);
       }
 
   /*--------------------------------------------------------------*/
   /* The above is only necessary on the 1 or XMP. On the Cray 2,  */
   /* we simply do the following macro. This is not very general   */
   /* though. If you need the above on the Cray 2, just compile    */
   /* insert "#undef FtoCPtr" after you #include this file.        */
   /*--------------------------------------------------------------*/
 
   #ifdef CRAY2
       #define FtoCPtr(x) x.contents
   #endif
 
 
/*==================================================================*/
/*                B A S I S  D E F I N I T I O N S                  */
/*==================================================================*/
 
/********************************************************************/
/*   data definitions                                               */
/********************************************************************/
 
  typedef struct {
     float real, imag;
     } complex;
 
  typedef int Filedes;
  typedef char Filename[8];
  typedef char Varname[VARNAMESIZE];
  typedef float doub[2];
 
 
/********************************************************************/
/*   Useful Constants                                               */
/********************************************************************/
 
   #define YES   1
   #define NO    0
   #define OK    0
   #define ERR   -1
   #define DONE  1
 
 
/********************************************************************/
/*   Interface to Basis Routines                                    */
/********************************************************************/
 
/*------------------------------------------------------------------*/
/* Dynamic Data Allocation                                          */
/*------------------------------------------------------------------*/
 
   fortran long ALLOT(fcd fcdval, long *len);
   static long allot(char *name, long length)
      {
      return (ALLOT(CtoFStr(name), &length));
      }
 
   fortran long FREE(fcd fcdval);
   static long bfree(char *name)
      {
      return (FREE(CtoFStr(name)));
      }
 
   fortran long CHANGE(fcd fcdval, long *newlen);
   static long change(char *name, long *newlen)
      {
      return (CHANGE(CtoFStr(name), &newlen));
      }
 
   fortran long GALLOT(fcd fcdval, long *length);
   static long gallot(char *name, long length)
      {
      return (GALLOT(CtoFStr(name), &length));
      }
 
   fortran long GFREE(fcd fcdval);
   static long gfree(char *name)
      {
      return (GFREE(CtoFStr(name)));
      }
 
   fortran long GCHANGE(fcd fcdval, long *newlen);
   static long gchange(char *name, long *newlen)
      {
      return (GCHANGE(CtoFStr(name), &newlen));
      }
 
 
/*------------------------------------------------------------------*/
/* I/O                                                              */
/*------------------------------------------------------------------*/
 
   fortran void REMARK(fcd fcdstr);
   #define remark(str) REMARK(CtoFStr(str))
 
 
/*------------------------------------------------------------------*/
/* Executing the Parser                                             */
/*------------------------------------------------------------------*/
 
   fortran void PARSESTR(fcd fcds);
   #define parsestr(s) PARSESTR(CtoFStr(s))
 
 
/*------------------------------------------------------------------*/
/* Executing a user function                                        */
/*------------------------------------------------------------------*/
 
   fortran void EXECUSER(fcd fcdf);
   #define execuser(f) EXECUSER(CtoFStr(f))
 
 
/*------------------------------------------------------------------*/
/* Error Recovery                                                   */
/*------------------------------------------------------------------*/
 
   fortran void KABOOM(long *iflag);
   static void kaboom(long iflag)
      {
      KABOOM(&iflag);
      }
 
 
/*------------------------------------------------------------------*/
/* Editing                                                          */
/*------------------------------------------------------------------*/
 
   fortran long EDIT(long *unit, fcd fcdname);
   static long edit(long unit, char *name)
      {
      return (EDIT(&unit, CtoFStr(name)));
      }
 
 
/*------------------------------------------------------------------*/
/* Plotting (GRAFL-III interface)                                   */
/*------------------------------------------------------------------*/
 
   fortran void PLTSTART(void);
   #define pltstart PLTSTART
 
   fortran void PBOX(void);
   #define pbox PBOX
 
   fortran void PEMPTY(void);
   #define pempty PEMPTY
 
   fortran void PFRAME(void);
   #define pframe PFRAME
 
   fortran void PCHRSIZE(long *nchar);
   #define pchrsize(nchar) PCHRSIZE(CtoFStr(nchar))
 
   fortran void PCLIP(fcd fcdcclip);
   #define pclip(cclip) PCLIP(CtoFStr(cclip))
 
   fortran void PCLIPT(fcd fcdiclipt);
   #define pclipt(iclipt) PCLIPT(CtoFStr(iclipt))
 
   fortran void PCOLOR(fcd fcdicolor);
   #define pcolor(icolor) PCOLOR(CtoFStr(icolor))
 
   fortran void PFONT(fcd fcdifont);
   #define pfont(ifont) PFONT(CtoFStr(ifont))
 
   fortran void PFRAMC(fcd fcdiframc);
   #define pframc(iframc) PFRAMC(CtoFStr(iframc))
 
   fortran void PGRID(long *igridx, long *igridy);
   static void pgrid(long igridx, long igridy)
      {
      PGRID(&igridx, &igridy);
      }
 
   fortran void PLABEL(fcd fcdilabel);
   #define plabel(ilabel) PLABEL(CtoFStr(ilabel))
 
   fortran void PLINE(float *x1, float *x2, float *y1, float *y2);
   static void pline(float x1, float x2, float y1, float y2)
      {
      PLINE(&x1, &x2, &y1, &y2);
      }
 
   fortran void PLINEWID(float *width);
   static void plinewid(float width)
      {
      PLINEWID(&width);
      }
 
   fortran void PSETCH(float *x, float *y, long *nchar, float *arg);
   static void psetch(float x, float y, long nchar, float arg)
      {
      PSETCH(&x, &y, &nchar, &arg);
      }
 
   fortran void PSTRID(long *isxp, long *isyp);
   static void pstrid(long isxp, long isyp)
      {
      PSTRID(&isxp, &isyp);
      }
 
   fortran void PSTYLE(fcd fcdistyle);
   #define pstyle(istyle) PSTYLE(CtoFStr(istyle))
 
   fortran void PTEXT(fcd fcdstr);
   #define ptext(str) PTEXT(CtoFStr(str))
 
   fortran void PTICKS(long *nticksx, long *nticksy);
   static void pticks(long nticksx, long nticksy)
      {
      PTICKS(&nticksx, &nticksy);
      }
 
   fortran void PTICKSR(fcd fcdcticksr);
   #define pticksr(cticksr) PTICKSR(CtoFStr(cticksr))
 
   fortran void PTITLE(fcd fcdit, fcd fcdix, fcd fcdiy, fcd fcdir);
   static void ptitle(char *it, char *ix, char *iy, char *ir)
      {
      PTITLE(CtoFStr(it), CtoFStr(ix), CtoFStr(iy), CtoFStr(ir));
      }
 
   fortran void PTRIPLE(fcd fcdifont1, fcd fcdifont2, fcd fcdifont3);
   static void ptriple(char *ifont1, char *ifont2, char *ifont3)
      {
      PTRIPLE(CtoFStr(ifont1), CtoFStr(ifont2), CtoFStr(ifont3));
      }
 
   fortran void P1234(long *iquad);
   static void p1234(long iquad)
      {
      P1234(&iquad);
      }
 
   fortran float PCURVE(fcd fcdll, float *x, float *y, long *n, long *iquad,
                       fcd fcdit, fcd fcdix, fcd fcdiy,
                       fcd fcdir, fcd fcdlab);
   static float pcurve(char *ll, float *x, float *y, long n, long iquad,
               char *it, char *ix, char *iy, char *ir, char *lab)
      {
      float okif0;
 
      okif0 = PCURVE(CtoFStr(ll), x, y, &n, &iquad, CtoFStr(it),
              CtoFStr(ix), CtoFStr(iy), CtoFStr(ir), CtoFStr(lab));
      return (okif0);
      }
 
   fortran float PLOT3D(fcd fcdll, float *phi, float *x, float *y, float *z,
            float *xl, float *yl, float *zl, long *nd, long *nx, long *ny,
            long *iquad, fcd fcdit, fcd fcdix, fcd fcdiy, fcd fcdir);
   static float plot3d(char *ll, float phi, float *x, float *y, float *z,
            float *xl, float *yl, float *zl, long nd, long nx, long ny,
            long iquad, char *it, char *ix, char *iy, char *ir)
      {
      float okif0;
      okif0 = PLOT3D(CtoFStr(ll), &phi, x, y, z,
              xl, yl, zl, &nd, &nx, &ny, &iquad,
              CtoFStr(it), CtoFStr(ix), CtoFStr(iy), CtoFStr(ir));
      return (okif0);
      }
 
   fortran float PMARKR(fcd fcdll, float *x, float *y, long *n, long *iquad,
                       fcd fcdit, fcd fcdix, fcd fcdiy,
                       fcd fcdir, fcd fcdmarker, float *size);
   static float pmarkr(char *ll, float *x, float *y, long n, long iquad,
         char *it, char *ix, char *iy, char *ir, char *marker, float size)
      {
      float okif0;
      okif0 = PMARKR(CtoFStr(ll), x, y, &n, &iquad, CtoFStr(it),
           CtoFStr(ix), CtoFStr(iy), CtoFStr(ir), CtoFStr(marker), &size);
      return (okif0);
      }
 
   fortran float PSCALE(fcd fcdll, float *xmin, float *xmax,
                       float *ymin, float *ymax, long *iquad);
   static float pscale(char *ll, float xmin, float xmax, float ymin, float ymax,
               long iquad)
      {
      float okif0;
 
      okif0 = PSCALE(CtoFStr(ll), &xmin, &xmax, &ymin, &ymax, &iquad);
      return (okif0);
      }
 
   fortran float PCONRD(fcd fcdll, float *x, float *y, float *z,
                      long *nd, float *alev,
                      long *ilev, long *ws, long *iquad,
                      fcd fcdit, fcd fcdix, fcd fcdiy, fcd fcdir);
   static float pconrd(char *ll, float *x, float *y, float *z,
               long n, float *alev, long ilev, long *ws, long iquad,
               char *it, char *ix, char *iy, char *ir)
      {
      float okif0;
      okif0 = PCONRD(CtoFStr(ll), x, y, z, &n, alev, &ilev, ws, &iquad,
              CtoFStr(it), CtoFStr(ix), CtoFStr(iy), CtoFStr(ir));
      return (okif0);
      }
 
   fortran float PCONRM(fcd fcdll, float *x, float *y, float *z,
                      long *nd, long *nx, long *ny, float *alev,
                      long *ilev, long *iquad,
                      fcd fcdit, fcd fcdix, fcd fcdiy, fcd fcdir);
   static float pconrm(char *ll, float *x, float *y, float *z,
               long nd, long nx, long ny, float *alev, long ilev, long iquad,
               char *it, char *ix, char *iy, char *ir)
      {
      float okif0;
      okif0 = PCONRM(CtoFStr(ll), x, y, z, &nd, &nx, &ny, alev, &ilev, &iquad,
              CtoFStr(it), CtoFStr(ix), CtoFStr(iy), CtoFStr(ir));
      return (okif0);
      }
 
   fortran float PCONT(fcd fcdll, float *x, float *y, float *z,
                      long *nd, long *nx, long *ny, float *alev,
                      long *ilev, long *iquad,
                      fcd fcdit, fcd fcdix, fcd fcdiy, fcd fcdir);
   static float pcont(char *ll, float *x, float *y, float *z,
               long nd, long nx, long ny, float *alev, long ilev, long iquad,
               char *it, char *ix, char *iy, char *ir)
      {
      float okif0;
      okif0 = PCONT(CtoFStr(ll), x, y, z, &nd, &nx, &ny, alev, &ilev, &iquad,
              CtoFStr(it), CtoFStr(ix), CtoFStr(iy), CtoFStr(ir));
      return (okif0);
      }
 
   fortran float PSHADE(fcd fcdll, float *x, float *y, long *n, long *iquad,
                       fcd fcdit, fcd fcdix, fcd fcdiy,
                       fcd fcdir, fcd fcdshade);
   static float pshade(char *ll, float *x, float *y, long n, long iquad,
               char *it, char *ix, char *iy, char *ir, char *shade)
      {
      float okif0;
      okif0 = PSHADE(CtoFStr(ll), x, y, &n, &iquad, CtoFStr(it),
              CtoFStr(ix), CtoFStr(iy), CtoFStr(ir), CtoFStr(shade));
      return (okif0);
      }
 
   fortran float PVEC(fcd fcdll, float *x, float *y, float *vx, float *vy,
          long *n, long *iquad, fcd fcdit, fcd fcdix, fcd fcdiy, fcd fcdir);
   static float pvec(char *ll, float *x, float *y, float *vx, float *vy,
         long n, long iquad, char *it, char *ix, char *iy, char *ir)
      {
      float okif0;
 
      okif0 = PVEC(CtoFStr(ll), x, y, vx, vy, &n, &iquad, CtoFStr(it),
              CtoFStr(ix), CtoFStr(iy), CtoFStr(ir));
      return (okif0);
      }
 
 
 
#endif /* __BASIS_H__ */
