/*
 * MATHF.C - math functions that do more what we need
 *
 */

#include "cfcn.h"

#define MEXP_CMP(cx, sx, ey, eyi, dc)                                       \
    {double x, y;                                                           \
     x   = dc.r;                                                            \
     y   = dc.i;                                                            \
     ey  = exp(y);                                                          \
     eyi = 1.0/ey;                                                          \
     cx  = cos(x);                                                          \
     sx  = sin(x);}

#define MEXP_CMPH(ex, exi, cy, sy, dc)                                      \
    {double x, y;                                                           \
     x   = dc.r;                                                            \
     y   = dc.i;                                                            \
     ex  = exp(x);                                                          \
     exi = 1.0/ex;                                                          \
     cy  = cos(y);                                                          \
     sy  = sin(y);}

#define MEXP(rc, dc)                                                        \
    {double x, y, a, c, s;                                                  \
     x  = dc.r;                                                             \
     y  = dc.i;                                                             \
     a  = exp(x);                                                           \
     c  = cos(y);                                                           \
     s  = sin(y);                                                           \
     rc.r = a*c;                                                            \
     rc.i = a*s;}

#define MLOG(rc, dc)                                                        \
    {double x, y, a, b;                                                     \
     x = dc.r;                                                              \
     y = dc.i;                                                              \
     a = log( HYPOT(x, y) );                                                \
     b = atan2(y, x);                                                       \
     rc.r = a;                                                              \
     rc.i = b;}

#define MULT_RC(rc, a)                                                      \
    {rc.r *= a;                                                             \
     rc.i *= a;}

#define MDIV(rc, ac, bc)                                                    \
    {double ar, ai, br, bi, u, brovu, biovu, d;                             \
     ar = ac.r;                                                             \
     ai = ac.i;                                                             \
     br = bc.r;                                                             \
     bi = bc.i;                                                             \
     u  = max(br,bi);                                                       \
     brovu = br/u;                                                          \
     biovu = bi/u;                                                          \
     d  = 1.0/(br*brovu + bi*biovu);                                        \
     rc.r = d*(ar*brovu + ai*biovu);                                        \
     rc.i = d*(ai*brovu - ar*biovu);}

extern double acosh(double x);

static double ln10e = 0.0;

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MASQRT - sqrt routine */

double masqrt(double x)
{
    double r;

    if (x < 0.0) {
        BasError("sqrt: negative argument.", 24);
        r = 0.0;
    } else
        r = sqrt(x);

    return (r);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MALOG - log routine */

double malog(double x)
{
    double r;

    if (x <= 0.0)
        r = logof0;
    else
        r = log(x);

    return (r);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MALOG10 - log10 routine */

double malog10(double x)
{
    double r;

    if (x <= 0.0)
        r = logof0;
    else
        r = log10(x);

    return (r);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MACOT - cot routine */

double macot(double x)
{
    double r;

    r = tan(x);

    return ((r == 0.0) ? HUGE : 1.0 / r);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MACOTH - coth routine */

double macoth(double x)
{
    double r;

    r = tanh(x);

    return ((r == 0.0) ? HUGE : 1.0 / r);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MAACOS - acos routine */

double maacos(double x)
{
    double o, r, rs, rb;

    o = 1.0;
    if (fabs(x) > 1.000001) {
        BasError("acos: abs(argument) > 1.000001", 30);
        return (-HUGE);
    }

    rs = min(1.0, fabs(x));
    rb = SIGNF(o, x);
    r = acos(rb * rs);

    return (r);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MAASIN - asin routine */

double maasin(double x)
{
    double o, r, rs, rb;

    o = 1.0;
    if (fabs(x) > 1.000001) {
        BasError("asin: abs(argument) > 1.000001", 30);
        return (-HUGE);
    }

    rs = min(1.0, fabs(x));
    rb = SIGNF(o, x);
    r = asin(rb * rs);

    return (r);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MAAINT - aint routine (largest integer less than x) */

double maaint(double x)
{
    double r;

    r = floor(x);

    return (r);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MAANINT - anint routine (round to nearest integer) */

double maanint(double x)
{
    double r;

    r = floor(x + 0.5);

    return (r);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MASIGN - sign routine */

double masign(double x, double y)
{
    double r;

    r = SIGNF(x, y);

    return (r);
}

/*--------------------------------------------------------------------------*/

/*                           DOUBLE COMPLEX FUNCTIONS                       */

/*--------------------------------------------------------------------------*/

/* MADCRANF - double complex ranf function */

void madcranf(dcomplex *c, int n)
{
    int i;
    double x;
    dcomplex dc;

    for (i = 0; i < n; i++) {
        x = 2.0 * Pi * Ranf();
        dc.r = cos(x);
        dc.i = sin(x);
        c[i] = dc;
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MADCABS - double complex abs function */

void madcabs(double *r, dcomplex *c, int n)
{
    int i;
    double rc, x, y;
    dcomplex dc;

    for (i = 0; i < n; i++) {
        dc = c[i];

        x = dc.r;
        y = dc.i;
        rc = HYPOT(x, y);

        r[i] = rc;
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MADCAINT - double complex aint function
 *          - interpreted to mean the greatest integer less than the absolute
 *          - value (i.e. magnitude) of the complex number
 */

void madcaint(double *r, dcomplex *c, int n)
{
    int i;
    double rc, x, y;
    dcomplex dc;

    for (i = 0; i < n; i++) {
        dc = c[i];

        x = dc.r;
        y = dc.i;
        rc = HYPOT(x, y);

        r[i] = floor(rc);
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MADCANINT - double complex anint function
 *           - interpreted to mean the nearest integer to the absolute
 *           - value (i.e. magnitude) of the complex number
 */

void madcanint(double *r, dcomplex *c, int n)
{
    int i;
    double rc, x, y;
    dcomplex dc;

    for (i = 0; i < n; i++) {
        dc = c[i];

        x = dc.r;
        y = dc.i;
        rc = HYPOT(x, y);

        r[i] = floor(rc + 0.5);
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MADCEXP - double complex exp function */

void madcexp(dcomplex *r, dcomplex *c, int n)
{
    int i;
    dcomplex rc, dc;

    for (i = 0; i < n; i++) {
        dc = c[i];

        MEXP(rc, dc);

        r[i] = rc;
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MADCLOG - double complex log function */

void madclog(dcomplex *r, dcomplex *c, int n)
{
    int i;
    dcomplex rc, dc;

    for (i = 0; i < n; i++) {
        dc = c[i];

        MLOG(rc, dc);

        r[i] = rc;
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MADCLOG10 - double complex log10 function */

void madclog10(dcomplex *r, dcomplex *c, int n)
{
    int i;
    dcomplex rc, dc;

    if (ln10e == 0.0)
        ln10e = log10(exp(1.0));

    for (i = 0; i < n; i++) {
        dc = c[i];

        MLOG(rc, dc);

        MULT_RC(rc, ln10e);

        r[i] = rc;
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MADCSQRT - double complex sqrt function */

void madcsqrt(dcomplex *r, dcomplex *c, int n)
{
    int i;
    dcomplex rc, dc;
    double x, y, a, b, cb;

    for (i = 0; i < n; i++) {
        dc = c[i];

        x = dc.r;
        y = dc.i;
        a = sqrt(HYPOT(x, y));
        b = atan2(y, x);
        cb = cos(b);

        rc.r = a * sqrt(0.5 * (1.0 + cb));
        rc.i = a * sqrt(0.5 * (1.0 - cb));

        r[i] = rc;
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MADCSIN - double complex sin function */

void madcsin(dcomplex *r, dcomplex *c, int n)
{
    int i;
    dcomplex rc, dc;
    double cx, sx, ey, eyi;

    for (i = 0; i < n; i++) {
        dc = c[i];

        MEXP_CMP(cx, sx, ey, eyi, dc);

        rc.r = 0.5 * sx * (ey + eyi);
        rc.i = 0.5 * cx * (ey - eyi);

        r[i] = rc;
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MADCCOS - double complex cos function */

void madccos(dcomplex *r, dcomplex *c, int n)
{
    int i;
    dcomplex rc, dc;
    double cx, sx, ey, eyi;

    for (i = 0; i < n; i++) {
        dc = c[i];

        MEXP_CMP(cx, sx, ey, eyi, dc);

        rc.r = 0.5 * cx * (ey + eyi);
        rc.i = 0.5 * sx * (eyi - ey);

        r[i] = rc;
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MADCTAN - double complex tan function */

void madctan(dcomplex *r, dcomplex *c, int n)
{
    int i;
    dcomplex rc, sc, cc, dc;
    double cx, sx, ey, eyi;

    for (i = 0; i < n; i++) {
        dc = c[i];

        MEXP_CMP(cx, sx, ey, eyi, dc);

        sc.r = sx * (ey + eyi);
        sc.i = cx * (ey - eyi);

        cc.r = cx * (ey + eyi);
        cc.i = sx * (eyi - ey);

        MDIV(rc, sc, cc);

        r[i] = rc;
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MADCCOT - double complex cot function */

void madccot(dcomplex *r, dcomplex *c, int n)
{
    int i;
    dcomplex rc, sc, cc, dc;
    double cx, sx, ey, eyi;

    for (i = 0; i < n; i++) {
        dc = c[i];

        MEXP_CMP(cx, sx, ey, eyi, dc);

        sc.r = sx * (ey + eyi);
        sc.i = cx * (ey - eyi);

        cc.r = cx * (ey + eyi);
        cc.i = sx * (eyi - ey);

        MDIV(rc, cc, sc);

        r[i] = rc;
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MADCSINH - double complex sinh function */

void madcsinh(dcomplex *r, dcomplex *c, int n)
{
    int i;
    dcomplex rc, dc;
    double ex, exi, cy, sy;

    for (i = 0; i < n; i++) {
        dc = c[i];

        MEXP_CMPH(ex, exi, cy, sy, dc);

        rc.r = 0.5 * (ex - exi) * cy;
        rc.i = 0.5 * (ex + exi) * sy;

        r[i] = rc;
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MADCCOSH - double complex cosh function */

void madccosh(dcomplex *r, dcomplex *c, int n)
{
    int i;
    dcomplex rc, dc;
    double ex, exi, cy, sy;

    for (i = 0; i < n; i++) {
        dc = c[i];

        MEXP_CMPH(ex, exi, cy, sy, dc);

        rc.r = 0.5 * (ex + exi) * cy;
        rc.i = 0.5 * (ex - exi) * sy;

        r[i] = rc;
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MADCTANH - double complex tanh function */

void madctanh(dcomplex *r, dcomplex *c, int n)
{
    int i;
    dcomplex rc, sc, cc, dc;
    double ex, exi, cy, sy;

    for (i = 0; i < n; i++) {
        dc = c[i];

        MEXP_CMPH(ex, exi, cy, sy, dc);

        sc.r = (ex - exi) * cy;
        sc.i = (ex + exi) * sy;

        cc.r = (ex + exi) * cy;
        cc.i = (ex - exi) * sy;

        MDIV(rc, sc, cc);

        r[i] = rc;
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MADCCOTH - double complex coth function */

void madccoth(dcomplex *r, dcomplex *c, int n)
{
    int i;
    dcomplex rc, sc, cc, dc;
    double ex, exi, cy, sy;

    for (i = 0; i < n; i++) {
        dc = c[i];

        MEXP_CMPH(ex, exi, cy, sy, dc);

        sc.r = (ex - exi) * cy;
        sc.i = (ex + exi) * sy;

        cc.r = (ex + exi) * cy;
        cc.i = (ex - exi) * sy;

        MDIV(rc, cc, sc);

        r[i] = rc;
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MADCASIN - double complex asin function */

void madcasin(dcomplex *r, dcomplex *c, int n)
{
    int i;
    dcomplex rc, dc;
    double x, y, xp, xm, pq, as, ac;

    for (i = 0; i < n; i++) {
        dc = c[i];

        x = dc.r;
        y = dc.i;

        xp = 1.0 + x;
        xm = 1.0 - x;
        pq = 0.5 * (HYPOT(xp, y) + HYPOT(xm, y));
        as = asin(x / pq);
        ac = acosh(pq);

        rc.r = as;
        rc.i = (y > 0.0) ? ac : -ac;

        r[i] = rc;
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MADCACOS - double complex acos function */

void madcacos(dcomplex *r, dcomplex *c, int n)
{
    int i;
    dcomplex rc, dc;
    double x, y, xp, xm, pq, as, ac;

    for (i = 0; i < n; i++) {
        dc = c[i];

        x = dc.r;
        y = dc.i;

        xp = 1.0 + x;
        xm = 1.0 - x;
        pq = 0.5 * (HYPOT(xp, y) + HYPOT(xm, y));
        as = acos(x / pq);
        ac = acosh(pq);

        rc.r = as;
        rc.i = (y > 0.0) ? -ac : ac;

        r[i] = rc;
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* MADCATAN - double complex atan function */

void madcatan(dcomplex *r, dcomplex *c, int n)
{
    int i;
    dcomplex rc, dc;
    double x, y, xs, yp, ym;

    for (i = 0; i < n; i++) {
        dc = c[i];

        x = dc.r;
        y = dc.i;

        xs = x * x;
        yp = 1.0 + y;
        ym = 1.0 - y;

        rc.r = 0.5 * atan2(2.0 * x, 1.0 - xs - y * y);
        rc.i = 0.25 * log((yp * yp + xs) / (ym * ym + xs));

        r[i] = rc;
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
