/*
 * PARMOVE.C
 *
 */

#include "rt.h"

#define OK 0
#define ERR -1

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Returns the number of bytes taken up by a particular typecode */
Integer Utbytesin(Integer tc)
{
    static char msg1[] = "Utbytesin: don't know the size of a function.";
    char msg[80];

    switch (tc) {
    case INTEGER:
    case REAL:
    case LOGICAL:
    case STRUCTURE:
        return (sizeof(Integer));
    case COMPLEX:
    case DOUBLE:
        return (sizeof(double));
    case DOUBLE_C:
#     if defined (C90)|defined(CRAY2)|defined(GENERIC)|defined(J90)|defined(MAC)\
        |defined(MACOS)|defined(NLTSS)|defined(UNICOS)|defined(XMP)|defined(YMP)
        return (sizeof(double));
#     else
        return (2 * sizeof(double));
#     endif
    case RANGE:
        return (3 * sizeof(Integer));
    case NULL_TYPE:
        return (0);
    case FWA:
        return (sizeof(Address));
    case FUNCTION_TYPE:
        BasError(msg1, strlen(msg1));
        Kaboom(0);
    default:
        if (tc < 0)
            return (-tc);
        sprintf(msg, "Utbytesin: unknown type: %ld.", (long) tc);
        BasError(msg, strlen(msg));
        Kaboom(0);
    }
    return (-1);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* ParMove moves an object from src to dest. The destination memory
 * is always consecutive. The parameters are as follows:
 *     tc: the type of the entries in the object
 *     ndim: the number of dimensions. Both src and dest have the same
 *         number of dimensions.
 *     ilow, ihi: arrays ndim long; entries contain the low and high
 *         subscripts of each dimension of src (in the FORTRAN sense).
 *     icol: basically just (ihi - ilow) + 1.
 *     sl, sh: arrays ndim long; these arrays specify the subarray
 *         of src to be assigned to dest. In the case of a range, these
 *         will be the low and high subscripts of the range. In the
 *         case of a 1d array of subscripts, sl will be DEFAULT + 1
 *         and sh will be the subscript into the bassubs array of
 *         stack elemnts describing the 1d arrays.
 *     si: in the range case, specifies a step between sl and sh.
 *         in the array case, it's just 1.
 *     subarr, lsubarr: if subscript number i is a 1d array, then
 *         subarr [i] is the address of the array and lsubarr [i]
 *         is its length. if subscript number i is a range, then
 *         subarr [i] and lsubarr [i] will be zero.
 * src and dest are guaranteed to be the same size and shape.
 */
Integer Parmove(Address dest, Address src, Integer tc, Integer ndim,
                Integer *ilow, Integer *ihi, Integer *icol,
                Integer *sl, Integer *sh, Integer *si,
                int **subarr, Integer *lsubarr)
{
    int nbytes;                 /* The number of bytes in a single item */
    int i;
    int i0, i1, i2, i3, i4, i5, i6;
    int j0, j1, j2, j3, j4, j5, j6;
    int upper[7], incr[7], size[7], istart[7];
    int lower[7];
    int hasarray;
    Address src0;

    nbytes = Utbytesin(tc);

    /* simple, scalar case first */
    if (ndim == 0) {
        memcpy(dest, src, nbytes);
        return (OK);
    }

    hasarray = 0;

    /* Check for an array subscript */
    for (i = 0; i < 7; i++)
        if (subarr[i] != 0) {
            hasarray = 1;
            break;
        }

    size[0] = nbytes;
    for (i = 0; i < ndim; i++) {
        if (icol[i] == 0)
            /* missing dimension--object has no size, nothing to move. */
            return OK;
        if (i > 0)
            size[i] = size[i - 1] * icol[i - 1];
        if (lsubarr[i] == 0) {
            lower[i] = (sl[i] - ilow[i]) * size[i];
            upper[i] = (sh[i] - ilow[i]) * size[i];
            incr[i] = si[i] * size[i];
        } else {
            upper[i] = lsubarr[i] - 1;
            incr[i] = 1;
        }
    }

    if (!hasarray) {
        switch (ndim) {
        case 1:
            if (si[0] == 1) {
                /* Just a consecutive copy */
                memcpy(dest, (char *) src + lower[0],
                       (sh[0] - sl[0] + 1) * nbytes);
            } else {
                for (src0 = (char *) src + lower[0];
                     incr[0] > 0 ?
                     src0 <= (void *) ((char *) src + upper[0]) :
                     src0 >= (void *) ((char *) src + upper[0]);
                     src0 = (void *) ((char *) src0 + incr[0]),
                     dest = (void *) ((char *) dest + nbytes))
                    memcpy(dest, src0, nbytes);
            }
            break;
        case 2:
            for (j1 = lower[1];
                 incr[1] < 0 ? j1 >= upper[1] : j1 <= upper[1];
                 j1 += incr[1])
                for (j0 = lower[0];
                     incr[0] > 0 ? j0 <= upper[0] : j0 >= upper[0];
                     j0 += incr[0], dest =
                     (void *) ((char *) dest + nbytes))
                    memcpy(dest, (char *) src + j0 + j1, nbytes);
            break;
        case 3:
            for (j2 = lower[2];
                 incr[2] < 0 ? j2 >= upper[2] : j2 <= upper[2];
                 j2 += incr[2])
                for (j1 = lower[1];
                     incr[1] < 0 ? j1 >= upper[1] : j1 <= upper[1];
                     j1 += incr[1])
                    for (j0 = lower[0];
                         incr[0] > 0 ? j0 <= upper[0] : j0 >= upper[0];
                         j0 += incr[0], dest =
                         (void *) ((char *) dest + nbytes))
                        memcpy(dest, (char *) src + j0 + j1 + j2, nbytes);
            break;
        case 4:
            for (j3 = lower[3];
                 incr[3] < 0 ? j3 >= upper[3] : j3 <= upper[3];
                 j3 += incr[3])
                for (j2 = lower[2];
                     incr[2] < 0 ? j2 >= upper[2] : j2 <= upper[2];
                     j2 += incr[2])
                    for (j1 = lower[1];
                         incr[1] < 0 ? j1 >= upper[1] : j1 <= upper[1];
                         j1 += incr[1])
                        for (j0 = lower[0];
                             incr[0] > 0 ? j0 <= upper[0] : j0 >= upper[0];
                             j0 += incr[0], dest =
                             (void *) ((char *) dest + nbytes))
                            memcpy(dest, (char *) src + j0 + j1 + j2 + j3,
                                   nbytes);
            break;
        case 5:
            for (j4 = lower[4];
                 incr[4] < 0 ? j4 >= upper[4] : j4 <= upper[4];
                 j4 += incr[4])
                for (j3 = lower[3];
                     incr[3] < 0 ? j3 >= upper[3] : j3 <= upper[3];
                     j3 += incr[3])
                    for (j2 = lower[2];
                         incr[2] < 0 ? j2 >= upper[2] : j2 <= upper[2];
                         j2 += incr[2])
                        for (j1 = lower[1];
                             incr[1] < 0 ? j1 >= upper[1] : j1 <= upper[1];
                             j1 += incr[1])
                            for (j0 = lower[0];
                                 incr[0] > 0 ? j0 <= upper[0] : j0 >=
                                 upper[0];
                                 j0 += incr[0], dest =
                                 (void *) ((char *) dest + nbytes))
                                memcpy(dest,
                                       (char *) src + j0 + j1 + j2 + j3 +
                                       j4, nbytes);
            break;
        case 6:
            for (j5 = lower[5];
                 incr[5] < 0 ? j5 >= upper[5] : j5 <= upper[5];
                 j5 += incr[5])
                for (j4 = lower[4];
                     incr[4] < 0 ? j4 >= upper[4] : j4 <= upper[4];
                     j4 += incr[4])
                    for (j3 = lower[3];
                         incr[3] < 0 ? j3 >= upper[3] : j3 <= upper[3];
                         j3 += incr[3])
                        for (j2 = lower[2];
                             incr[2] < 0 ? j2 >= upper[2] : j2 <= upper[2];
                             j2 += incr[2])
                            for (j1 = lower[1];
                                 incr[1] < 0 ? j1 >= upper[1] : j1 <=
                                 upper[1]; j1 += incr[1])
                                for (j0 = lower[0];
                                     incr[0] > 0 ? j0 <= upper[0] : j0 >=
                                     upper[0];
                                     j0 += incr[0], dest =
                                     (void *) ((char *) dest + nbytes))
                                    memcpy(dest,
                                           (char *) src + j0 + j1 + j2 +
                                           j3 + j4 + j5, nbytes);
            break;
        case 7:
            for (j6 = lower[6];
                 incr[6] < 0 ? j6 >= upper[6] : j6 <= upper[6];
                 j6 += incr[6])
                for (j5 = lower[5];
                     incr[5] < 0 ? j5 >= upper[5] : j5 <= upper[5];
                     j5 += incr[5])
                    for (j4 = lower[4];
                         incr[4] < 0 ? j4 >= upper[4] : j4 <= upper[4];
                         j4 += incr[4])
                        for (j3 = lower[3];
                             incr[3] < 0 ? j3 >= upper[3] : j3 <= upper[3];
                             j3 += incr[3])
                            for (j2 = lower[2];
                                 incr[2] < 0 ? j2 >= upper[2] : j2 <=
                                 upper[2]; j2 += incr[2])
                                for (j1 = lower[1];
                                     incr[1] < 0 ? j1 >= upper[1] : j1 <=
                                     upper[1]; j1 += incr[1])
                                    for (j0 = lower[0];
                                         incr[0] > 0 ? j0 <=
                                         upper[0] : j0 >= upper[0];
                                         j0 += incr[0], dest =
                                         (void *) ((char *) dest + nbytes))
                                        memcpy(dest,
                                               (char *) src + j0 + j1 +
                                               j2 + j3 + j4 + j5 + j6,
                                               nbytes);
            break;
        }
    } else {                    /* hasarray */
        /* Unfortunately the cost is heavy if even one subscript is a 1d array,
         * because the number of combinations of ranges and 1d arrays in up to
         * seven subscripts is prohibitively large (7! + 6! + 5! + 4! + 3! + 2 + 1),
         * and so we have to handle both cases in one.
         * First, we set up start, stop, and increment values in advance;
         * and then we have to test whether to compute the true offset. If
         * tests inside possible deeply nested loops are not a real good idea,
         * I really don't know a better way to do it.. */
        /* Set up limits for loops */
        for (i = 0; i < ndim; i++)
            istart[i] = lsubarr[i] ? 0 : lower[i];

        switch (ndim) {
        case 1:
            for (j0 = 0; j0 < lsubarr[0];
                 j0++, dest = (void *) ((char *) dest + nbytes))
                memcpy(dest,
                       (char *) src + (subarr[0][j0] - ilow[0]) * nbytes,
                       nbytes);
            break;
        case 2:
            for (j1 = istart[1];
                 incr[1] < 0 ? j1 >= upper[1] : j1 <= upper[1];
                 j1 += incr[1]) {
                i1 = lsubarr[1] ? (subarr[1][j1] - ilow[1]) * size[1] : j1;
                for (j0 = istart[0];
                     incr[0] > 0 ? j0 <= upper[0] : j0 >= upper[0];
                     j0 += incr[0]) {
                    i0 = lsubarr[0] ? (subarr[0][j0] -
                                       ilow[0]) * size[0] : j0;
                    memcpy(dest, (char *) src + i0 + i1, nbytes);
                    dest = (void *) ((char *) dest + nbytes);
                }
            }
            break;
        case 3:
            for (j2 = istart[2];
                 incr[2] < 0 ? j2 >= upper[2] : j2 <= upper[2];
                 j2 += incr[2]) {
                i2 = lsubarr[2] ? (subarr[2][j2] - ilow[2]) * size[2] : j2;
                for (j1 = istart[1];
                     incr[1] < 0 ? j1 >= upper[1] : j1 <= upper[1];
                     j1 += incr[1]) {
                    i1 = lsubarr[1] ? (subarr[1][j1] -
                                       ilow[1]) * size[1] : j1;
                    for (j0 = istart[0];
                         incr[0] > 0 ? j0 <= upper[0] : j0 >= upper[0];
                         j0 += incr[0]) {
                        i0 = lsubarr[0] ? (subarr[0][j0] -
                                           ilow[0]) * size[0] : j0;
                        memcpy(dest, (char *) src + i0 + i1 + i2, nbytes);
                        dest = (void *) ((char *) dest + nbytes);
                    }
                }
            }
            break;
        case 4:
            for (j3 = istart[3];
                 incr[3] < 0 ? j3 >= upper[3] : j3 <= upper[3];
                 j3 += incr[3]) {
                i3 = lsubarr[3] ? (subarr[3][j3] - ilow[3]) * size[3] : j3;
                for (j2 = istart[2];
                     incr[2] < 0 ? j2 >= upper[2] : j2 <= upper[2];
                     j2 += incr[2]) {
                    i2 = lsubarr[2] ? (subarr[2][j2] -
                                       ilow[2]) * size[2] : j2;
                    for (j1 = istart[1];
                         incr[1] < 0 ? j1 >= upper[1] : j1 <= upper[1];
                         j1 += incr[1]) {
                        i1 = lsubarr[1] ? (subarr[1][j1] -
                                           ilow[1]) * size[1] : j1;
                        for (j0 = istart[0];
                             incr[0] > 0 ? j0 <= upper[0] : j0 >= upper[0];
                             j0 += incr[0]) {
                            i0 = lsubarr[0] ? (subarr[0][j0] -
                                               ilow[0]) * size[0] : j0;
                            memcpy(dest, (char *) src + i0 + i1 + i2 + i3,
                                   nbytes);
                            dest = (void *) ((char *) dest + nbytes);
                        }
                    }
                }
            }
            break;
        case 5:
            for (j4 = istart[4];
                 incr[4] < 0 ? j4 >= upper[4] : j4 <= upper[4];
                 j4 += incr[4]) {
                i4 = lsubarr[4] ? (subarr[4][j4] - ilow[4]) * size[4] : j4;
                for (j3 = istart[3];
                     incr[3] < 0 ? j3 >= upper[3] : j3 <= upper[3];
                     j3 += incr[3]) {
                    i3 = lsubarr[3] ? (subarr[3][j3] -
                                       ilow[3]) * size[3] : j3;
                    for (j2 = istart[2];
                         incr[2] < 0 ? j2 >= upper[2] : j2 <= upper[2];
                         j2 += incr[2]) {
                        i2 = lsubarr[2] ? (subarr[2][j2] -
                                           ilow[2]) * size[2] : j2;
                        for (j1 = istart[1];
                             incr[1] < 0 ? j1 >= upper[1] : j1 <= upper[1];
                             j1 += incr[1]) {
                            i1 = lsubarr[1] ? (subarr[1][j1] -
                                               ilow[1]) * size[1] : j1;
                            for (j0 = istart[0];
                                 incr[0] > 0 ? j0 <= upper[0] : j0 >=
                                 upper[0]; j0 += incr[0]) {
                                i0 = lsubarr[0] ? (subarr[0][j0] -
                                                   ilow[0]) * size[0] : j0;
                                memcpy(dest,
                                       (char *) src + i0 + i1 + i2 + i3 +
                                       i4, nbytes);
                                dest = (void *) ((char *) dest + nbytes);
                            }
                        }
                    }
                }
            }
            break;
        case 6:
            for (j5 = istart[5];
                 incr[5] < 0 ? j5 >= upper[5] : j5 <= upper[5];
                 j5 += incr[5]) {
                i5 = lsubarr[5] ? (subarr[5][j5] - ilow[5]) * size[5] : j5;
                for (j4 = istart[4];
                     incr[4] < 0 ? j4 >= upper[4] : j4 <= upper[4];
                     j4 += incr[4]) {
                    i4 = lsubarr[4] ? (subarr[4][j4] -
                                       ilow[4]) * size[4] : j4;
                    for (j3 = istart[3];
                         incr[3] < 0 ? j3 >= upper[3] : j3 <= upper[3];
                         j3 += incr[3]) {
                        i3 = lsubarr[3] ? (subarr[3][j3] -
                                           ilow[3]) * size[3] : j3;
                        for (j2 = istart[2];
                             incr[2] < 0 ? j2 >= upper[2] : j2 <= upper[2];
                             j2 += incr[2]) {
                            i2 = lsubarr[2] ? (subarr[2][j2] -
                                               ilow[2]) * size[2] : j2;
                            for (j1 = istart[1];
                                 incr[1] < 0 ? j1 >= upper[1] : j1 <=
                                 upper[1]; j1 += incr[1]) {
                                i1 = lsubarr[1] ? (subarr[1][j1] -
                                                   ilow[1]) * size[1] : j1;
                                for (j0 = istart[0];
                                     incr[0] > 0 ? j0 <= upper[0] : j0 >=
                                     upper[0]; j0 += incr[0]) {
                                    i0 = lsubarr[0] ? (subarr[0][j0] -
                                                       ilow[0]) *
                                        size[0] : j0;
                                    memcpy(dest,
                                           (char *) src + i0 + i1 + i2 +
                                           i3 + i4 + i5, nbytes);
                                    dest =
                                        (void *) ((char *) dest + nbytes);
                                }
                            }
                        }
                    }
                }
            }
            break;
        case 7:
            for (j6 = istart[6];
                 incr[6] < 0 ? j6 >= upper[6] : j6 <= upper[6];
                 j6 += incr[6]) {
                i6 = lsubarr[6] ? (subarr[6][j6] - ilow[6]) * size[6] : j6;
                for (j5 = istart[5];
                     incr[5] < 0 ? j5 >= upper[5] : j5 <= upper[5];
                     j5 += incr[5]) {
                    i5 = lsubarr[5] ? (subarr[5][j5] -
                                       ilow[5]) * size[5] : j5;
                    for (j4 = istart[4];
                         incr[4] < 0 ? j4 >= upper[4] : j4 <= upper[4];
                         j4 += incr[4]) {
                        i4 = lsubarr[4] ? (subarr[4][j4] -
                                           ilow[4]) * size[4] : j4;
                        for (j3 = istart[3];
                             incr[3] < 0 ? j3 >= upper[3] : j3 <= upper[3];
                             j3 += incr[3]) {
                            i3 = lsubarr[3] ? (subarr[3][j3] -
                                               ilow[3]) * size[3] : j3;
                            for (j2 = istart[2];
                                 incr[2] < 0 ? j2 >= upper[2] : j2 <=
                                 upper[2]; j2 += incr[2]) {
                                i2 = lsubarr[2] ? (subarr[2][j2] -
                                                   ilow[2]) * size[2] : j2;
                                for (j1 = istart[1];
                                     incr[1] < 0 ? j1 >= upper[1] : j1 <=
                                     upper[1]; j1 += incr[1]) {
                                    i1 = lsubarr[1] ? (subarr[1][j1] -
                                                       ilow[1]) *
                                        size[1] : j1;
                                    for (j0 = istart[0];
                                         incr[0] > 0 ? j0 <=
                                         upper[0] : j0 >= upper[0];
                                         j0 += incr[0]) {
                                        i0 = lsubarr[0] ? (subarr[0][j0] -
                                                           ilow[0]) *
                                            size[0] : j0;
                                        memcpy(dest,
                                               (char *) src + i0 + i1 +
                                               i2 + i3 + i4 + i5 + i6,
                                               nbytes);
                                        dest =
                                            (void *) ((char *) dest +
                                                      nbytes);
                                    }
                                }
                            }
                        }
                    }
                }
            }
            break;
        }
    }
    return (OK);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Parstore moves the object described by stack element s, which is
 * stored consecutively, into dest. The parameters are as follows:
 *     tc: the type of the entries in the object
 *     ndim: the number of dimensions of dest.
 *     ilow, ihi: arrays ndim long; entries contain the low and high
 *         subscripts of each dimension of dest (in the FORTRAN sense).
 *     icol: basically just (ihi - ilow) + 1.
 *     sl, sh: arrays ndim long; these arrays specify the subarray
 *         of dest which is to be assigned to. In the case of a range, these
 *         will be the low and high subscripts of the range. In the
 *         case of a 1d array of subscripts, sl will be DEFAULT + 1
 *         and sh will be the subscript into the bassubs array of
 *         stack elemnts describing the 1d arrays.
 *     si: in the range case, specifies a step between sl and sh.
 *         in the array case, it's just 1.
 *     s:  a stack element describing the object to be assigned.
 *         it is stored consecutively in memory. All size/shape
 *         checking has been done prior to calling Parstore.
 *     subarr, lsubarr: if subscript number i is a 1d array, then
 *         subarr [i] is the address of the array ald lsubarr [i]
 *         is its length. if subscript number i is a range, then
 *         subarr [i] and lsubarr [i] will be zero.
 */
Integer Parstore(Address dest, Integer tc, Integer ndim,
                 Integer *ilow, Integer *ihi, Integer *icol, Integer *sl,
                 Integer *sh, Integer *si, Integer *s, int **subarr,
                 Integer *lsubarr)
{
    int nbytes;                 /* The number of bytes in a single item */
    int i;
    int i0, i1, i2, i3, i4, i5, i6;
    int j0, j1, j2, j3, j4, j5, j6;
    int upper[7], incr[7], size[7], istart[7];
    int lower[7];
    int hasarray;
    Address src;
    Address dest0;

    /* Note: the case when s is a scalar to be broadcast , and the case *
     * ndim ==0, are handled in parstors.                               */

    nbytes = Utbytesin(tc);
    src = Paraddr(s);

    hasarray = 0;

    /* Check for an array subscript */
    for (i = 0; i < 7; i++)
        if (subarr[i] != 0) {
            hasarray = 1;
            break;
        }

    size[0] = nbytes;
    for (i = 0; i < ndim; i++) {
        if (icol[i] == 0)
            /* missing dimension--object has no size, nothing to move. */
            return OK;
        if (i > 0)
            size[i] = size[i - 1] * icol[i - 1];
        if (lsubarr[i] == 0) {
            lower[i] = (sl[i] - ilow[i]) * size[i];
            upper[i] = (sh[i] - ilow[i]) * size[i];
            incr[i] = si[i] * size[i];
        } else {
            upper[i] = lsubarr[i] - 1;
            incr[i] = 1;
        }
    }

    if (!hasarray) {
        switch (ndim) {
        case 1:
            if (si[0] == 1) {
                /* Just a consecutive copy */
                memcpy((char *) dest + lower[0], src,
                       (sh[0] - sl[0] + 1) * nbytes);
            } else {
                for (dest0 = (char *) dest + lower[0];
                     incr[0] < 0 ?
                     dest0 >= (void *) ((char *) dest + upper[0]) :
                     dest0 <= (void *) ((char *) dest + upper[0]);
                     dest0 = (void *) ((char *) dest0 + incr[0]),
                     src = (void *) ((char *) src + nbytes))
                    memcpy(dest0, src, nbytes);
            }
            break;
        case 2:
            for (j1 = lower[1];
                 incr[1] < 0 ? j1 >= upper[1] : j1 <= upper[1];
                 j1 += incr[1])
                for (j0 = lower[0];
                     incr[0] > 0 ? j0 <= upper[0] : j0 >= upper[0];
                     j0 += incr[0], src = (void *) ((char *) src + nbytes))
                    memcpy((char *) dest + j0 + j1, src, nbytes);
            break;
        case 3:
            for (j2 = lower[2];
                 incr[2] < 0 ? j2 >= upper[2] : j2 <= upper[2];
                 j2 += incr[2])
                for (j1 = lower[1];
                     incr[1] < 0 ? j1 >= upper[1] : j1 <= upper[1];
                     j1 += incr[1])
                    for (j0 = lower[0];
                         incr[0] > 0 ? j0 <= upper[0] : j0 >= upper[0];
                         j0 += incr[0], src =
                         (void *) ((char *) src + nbytes))
                        memcpy((char *) dest + j0 + j1 + j2, src, nbytes);
            break;
        case 4:
            for (j3 = lower[3];
                 incr[3] < 0 ? j3 >= upper[3] : j3 <= upper[3];
                 j3 += incr[3])
                for (j2 = lower[2];
                     incr[2] < 0 ? j2 >= upper[2] : j2 <= upper[2];
                     j2 += incr[2])
                    for (j1 = lower[1];
                         incr[1] < 0 ? j1 >= upper[1] : j1 <= upper[1];
                         j1 += incr[1])
                        for (j0 = lower[0];
                             incr[0] > 0 ? j0 <= upper[0] : j0 >= upper[0];
                             j0 += incr[0], src =
                             (void *) ((char *) src + nbytes))
                            memcpy((char *) dest + j0 + j1 + j2 + j3, src,
                                   nbytes);
            break;
        case 5:
            for (j4 = lower[4];
                 incr[4] < 0 ? j4 >= upper[4] : j4 <= upper[4];
                 j4 += incr[4])
                for (j3 = lower[3];
                     incr[3] < 0 ? j3 >= upper[3] : j3 <= upper[3];
                     j3 += incr[3])
                    for (j2 = lower[2];
                         incr[2] < 0 ? j2 >= upper[2] : j2 <= upper[2];
                         j2 += incr[2])
                        for (j1 = lower[1];
                             incr[1] < 0 ? j1 >= upper[1] : j1 <= upper[1];
                             j1 += incr[1])
                            for (j0 = lower[0];
                                 incr[0] > 0 ? j0 <= upper[0] : j0 >=
                                 upper[0];
                                 j0 += incr[0], src =
                                 (void *) ((char *) src + nbytes))
                                memcpy((char *) dest + j0 + j1 + j2 + j3 +
                                       j4, src, nbytes);
            break;
        case 6:
            for (j5 = lower[5];
                 incr[5] < 0 ? j5 >= upper[5] : j5 <= upper[5];
                 j5 += incr[5])
                for (j4 = lower[4];
                     incr[4] < 0 ? j4 >= upper[4] : j4 <= upper[4];
                     j4 += incr[4])
                    for (j3 = lower[3];
                         incr[3] < 0 ? j3 >= upper[3] : j3 <= upper[3];
                         j3 += incr[3])
                        for (j2 = lower[2];
                             incr[2] < 0 ? j2 >= upper[2] : j2 <= upper[2];
                             j2 += incr[2])
                            for (j1 = lower[1];
                                 incr[1] < 0 ? j1 >= upper[1] : j1 <=
                                 upper[1]; j1 += incr[1])
                                for (j0 = lower[0];
                                     incr[0] > 0 ? j0 <= upper[0] : j0 >=
                                     upper[0];
                                     j0 += incr[0], src =
                                     (void *) ((char *) src + nbytes))
                                    memcpy((char *) dest + j0 + j1 + j2 +
                                           j3 + j4 + j5, src, nbytes);
            break;
        case 7:
            for (j6 = lower[6];
                 incr[6] < 0 ? j6 >= upper[6] : j6 <= upper[6];
                 j6 += incr[6])
                for (j5 = lower[5];
                     incr[5] < 0 ? j5 >= upper[5] : j5 <= upper[5];
                     j5 += incr[5])
                    for (j4 = lower[4];
                         incr[4] < 0 ? j4 >= upper[4] : j4 <= upper[4];
                         j4 += incr[4])
                        for (j3 = lower[3];
                             incr[3] < 0 ? j3 >= upper[3] : j3 <= upper[3];
                             j3 += incr[3])
                            for (j2 = lower[2];
                                 incr[2] < 0 ? j2 >= upper[2] : j2 <=
                                 upper[2]; j2 += incr[2])
                                for (j1 = lower[1];
                                     incr[1] < 0 ? j1 >= upper[1] : j1 <=
                                     upper[1]; j1 += incr[1])
                                    for (j0 = lower[0];
                                         incr[0] > 0 ? j0 <=
                                         upper[0] : j0 >= upper[0];
                                         j0 += incr[0], src =
                                         (void *) ((char *) src + nbytes))
                                        memcpy((char *) dest + j0 + j1 +
                                               j2 + j3 + j4 + j5 + j6, src,
                                               nbytes);
            break;
        }
    } else {                    /* hasarray */
        /* Unfortunately the cost is heavy if even one subscript is a 1d array,
         * because the number of combinations of ranges and 1d arrays in up to
         * seven subscripts is prohibitively large (7! + 6! + 5! + 4! + 3! + 2 + 1),
         * and so we have to handle both cases in one.
         * First, we set up start, stop, and increment values in advance;
         * and then we have to test whether to compute the true offset. If
         * tests inside possible deeply nested loops are not a real good idea,
         * I really don't know a better way to do it.. */
        /* Set up limits for loops */
        for (i = 0; i < ndim; i++)
            istart[i] = lsubarr[i] ? 0 : lower[i];

        switch (ndim) {
        case 1:
            for (j0 = 0; j0 < lsubarr[0];
                 j0++, src = (void *) ((char *) src + nbytes))
                memcpy((char *) dest + (subarr[0][j0] - ilow[0]) * nbytes,
                       src, nbytes);
            break;
        case 2:
            for (j1 = istart[1];
                 incr[1] < 0 ? j1 >= upper[1] : j1 <= upper[1];
                 j1 += incr[1]) {
                i1 = lsubarr[1] ? (subarr[1][j1] - ilow[1]) * size[1] : j1;
                for (j0 = istart[0];
                     incr[0] > 0 ? j0 <= upper[0] : j0 >= upper[0];
                     j0 += incr[0]) {
                    i0 = lsubarr[0] ? (subarr[0][j0] -
                                       ilow[0]) * size[0] : j0;
                    memcpy((char *) dest + i0 + i1, src, nbytes);
                    src = (void *) ((char *) src + nbytes);
                }
            }
            break;
        case 3:
            for (j2 = istart[2];
                 incr[2] < 0 ? j2 >= upper[2] : j2 <= upper[2];
                 j2 += incr[2]) {
                i2 = lsubarr[2] ? (subarr[2][j2] - ilow[2]) * size[2] : j2;
                for (j1 = istart[1];
                     incr[1] < 0 ? j1 >= upper[1] : j1 <= upper[1];
                     j1 += incr[1]) {
                    i1 = lsubarr[1] ? (subarr[1][j1] -
                                       ilow[1]) * size[1] : j1;
                    for (j0 = istart[0];
                         incr[0] > 0 ? j0 <= upper[0] : j0 >= upper[0];
                         j0 += incr[0]) {
                        i0 = lsubarr[0] ? (subarr[0][j0] -
                                           ilow[0]) * size[0] : j0;
                        memcpy((char *) dest + i0 + i1 + i2, src, nbytes);
                        src = (void *) ((char *) src + nbytes);
                    }
                }
            }
            break;
        case 4:
            for (j3 = istart[3];
                 incr[3] < 0 ? j3 >= upper[3] : j3 <= upper[3];
                 j3 += incr[3]) {
                i3 = lsubarr[3] ? (subarr[3][j3] - ilow[3]) * size[3] : j3;
                for (j2 = istart[2];
                     incr[2] < 0 ? j2 >= upper[2] : j2 <= upper[2];
                     j2 += incr[2]) {
                    i2 = lsubarr[2] ? (subarr[2][j2] -
                                       ilow[2]) * size[2] : j2;
                    for (j1 = istart[1];
                         incr[1] < 0 ? j1 >= upper[1] : j1 <= upper[1];
                         j1 += incr[1]) {
                        i1 = lsubarr[1] ? (subarr[1][j1] -
                                           ilow[1]) * size[1] : j1;
                        for (j0 = istart[0];
                             incr[0] > 0 ? j0 <= upper[0] : j0 >= upper[0];
                             j0 += incr[0]) {
                            i0 = lsubarr[0] ? (subarr[0][j0] -
                                               ilow[0]) * size[0] : j0;
                            memcpy((char *) dest + i0 + i1 + i2 + i3, src,
                                   nbytes);
                            src = (void *) ((char *) src + nbytes);
                        }
                    }
                }
            }
            break;
        case 5:
            for (j4 = istart[4];
                 incr[4] < 0 ? j4 >= upper[4] : j4 <= upper[4];
                 j4 += incr[4]) {
                i4 = lsubarr[4] ? (subarr[4][j4] - ilow[4]) * size[4] : j4;
                for (j3 = istart[3];
                     incr[3] < 0 ? j3 >= upper[3] : j3 <= upper[3];
                     j3 += incr[3]) {
                    i3 = lsubarr[3] ? (subarr[3][j3] -
                                       ilow[3]) * size[3] : j3;
                    for (j2 = istart[2];
                         incr[2] < 0 ? j2 >= upper[2] : j2 <= upper[2];
                         j2 += incr[2]) {
                        i2 = lsubarr[2] ? (subarr[2][j2] -
                                           ilow[2]) * size[2] : j2;
                        for (j1 = istart[1];
                             incr[1] < 0 ? j1 >= upper[1] : j1 <= upper[1];
                             j1 += incr[1]) {
                            i1 = lsubarr[1] ? (subarr[1][j1] -
                                               ilow[1]) * size[1] : j1;
                            for (j0 = istart[0];
                                 incr[0] > 0 ? j0 <= upper[0] : j0 >=
                                 upper[0]; j0 += incr[0]) {
                                i0 = lsubarr[0] ? (subarr[0][j0] -
                                                   ilow[0]) * size[0] : j0;
                                memcpy((char *) dest + i0 + i1 + i2 + i3 +
                                       i4, src, nbytes);
                                src = (void *) ((char *) src + nbytes);
                            }
                        }
                    }
                }
            }
            break;
        case 6:
            for (j5 = istart[5];
                 incr[5] < 0 ? j5 >= upper[5] : j5 <= upper[5];
                 j5 += incr[5]) {
                i5 = lsubarr[5] ? (subarr[5][j5] - ilow[5]) * size[5] : j5;
                for (j4 = istart[4];
                     incr[4] < 0 ? j4 >= upper[4] : j4 <= upper[4];
                     j4 += incr[4]) {
                    i4 = lsubarr[4] ? (subarr[4][j4] -
                                       ilow[4]) * size[4] : j4;
                    for (j3 = istart[3];
                         incr[3] < 0 ? j3 >= upper[3] : j3 <= upper[3];
                         j3 += incr[3]) {
                        i3 = lsubarr[3] ? (subarr[3][j3] -
                                           ilow[3]) * size[3] : j3;
                        for (j2 = istart[2];
                             incr[2] < 0 ? j2 >= upper[2] : j2 <= upper[2];
                             j2 += incr[2]) {
                            i2 = lsubarr[2] ? (subarr[2][j2] -
                                               ilow[2]) * size[2] : j2;
                            for (j1 = istart[1];
                                 incr[1] < 0 ? j1 >= upper[1] : j1 <=
                                 upper[1]; j1 += incr[1]) {
                                i1 = lsubarr[1] ? (subarr[1][j1] -
                                                   ilow[1]) * size[1] : j1;
                                for (j0 = istart[0];
                                     incr[0] > 0 ? j0 <= upper[0] : j0 >=
                                     upper[0]; j0 += incr[0]) {
                                    i0 = lsubarr[0] ? (subarr[0][j0] -
                                                       ilow[0]) *
                                        size[0] : j0;
                                    memcpy((char *) dest + i0 + i1 + i2 +
                                           i3 + i4 + i5, src, nbytes);
                                    src = (void *) ((char *) src + nbytes);
                                }
                            }
                        }
                    }
                }
            }
            break;
        case 7:
            for (j6 = istart[6];
                 incr[6] < 0 ? j6 >= upper[6] : j6 <= upper[6];
                 j6 += incr[6]) {
                i6 = lsubarr[6] ? (subarr[6][j6] - ilow[6]) * size[6] : j6;
                for (j5 = istart[5];
                     incr[5] < 0 ? j5 >= upper[5] : j5 <= upper[5];
                     j5 += incr[5]) {
                    i5 = lsubarr[5] ? (subarr[5][j5] -
                                       ilow[5]) * size[5] : j5;
                    for (j4 = istart[4];
                         incr[4] < 0 ? j4 >= upper[4] : j4 <= upper[4];
                         j4 += incr[4]) {
                        i4 = lsubarr[4] ? (subarr[4][j4] -
                                           ilow[4]) * size[4] : j4;
                        for (j3 = istart[3];
                             incr[3] < 0 ? j3 >= upper[3] : j3 <= upper[3];
                             j3 += incr[3]) {
                            i3 = lsubarr[3] ? (subarr[3][j3] -
                                               ilow[3]) * size[3] : j3;
                            for (j2 = istart[2];
                                 incr[2] < 0 ? j2 >= upper[2] : j2 <=
                                 upper[2]; j2 += incr[2]) {
                                i2 = lsubarr[2] ? (subarr[2][j2] -
                                                   ilow[2]) * size[2] : j2;
                                for (j1 = istart[1];
                                     incr[1] < 0 ? j1 >= upper[1] : j1 <=
                                     upper[1]; j1 += incr[1]) {
                                    i1 = lsubarr[1] ? (subarr[1][j1] -
                                                       ilow[1]) *
                                        size[1] : j1;
                                    for (j0 = istart[0];
                                         incr[0] > 0 ? j0 <=
                                         upper[0] : j0 >= upper[0];
                                         j0 += incr[0]) {
                                        i0 = lsubarr[0] ? (subarr[0][j0] -
                                                           ilow[0]) *
                                            size[0] : j0;
                                        memcpy((char *) dest + i0 + i1 +
                                               i2 + i3 + i4 + i5 + i6, src,
                                               nbytes);
                                        src =
                                            (void *) ((char *) src +
                                                      nbytes);
                                    }
                                }
                            }
                        }
                    }
                }
            }
            break;
        }
    }
    return (OK);
}                               /* end of Parstore */

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Parstors moves the object described by stack element s, which is
 * a single scalar, into dest. The parameters are as follows:
 *     tc: the type of the entries in the object
 *     ndim: the number of dimensions of dest.
 *     ilow, ihi: arrays ndim long; entries contain the low and high
 *         subscripts of each dimension of dest (in the FORTRAN sense).
 *     icol: basically just (ihi - ilow) + 1.
 *     sl, sh: arrays ndim long; these arrays specify the subarray
 *         of dest which is to be assigned to. In the case of a range, these
 *         will be the low and high subscripts of the range. In the
 *         case of a 1d array of subscripts, sl will be DEFAULT + 1
 *         and sh will be the subscript into the bassubs array of
 *         stack elemnts describing the 1d arrays.
 *     si: in the range case, specifies a step between sl and sh.
 *         in the array case, it's just 1.
 *     s:  a stack element describing the object to be assigned.
 *         which is a single scalar of type tc.
 *     subarr, lsubarr: if subscript number i is a 1d array, then
 *         subarr [i] is the address of the array ald lsubarr [i]
 *         is its length. if subscript number i is a range, then
 *         subarr [i] and lsubarr [i] will be zero.
 * Note:      Save the address of the target in dofwa in Actions. This is
 *            only used in the case of assignment to a do loop control
 *            variable; lrsetdo places this address in the stack element
 *            so that it can be used subsequently for incrementing and
 *            testing. This value is always saved, since we don't know
 *            who called us.
 */

Integer Parstors(Address dest, Integer tc, Integer ndim,
                 Integer *ilow, Integer *ihi, Integer *icol, Integer *sl,
                 Integer *sh, Integer *si, Integer *s, int **subarr,
                 Integer *lsubarr)
{
    int nbytes;                 /* The number of bytes in a single item */
    int i;
    int i0, i1, i2, i3, i4, i5, i6;
    int j0, j1, j2, j3, j4, j5, j6;
    int upper[7], incr[7], size[7], istart[7];
    int lower[7];
    int hasarray;
    Address src;
    Address dest0;


    nbytes = Utbytesin(tc);
    src = Paraddr(s);

    if (ndim == 0) {
        /* Move a single scalar from src to dest */
        dofwa = dest;
        memcpy(dest, src, nbytes);
        return (OK);
    }

    hasarray = 0;

    /* Check for an array subscript */
    for (i = 0; i < 7; i++)
        if (subarr[i] != 0) {
            hasarray = 1;
            break;
        }

    size[0] = nbytes;
    for (i = 0; i < ndim; i++) {
        if (icol[i] == 0)
            /* missing dimension--object has no size, nothing to move. */
            return OK;
        if (i > 0)
            size[i] = size[i - 1] * icol[i - 1];
        if (lsubarr[i] == 0) {
            lower[i] = (sl[i] - ilow[i]) * size[i];
            upper[i] = (sh[i] - ilow[i]) * size[i];
            incr[i] = si[i] * size[i];
        } else {
            upper[i] = lsubarr[i] - 1;
            incr[i] = 1;
        }
    }

    if (!hasarray) {
        switch (ndim) {
        case 1:
            for (dofwa = dest0 = (char *) dest + lower[0];
                 dest0 <= (void *) ((char *) dest + upper[0]);
                 dest0 = (void *) ((char *) dest0 + incr[0]))
                memcpy(dest0, src, nbytes);
            break;
        case 2:
            dofwa = (char *) dest + lower[0] + lower[1];
            for (j1 = lower[1];
                 incr[1] < 0 ? j1 >= upper[1] : j1 <= upper[1];
                 j1 += incr[1])
                for (j0 = lower[0];
                     incr[0] > 0 ? j0 <= upper[0] : j0 >= upper[0];
                     j0 += incr[0])
                    memcpy((char *) dest + j0 + j1, src, nbytes);
            break;
        case 3:
            dofwa = (char *) dest + lower[0] + lower[1] + lower[2];
            for (j2 = lower[2];
                 incr[2] < 0 ? j2 >= upper[2] : j2 <= upper[2];
                 j2 += incr[2])
                for (j1 = lower[1];
                     incr[1] < 0 ? j1 >= upper[1] : j1 <= upper[1];
                     j1 += incr[1])
                    for (j0 = lower[0];
                         incr[0] > 0 ? j0 <= upper[0] : j0 >= upper[0];
                         j0 += incr[0])
                        memcpy((char *) dest + j0 + j1 + j2, src, nbytes);
            break;
        case 4:
            dofwa = (char *) dest + lower[0] + lower[1] + lower[2]
                + lower[3];
            for (j3 = lower[3];
                 incr[3] < 0 ? j3 >= upper[3] : j3 <= upper[3];
                 j3 += incr[3])
                for (j2 = lower[2];
                     incr[2] < 0 ? j2 >= upper[2] : j2 <= upper[2];
                     j2 += incr[2])
                    for (j1 = lower[1];
                         incr[1] < 0 ? j1 >= upper[1] : j1 <= upper[1];
                         j1 += incr[1])
                        for (j0 = lower[0];
                             incr[0] > 0 ? j0 <= upper[0] : j0 >= upper[0];
                             j0 += incr[0])
                            memcpy((char *) dest + j0 + j1 + j2 + j3, src,
                                   nbytes);
            break;
        case 5:
            dofwa = (char *) dest + lower[0] + lower[1] + lower[2]
                + lower[3] + lower[4];
            for (j4 = lower[4];
                 incr[4] < 0 ? j4 >= upper[4] : j4 <= upper[4];
                 j4 += incr[4])
                for (j3 = lower[3];
                     incr[3] < 0 ? j3 >= upper[3] : j3 <= upper[3];
                     j3 += incr[3])
                    for (j2 = lower[2];
                         incr[2] < 0 ? j2 >= upper[2] : j2 <= upper[2];
                         j2 += incr[2])
                        for (j1 = lower[1];
                             incr[1] < 0 ? j1 >= upper[1] : j1 <= upper[1];
                             j1 += incr[1])
                            for (j0 = lower[0];
                                 incr[0] > 0 ? j0 <= upper[0] : j0 >=
                                 upper[0]; j0 += incr[0])
                                memcpy((char *) dest + j0 + j1 + j2 + j3 +
                                       j4, src, nbytes);
            break;
        case 6:
            dofwa = (char *) dest + lower[0] + lower[1] + lower[2]
                + lower[3] + lower[4] + lower[5];
            for (j5 = lower[5];
                 incr[5] < 0 ? j5 >= upper[5] : j5 <= upper[5];
                 j5 += incr[5])
                for (j4 = lower[4];
                     incr[4] < 0 ? j4 >= upper[4] : j4 <= upper[4];
                     j4 += incr[4])
                    for (j3 = lower[3];
                         incr[3] < 0 ? j3 >= upper[3] : j3 <= upper[3];
                         j3 += incr[3])
                        for (j2 = lower[2];
                             incr[2] < 0 ? j2 >= upper[2] : j2 <= upper[2];
                             j2 += incr[2])
                            for (j1 = lower[1];
                                 incr[1] < 0 ? j1 >= upper[1] : j1 <=
                                 upper[1]; j1 += incr[1])
                                for (j0 = lower[0];
                                     incr[0] > 0 ? j0 <= upper[0] : j0 >=
                                     upper[0]; j0 += incr[0])
                                    memcpy((char *) dest + j0 + j1 + j2 +
                                           j3 + j4 + j5, src, nbytes);
            break;
        case 7:
            dofwa = (char *) dest + lower[0] + lower[1] + lower[2]
                + lower[3] + lower[4] + lower[5] + lower[6];
            for (j6 = lower[6];
                 incr[6] < 0 ? j6 >= upper[6] : j6 <= upper[6];
                 j6 += incr[6])
                for (j5 = lower[5];
                     incr[5] < 0 ? j5 >= upper[5] : j5 <= upper[5];
                     j5 += incr[5])
                    for (j4 = lower[4];
                         incr[4] < 0 ? j4 >= upper[4] : j4 <= upper[4];
                         j4 += incr[4])
                        for (j3 = lower[3];
                             incr[3] < 0 ? j3 >= upper[3] : j3 <= upper[3];
                             j3 += incr[3])
                            for (j2 = lower[2];
                                 incr[2] < 0 ? j2 >= upper[2] : j2 <=
                                 upper[2]; j2 += incr[2])
                                for (j1 = lower[1];
                                     incr[1] < 0 ? j1 >= upper[1] : j1 <=
                                     upper[1]; j1 += incr[1])
                                    for (j0 = lower[0];
                                         incr[0] > 0 ? j0 <=
                                         upper[0] : j0 >= upper[0];
                                         j0 += incr[0])
                                        memcpy((char *) dest + j0 + j1 +
                                               j2 + j3 + j4 + j5 + j6, src,
                                               nbytes);
            break;
        }
    } else {                    /* hasarray */
        /* Unfortunately the cost is heavy if even one subscript is a 1d array,
         * because the number of combinations of ranges and 1d arrays in up to
         * seven subscripts is prohibitively large (7! + 6! + 5! + 4! + 3! + 2 + 1),
         * and so we have to handle both cases in one.
         * First, we set up start, stop, and increment values in advance;
         * and then we have to test whether to compute the true offset. If
         * tests inside possible deeply nested loops are not a real good idea,
         * I really don't know a better way to do it.. */
        /* Set up limits for loops */
        for (i = 0; i < ndim; i++)
            istart[i] = lsubarr[i] ? 0 : lower[i];

        switch (ndim) {
        case 1:
            dofwa = (char *) dest + (subarr[0][0] - ilow[0]) * nbytes;
            for (j0 = 0; j0 < lsubarr[0]; j0++)
                memcpy((char *) dest + (subarr[0][j0] - ilow[0]) * nbytes,
                       src, nbytes);
            break;
        case 2:
            dofwa = (char *) dest + istart[1] + istart[0];
            for (j1 = istart[1];
                 incr[1] < 0 ? j1 >= upper[1] : j1 <= upper[1];
                 j1 += incr[1]) {
                i1 = lsubarr[1] ? (subarr[1][j1] - ilow[1]) * size[1] : j1;
                for (j0 = istart[0];
                     incr[0] > 0 ? j0 <= upper[0] : j0 >= upper[0];
                     j0 += incr[0]) {
                    i0 = lsubarr[0] ? (subarr[0][j0] -
                                       ilow[0]) * size[0] : j0;
                    memcpy((char *) dest + i0 + i1, src, nbytes);
                }
            }
            break;
        case 3:
            dofwa = (char *) dest + istart[2] + istart[1] + istart[0];
            for (j2 = istart[2];
                 incr[2] < 0 ? j2 >= upper[2] : j2 <= upper[2];
                 j2 += incr[2]) {
                i2 = lsubarr[2] ? (subarr[2][j2] - ilow[2]) * size[2] : j2;
                for (j1 = istart[1];
                     incr[1] < 0 ? j1 >= upper[1] : j1 <= upper[1];
                     j1 += incr[1]) {
                    i1 = lsubarr[1] ? (subarr[1][j1] -
                                       ilow[1]) * size[1] : j1;
                    for (j0 = istart[0];
                         incr[0] > 0 ? j0 <= upper[0] : j0 >= upper[0];
                         j0 += incr[0]) {
                        i0 = lsubarr[0] ? (subarr[0][j0] -
                                           ilow[0]) * size[0] : j0;
                        memcpy((char *) dest + i0 + i1 + i2, src, nbytes);
                    }
                }
            }
            break;
        case 4:
            dofwa = (char *) dest + istart[3] + istart[2] + istart[1] +
                istart[0];
            for (j3 = istart[3];
                 incr[3] < 0 ? j3 >= upper[3] : j3 <= upper[3];
                 j3 += incr[3]) {
                i3 = lsubarr[3] ? (subarr[3][j3] - ilow[3]) * size[3] : j3;
                for (j2 = istart[2];
                     incr[2] < 0 ? j2 >= upper[2] : j2 <= upper[2];
                     j2 += incr[2]) {
                    i2 = lsubarr[2] ? (subarr[2][j2] -
                                       ilow[2]) * size[2] : j2;
                    for (j1 = istart[1];
                         incr[1] < 0 ? j1 >= upper[1] : j1 <= upper[1];
                         j1 += incr[1]) {
                        i1 = lsubarr[1] ? (subarr[1][j1] -
                                           ilow[1]) * size[1] : j1;
                        for (j0 = istart[0];
                             incr[0] > 0 ? j0 <= upper[0] : j0 >= upper[0];
                             j0 += incr[0]) {
                            i0 = lsubarr[0] ? (subarr[0][j0] -
                                               ilow[0]) * size[0] : j0;
                            memcpy((char *) dest + i0 + i1 + i2 + i3, src,
                                   nbytes);
                        }
                    }
                }
            }
            break;
        case 5:
            dofwa = (char *) dest + istart[4] + istart[3] + istart[2] +
                istart[1] + istart[0];
            for (j4 = istart[4];
                 incr[4] < 0 ? j4 >= upper[4] : j4 <= upper[4];
                 j4 += incr[4]) {
                i4 = lsubarr[4] ? (subarr[4][j4] - ilow[4]) * size[4] : j4;
                for (j3 = istart[3];
                     incr[3] < 0 ? j3 >= upper[3] : j3 <= upper[3];
                     j3 += incr[3]) {
                    i3 = lsubarr[3] ? (subarr[3][j3] -
                                       ilow[3]) * size[3] : j3;
                    for (j2 = istart[2];
                         incr[2] < 0 ? j2 >= upper[2] : j2 <= upper[2];
                         j2 += incr[2]) {
                        i2 = lsubarr[2] ? (subarr[2][j2] -
                                           ilow[2]) * size[2] : j2;
                        for (j1 = istart[1];
                             incr[1] < 0 ? j1 >= upper[1] : j1 <= upper[1];
                             j1 += incr[1]) {
                            i1 = lsubarr[1] ? (subarr[1][j1] -
                                               ilow[1]) * size[1] : j1;
                            for (j0 = istart[0];
                                 incr[0] > 0 ? j0 <= upper[0] : j0 >=
                                 upper[0]; j0 += incr[0]) {
                                i0 = lsubarr[0] ? (subarr[0][j0] -
                                                   ilow[0]) * size[0] : j0;
                                memcpy((char *) dest + i0 + i1 + i2 + i3 +
                                       i4, src, nbytes);
                            }
                        }
                    }
                }
            }
            break;
        case 6:
            dofwa = (char *) dest + istart[5] + istart[4] + istart[3] +
                istart[2] + istart[1] + istart[0];
            for (j5 = istart[5];
                 incr[5] < 0 ? j5 >= upper[5] : j5 <= upper[5];
                 j5 += incr[5]) {
                i5 = lsubarr[5] ? (subarr[5][j5] - ilow[5]) * size[5] : j5;
                for (j4 = istart[4];
                     incr[4] < 0 ? j4 >= upper[4] : j4 <= upper[4];
                     j4 += incr[4]) {
                    i4 = lsubarr[4] ? (subarr[4][j4] -
                                       ilow[4]) * size[4] : j4;
                    for (j3 = istart[3];
                         incr[3] < 0 ? j3 >= upper[3] : j3 <= upper[3];
                         j3 += incr[3]) {
                        i3 = lsubarr[3] ? (subarr[3][j3] -
                                           ilow[3]) * size[3] : j3;
                        for (j2 = istart[2];
                             incr[2] < 0 ? j2 >= upper[2] : j2 <= upper[2];
                             j2 += incr[2]) {
                            i2 = lsubarr[2] ? (subarr[2][j2] -
                                               ilow[2]) * size[2] : j2;
                            for (j1 = istart[1];
                                 incr[1] < 0 ? j1 >= upper[1] : j1 <=
                                 upper[1]; j1 += incr[1]) {
                                i1 = lsubarr[1] ? (subarr[1][j1] -
                                                   ilow[1]) * size[1] : j1;
                                for (j0 = istart[0];
                                     incr[0] > 0 ? j0 <= upper[0] : j0 >=
                                     upper[0]; j0 += incr[0]) {
                                    i0 = lsubarr[0] ? (subarr[0][j0] -
                                                       ilow[0]) *
                                        size[0] : j0;
                                    memcpy((char *) dest + i0 + i1 + i2 +
                                           i3 + i4 + i5, src, nbytes);
                                }
                            }
                        }
                    }
                }
            }
            break;
        case 7:
            dofwa = (char *) dest + istart[6] + istart[5] + istart[4] +
                istart[3] + istart[2] + istart[1] + istart[0];
            for (j6 = istart[6];
                 incr[6] < 0 ? j6 >= upper[6] : j6 <= upper[6];
                 j6 += incr[6]) {
                i6 = lsubarr[6] ? (subarr[6][j6] - ilow[6]) * size[6] : j6;
                for (j5 = istart[5];
                     incr[5] < 0 ? j5 >= upper[5] : j5 <= upper[5];
                     j5 += incr[5]) {
                    i5 = lsubarr[5] ? (subarr[5][j5] -
                                       ilow[5]) * size[5] : j5;
                    for (j4 = istart[4];
                         incr[4] < 0 ? j4 >= upper[4] : j4 <= upper[4];
                         j4 += incr[4]) {
                        i4 = lsubarr[4] ? (subarr[4][j4] -
                                           ilow[4]) * size[4] : j4;
                        for (j3 = istart[3];
                             incr[3] < 0 ? j3 >= upper[3] : j3 <= upper[3];
                             j3 += incr[3]) {
                            i3 = lsubarr[3] ? (subarr[3][j3] -
                                               ilow[3]) * size[3] : j3;
                            for (j2 = istart[2];
                                 incr[2] < 0 ? j2 >= upper[2] : j2 <=
                                 upper[2]; j2 += incr[2]) {
                                i2 = lsubarr[2] ? (subarr[2][j2] -
                                                   ilow[2]) * size[2] : j2;
                                for (j1 = istart[1];
                                     incr[1] < 0 ? j1 >= upper[1] : j1 <=
                                     upper[1]; j1 += incr[1]) {
                                    i1 = lsubarr[1] ? (subarr[1][j1] -
                                                       ilow[1]) *
                                        size[1] : j1;
                                    for (j0 = istart[0];
                                         incr[0] > 0 ? j0 <=
                                         upper[0] : j0 >= upper[0];
                                         j0 += incr[0]) {
                                        i0 = lsubarr[0] ? (subarr[0][j0] -
                                                           ilow[0]) *
                                            size[0] : j0;
                                        memcpy((char *) dest + i0 + i1 +
                                               i2 + i3 + i4 + i5 + i6, src,
                                               nbytes);
                                    }
                                }
                            }
                        }
                    }
                }
            }
            break;
        }
    }
    return (OK);
}                               /* end of Parstore */

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
