/*
 * rdb.c
 *
 */

#include "rt.h"

/*############################ Group ############################*/
/* Groups and Variables will be stored in dynamic arrays whose   */
/* sizes can be changed if necessary. The index of a group and   */
/* of a variable will be 0-based, i. e., one less than what      */
/* FORTRAN expects.                                              */
/*###############################################################*/

typedef struct Group Group;

struct Group {
    Integer grpnam;             /* Index into the string table            */
    Integer remark;             /* Remarks about Groups                   */
    Integer gvmap;              /* Index to this Group's first Variable   */
    Integer gattr;              /* Group attributes                       */
};

/*########################### Variable ##########################*/
/* Groups and Variables will be stored in dynamic arrays whose   */
/* sizes can be changed if necessary. The index of a group and   */
/* of a variable will be 0-based, i. e., one less than what      */
/* FORTRAN expects.                                              */
/*###############################################################*/

typedef struct Variable Variable;

struct Variable {
    Integer vattr;              /* Variable attributes                                 */
    Integer varnum;             /* slot number for name in string table                */
    Integer types;              /* type of this object                                 */
    Integer dimsto;             /* slot number for original dimension string           */
    Integer units;              /* slot number for units string                        */
    Integer comm;               /* comments about this object                          */
    Integer dimen;              /* number of dimensions of this object                 */
    Integer limstr;             /* slot number of dimension-limiting string            */
    Integer actstr;             /* slot number of assignment action string             */
    Integer varprv;             /* hash table previous entry                           */
    Integer varnxt;             /* hash table next entry                              */
    Integer elems;              /* Number of words of dynamic storage including pad    */
    Integer rely;               /* Will array change size?                             */
    Integer funtyp;             /* Tells type of a function (<0, compiled; =0, builtin;
                                 *                           >0, user)                  */
    Integer origin[7];          /* Low value of each subscript                          */
    Integer dimstr[7];          /* High value of each subscript                         */
    Integer dynflg;             /* 0=static, 1=dynamic data, static pointer
                                 * 2=dynamic data, dynamic pointer                      */
    Integer refcount;           /* Number of internal basis references.                 */
    logical chamel;             /* Is this variable a chameleon?                        */
    Address addr;               /* addresses of variables/pointers/functions            */
    Address argn;               /* addresses of function and argument names             */
    Address argt;               /* addresses of type information                        */
    Address setval;             /* pointer to value to use to fill dynamic variable;    */
    /* points to text of a user function;                   */
    /* points to calling sequence of compiled function      */
    /* while being gathered 3/17/95 ZCM                     */
};

/*########################## Package ##########################*/
/* Packages will be in a static array which is never changed   */
/* in size, and which is(essentially) 1-based, as in FORTRAN. */
/*#############################################################*/

typedef struct Package Package;

struct Package {
    Integer packno;             /* The number of this package                        */
    Integer nvars0;             /* The number of permanent variables in this package */
    Integer ngrps;              /* The number of groups currently in this package    */
    Integer maxgrps;            /* The maximum number of groups in this package      */
    Integer nvars;              /* The number of variables in this package           */
    Integer maxvari;            /* The maximum number of variables ever in package   */
    Integer pripkg;             /* The priority of this package                      */
    Integer maxvars;            /* The maximum number of variables this pkg holds    */
    Group *group1;              /* Pointer to the first Group in this Package        */
    Variable *var1;             /* Pointer to the first variable in this package     */
};

/*########### Here's the permanent part of the database ###########*/
/* Note, since database[0] is a dummy, packages are indexed from  */
/* 1, exactly as in FORTRAN.                                       */
/*#################################################################*/

static Package database[MAXPAK + 3];    /* Note, package 0 is a dummy. */

#define dblocal database[LOCALDB]
#define dbuser  database[USERDB]

static Integer currpkg = 1;     /* The current package number */
static Integer prihigh = 2;     /* The highest current priority */
static Integer locdepth = 0;    /* Number of "moms" one can see in local db */
static Integer rtexit = 0;      /* Set for final delete at exit time */


/*#################### Comments Group ####################*/

typedef struct Commentstruct Commentstruct;

struct Commentstruct {
    char comms[COMMENTSIZE];    /* The actual text of the comment */
    Integer commn;              /* Link to the next comment       */
};

/* The current maximum number of comments */
static Integer maxcomms = 0;

/* never used */
/* The current actual number of comments */
/* static Integer ncomms = 0;  */

/* Array containing all comments */
static Commentstruct *dcomm;

/* points to the slot in commn which tells us what is the
 * next available space for comments.
 */
static Integer freecom = -1;

/*######### Package Information Formerly in Code_Info #######*/
         /* Note: package number 0 is a dummy. */
/* The number of packages installed: */
static Integer numpak = 0;

/* Default maximum number of iterations for execution: */
static Integer pakmaxi[MAXPAK + 1];

/* Abbreviated names of Packages: */
static char paknam[MAXPAK + 1][PNSIZE + 1];

/* Full names of packages, used for edits, etc. */
static char lngnam[MAXPAK + 1][LNGNAMSIZE + 1];

/* Package status. LOCALDB and USERDB are UP initially. */
static Integer pakstat[MAXPAK + 3];

/* number of firstpkg's including par. */
static Integer nfirstps = 0;

/* The first package to be made current, if any */
static char firstpkg[MAXPAK + 1][PNSIZE + 1];

/* number of foreign packages */
static integer numfrgn = 0;

/* package numbers of the foreign packages.
 * a foreign package is one whose variable database is not entirely
 * handled by basis.  A package is declared foreign via config. */
static Integer foreign[MAXPAK + 1];


/*#################### Static Variables #####################*/
static Integer zero = 0;

/*############ "Forward" function neclarations #############*/

void
    Rtguinfo(Integer jvar, Integer ndb, Address *fwa,
             Integer *codelen, Address *nname),
    Rtstc(Integer jvar, Integer ndb, Integer tc),
    Glbpknam(Integer i, char *pknam),
    Rtstgget(Integer jvar, Integer ndb, Integer jdimen,
             Integer *larg, Integer *jlow, Integer pad);

Integer
    Getfrgn(Integer k), Getnumf(), Rtgtc(Integer jvar, Integer ndb);

Address
    Rtgaddr(Integer jvar, Integer ndb);


/*--------------------------------------------------------------------------*/

/*                      STATIC AUXILIARY FUNCTIONS                          */

/*--------------------------------------------------------------------------*/

/* Create a variable's unique key from its jvar and ndb. */

Integer Rtkey(Integer jvar, Integer ndb)
{
    return jvar << LHASH | ndb;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void rdb_string_set(char *s, Integer n)
{
    Integer j;

    for (j = strlen(s); j < n; j++)
        s[j] = ' ';

    s[n] = '\0';

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void rdb_string_copy(char *d, char *s, Integer n)
{
    strncpy(d, s, n);
    d[n] = '\0';

    return;
}

/*--------------------------------------------------------------------------*/

/*                            DELETE A COMMENT                              */

/*--------------------------------------------------------------------------*/

 /*      Programmed by : Zane C. Motteler
  *      Date          : September 1989(In C, September 1999)
  *      Purpose       : Delete the comments of a variable or function when
  *                      it is about to be redefined. This is done recursively.
  *      Note on the C version: C subscripots are one less tha FORTRAN subscripts.
  *
  */
static void Rtcomrm(Integer *ncom)
{
    if (*ncom == -1)
        return;                 /*(no comments to remove) */
    if (dcomm[*ncom].commn != -1)
        /* recursively remove linked comments */
        Rtcomrm(&(dcomm[*ncom].commn));
    /* At this point only one comment is left, or there was only one to start
     * with. Remove it by adding the current comment to the free comment list.
     */
    dcomm[*ncom].commn = freecom;
    freecom = *ncom;
    *ncom = -1;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Rtcrack(Integer key, Integer *jvar, Integer *ndb)
{
    *ndb = key & MHASH;
    *jvar = key >> LHASH;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

 /* calculate n where typestr contains the string
  * character, character*n or character*(n) where n is an integer.
  * note that n cannot be an integer expression.
  */
static Integer Utchlen(char *typestr)
{
    int n;
    char *number;
    char msg[80];

    if (!strpbrk(typestr, "*")) {
        /* can't determine the length, so give 1 */
        return 1;
    } else if (!(number = strpbrk(typestr, "0123456789")) ||
             !(n = atoi(number))) {
        sprintf(msg, "Error in utchlen: type input was %s.", typestr);
        BasError(msg, strlen(msg));
        Kaboom(0);
    }
    return n;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTSETXIT - 
 * Turn on the variable rtexit. When rtexit is set, Rtdelet knows that
 * it is being called at program termination and therefore *all*
 * dynamic addresses must be freed. This is dangerous in the extreme
 * if you are not exiting
 */

void Rtsetxit()
{
    rtexit = 1;

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

Integer Utstrcod(char *typestr)
{
    char msg[80];
    int lstr;

    lstr = strlen(typestr);

    if (!strcmp(typestr, "integer"))
        return INTEGER;

    if (!strcmp(typestr, "real"))
        return (REAL);

    if (!strcmp(typestr, "complex"))
        return (COMPLEX);

    if (!strcmp(typestr, "logical"))
        return (LOGICAL);

    if (!strcmp(typestr, "range"))
        return (RANGE);

    if (!strncmp(typestr, "character", 9))
        return (-Utchlen(typestr));

    if (!strncmp(typestr, "double", 6)) {

/* double or double complex */
        if (lstr >= 13 && !strncmp(typestr + lstr - 7, "complex", 7))
            return (DOUBLE_C);
        else
            return (DOUBLE);
    }

    if (!strcmp(typestr, "name"))
        return (NAME);

    if (!strcmp(typestr, "string"))
        return (-MAXSTRING);

    if (!strcmp(typestr, "String"))
        return (STRING_ADDRESS);

    if (!strcmp(typestr, "function"))
        return (FUNCTION_TYPE);

    if (!strcmp(typestr, "chameleon"))
        return (INTEGER);

    if (!strcmp(typestr, "structure"))
        return (STRUCTURE);

    if (!strcmp(typestr, "external"))
        return (FUNCTION_TYPE);

    if (!strcmp(typestr, "Address") || !strcmp(typestr, "address"))
        return (FWA);

    if (!strcmp(typestr, "null"))
        return (NULL_TYPE);

    if (!strcmp(typestr, "indirect"))
        return (INDIRECT);

    if (!strcmp(typestr, "group"))
        return (GROUP);

    if (!strcmp(typestr, "left-hand side"))
        return (LHS);

    if (!strcmp(typestr, "word address") | !strcmp(typestr, "integer*8"))
        return (FWA);

    if (!strcmp(typestr, "variable in file"))
        return (FILE_VAR);

    sprintf(msg, "utstrcod: unknown type -- %s.", typestr);
    BasError(msg, strlen(msg));
    Kaboom(0);

    return ERR;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

 /* get min, max number of args from units for builtin function
  * units contains either a number or number1-number2
  * return ilow = number ihi=number or number1 and number2 resp.
  */
static void Rtfcnbc(char *units, Integer *ilow, Integer *ihi, char *vname)
{
    char *ih, *u1;
    char msg[80];

    ih = strpbrk(units, "-");

    u1 = strpbrk(units, "0123456789");
    if (!u1) {
        sprintf(msg, "Bad number of variables spec on builtin: %s.",
                units);
        BasError(msg, strlen(msg));
        sprintf(msg, "Function name is: %s.", vname);
        Kaboom(0);
    }
    ilow[0] = atoi(u1);
    if (!ih) {
        ihi[0] = ilow[0];
    } else {
        u1 = strpbrk(ih, "0123456789");
        if (!u1) {
            sprintf(msg, "Bad number of variables spec on builtin: %s.",
                    units);
            BasError(msg, strlen(msg));
            sprintf(msg, "Function name is: %s.", vname);
            Kaboom(0);
        }
        ihi[0] = atoi(u1);
    }
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

static void err9000(Integer jvar, Integer ndb, Integer *dlim)
{
    char *limstr, msg1[80],
        msg2[] = "Removing bad limit string; use setlimit to reset",
        vname[VARNAMESIZE + 1];

    INIT_BLANK_STRING(vname, VARNAMESIZE);

    Rtgvname(jvar, ndb, vname);
    sprintf(msg1, "rtxdb: unhappy with the limit string for %s", vname);
    BasError(msg1, strlen(msg1));

    limstr = FccMakeString((char *) 0, DIMSTRSIZE, 1);
    StrgGet(*dlim, limstr);
    BasError(limstr, strlen(limstr));

    FccFree(limstr);
    BasError(msg2, strlen(msg2));

    *dlim = 0;

    Kaboom(0);

    return;
}

/*--------------------------------------------------------------------------*/

/*                       DATABASE CREATION                                  */

/*--------------------------------------------------------------------------*/

/* RTDBCR -
 *       This routine creates a database holding ng groups
 *       and nv variables
 *
 *     AUTHOR...Paul F. Dubois
 *              Zane C. Motteler(C version, Sept. 1999)
 */

void Rtdbcr(Integer ndb, Integer ngarg, Integer nvarg)
{
    Integer ng, nv, grpsize, varsize;
    char msg[80];
    Package *pd;

/* Other routines assume dummy package 0 has priority 0. */
    if (ndb == 0) {
        database[0].pripkg = 0;
        return;
    }

    ng = (ngarg > 1) ? ngarg : 1;
    grpsize = (ng + 1) * sizeof(Group) / sizeof(Integer);
    nv = (nvarg > 1) ? nvarg : 1;
    varsize = nv * sizeof(Variable) / sizeof(Integer);

    pd = &database[ndb];

    pd->packno = ndb;
    pd->nvars0 = nvarg;
    pd->ngrps = 0;
    pd->maxgrps = ng;
    pd->nvars = 0;
    pd->maxvars = nv;
    pd->maxvari = 0;

/* Allocate space for Groups in this database */
    sprintf(msg, "Groups in pkg %ld", (long) ndb);
    Osallotn((void *) &(pd->group1), grpsize, "PERM|RTDBCR:pd->group1");
    pd->group1[0].gvmap = 0;

/* Allocate space for variables in this database */
    sprintf(msg, "Variables in %ld", (long) ndb);
    Osallotn((void *) &(pd->var1), varsize, "PERM|RTDBCR:pd->var1");

    return;
}

/*--------------------------------------------------------------------------*/

/*                        CREATE A GROUP                                    */

/*--------------------------------------------------------------------------*/

/* RTGRPE - begin a new group in a database */

Integer Rtgrpe(Integer ndb, char *group)
{
    Integer ng, mg, oldsize, newsize;
    char msg[80];
    Group *pg;

    ng = ++(database[ndb].ngrps);
    mg = database[ndb].maxgrps;
    if (ng > mg) {
        oldsize = (mg + 1) * sizeof(Group) / sizeof(Integer);
        mg += 10;
        newsize = (mg + 1) * sizeof(Group) / sizeof(Integer);

        database[ndb].maxgrps = mg;
        sprintf(msg, "Groups in pkg %ld", (long) ndb);

        Oschangn((Address *) &(database[ndb].group1), newsize, oldsize,
                 msg);
    }

    pg = database[ndb].group1;

/* start with no variables in the group: */
    pg[ng].gvmap = pg[ng - 1].gvmap;
    pg[ng - 1].grpnam = StrgFind(group);
    pg[ng - 1].remark = -1;
    pg[ng - 1].gattr = 0;

    return ng;
}

/*--------------------------------------------------------------------------*/

/*                  DELETE A VARIABLE'S STORAGE SPACE                       */

/*--------------------------------------------------------------------------*/

/* RTDELET - delete storage corresponding to entry nv in db
 * Revision 9/89(ZCM) -- remove a variable's comments when it is deleted
 * Revision 8/93(ZCM) -- remove a function's text when it is deleted
 * Rewritten in C 9/1999(ZCM)
 */

static void Rtdelet(Integer nv, Integer ndb)
{
    Integer df;
    Address ip;
    Address *ipp;
    Variable *pv;

    pv = &database[ndb].var1[nv - 1];
    df = pv->dynflg;

    if (df) {
        ipp = (Address *) (pv->addr);
        ip = *ipp;
        pv->elems = -1;

        if (pv->refcount > 0) {
            Osfreen(&ip, "Pointer to data.");
            pv->refcount--;
        }

        if (df == 2) {
            Osfreen((Address *) (&ipp), "Pointer to pointer to data.");
            pv->addr = (Address) NOTSET;
            pv->dynflg = 0;
        }

        if ((ndb == USERDB) || (ndb == LOCALDB) || (rtexit == 1)) {
            if (pv->setval != (Address) NOTSET) {
                Osfreen(&pv->setval, "initial value of above");
                pv->setval = (Address) NOTSET;
            }
            Rtcomrm(&pv->comm);
        }
        return;
    }

/* only in dbuser do we delete function space */
    if (database[ndb].packno != USERDB)
        return;

/* release other space attached to function */
    if (pv->types == FUNCTION_TYPE) {
        ip = pv->addr;
        pv->addr = (Address) NOTSET;
        Osfreen(&ip, "function code");

        ip = pv->argn;
        pv->argn = (Address) NOTSET;
        Osfreen(&ip, "function name");

        ip = pv->setval;
        pv->setval = (Address) NOTSET;
        Osfreen(&ip, "function text");

        pv->types = NULL_TYPE;
        Rtcomrm(&pv->comm);

        return;
    }
}

/*--------------------------------------------------------------------------*/

/*                       GET A SLOT IN THE DATABASE                         */

/*--------------------------------------------------------------------------*/

/* RTSLOT -
 * find a slot for a name in a database, return slot number
 * return 0 if error, otherwise return index of new variable
 *           Modified by P. A. Willmann 12/20/86(dimstr, origin and effdim).
 *     AUTHOR ...Paul Dubois 1/7/87
 *     MODIFIED...ZCM 7/91(remove direct references to verbose & debug)
 *                ZCM 2/96(in debug/verbose mode, check for redefining
 *                          any variable--users sometimes mask a Basis
 *                          function or variable by redefining.)
 *             FNF 7/29/97(Changed "write(STDERR,...)" statements to
 *                          "call baserror".)
 *             FNF 9/02/97(Copied vname to satisfy Cray compiler.)
 *             ZCM 9/26/1999 Reprogrammed in C.
 */

static Integer Rtslot(Integer ndb, char *vname)
{
    Integer nv, tc, ndb1, jvar1, jentry, mvold;
    Integer oldsize, mv, newsize, j, key, jj, nn;
    char msg[80], pk[PNSIZE];
    char *blk;
    Package *pd;
    Variable *pv;

    if (bashide == 1) {
        ndb1 = 0;
        jvar1 = Rtlookup(vname, &ndb1, &tc, -1);
        if (jvar1 != 0 && ndb1 != 0) {
            Glbpknam(ndb1, pk);
            if (strncmp(pk, "global  ", PNSIZE) &&
                strncmp(pk, "local   ", PNSIZE)) {
                /* user is attempting to redefine a variable in some pkg */
                if (pk[PNSIZE - 1] != ' ') {
                    sprintf(msg, "Redefined: %s will hide name %.8s%s.",
                            vname, pk, vname);
                } else {
                    blk = strchr(pk, ' ');
                    *blk = '\0';
                    sprintf(msg, "Redefined: %s will hide name %s%s.",
                            vname, pk, vname);
                }
                BasError(msg, strlen(msg));
            }
        }
    }

/* if this variable already in the database, replace */
    nv = Rtlookup(vname, &ndb, &tc, -1);

    pd = &database[ndb];

/* don't allow redefinition of a permanent variable */
    if (nv == 0)
        nv = pd->nvars + 1;

    else if (nv < pd->nvars0)
        return 0;

    else if (nv < pd->group1[pd->ngrps - 1].gvmap)
/* don't accept a match that is not in the last group
 * this forces variables declared inside functions to be local
 * and go into the function stack frame
 */
        nv = pd->nvars + 1;

    if (nv > pd->maxvari)
        pd->maxvari = nv;

    pv = &pd->var1[nv - 1];
    if (nv > pd->nvars) {       /* new variable */
        pd->nvars = nv;
        if (nv > pd->maxvars) {

/* variable space is used up; need some more. */
            mvold = pd->maxvars;
            oldsize = mvold * sizeof(Variable);
            mv = max(16, pd->maxvars * 3 / 2);
            newsize = mv * sizeof(Variable);
            pd->maxvars = mv;
            sprintf(msg, "variables in pkg %ld", (long) ndb);
            Oschangn((Address *) (&(pd->var1)), newsize, oldsize, msg);
            pv = &pd->var1[nv - 1];
        }

        pd->group1[pd->ngrps].gvmap += 1;

        if (pv->dynflg == 2) {
            Osfreen(&pv->addr, "RTSLOT:addr");
        }

        pv->setval = (Address) NOTSET;
        pv->elems = -1;
        pv->rely = -1;
        pv->addr = (Address) NOTSET;
        pv->refcount = 0;
        pv->comm = -1;
        pv->varnum = StrgFind(vname);
        pv->varnxt = 0;
        pv->varprv = 0;

        key = Rtkey(nv, ndb);
        jentry = IhashFind(vartable, pv->varnum);
        if (jentry > 0) {
            j = IhashGet(vartable, jentry);
            if (j > 0) {
/* already a variable by that name
 * --chain the new one in at the HEAD of the chain, ZCM 9/23/94
 */
                Rtcrack(j, &jj, &nn);
                pv->varnxt = j;
                database[nn].var1[jj - 1].varprv = key;
            }
        }

/* new logic, ZCM 9/23/94
 * if jentry <= 0, there never was a variable by that name.
 * if jentry > 0 and j <= 0, there used to be a variable by that name
 *    but it was FORGOTten.
 * if jentry > 0 and j > 0, there is a variable by this name but we
 *    want this new one to be put at the head of the hash chain.
 * So in all cases, we call ihashput with the key.
 */
        IhashPut(vartable, pv->varnum, key);
    } else {
/* old variable redefined
 * due to the above cautions, this can only be a user-defined function or
 * variable
 */
        if (debug == 1 && verbose == 1) {
            sprintf(msg, "Redefined: %s", vname);
            BasError(msg, strlen(msg));
        }
        Rtdelet(nv, ndb);
    }

/* inherit attributes of the group */
    pv->vattr = pd->group1[pd->ngrps - 1].gattr;

    return nv;
}

/*--------------------------------------------------------------------------*/

/*                       CREATE A VARIABLE                                  */

/*--------------------------------------------------------------------------*/

/* RTVARE - enter a variable into a database
 *     AUTHOR ...Paul Dubois 1/7/86
 *           Modified by P. A. Willmann 12/20/86(dimstr, origin and effdim).
 * return 0 if error, otherwise return index of new variable
 * Reprogrammed in C by Zane C. Motteler, September 1999
 */

Integer Rtvare(Integer ndb, char *vname, Address addr, Integer dynflg,
               char *type, char *dimsto, char *units)
{
    Address *iptr;
    Integer nv, ierr, ilow[7], ihi[7], ndim, nelts, i;
    char msg[80];
    Variable *pv;

    nv = Rtslot(ndb, vname);
    if (nv == 0)
        return 0;

    pv = &database[ndb].var1[nv - 1];

#if 0
    if (pv->addr != (Address) NOTSET) Kaboom(0);
#endif

    pv->addr = addr;
    pv->dynflg = dynflg;
    pv->types = Utstrcod(type);
    pv->dimsto = StrgFind(dimsto);
    if (dynflg) {
        iptr = (Address *) addr;
        *iptr = (Address) NOTSET;
    } else {
        Fpullapt(&ndb, &pv->dimsto, &zero, ilow, ihi, &ndim, &ierr);
        if (ierr == ERR) {
            sprintf(msg, "%s is a static variable which has been", vname);
            BasError(msg, strlen(msg));
            sprintf(msg, " improperly dimensioned %s.", dimsto);
            BasError(msg, strlen(msg));
            Kaboom(0);
        }
        nelts = 1;
        for (i = 0; i < ndim; i++) {
            pv->origin[i] = ilow[i];
            pv->dimstr[i] = ihi[i];
            nelts = nelts * (ihi[i] - ilow[i]) + 1;
        }
        pv->dimen = ndim;
        pv->elems = nelts;
    }

    pv->rely = -1;
    pv->units = StrgFind(units);
    pv->comm = -1;
    pv->chamel = (logical) (!strcmp(type, "chameleon"));
    pv->limstr = 0;
    pv->actstr = 0;

    return nv;
}

/*--------------------------------------------------------------------------*/

/*              INITIALIZE STORAGE FOR A NEW VARIABLE OR INDIRECT           */

/*--------------------------------------------------------------------------*/

/* RTSTGINI - Initialize storage to default values
 *
 * REVISION(ZCM) 08/19/97
 *    remove references to Domestic and replace with access functions
 * REWRITTEN IN C(ZCM) 10/06/1999
 */

void Rtstgini(Integer jvar, Integer ndb, Integer lenwrds)
{
    char vname[VARNAMESIZE + 1], msg[100];
    Integer tc;
    Address dval, fwa;
    Integer *ip;
    Address *ap;
    float *rp;
    double *fp;
    logical *lp;
    char *sp, *sv;
    int i;
    Variable *pv;

    pv = &database[ndb].var1[jvar - 1];

    tc = pv->types;

/* pointer to default value, if any: */
    dval = pv->setval;
    fwa = pv->addr;

    if (pv->dynflg)
        fwa = *((Address *) fwa);

    if (dval == (Address) NOTSET) {
/* no initial value was prescribed; so set it to zero or blanks */
/* (ZCM 12/15/1999) Address variables get set to NOTSET         */
/* (ZCM 3/27/2000)  That was the wrong thing to do! Make it NULL*/
        if (tc == FWA) {
            lenwrds = lenwrds * sizeof(Integer) / sizeof(Address);
            ap = (Address *) fwa;
            for (i = 0; i < lenwrds; i++)
                ap[i] = (Address) NULL;
        } else if (tc > 0) {
            memset(fwa, '\0', lenwrds * sizeof(Integer));
        } else {
            memset(fwa, ' ', lenwrds * sizeof(Integer));
        }

    } else if (tc >= 0) {
/* initial value is a numeric type, so case it out */
        switch (tc) {
        case INTEGER:
            ip = (Integer *) fwa;
            for (i = 0; i < lenwrds; i++)
                ip[i] = *((Integer *) dval);
            break;

        case REAL:
            rp = (float *) fwa;
            for (i = 0; i < lenwrds; i++)
                rp[i] = *((float *) dval);
            break;

#   if defined HAS_DOUBLE_COMPLEX
        case COMPLEX:
            rp = (float *) fwa;
            for (i = 0; i < lenwrds / 2; i++) {
                rp[2 * i] = *((float *) dval);
                rp[2 * i + 1] = (((float *) dval)[1]);
            }
            break;

        case DOUBLE_C:
            fp = (double *) fwa;
            for (i = 0; i < lenwrds / 2 * sizeof(Integer) / sizeof(double);
                 i++) {
                fp[2 * i] = *((double *) dval);
                fp[2 * i + 1] = ((double *) dval)[1];
            }
            break;

#   else

        case COMPLEX:
        case DOUBLE_C:
            rp = (float *) fwa;
            for (i = 0; i < lenwrds / 2; i++) {
                rp[2 * i] = *((float *) dval);
                rp[2 * i + 1] = *(((float *) dval)[1]);
            }
            break;
#   endif

        case DOUBLE:
            fp = (double *) fwa;
            for (i = 0; i < lenwrds * sizeof(Integer) / sizeof(double);
                 i++)
                fp[i] = *((double *) dval);
            break;

        case LOGICAL:
            lp = (logical *) fwa;
            for (i = 0; i < lenwrds; i++)
                lp[i] = *((logical *) dval);
            break;

        case FWA:
            lenwrds = lenwrds * sizeof(Integer) / sizeof(Address);
            ap = (Address *) fwa;
            for (i = 0; i < lenwrds; i++)
                ap[i] = *((Address *) dval);
            break;

        default:
/* unrecognizable type */
            INIT_BLANK_STRING(vname, VARNAMESIZE);

            Rtgvname(jvar, ndb, vname);
            sprintf(msg,
                    "rtstgini: Unable to initialize dynamic variable %s because of poor type.",
                    vname);
            BasError(msg, strlen(msg));
        }
    } else {
/* string variable */
        sp = (char *) Rtgaddr(jvar, ndb);
        sv = (char *) fwa;
        memset((void *) sp, *sv, lenwrds * sizeof(Integer));
    }
}

/*--------------------------------------------------------------------------*/

/*      MEMORY CHECKER CALLED BY GALLOT--ONLY PRINTS IF MEMTRACE == YES     */

/*--------------------------------------------------------------------------*/

void Rtmemchk(Integer ndb, Integer jvar1, Integer jvar2, Integer force)
{
    int i;
    char msg[MAXOUT];
    Address *fwa;
    Variable *pv;

    if (memtrace != 1 && force != 1)
        return;

    for (i = jvar1; i <= jvar2; i++) {
        pv = &database[ndb].var1[i - 1];

        if (pv->dynflg) {

/* Check addresses of dynamics */
            fwa = pv->addr;
            sprintf(msg, "Variable(%ld, %d): Address %p, Points to %p.",
                    (long) ndb, i, fwa, *fwa);

            if ((*fwa != (Address) NOTSET) && (*((Integer *) fwa) < 1000))
                strcat(msg, " <---######## SUSPICIOUS!\n");
            else
                strcat(msg, "\n");
            Remark(msg, strlen(msg));
        }
    }

    return;
}

/*--------------------------------------------------------------------------*/

/*                 GET STORAGE FOR A NEW VARIABLE OR INDIRECT               */

/*--------------------------------------------------------------------------*/

/*
 *     NAME OF ROUTINE...RTSTGGET
 *
 *     AUTHOR...Ron Litterst  8/29/84
 *                       modified 12/20/84 to make allocation of allocated
 *                       array not an error.  Release old instead.
 *       added 9/8/86 the allocated space is cleared to 0
 *       Modified 11/11/1986 to simplify
 *           Modified by     P. A. Willmann 12/20/86(dimstr, origin and effdim).
 *                       P. F. Dubois   12/16/87( jlow)
 *                       Zane C. Motteler 7/5/89(initialize dynamic
 *                             variables to specified value)
 *        Modified 8/23/89 by Dubois to add padding
 *
 *     PURPOSE...
 *
 *        This routine gets storage for a data base variable
 *        Dynamic storage for the array is allocated by calling the System
 *        Memory Manager Routines.  To do that it needs the number of words
 *        to allocate.  The length can be calculated by determining the
 *        dimensionality of the array from input arguments jdimen and jarg.
 *        Checks are made on the variable attributes and storage is
 *        alloted by calling mzmalloc.  If the storage is alloted
 *        then certain variable attributes will be set in the RT data
 *        base.  If the storage is not alloted then the remark
 *        subroutine will be called with an appropriate message,
 *        and basis will kaboom.
 *
 *     GLOSSARY OF IMPORTANT VARIABLES...
 *
 *        db         = callers RT data base manager
 *        jvar       = index of the variable
 *        jdimen     = dimensionality of variable
 *        dtypes     = variable data type, real, integer, etc
 *        daddr      = variable address or variable pointer address
 *        ddimstr    = variable upper subscripts
 *        lenelms    = allocation length in elements
 *        lenwrds    = allocation length in machine words
 *        jarg, jlow = lengths, origins desired
 *
 *# REVISION(ZCM) 08/19/97
 *#    remove references to Domestic and replace with access functions
 *# REWRITTEN IN C(ZCM 10/06/1999)
 */

void Rtstgget(Integer jvar, Integer ndb, Integer jdimen, Integer *jarg,
              Integer *jlow, Integer pad)
{
    Integer lenelms, i, lenbytes, lenwrds;
    char vname[VARNAMESIZE + 1], msg[80];
    Address *fwa;
    Variable *pv;

    pv = &database[ndb].var1[jvar - 1];

/* 
 *        set local pointer arrays and scalars that point to the caller's
 *        RT data base, needed for this routine
 *
 *        make sure variable is dynamic
 */

    if (!pv->dynflg) {
        INIT_BLANK_STRING(vname, VARNAMESIZE);

        Rtgvname(jvar, ndb, vname);
        sprintf(msg, "rtstgget: array is not a dynamic array -- %s.",
                vname);
        BasError(msg, strlen(msg));
        Kaboom(0);
    }
/*
 *        loop over array of arguments and calculate element array length
 */
    lenelms = 1;
    for (i = 0; i < jdimen; i++)
        lenelms *= jarg[i];
    lenelms += pad;
/*
 *        at this point lenelms is the number of elements to allocate and
 *        jdimen is the dimensionality of the array...
 *
 *        set fwa to be the pointer to the variable to be allocated
 */
    fwa = (Address *) (pv->addr);
/*
 *        if array already allotted, release old storage
 */
    if (*fwa != (Address) NOTSET) {
        Osfreen(fwa, "fwa in database");
        pv->refcount--;
    }
/*
 *       (lenelms) is the number of elements in the array, but not
 *        necessarily the number of MACHINE words to allocate on
 *        account of array type, use the function UTSIZEOF to get actual
 *        number of machine words to allocate
 */
    lenbytes = lenelms * Utbytesin(pv->types);
    lenwrds =
        lenbytes / sizeof(Integer) + (lenbytes % sizeof(Integer) ? 1 : 0);
/*
 *     Call the System Memory Manager to allocate requested storage
 *     for the array
 *
 *    (ZCM 9/2/1999) Don't twiddle dimensions if they involve arbitrary arrays.
 */
    if (lenwrds >= 0) {
        lenwrds = max(1, lenwrds);
        Osallotn(fwa, lenwrds, "RTSTGGET:fwa");
        pv->elems = lenelms;
        pv->rely = lenelms;
        pv->refcount++;
    } else {
        INIT_BLANK_STRING(vname, VARNAMESIZE);

        Rtgvname(jvar, ndb, vname);
        sprintf(msg, "rtstgget: Trying to allot array, %s of length < 0",
                vname);
        BasError(msg, strlen(msg));
        *fwa = (Address) NOTSET;
        Kaboom(0);
    }
    pv->dimen = jdimen;
    for (i = 0; i < jdimen; i++) {
        if (jlow[i] != DEFAULT + 1) {
            pv->origin[i] = jlow[i];
            pv->dimstr[i] = jarg[i] - 1 + jlow[i];
        } else {
            pv->origin[i] = 1;
            pv->dimstr[i] = jarg[i];
        }
    }
/* initialize the new space */
    Rtstgini(jvar, ndb, lenwrds);

    return;
}

/*--------------------------------------------------------------------------*/

/*               RELEASE DYNAMIC STORAGE PREVIOUSLY ALLOCATED               */

/*--------------------------------------------------------------------------*/

/*
 *     NAME OF ROUTINE...RTSTGDEL
 *
 *     AUTHOR...Ron Litterst  8/29/84
 *           Modified by P. A. Willmann 12/20/86(dimstr, origin and effdim).
 *
 *     PURPOSE...
 *
 *        This routine releases dynamic storage previously allocated
 *        to variable(name), which is in caller's RT data base.  This
 *        is accomplished by calling the System Memory Manager and
 *        resetting variable attributes in the RT data base.
 *
 *     GLOSSARY OF IMPORTANT VARIABLES...
 *
 *        ndb        = The package number
 *        jvar       = index of variable
 *
 * REPROGRAMMED IN C(ZCM) 10/07/1999
 */

void Rtstgdel(Integer jvar, Integer ndb)
{
    char vname[VARNAMESIZE + 1], msg[100];
    Address *pntname;
    Variable *pv;

    pv = &database[ndb].var1[jvar - 1];

    if (!pv->dynflg) {
        INIT_BLANK_STRING(vname, VARNAMESIZE);

        Rtgvname(jvar, ndb, vname);
        sprintf(msg, "rtstgdel: variable not dynamic -- %s.", vname);
        BasError(msg, strlen(msg));
        Kaboom(0);
    }

    pntname = (Address *) pv->addr;
    if (*pntname != (Address) NOTSET) {

/*
 *        array is actively allocated...
 *        now call the System Memory Manager to release storage
 */
        pv->elems = -1;
        Osfreen((Address *) (pntname), "data in database");
/*
 *        storage for variable(name) has been released, now set
 *        variable attributes in RT data base
 */
        *pntname = (Address) NOTSET;
        pv->refcount--;
    }

    return;
}

/*--------------------------------------------------------------------------*/

/*               CREATE A NEW INDIRECT(IMPLEMENTS DECLARATION)             */

/*--------------------------------------------------------------------------*/

/* RTNEWIND - given the stack elements svar, svals, creates a new
 * indirect variable in the database db
 * svar describes a name without subscripts
 * svals is an expression, must be supplied and of character type
 * for reduction of
 *       indirect name = exp
 *
 *     AUTHOR...Paul F. Dubois 5/31/86
 * ZCM 8/19/97 Replace Domestic references with access functions
 * revised 8/12/1999(ZCM) no longer needs sp as argument
 * rewritten in C 10/06/1999(ZCM)
 */

Integer Rtnewind(Integer ndb, Integer *svari, Integer *svalsi)
{
    Integer ndim, tc, jlen[7], jorg[7], jvar;
    static char
        msg1[] = "Dimensions not allowed on indirect declaration",
        msg2[] = "Incorrect indirect declaration.  Correct is:",
        msg3[] = "indirect varname = character-expression",
        msg4[] = "rtnewind: improper assignment of values";

/* 1. Consistency checks */
    if (svari[NS - 1] != 0) {
        BasError(msg1, strlen(msg1));
        return 0;
    }

    if (svalsi[TC - 1] >= 0) {
        BasError(msg2, strlen(msg2));
        BasError(msg3, strlen(msg3));
        return 0;
    }

/* 2. place name in database */
    tc = -(VARNAMESIZE);
    jvar = Rtinname(ndb, tc, svari);
    if (jvar == 0)
        return 0;
    ndim = 0;

/* 3. allocate storage */
    Rtstgget(jvar, ndb, ndim, jlen, jorg, 0);

/* 4. store initial value */
/* ##### ZCM 8/3/95 release space if in literal table */
    if (svari[TC - 1] == NAME && svari[NAML - 1] < 0)
        Parrel(svari);

    svari[TC - 1] = LHS;
    svari[N - 1] = ndb;
    svari[TCSAVE - 1] = tc;
    svari[PTR - 1] = jvar;
    svari[NS - 1] = 0;

/* call parasgn to do the work; parasgn releases storage for svals */
    if (Parasgn(svari, svalsi) == ERR) {
        BasError(msg4, strlen(msg4));
        return 0;
    }

/* 5. Throw the indirect switch */
    Rtstc(jvar, ndb, INDIRECT);

/* in case name was calculated... */
    Parkill(svari);

    return jvar;
}

/*--------------------------------------------------------------------------*/

/*               CREATE A NEW VARIABLE(IMPLEMENTS DECLARATION)             */

/*--------------------------------------------------------------------------*/

/* RTNEWVAR - given the stack elements svar, svals, creates a
 * new variable in the database db of type tc
 * svar describes a name with or without subscripts
 * svals is an expression, or maybe a NULL_TYPE if not supplied
 * for reduction of
 *       type name(dims) = exp
 * where =exp and(dims) is optional
 *
 *     AUTHOR...Paul F. Dubois 7/28/86
 *           Modified by P. A. Willmann 12/20/86(dimstr, origin and effdim).
 *           Modified by Z. C. Motteler 6/24/87(rtinname extracted)
 * revised 8/29/86 to make vals an expression
 * revised 6/15/86 to make db an argument
 * revised 5/31/88 to add indirect
 * revised 6/23/89 to return jvar instead of OK
 * revised 8/8/89 to return 0 if more than 7 dimensions are specified(ZCM)
 * revised 8/12/1999(ZCM) no longer needs sp as argument
 * rewritten in C 10/06/1999(ZCM)
 */

Integer Rtnewvar(Integer ndb, Integer tc, Integer *svari, Integer *svalsi)
{
    Integer ndim, i, jlen[7], jorg[7], jvar;
    static char
        msg1[] = "rtnewvar: variable declared with too many subscripts.",
        msg2[] = "rtnewvar: improper assignment of values",
        msg3[] = "(but variable was created o.k.)";

/* 0. check for indirect, handled differently */
    if (tc == INDIRECT)
        return Rtnewind(ndb, svari, svalsi);

/* 0.5 allow type name=array */
    if (svari[NS - 1] == 0)
        if (svalsi[TC - 1] != NULL_TYPE) {
            svari[NS - 1] = svalsi[N - 1];
            for (i = 0; i < svalsi[N - 1]; i++) {
                svari[N1 - 1 + i] = svalsi[M1 - 1 + i];
                svari[M1 - 1 + i] =
                    svalsi[M1 - 1 + i] + svalsi[N1 - 1 + i] - 1;
                svari[I1 - 1 + i] = 1;
            }
        }

/* 1. place name in database */
    if (svari[NS - 1] > 7) {
        BasError(msg1, strlen(msg1));
        return 0;
    }
    jvar = Rtinname(ndb, tc, svari);
    if (jvar == 0)
        return 0;
    ndim = svari[NS - 1];
    for (i = 0; i < ndim; i++) {
        jorg[i] = svari[N1 - 1 + i];
        jlen[i] = svari[M1 - 1 + i] - svari[N1 - 1 + i] + 1;
    }

/* 2. allocate storage */
    Rtstgget(jvar, ndb, ndim, jlen, jorg, 0);

/* 3. store initial values supplied if any */
    if (svalsi[TC - 1] != NULL_TYPE) {

/* Call parrel or it will stick around forever(ZCM 10/2/95) */
        Parrel(svari);

/*  a. make the variable name into a <lhs> with no subscripts supplied */
        svari[NS - 1] = 0;
        svari[N - 1] = ndb;
        svari[PTR - 1] = jvar;
        svari[TC - 1] = LHS;
        svari[TCSAVE - 1] = tc;

/*  b. call parasgn to do the work; parasgn releases storage for svalsi */
        if (Parasgn(svari, svalsi) == ERR) {
            BasError(msg2, strlen(msg2));
            BasError(msg3, strlen(msg3));
            return 0;
        }
    }

/* in case name was calculated... */
    Parkill(svari);
    return jvar;
}

/*--------------------------------------------------------------------------*/

/*                        CREATE A BUILTIN FUNCTION                         */

/*--------------------------------------------------------------------------*/

/* RTFCNEB - enter a built-in function into a database
 *     AUTHOR ...Paul Dubois 1/7/86
 *     REWRITTEN in C ...Zane C. Motteler 9/1999
 * return 0 if error, otherwise return index where installed
 * dimsto contains symbolic but otherwised unused calling seq
 * units contains number of arguments as int or int-int
 */

Integer Rtfcneb(Integer ndb, char *vname, char *dimsto, char *units)
{
    Integer nv, i, j;
    Integer vnameindex;
    Variable *pv;

    nv = Rtslot(ndb, vname);
    if (nv == 0)
        return nv;

    vnameindex = StrgFind(vname);

    pv = &database[ndb].var1[nv - 1];

    pv->units = StrgFind(units);
    Rtfcnbc(units, &i, &j, vname);
    pv->types = FUNCTION_TYPE;
    pv->dimsto = StrgFind(dimsto);
    pv->dynflg = 0;
    pv->comm = -1;
    pv->funtyp = 0;             /*builtin */
    pv->origin[0] = i;
    pv->origin[1] = ndb;
    pv->dimstr[0] = j;
    pv->dimen = Funparin(vnameindex, zero);

    return nv;
}

/*--------------------------------------------------------------------------*/

/*                    CREATE A USER-DEFINED FUNCTION                        */

/*--------------------------------------------------------------------------*/

/* RTFCNEU - create database entry for user function
 * return index in dbuser
 * Reprogrammed in C(9/1999, ZCM)
 */

Integer Rtfcneu(logical *hastext)
{
    NameType *name;
    char vname[VARNAMESIZE + 1];
    char dimsto[DIMSTRSIZE + 1];
    static char msg[] = "rtfcneu: dimension string > 500 characters long.";
    int i, nv;
    Variable *pv;

    Rtufnam((Address *) (&name));
    strncpy(vname, name->names, VARNAMESIZE - 1);

/* Truncate trailing blanks off name: */
    for (i = VARNAMESIZE - 2; i >= 0; i--) {
        if (vname[i] != ' ') {
            vname[i + 1] = '\0';
            break;
        }
    }
    nv = Rtslot(USERDB, vname);
    if (nv == 0)
        return 0;

    pv = &dbuser.var1[nv - 1];

    pv->funtyp = 1;             /* user */
    pv->types = FUNCTION_TYPE;
    pv->comm = -1;
    pv->dynflg = 0;
    pv->units = StrgFind(" ");
    pv->origin[1] = USERDB;
    Rtufcn(&pv->addr, &pv->dimen, &pv->origin[0],
           &pv->dimstr[0], &pv->dimstr[2],
           &pv->argn, &pv->dimstr[5], &pv->setval, *hastext);

    if (pv->dimstr[0] == 0) {
        pv->dimsto = StrgFind("()");
        return nv;
    } else {
        dimsto[0] = '(';
        dimsto[1] = '\0';
        for (i = 1; i <= pv->dimstr[0]; i++) {
            if (i != 1)
                strcat(dimsto, ",");
            if (strlen(dimsto) + strlen(name[i].names) + 1 > DIMSTRSIZE) {
                BasError(msg, strlen(msg));
                Kaboom(0);
            }
            strcat(dimsto, name[i].names);
        }
        strcat(dimsto, ")");
        pv->dimsto = StrgFind(dimsto);
        return nv;
    }
}

/*--------------------------------------------------------------------------*/

/*           CREATE A FUNCTION THAT WAS AN ARGUMENT TO A USER FUNCTION      */

/*--------------------------------------------------------------------------*/

/* RTFCNFP - create database entry for a function that was an argument to a
 *         - user-defined function
 */

Integer Rtfcnfp(char *vname, Integer *svalsi)
{
    Integer nv;
    Variable *pv;

    nv = Rtslot(LOCALDB, vname);
    if (nv == 0)
        return (ERR);

    pv = &dblocal.var1[nv - 1];

    pv->types = FUNCTION_TYPE;
    pv->dynflg = 0;
    pv->comm = -1;

    if (svalsi[N2 - 1] > 0)
/* special stuff for user function */
        Rtguinfo(svalsi[PTR - 1], svalsi[N - 1], &pv->addr,
                 &pv->dimstr[2], &pv->argn);

    else
/* allows for indirection, points to the database entry for this function */
        pv->dimen = Rtkey(svalsi[PTR - 1], svalsi[N - 1]);

    pv->funtyp = svalsi[N2 - 1];
    pv->origin[0] = svalsi[N1 - 1];     /* minargs */
    pv->origin[1] = svalsi[N - 1];      /* ndb */
    pv->dimstr[0] = svalsi[M1 - 1];     /* maxargs */
    pv->dimsto = StrgFind("(?)");
    pv->units = StrgFind(" ");

    return OK;
}

/*--------------------------------------------------------------------------*/

/*               CREATE A COMPILED FUNCTION                             
 *           (1) INITIALIZE THE DATABASE ENTRY                         */

/*--------------------------------------------------------------------------*/

/* RTFCNE1 - enter a compiled function into a database
 *     AUTHOR ...Paul Dubois 1/26/86
 * revised and broken up into rtfcne1, rtfcne2, and rtfcne3
 *     By     ...Zane C. Motteler August 1994
 * rtfcne1 enters preliminary information into database
 * rtfcne2 is called once for each parameter string and opening and closing
 *         parens
 * rtfcne3 is called to complete the work(parse the parameter string,
 *         set up parameter descriptions, and deallocate space)
 *     Bug fix 4/10/95 to set dargn, dargt to point to list of
 *                     names and types
 * Reprogrammed in C 9/1999(ZCM)
 */

Integer Rtfcne1(Integer ndb, char *vname, Integer iseg,
                char *typestr, char *units)
{
    Integer nv, itype, idum, idim, dimlen;
    char msg[80], *ch;
    Variable *pv;

    nv = Rtslot(ndb, vname);
    if (nv <= 0) {
        sprintf(msg, "rtfcne1: attempt to redefine %s.", vname);
        BasError(msg, strlen(msg));
        return ERR;
    }

    itype = Utstrcod(typestr);
    pv = &database[ndb].var1[nv - 1];

    pv->types = FUNCTION_TYPE;
    pv->dynflg = 0;
    pv->comm = -1;
    pv->dimsto = 0;
    pv->units = StrgFind(units);
    pv->funtyp = -iseg;
    idum = StrgFind(vname);
    pv->dimen = Funparin(idum, itype);
    pv->argn = (Address) (funpar + pv->dimen - 1);

/* the philosophy is now: max. and min. of args will be kept
 * to -1 until call to rtfcne3, when parameter string will be
 * parsed and they will be evaluated.
 */
    pv->origin[0] = -1;         /* max no of args */
    pv->dimstr[0] = -1;         /* min no of args */
    pv->origin[1] = ndb;

    dimlen = MAXSTRING;         /* Initially */
    pv->dimstr[1] = dimlen;     /* length of dimension string */
    idim = dimlen / sizeof(Integer);
    if (dimlen % sizeof(Integer))
        idim += 1;

    Osallotn((Address) (&ch), idim, "RTFCNE1:ch");
    ch[0] = '\0';

    pv->setval = (Address) ch;
    return nv;
}

/*--------------------------------------------------------------------------*/

/*               CREATE A COMPILED FUNCTION                             
 *           (2) ENTER AN ARGUMENT SPECIFICATION                       */

/*--------------------------------------------------------------------------*/

/* RTFCNE2 
 *     Programmed by Zane C. Motteler
 *                   August 15, 1994
 *     Rewritten in C by Zane C. Motteler
 *                   September 28, 1999
 *     Purpose--get around the 500 character limitation by storing the
 *              parameter descriptions in a large character*1 array.
 *              space already allocated in rtfcne1. This routine needs
 *              to be called at least twice(once for opening paren,
 *              once for closing paren).
 */

Integer Rtfcne2(Integer jvar, Integer ndb, char *dimen)
{
    char vname[VARNAMESIZE + 1], msg[80];
    Integer crdimlen, idim, idimnew, dimlen;
    char *ch;
    Variable *pv;

    pv = &database[ndb].var1[jvar - 1];

    if (pv->types != FUNCTION_TYPE) {
        INIT_BLANK_STRING(vname, VARNAMESIZE);

        Rtgvname(jvar, ndb, vname);
        sprintf(msg, "rtfcne2: called on behalf of non-function %s.",
                vname);
        BasError(msg, strlen(msg));
        return 0;
    }

    if (pv->funtyp >= 0) {
        INIT_BLANK_STRING(vname, VARNAMESIZE);

        Rtgvname(jvar, ndb, vname);
        sprintf(msg,
                "rtfcne2: called on behalf of non-compiled function %s.",
                vname);
        BasError(msg, strlen(msg));
        return ERR;
    }

    crdimlen = strlen(dimen);
    ch = (char *) pv->setval;
    if (crdimlen + strlen(ch) + 1 > pv->dimstr[1]) {
        dimlen = pv->dimstr[1];
        idim = dimlen / sizeof(Integer);
        if (dimlen % sizeof(Integer))
            idim += 1;
        idimnew = (dimlen + MAXSTRING) / sizeof(Integer);

        if ((dimlen + MAXSTRING) % sizeof(Integer))
            idimnew += 1;

        Oschangn((Address *) (&ch), idimnew, idim, "dimension string");
        pv->dimstr[1] += MAXSTRING;
        pv->setval = (Address) ch;
    }

    strcat(ch, dimen);
    return (jvar);
}

/*--------------------------------------------------------------------------*/

/*               CREATE A COMPILED FUNCTION                             
 *           (3) PARSE THE ARGUMENT STRING & FINISH                    */

/*--------------------------------------------------------------------------*/

/* RTFCNE3
 *     Programmed by Zane C. Motteler
 *                   August 15, 1994
 *     Reprogrammed in C by Zane C. Motteler
 *                   September 28, 1999
 *     Purpose--Parse the parameter string of a compiled function to
 *              count arguments and give the right data types.
 *              presumes that mac squeezed all blanks out of dimension
 *              information before sending it over, and that all
 *              variables have had their types(maybe default)
 *              filled in by mac. Thus the string should be of the
 *              form "()" or "(spec[,spec...])" where spec is
 *                    name:type
 *     Revised 9/1/94 to use character instead of integer(fin/ifin)
 */

Integer Rtfcne3(Integer jvar, Integer ndb)
{
    char *ch, *chi, msg[90], fin;
    Integer crdimlen, nargs, nargsm, idone, istart, itype;
    Integer typelen, namlen, idum, jextra;
    char vname[VARNAMESIZE + 1], ptype[17];
    Variable *pv;

    pv = &database[ndb].var1[jvar - 1];

    if (pv->types != FUNCTION_TYPE) {
        INIT_BLANK_STRING(vname, VARNAMESIZE);

        Rtgvname(jvar, ndb, vname);
        sprintf(msg, "rtfcne3: called on behalf of non-function %s.",
                vname);
        BasError(msg, strlen(msg));
        return 0;
    }

    if (pv->funtyp >= 0) {
        INIT_BLANK_STRING(vname, VARNAMESIZE);

        Rtgvname(jvar, ndb, vname);
        sprintf(msg,
                "rtfcne3: called on behalf of non-compiled function %s.",
                vname);
        BasError(msg, strlen(msg));
        return ERR;
    }

    ch = (char *) pv->setval;
    crdimlen = strlen(ch);
    nargs = 0;
    nargsm = MAXCFARGS;
    jextra = 0;
    if (crdimlen != 2) {        /*(The case "()") */
        idone = 0;              /* marks end of last token */
        istart = 1;
        fin = ch[istart];
        if (fin == ';')
            idone = istart;
        do {
            if (fin == ';') {
                nargsm = min(nargsm, nargs);    /* first semicolon is one that counts */
                istart++;
            }
            idone = istart;
            if (idone >= crdimlen - 1) {
                INIT_BLANK_STRING(vname, VARNAMESIZE);

                Rtgvname(jvar, ndb, vname);
                sprintf(msg,
                        "rtfcne3: dimension string for %s not terminated by ')'??",
                        vname);
                BasError(msg, strlen(msg));
                return (0);
            }

/* NOTE: blank, comma, rparen, and semicolon not possible
 *       since mac always fills in a type.
 */
            chi = strpbrk(ch + idone, ":");
            namlen = chi - (ch + idone);
            strncpy(vname, ch + idone, namlen);
            vname[namlen] = '\0';
            idone += namlen + 1;        /* move past colon */
            chi = strpbrk(ch + idone, ",;)");
            typelen = chi - (ch + idone);
            fin = *chi;

            strncpy(ptype, ch + idone, typelen);
            ptype[typelen] = '\0';
            idone += typelen;

            itype = Utstrcod(ptype);
            if (itype == STRING_ADDRESS)
                jextra += 1;
            nargs += 1;
            idum = StrgFind(vname);
            Funparin(idum, itype);
            istart = fin == ';' ? idone : idone + 1;
        } while (fin != ')');
    }

    if (nargs > MAXCFARGS) {
        INIT_BLANK_STRING(vname, VARNAMESIZE);

        Rtgvname(jvar, ndb, vname);
        sprintf(msg, "rtfcne3: Too many arguments for %s!", vname);
        BasError(msg, strlen(msg));
        return 0;
    }

    nargsm = min(nargsm, nargs);
    pv->origin[0] = nargsm;
    pv->origin[2] = jextra;
    pv->dimstr[0] = nargs;

/* now free up the character array; we're done with it */
    Osfreen((Address *) (&ch), "dimension string");

    pv->dimstr[1] = 0;
    pv->dimstr[2] = 0;
    pv->setval = (Address) NOTSET;

    return jvar;
}

/*--------------------------------------------------------------------------*/

/*                          SET THE SHAPE OF AN ARRAY                       */

/*--------------------------------------------------------------------------*/

/*   NAME OF ROUTINE...SETSHAPE
 *
 *   PURPOSE: To reset shape of array, permanently
 *
 *   NAME OF AUTHOR...Paul F. Dubois 6/11/88
 *
 *   3/29/1999(ZCM) Revised for new dynamic string paradigm
 *   9/28/1999(ZCM) Rewritten in C
 */

Integer Csetshap(char *name, char *limits)
{
    Integer jvar, ndb, tc, ndim, nelt;
    Integer ilow[7], ihi[7], ierr, i;
    Address ip;
    Variable *pv;
    static char
        msg1[] = "setshape: cannot call with a group or function name:",
        msg2[] = "setshape-- variable:",
        msg3[] = "could not evaluate dimensions:";

    if (Rtfinder(name, &jvar, &ndb, &tc, "setshape") == ERR)
        return ERR;

    if (tc == GROUP || tc == FUNCTION_TYPE) {
        BasError(msg1, strlen(msg1));
        BasError(name, strlen(name));
        return ERR;
    }

/* this service not available to foreigners */
    if (Rtnofrgn("setshape", jvar, ndb) == ERR) {
        BasError(name, strlen(name));
        return ERR;
    }

    pv = &database[ndb].var1[jvar - 1];

    pv->dimsto = StrgFind(limits);
    Fpullapt(&ndb, &(pv->dimsto), &zero, ilow, ihi, &ndim, &ierr);
    if (ierr == ERR) {
        BasError(msg2, strlen(msg2));
        BasError(name, strlen(name));
        BasError(msg3, strlen(msg3));
        BasError(limits, strlen(limits));
        return ERR;
    }

    pv->dimen = ndim;
    nelt = 1;
    for (i = 0; i < ndim; i++) {
        pv->origin[i] = ilow[i];
        pv->dimstr[i] = ihi[i];

        nelt *= (ihi[i] - ilow[i] + 1);
    }

    if (pv->elems == -1) {
/* user allocated space behind our back?
 * variable must be dynamic; if the FWA is set, believe user about space.
 */
        ip = pv->addr;
        if (ip != (Address) NOTSET) {
            pv->elems = nelt;
            pv->rely = nelt;
        }
    }

    return OK;
}

/*--------------------------------------------------------------------------*/

/*                    SET THE MINIMUM NUMBER OF ARGUMENTS                   */

/*--------------------------------------------------------------------------*/

/*
 *   NAME OF ROUTINE...SETMNARG
 *
 *   PURPOSE: To reset minimum number of permitted arguments
 *
 *   NAME OF AUTHOR...Paul Dubois 2/12/87
 *   REWRITTEN IN C...Zane C. Motteler 9/29/1999
 *
 *(FNF 7/28/97) Changed "write(STDERR,...)" statement to "call baserror".
 */

Integer Csetmnar(char *name, Integer *imin)
{
    Integer tc, jvar, ndb, imn;
    Variable *pv;
    char errmsg[80];
    static char
        msg1[] = "setmnarg: was not called with a function name:",
        msg2[] = "setmnarg: argument was a built-in function name:",
        msg3[] = "setmnarg: minimum requested exceeds maximum";

    imn = *imin;

    if (Rtfinder(name, &jvar, &ndb, &tc, "setmnarg") == ERR)
        return ERR;

    if (tc != FUNCTION_TYPE) {
        BasError(msg1, strlen(msg1));
        BasError(name, strlen(name));
        return ERR;
    }

/* this service not available to foreigners */
    if (Rtnofrgn("setmnarg", jvar, ndb) == ERR) {
        BasError(name, strlen(name));
        return ERR;
    }

    pv = &database[ndb].var1[jvar - 1];

    if (pv->funtyp == 0) {
        BasError(msg2, strlen(msg2));
        BasError(name, strlen(name));
        return ERR;
    }

    if (imn > pv->dimstr[0]) {
        BasError(msg3, strlen(msg3));
        BasError(name, strlen(name));
        sprintf(errmsg,
                "Minimum number of arguments requested = %ld, max = %ld.",
                (long) imn, (long) pv->dimstr[0]);
        BasError(errmsg, strlen(errmsg));
        return ERR;
    }

    pv->origin[0] = max(0, imn);
    return OK;

}

/*--------------------------------------------------------------------------*/

/*                    SET THE LENGTH OF THE LAST SUBSCRIPT                  */

/*--------------------------------------------------------------------------*/

/* RTSETDL - set the length of the last subscript to irec
 * increase storage if required so that this becomes legal
 * remove reference to autobump(ZCM, 7/91)
 * change to exponential growth model(PFD, 8/98)
 * Reprogram in C(ZCM, 9/1999)
 */

Integer Rtsetdl(Integer jvar, Integer ndb, Integer irec)
{
    Integer n, ndim, newsize;
    Variable *pv;
    static char errmsg[] = "rtsetdl: called on a scalar!";

    pv = &database[ndb].var1[jvar - 1];
    if (pv->dimen == 0) {
        BasError(errmsg, strlen(errmsg));
        return ERR;
    }

    newsize = max(irec, 0);
    pv->limstr = -(newsize + 1);
    ndim = pv->dimen;
    n = pv->dimstr[ndim - 1] - pv->origin[ndim - 1] + 1;

    if (newsize > n) {
/* exponential growth, growth factor less than golden ratio
 * to minimize fragmentation
 */
        n *= 1.5;
        n = max(max(n, 16), newsize);
        return Changej(&jvar, &ndb, &n);
    }

    return OK;
}

/*--------------------------------------------------------------------------*/

/*                    ADD A DIMENSION OF 1 AS NEXT SUBSCRIPT                */

/*--------------------------------------------------------------------------*/

/* RTUP1D - Add a dimension of 1 as the next subscript */

void Rtup1d(Integer jvar, Integer ndb)
{
    char newd[DIMSTRSIZE + 1];
    int ndim, j;
    Variable *pv;

    pv = &database[ndb].var1[jvar - 1];

    ndim = pv->dimen += 1;
    pv->origin[ndim - 1] = 1;
    pv->dimstr[ndim - 1] = 1;
    if (ndim == 1) {
        pv->dimsto = StrgFind("(1)");
    } else {
        INIT_BLANK_STRING(newd, DIMSTRSIZE + 1);

        StrgGet(pv->dimsto, newd);

        j = Utgetcl(newd);
        strcpy(newd + j - 1, ",1)");

        pv->dimsto = StrgFind(newd);
    }

    return;
}

/*--------------------------------------------------------------------------*/

/*                    ADD A DIMENSION OF 1 TO THE VARIABLE                  */

/*--------------------------------------------------------------------------*/

/* CRTADDIM - adds a dimension of length one to the variable */

Integer Crtaddim(char *name)
{
    Integer jvar, tc, ndb;
    char msg1[] = "rtadddim: cannot call with a group or function name:",
        msg2[] = "rtadddim: already 7-D!";

    if (Rtfinder(name, &jvar, &ndb, &tc, "rtadddim") == ERR)
        return ERR;

    if (tc == GROUP || tc == FUNCTION_TYPE) {
        BasError(msg1, strlen(msg1));
        BasError(name, strlen(name));
        return ERR;
    }

/* this service not available to foreigners */
    if (Rtnofrgn("rtadddim", jvar, ndb) == ERR) {
        BasError(name, strlen(name));
        return ERR;
    }

    if (database[ndb].var1[jvar - 1].dimen == 7) {
        BasError(msg2, strlen(msg2));
        BasError(name, strlen(name));
        return ERR;
    }

    Rtup1d(jvar, ndb);

    return OK;
}

/*--------------------------------------------------------------------------*/

/*                      ADD ATTRIBUTE TO A GROUP                            */

/*--------------------------------------------------------------------------*/

/* RTGATTRE - add the attribute to the group jgrp in database ndb */

void Rtgattre(Integer jgrp, Integer ndb, char *attr)
{
    int n;

    n = Rtattrin(attr);

    Rtsattr(&(database[ndb].group1[jgrp - 1].gattr), n);

    return;
}

/*--------------------------------------------------------------------------*/

/* THE FOLLOWING SUBROUTINES BEGINNING WITH "rts" SET CERTAIN VALUES  
 * IN THE DATABASE.                                                         */

/*--------------------------------------------------------------------------*/

/* RTSDELMS -
 *       Programmed by Zane C. Motteler
 *       19 August 1997
 *      (Reprogrammed in C 29 September 1999)
 *       purpose -- set the value of delems in the database
 */

void Rtsdelms(Integer jvar, Integer ndb, Integer ival)
{
    database[ndb].var1[jvar - 1].elems = ival;

    return;
}

/*--------------------------------------------------------------------------*/

/*                      ROUTINES WHICH CHANGE THE DATABASE                  */

/*--------------------------------------------------------------------------*/

void Rtsetuf(Integer n)
{
    if (n == 1) {
        database[USERDB].pripkg = 1;
    } else {
        database[USERDB].pripkg = prihigh + 1;
        database[LOCALDB].pripkg = prihigh + 2;
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTSNV0 - set nvars0 in the database */

void Rtsnv0(Integer npar)
{
    database[npar].nvars0 = database[npar].nvars;

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTSLIMST - set the string number of the limits string */

void Rtslimst(Integer jvar, Integer ndb, Integer ival)
{
    database[ndb].var1[jvar - 1].limstr = ival;

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTSVATTR - set the attribute word of a variable */

void Rtsvattr(Integer jvar, Integer ndb, Integer iattr)
{
    database[ndb].var1[jvar - 1].vattr = iattr;

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTSNASTR - set the number of the action string in the string table */

void Rtsnastr(Integer jvar, Integer ndb, Integer ival)
{
    database[ndb].var1[jvar - 1].actstr = ival;

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Rtsloc(Integer n)
{
    locdepth = n;

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTSTC - set the type code of a database entry */

void Rtstc(Integer jvar, Integer ndb, Integer tc)
{
    database[ndb].var1[jvar - 1].types = tc;

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTINCREF - Increment a variable's reference count.
 */

void Rtincref(Integer jvar, Integer ndb)
{
    database[ndb].var1[jvar - 1].refcount++;

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTSGATTR - replace the group's attribute word with the new one */

void Rtsgattr(Integer ndb, Integer ngrp, Integer iattr)
{
    database[ndb].group1[ngrp - 1].gattr = iattr;

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Seemom()
{
    locdepth += 1;

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/*
 * rtdynset                                                              #
 * Programmed By: Zane C. Motteler                                       #
 *                                                                       #
 * Date:          30 June 1989                                           #
 *                                                                       #
 * Purpose:       Set the initial value for a dynamic variable, so that  #
 *                when it is allocated, it can be initialized to the     #
 *                specified value. This routine works for all types of   #
 *                variables except strings.                              #
 *                                                                       #
 */

void Rtdynset(Integer jvar, Integer ndb, Address value)
{
    Integer nbytes, nwords;
    Variable *pv;
    Address dval;

    pv = &database[ndb].var1[jvar - 1];
    nbytes = Utbytesin(pv->types);
    nwords = nbytes / sizeof(Integer) + (nbytes % sizeof(Integer) ? 1 : 0);

    Osallotn(&dval, nwords, "RTDYNSET:dval");
    pv->setval = dval;

    memcpy(dval, value, nbytes);

    return;
}

/*--------------------------------------------------------------------------*/

/*              SET INITIAL VALUE OF A DYNAMIC STRING VARIABLE          */

/*--------------------------------------------------------------------------*/

/*
 * rtdynstc
 * Programmed By: Zane C. Motteler
 *
 * Date:        30 June 1989
 *              Reprogrammed in C 2 December 1999
 *
 * Purpose:     Set the initial value for a dynamic variable, so that
 *              when it is allocated, it can be initialized to the
 *              specified value. This routine works for string  
 *              variables.
 *
 */

void Rtdynstc(Integer jvar, Integer ndb, char *value)
{
    Integer slen, ilen;
    Address pval;

    slen = -Rtgtc(jvar, ndb);
    ilen = slen / sizeof(Integer) + (slen % sizeof(Integer) ? 1 : 0);
    Osallotn(&pval, ilen, "RTDYNSTC:pval");
    database[ndb].var1[jvar - 1].setval = pval;

    if (strlen(value) < slen) {
        memset((char *) pval, (int) ' ', slen);
        strncpy((char *) pval, value, strlen(value));

    } else {
        strncpy((char *) pval, value, slen);
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTFIN - call the package fins */

void Rtfin(Integer ies)
{
    Integer i, idum;
    char msg[120];
    static int first = TRUE;

/* make sure this is one just once no matter how many times its called */
    if (first) {
        first = FALSE;
        idum  = 0;

        for (i = 2; i <= numpak; i++) {
            if (pakstat[i] != UP) continue;
            if (Glbfin(&i, &ies, &idum) == ERR) {
                sprintf(msg,
                        "glbfin: problems terminating package %ld.\n",
                        (long) i);
                BasError(msg, strlen(msg));
            }
        }
    }


    return;
}

/*--------------------------------------------------------------------------*/

/*              DELETE ALL DYNAMIC VARIABLES FROM DATABASES              
 * DANGER: THIS MUST ONLY BE DONE IMMEDIATELY PRIOR TO TERMINATION!      
 * IT DESTROYS THE BASIS DATABASE FOR ANY FURTHER EXECUTION.             */

/*--------------------------------------------------------------------------*/

/* RTKILL - 
 * The purpose of this is to aid in detection of memory leaks.
 * Many Basis dynamic variables are still hanging around at
 * termination and do not really represent memory leaks. This
 * routine frees all(known) dynamic storage so should be a help
 * in locating real memory leaks. Note that Rtzapall kills all
 * Basis user variables. This takes care of permanent dynamics.
 * Written by Zane C. Motteler 3/31/2000
 */

void Rtkill()
{
    Integer i, j, nvar;

    for (i = 1; i <= numpak; i++) {
        nvar = Rtgnvari(i);
        for (j = 1; j <= nvar; j++)
            Rtdelet(j, i);
    }

    if (rtexit == 1) {

        Rtfin(0);

/* get rid of hash table nodes */
        FreeNodes();

/* get rid of string nodes */
        StrgDone();

/* free all the attribute handling */
        Rtdelcod();

/* free all the io table space */
        Osdeliotab();

/* get rid of default values */
        Osfreen(&defaddr, "RTKILL:defaddr");
        Osfreen(&defint, "RTKILL:defint");
        Osfreen(&defdbl, "RTKILL:defdbl");
        Osfreen(&defcmp, "RTKILL:defcmp");
        Osfreen((Address *) &defdcmp, "RTKILL:defdcmp");
        Osfreen(&defreal, "RTKILL:defreal");
        Osfreen(&deflog, "RTKILL:deflog");
        Osfreen(&defch, "RTKILL:defch");
        Freeout();
    }

    return;
}

/*--------------------------------------------------------------------------*/

/*          UNLINK A VARIABLE FROM THE HASH TABLE                     
 *          AND REMOVE IT FROM THE DB IF POSSIBLE                     */

/*--------------------------------------------------------------------------*/

/* RTUNLINK - unlink entry jvar of ndb from hash table.
 */

static void Rtunlink(Integer jvar, Integer ndb)
{
    Integer keynext, keyprev, jprev, nprev, jnext, nnext;
    Variable *pv;
    Package *pd;

    pd = &database[ndb];
    pv = &pd->var1[jvar - 1];

    keynext = pv->varnxt;
    keyprev = pv->varprv;

    if (keynext == 0 && keyprev == 0)
        IhashPut(vartable, pv->varnum, 0);

    else if (keynext > 0) {
        Rtcrack(keynext, &jnext, &nnext);
        database[nnext].var1[jnext - 1].varprv = keyprev;

/* this may be new head of hash chain */
        if (keyprev == 0)
            IhashPut(vartable, pv->varnum, keynext);
    }

    if (keyprev > 0) {
        Rtcrack(keyprev, &jprev, &nprev);
        database[nprev].var1[jprev - 1].varnxt = keynext;
    }

/* Mark it as dead just in case we can't permanently remove it */
    pv->varnum = DEAD;

/* but remove it from the database if it's the last variable. */
    if (jvar == pd->nvars) {
        pd->nvars--;
        pd->group1[pd->ngrps].gvmap -= 1;
    }

    return;
}

/*--------------------------------------------------------------------------*/

/*                REMOVE THE LAST GROUP FROM A DATABASE                     */

/*--------------------------------------------------------------------------*/

/* RTPOPLOC - Remove the last group in the local database. Upon return from
 * a user function, this group constitutes its local variables.
 * They must now be removed as they go out of scope.
 */

void Rtpoploc()
{
    Integer ng, nlg, n;
    Integer i;

    ng = dblocal.ngrps;
    nlg = dblocal.group1[ng - 1].gvmap + 1;
    n = dblocal.group1[ng].gvmap;

/* nlg..n are the indices of the first and last variables in this group. */
    for (i = n; i >= nlg; i--) {
        Rtunlink(i, LOCALDB);
        Rtdelet(i, LOCALDB);
    }
    dblocal.ngrps -= 1;

    return;
}

/*--------------------------------------------------------------------------*/

/*                          "FORGET" A VARIABLE                             */

/*--------------------------------------------------------------------------*/

/* RTFORGET - do a forget on a given variable/function.
 */

void Rtforget(Integer jvar, Integer ndb)
{
    char name[VARNAMESIZE + 1];
    char msg[80];

    if (jvar < 0)
        return;                 /* can not FORGET foreign variables */

    if (!Rtdattr(jvar, ndb, getreadonly())) {
        if (jvar <= database[ndb].nvars0) {

            INIT_BLANK_STRING(name, VARNAMESIZE);

            Rtgvname(jvar, ndb, name);
            sprintf(msg, "Cannot forget a permanent variable: %s.", name);
            BasError(msg, strlen(msg));
            Kaboom(0);
        }
        Rtdelet(jvar, ndb);
        Rtunlink(jvar, ndb);
    }

    return;
}

/*--------------------------------------------------------------------------*/

/*                  REMOVE HIERARCHY OF LOCAL DB'S                          */

/*--------------------------------------------------------------------------*/

/* RTCLEAN - Remove the hierarchy of local databases when a runtime error
 * occurs deep within a series of function calls.
 */

void Rtclean()
{
    int i;

    for (i = dblocal.ngrps; i > 0; i--)
        Rtpoploc();

    return;
}

/*--------------------------------------------------------------------------*/

/*                  MAKE PKG WITH HIGHEST PRIORITY CURRENT                  */

/*--------------------------------------------------------------------------*/

/* RTPARPOP - Compute the maximum priority of the first n packages and make
 * the one with the highest priority be the current one.
 */

void Rtparpop(Integer n)
{
    Integer maxpri;
    int i;

    if (currpkg == 1)
        database[currpkg].pripkg = 3;
    else
        database[currpkg].pripkg = 0;

    currpkg = 1;
    maxpri = database[1].pripkg;

    for (i = 2; i <= n; i++)
        if (database[i].pripkg > maxpri) {
            currpkg = i;
            maxpri = database[i].pripkg;
        }

    return;
}

/*--------------------------------------------------------------------------*/

/*                    GIVE PKG n THE HIGHEST PRIORITY                       */

/*--------------------------------------------------------------------------*/

/* RTPARPP - Give package n the highest priority; if u is true, set USERDB
 * to the next higher priority.
 */

void Rtparpp(Integer n, logical u)
{
    currpkg = n;
    database[n].pripkg = ++prihigh;

    if (u)
        dbuser.pripkg = prihigh + 1;

    dblocal.pripkg = prihigh + 2;

    return;
}

/*--------------------------------------------------------------------------*/

/*              EXTRACTING INFORMATION                                
 *                 FROM THE DATABASE                                  */

/*--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/

/*              GET NAME AND NUMBER OF VARIABLES OF A GROUP                 */

/*--------------------------------------------------------------------------*/

/* RTLSTG - Return the name and the number of variables in group jgrp.
 */

void Rtlstg(Integer jgrp, Integer *jvar, Integer ndb, char *name)
{
    Group *pg;

    pg = database[ndb].group1;

    StrgGet(pg[jgrp - 1].grpnam, name);

    *jvar = pg[jgrp].gvmap - pg[jgrp - 1].gvmap;

    return;
}

/*--------------------------------------------------------------------------*/

/*                EXTRACT INFORMATION ABOUT A DB VARIABLE                   */

/*--------------------------------------------------------------------------*/

/* RTXDB - crack open db and get out facts about variable number jvar
 * returns
 *       fwa             address of variable
 *       ilow, ihi       low, high subscripts
 *       icol            column lengths in memory
 *       ndim            number of dimensions
 * access: real access or not, limited or not
 * remove direct references to autodyn variables(ZCM, 7/91)
 * Rewrite in C(ZCM 9/30/1999)
 */

void Rtxdb(Integer jvar, Integer ndb, Address *fwa, Integer *ndim,
           Integer *ilow, Integer *ihi, Integer *icol, Integer access)
{
    Address ip;
    Integer ierr, i, ndiml, noff, nwords, nbytes, mjvar;
    char msg[80], vname[VARNAMESIZE + 1], cname[VARNAMESIZE + 1], *limstr;
    Variable *pv;
    static char
        msg2[] = "Error in form of limit string.",
        msg3[] = "Dimension mismatch with limit string.",
        msg4[] = "Negative column length in limit string",
        msg5[] = "rtxdb called with jvar = 0";


    INIT_BLANK_STRING(cname, VARNAMESIZE);

    pv = &database[ndb].var1[jvar - 1];

/*################## DOMESTIC ##########################################*/
    if (jvar > 0) {
        *ndim = pv->dimen;

/* set address depending on whether variable is static or dynamic */
        if (pv->dynflg) {
            ip = pv->addr;
            *fwa = *((Address *) ip);
            if (*fwa == (Address) NOTSET) {
                Fpullapt(&ndb, &(pv->dimsto), &zero,
                         ilow, ihi, ndim, &ierr);
                if (ierr == ERR) {
                    INIT_BLANK_STRING(vname, VARNAMESIZE);

                    Rtgvname(jvar, ndb, vname);
                    sprintf(msg,
                            "rtxdb: bad dimension string for dynamic variable %s.",
                            vname);
                    BasError(msg, strlen(msg));
                    limstr = FccMakeString(NULL, StrgLen(pv->dimsto), 1);
                    StrgGet(pv->dimsto, limstr);
                    BasError(limstr, strlen(limstr));
                    FccFree(limstr);
                    Kaboom(0);
                }

                for (i = 1; i <= *ndim; i++)
                    icol[i - 1] = ihi[i - 1] - ilow[i - 1] + 1;

                if ((autodyn == 1) && (access > 0)) {
                    Rtstgget(jvar, ndb, *ndim, icol, ilow, autodynp);
                    if (strncmp(autodyna, cname, VARNAMESIZE - 1)) {
                        limstr =
                            FccMakeString(autodyna, VARNAMESIZE - 1, 0);
                        Rtvattre(jvar, ndb, limstr);
                        FccFree(limstr);
                    }
                    *fwa = *((Address *) ip);
                } else {
                    return;
                }
            }
        } else {
            *fwa = pv->addr;
        }
        if (pv->elems == -1) {
/* If delems(jvar) >= 0 this is a normal Basis array.
 * Since fwa is not NOTSET, SOMEBODY has allocated some space but we don't
 * know anything about it.  If a limiting string exists, we could use it,
 * but we can't trust dorigin/ddimstr to contain anything meaningful.
 * It isn't obvious what policies to apply here.  What we hope for is that
 * the author will have done setshape if it is desired to have these
 * variables be visible.  We know he didn't call setshape.
 * This version reports the variable as a scalar.
 */
            *ndim = 0;
        }

        if (pv->limstr == 0 || access % 2 == 0) {
/* accessing with no limiting string */
            for (i = 1; i <= *ndim; i++) {
                ilow[i - 1] = pv->origin[i - 1];
                ihi[i - 1] = pv->dimstr[i - 1];
                icol[i - 1] = pv->dimstr[i - 1] - pv->origin[i - 1] + 1;
            }

        } else if (pv->limstr > 0) {
/* accessing with a limiting string */
            Fpullapt(&ndb, &(pv->limstr), &jvar, ilow, ihi, &ndiml, &ierr);
            if (ierr == ERR) {
                BasError(msg2, strlen(msg2));
                err9000(jvar, ndb, &(pv->limstr));
            }
            if (ndiml != *ndim) {
                BasError(msg3, strlen(msg3));
                err9000(jvar, ndb, &(pv->limstr));
            }
            for (i = 1; i <= *ndim; i++) {

/* allow zero length in a dimension but otherwise no monkey business */
                if (ilow[i - 1] < pv->origin[i - 1] ||
                    ihi[i - 1] > pv->dimstr[i - 1] ||
                    ihi[i - 1] < ilow[i - 1] - 1) {
                    BasError(msg4, strlen(msg4));
                    sprintf(msg, "%ld %ld %ld %ld %ld",
                            (long) i, (long) ilow[i - 1],
                            (long) ihi[i - 1], (long) pv->origin[i - 1],
                            (long) pv->dimstr[i - 1]);
                    BasError(msg, strlen(msg));
                    err9000(jvar, ndb, &(pv->limstr));
                }
                icol[i - 1] = pv->dimstr[i - 1] - pv->origin[i - 1] + 1;
            }

/* adjust fwa to first active location */
            noff = ilow[0] - pv->origin[0];
            for (i = 2; i <= *ndim; i++)
                noff += icol[i - 2] * (ilow[i - 1] - pv->origin[i - 1]);
            if (noff > 0) {
                nbytes = Utbytesin(pv->types) * noff;
                nwords =
                    nbytes / sizeof(Integer) +
                    (nbytes % sizeof(Integer) ? 1 : 0);
                *fwa = (Address) (((Integer *) fwa) + nwords);
            }
        } else {

/* limstr < 0 indicates a limit on the length of the LAST dimension
 * it is one larger in absolute value than the limit, so that 0 can be imposed
 */
            for (i = 1; i <= *ndim - 1; i++) {
                ilow[i - 1] = pv->origin[i - 1];
                ihi[i - 1] = pv->dimstr[i - 1];
                icol[i - 1] = ihi[i - 1] - ilow[i - 1] + 1;
            }
            if (*ndim > 0) {
                ilow[*ndim - 1] = pv->origin[*ndim - 1];
                ihi[*ndim - 1] = -pv->limstr + pv->origin[*ndim - 1] - 2;
                icol[*ndim - 1] = -pv->limstr - 1;
            }
        }
    } else if (jvar < 0) {
/*####################### FOREIGN #######################################*/
        mjvar = -jvar;
        Glbxdb(&ndb, &mjvar, fwa, ndim, ilow, ihi, icol);
    } else {
        BasError(msg5, strlen(msg5));
        Kaboom(0);
    }

    return;
}

/*--------------------------------------------------------------------------*/

/*                EXTRACT INFORMATION ABOUT A DB FUNCTION                   */

/*--------------------------------------------------------------------------*/

/* RTXDBF - crack open db and get out facts about function number jvar
 * Revised July, 1989(ZCM) -- eliminate ftok
 * Revised December, 1993 add possible negative jvar for psuedo array

 * Revised March 1995 to take into account new database organization
 *                   (separating addresses from integers).
 *                    shorter calling sequence; all function stack elements
 *                    now look the same.(ZCM)
 * Rewritten in C September 1999(ZCM)
 * returns
 *     ftyp =(0 = builtin, -(segment) = compiled, 1 = user)
 *     ilow, ihi   = min, max number of arguments permitted
 *     npack = package number, or 0 if builtin
 *     njvar = new jvar if this is a formal parameter standing for a
 *             compiled function
 */

void Rtxdbf(Integer jvar, Integer ndb, Integer *ftyp, Integer *ilow,
            Integer *ihi, Integer *npack, Integer *njvar)
{
    Variable *pv;

    if (jvar > 0) {
        pv = &database[ndb].var1[jvar - 1];

        *ftyp = pv->funtyp;
        *ilow = pv->origin[0];

/* This is user param. alias for compiled or builtin function */
        if (*ftyp <= 0 && ndb == LOCALDB)
            Rtcrack(pv->dimen, njvar, npack);

        *ihi = pv->dimstr[0];
    } else {
/* pseudo array, results in builtin function call */
        *ftyp = 0;
        *ilow = 0;
        *ihi = 0;
    }

    return;
}

/*--------------------------------------------------------------------------*/

/*                LOOK UP A VARIABLE IN THE DATABASE                        */

/*--------------------------------------------------------------------------*/

/* RTLOOKUP - 
 * Programming notes by Paul F. Dubois, August 1994
 *   The name space is organized into packages each of which has a number, ndb.
 *   Each dvarnum(jvar) is an integer(jentry) into the strg table:
 *      jentry=strgfind(name)  to put a name in
 *      call strgget(jentry, name) to get it out
 *  (strgget is smart and won't overrun name if it is too short)
 *   The names across the set of all packages are linked together for rapid
 *   searching in this routine, rtlookup.  The way this is done is to
 *   create a ihash table with index vartable in parinit; then, as each name
 *   is added an entry(dvarnum(jvar),rtkey(jvar,ndb)) is added to vartable.
 *   rtkey changes a jvar, ndb pair into an integer "key", and rtcrack
 *   changes such a key back into its parts.
 *   Starting from the vartable, we follow the chain of keys which point
 *   to variables with this name; the chain is double-threaded
 *   using dvarnxt and dvarprev.  So, one starts by changing the name into
 *   a slot number with strgfind, then looks up the slot number in the vartable
 *   by doing
 *           jentry = ihashfind(jtable, slotno)
 *   where jtable is vartable for variables; either it isn't there(-1
 *   returned by ihashfind) or it is there, in which case we get the associated
 *   value, a key,
 *           key = ihashget(jtable, jentry)
 *   and use it to start down the chain.  If the value is 0,
 *   this indicates that a variable by that name used to exist but no longer
 *   does.  To keep the ihash tables efficient, we don't try to delete entries
 *   from it but merely change the value associated with a name to 0, a key
 *   which we take to mean "nowhere".
 *
 *   macros use a second hashtable numbered 'mactable'. The key value
 *   in that table, if found and positive, is the slot number of the
 *   macro body in the string table. as with variables, ihashfind
 *   return -1 if a name isn't there, and if it is, and ihashget
 *   return 0, then it used to be a macro but has been undefined.
 *
 *   The point of this complex scheme is to compare integers rather than
 *   strings, and to have a search time which is proportional to the
 *   number of variables with the same name rather than the number of
 *   variables or the number of packages.  We also hash the string only once.
 *
 * Note on a trick: pripkg(0) is 0; everything else is higher
 * LOCALDB is unusual in that hits are not acceptable except in the
 * last group(s).
 * limit the lookup to the Varname size...
 */

Integer Rtlookup(char *var, Integer *ndb, Integer *tc, Integer intable)
{
    Integer kslot, tcsave, ientry, numf, jvar, nvar;
    Integer ntry, jtry, nlow, k, ng, key;
    char vvar[VARNAMESIZE + 1];
    Variable *pv, *pu;

    if (strlen(var) > VARNAMESIZE - 1) {
        strncpy(vvar, var, VARNAMESIZE - 1);
    } else {
        strcpy(vvar, var);
    }
    vvar[Utgetcl(vvar)] = '\0'; /* Better not have trailing whitespace. */

/* 7/28/95 ZCM in many cases during execution, the slot number has
 * already been found and is in the stack element. This prevents
 * redundant lookups.
 */
    if (intable > 0) {
        kslot = intable;
    } else {
        kslot = StrgFind(vvar);
    }

    ientry = IhashFind(vartable, kslot);
    if (ientry <= 0) {
        key = 0;
    } else {
        key = IhashGet(vartable, ientry);
    }

    jvar = 0;
    ntry = 0;
    *tc = 0;

    if (*ndb == 0) {

/* search all databases, find match highest priority */
        while (key > 0) {
            Rtcrack(key, &jtry, &ntry);
            pv = &database[ntry].var1[jtry - 1];
            key = pv->varnxt;
            if (database[ntry].pripkg <= database[*ndb].pripkg)
                continue;
            if (pv->varnum == kslot) {
                if (ntry == LOCALDB) {
                    ng = dblocal.ngrps;
                    if (ng == 0) {
                        continue;
                    } else {
                        if (jtry <
                            dblocal.group1[max(1, ng - locdepth) -
                                           1].gvmap + 1)
                            continue;
                    }
                }
                *ndb = ntry;
                jvar = jtry;
                *tc = pv->types;
            }
        }

/* foreign? */
        numf = Getnumf();
        nvar = StrgFind(var);
        for (k = 1; k <= numf; k++) {
            ntry = Getfrgn(k);
            if (database[ntry].pripkg > database[*ndb].pripkg) {
                jtry = Glbfind(&ntry, &nvar, &tcsave);
                if (jtry != 0) {
                    *ndb = ntry;
                    jvar = -jtry;
                    *tc = tcsave;
                }
            }
        }
        return jvar;

/* search a specified package other than the locals */
    } else if (*ndb != LOCALDB) {
        while (key > 0) {
            Rtcrack(key, &jtry, &ntry);
            pv = &database[ntry].var1[jtry - 1];
            key = pv->varnxt;
            if (ntry != *ndb)
                continue;

            pu = &database[*ndb].var1[jtry - 1];
            if (pu->varnum == kslot) {
                *tc = pu->types;
                return jtry;
            }
        }

/* foreign? */
        numf = Getnumf();
        nvar = StrgFind(var);
        for (k = 1; k <= numf; k++) {
            ntry = Getfrgn(k);
            if (ntry == *ndb)
                return -Glbfind(&ntry, &nvar, tc);
        }
        return 0;
    } else {
/* search LOCALDB only */
        ng = dblocal.ngrps;
        if (ng == 0)
            return 0;
        nlow = dblocal.group1[max(1, ng - locdepth) - 1].gvmap;
        jvar = 0;
        while (key > 0) {
            Rtcrack(key, &jtry, &ntry);
            key = database[ntry].var1[jtry - 1].varnxt;
            if (ntry != LOCALDB)
                continue;

            if (jtry < nlow + 1)
                continue;

            pu = &dblocal.var1[jtry - 1];
            if (pu->varnum == kslot) {
                jvar = jtry;
                *tc = pu->types;
            }
        }
        return jvar;
    }
}

/*--------------------------------------------------------------------------*/

/*                   LOOK UP A GROUP IN THE DATABASE                        */

/*--------------------------------------------------------------------------*/

/* RTLOOKG - looks for a group called "group".
 */

Integer Rtlookg(char *group, Integer *ndb)
{
    Integer jgrp, ntry, jtry, ng, ln;
    char *tgroup;
    Package *pd, *pr;
    Group *pg;

    jgrp = 0;

    if (*ndb == 0) {
        pr = &database[*ndb];
        ln = strlen(group);
        for (ntry = 1; ntry <= MAXPAK + 2; ntry++) {
            pd = &database[ntry];
            if (pd->pripkg <= pr->pripkg)
                continue;
            ng = pd->ngrps;
            for (jtry = ng; jtry >= 1; jtry--) {
                pg = &pd->group1[jtry - 1];
                tgroup = CStrgGet(pg->grpnam);
                if (!strcmp(group, tgroup)) {
                    jgrp = jtry;
                    *ndb = ntry;
                    pr = &database[*ndb];
                    break;
                }
            }
        }

        if (jgrp != 0)
            return jgrp;
        for (ntry = 1; ntry <= MAXPAK + 2; ntry++) {
            pd = &database[ntry];
            if (pd->pripkg <= pr->pripkg)
                continue;
            ng = pd->ngrps;

/* same as above, but just look at first ln characters. */
            for (jtry = ng; jtry >= 1; jtry--) {
                pg = &pd->group1[jtry - 1];
                tgroup = CStrgGet(pg->grpnam);
                if (!strncmp(group, tgroup, ln)) {
                    jgrp = jtry;
                    *ndb = ntry;
                    pr = &database[*ndb];
                    break;
                }
            }
        }
    } else {
        pr = &database[*ndb];
        ln = strlen(group);
        ng = pr->ngrps;
        for (jtry = ng; jtry >= 1; jtry--) {
            pg = &pr->group1[jtry - 1];
            tgroup = CStrgGet(pg->grpnam);
            if (!strcmp(group, tgroup))
                return jtry;
        }

/* same as above, but just look at first ln characters. */
        for (jtry = ng; jtry >= 1; jtry--) {
            pg = &pr->group1[jtry - 1];
            tgroup = CStrgGet(pg->grpnam);
            if (!strncmp(group, tgroup, ln))
                return jtry;
        }
    }

    return jgrp;
}

/*--------------------------------------------------------------------------*/

/*                   FIND ALL OCCURRENCES OF VAR IN DATABASE                */

/*--------------------------------------------------------------------------*/

/* RTLOOKA - Find all occurrences of var in databases--up to USERDB, anyhow.
 */

void Rtlooka(char *var, Integer *jvar, Integer *ndb, Integer *tc,
             Integer *pri, Integer *no)
{
    Integer kslot, ntry, jtry, ientry, key, k, nvar, tcsave, numf;
    Integer nl;
    Variable *pv;

    kslot = StrgFind(var);
    ientry = IhashFind(vartable, kslot);
    if (ientry <= 0)
        key = 0;
    else
        key = IhashGet(vartable, ientry);

    jvar[0] = 0;
    ntry = 0;
    tc[0] = 0;
    nl = 0;

    while (key > 0) {
        Rtcrack(key, &jtry, &ntry);
        pv = &database[ntry].var1[jtry - 1];
        key = pv->varnxt;

        if (pv->varnum == kslot) {
/* We've found one!! */
            if (nl >= USERDB)
                break;

            ndb[nl] = ntry;
            jvar[nl] = jtry;
            tc[nl] = pv->types;
            pri[nl] = database[ntry].pripkg;
            nl++;
        }
    }

    if (nl < USERDB) {
        numf = Getnumf();
        for (k = 1; k <= numf; k++) {
            ntry = Getfrgn(k);
            nvar = StrgFind(var);
            jtry = Glbfind(&ntry, &nvar, &tcsave);
            if (jtry != 0) {
                if (nl >= USERDB)
                    break;

                ndb[nl] = ntry;
                jvar[nl] = -jtry;
                tc[nl] = tcsave;
                pri[nl] = database[ntry].pripkg;
                nl++;
            }
        }
    }

    *no = nl;

    return;
}

/*--------------------------------------------------------------------------*/

/*                     EXTRACT DIMENSIONING STRING FOR VAR                  */

/*--------------------------------------------------------------------------*/

void Rtxdso(Integer jvar, Integer ndb, char *dimsto)
{
    StrgGet(database[ndb].var1[jvar - 1].dimsto, dimsto);

    return;
}

/*--------------------------------------------------------------------------*/

/*                     GET SUBSCRIPTS FOR A STACK ELEMENT                   */

/*--------------------------------------------------------------------------*/

/* RTGSUBS - get correct subscripts for an lhs stack element
 *         - Moved here to isolate database references 7/91
 */

void Rtgsubs()
{
    Integer ndim, i, jvar, ndb;
    Variable *pv;

    jvar = ssl(PTR, rsp);
    ndb = ssl(N, rsp);

    pv = &database[ndb].var1[jvar - 1];
    ndim = pv->dimen;
    for (i = 1; i <= ndim; i++) {
        ssl(N1 + i - 1, rsp) = pv->origin[i - 1];
        ssl(M1 + i - 1, rsp) = pv->dimstr[i - 1];
        ssl(I1 + i - 1, rsp) = 0;
    }

    for (i = ndim + 1; i <= 7; i++) {
        ssl(N1 + i - 1, rsp) = 0;
        ssl(M1 + i - 1, rsp) = 0;
        ssl(I1 + i - 1, rsp) = 0;
    }

    return;
}

/*--------------------------------------------------------------------------*/

/*                     GET INFORMATION ON USER FUNCTION                     */

/*--------------------------------------------------------------------------*/

/* RTGUINFO - gets information out of database for user functions
 * fwa      : address of the function's intermediate code
 * codelen  : length of this code in words
 * nname    : pointer to the array containing names of function and args
 * Written by Zane C. Motteler 3/27/95; in C 9/23/1999
 */

void Rtguinfo(Integer jvar, Integer ndb, Address *fwa, Integer *codelen,
              Address *nname)
{
    Variable *pv;

    pv = &database[ndb].var1[jvar - 1];

    *fwa = pv->addr;
    *codelen = pv->dimstr[2];
    *nname = pv->argn;

    return;
}

/*--------------------------------------------------------------------------*/

/*                     GET ARGUMENT INFORMATION FOR  FUNCTION               */

/*--------------------------------------------------------------------------*/

/* RTGARGS - return argument information for functions.
 */

void Rtgargs(Integer jvar, Integer ndb, Integer *minargs, Integer *maxargs,
             Integer *namindx)
{
    Variable *pv;

    pv = &database[ndb].var1[jvar - 1];

    *minargs = pv->origin[0];
    *maxargs = pv->dimstr[0];
    *namindx = pv->dimen;

    return;
}

/*--------------------------------------------------------------------------*/

/*      THE FOLLOWING FUNCTIONS BEGINNING WITH "rtis" CHECK WHETHER A      
 *      CERTAIN CONDITION IS TRUE OF A VARIABLE.                           */

/*--------------------------------------------------------------------------*/

/* RTISDEAD - so that routines can check for "forgotten" variables.
 */

logical Rtisdead(Integer jvar, Integer ndb)
{
    return database[ndb].var1[jvar - 1].varnum == DEAD;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTISCHAM - check whether a variable is a chameleon.
 */

logical Rtischam(Integer jvar, Integer ndb)
{
    return database[ndb].var1[jvar - 1].chamel;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTISDYN - check whether a variable is dynamic.
 */

logical Rtisdyn(Integer jvar, Integer ndb)
{
    Integer df;

    df = database[ndb].var1[jvar - 1].dynflg;

    return (df != 0);
}

/*--------------------------------------------------------------------------*/

/* THE FOLLOWING FUNCTIONS AND SUBROUTINES BEGINNING WITH "rtg"       
 * RETRIEVE(get) SOME PIECE OF INFORMATION FROM THE DATABASE.        */

/*--------------------------------------------------------------------------*/

/* RTGEXTRA - return the count of "extra" variables(string lengths) among
 * arguments of functions with dynamic string arguments
 */

Integer Rtgextra(Integer jvar, Integer ndb)
{
    return database[ndb].var1[jvar - 1].origin[2];
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGPKGNO - get the package number associated with ndb
 * In this implementation, it's just ndb)
 */

Integer Rtgpkgno(ndb)
{
    return database[ndb].packno;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGDRELY - return the value of rely from the database.
 */

Integer Rtgdrely(Integer jvar, Integer ndb)
{
    return database[ndb].var1[jvar - 1].rely;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGDELMS - get the number of elements in a database entry.
 */

Integer Rtgdelms(Integer jvar, Integer ndb)
{
    return database[ndb].var1[jvar - 1].elems;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGNVAR - get number of variables in a database.
 */

Integer Rtgnvar(Integer ndb)
{
    return database[ndb].nvars;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGNVARI - get maximimum number of variables ever in a database.
 */

Integer Rtgnvari(Integer ndb)
{
    return database[ndb].maxvari;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGLOW - get the low value of the ith subscript of the variable.
 */

Integer Rtglow(Integer jvar, Integer ndb, Integer i)
{
    return database[ndb].var1[jvar - 1].origin[i - 1];
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGHIGH - get the high value of the ith subscript of the variable.
 */

Integer Rtghigh(Integer jvar, Integer ndb, Integer i)
{
    return database[ndb].var1[jvar - 1].dimstr[i - 1];
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGADDR - return the address of the variable denoted by jvar,ndb
 *         - (if it has one)
 * Bug fix by ZCM 7/13/94 to handle the dynamic case
 */

Address Rtgaddr(Integer jvar, Integer ndb)
{
    Address *ipp;
    Variable *pv;

    pv = &database[ndb].var1[jvar - 1];

    if (pv->dynflg) {
        ipp = (Address *) pv->addr;
        return *ipp;
    } else {
        return pv->addr;
    }

}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGDADDR - return the address from the database, rather than
 *          - following the pointer in the dynamic case.
 */

Address Rtgdaddr(Integer jvar, Integer ndb)
{
    return database[ndb].var1[jvar - 1].addr;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGWORDS - return the number of machine words taken by the variable,
 * but always return at least 1.
 */

Integer Rtgwords(Integer jvar, Integer ndb)
{
    Integer nbytes, nwords;
    Variable *pv;

    pv = &database[ndb].var1[jvar - 1];

    nbytes = pv->elems * Utbytesin(pv->types);

    nwords = nbytes / sizeof(Integer) + (nbytes % sizeof(Integer) ? 1 : 0);

    return (max(1, nwords));
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGLIMST - return the string number of the limits string.
 */

Integer Rtglimst(Integer jvar, Integer ndb)
{
    return database[ndb].var1[jvar - 1].limstr;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGUNITS - get the units string of a variable.
 */

void Rtgunits(Integer jvar, Integer ndb, char *runits)
{
    StrgGet(database[ndb].var1[jvar - 1].units, runits);

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGVATTR - return the attribute word of a variable.
 */

Integer Rtgvattr(Integer jvar, Integer ndb)
{
    return database[ndb].var1[jvar - 1].vattr;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGGATTR - return the attribute of a Group.
 */

Integer Rtggattr(Integer jgrp, Integer ndb)
{
    return database[abs(ndb)].group1[jgrp - 1].gattr;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGFTYP - get the function type of a function in the database.
 */

Integer Rtgftyp(Integer jvar, Integer ndb)
{
    return database[ndb].var1[jvar - 1].funtyp;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGPRI - get the priority of package npk.
 */

Integer Rtgpri(Integer npk)
{
    return database[npk].pripkg;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGNGRP - get the number of groups in package ndb.
 */

Integer Rtgngrp(Integer ndb)
{
    return database[ndb].ngrps;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGNNAME - return the slot number of a variable's name.
 */

Integer Rtgnname(Integer jvar, Integer ndb)
{
    return database[ndb].var1[jvar - 1].varnum;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGVNAME - return name of variable in this slot.
 */

void Rtgvname(Integer jvar, Integer ndb, char *name)
{
    char msg[100], pk[PNSIZE];

    if (ndb <= 0 || ndb >= MAXPAK + 3) {
        sprintf(msg, "Rtgvname: database specification(%ld) out of range.",
                (long) ndb);
        BasError(msg, strlen(msg));
        strcpy(name, "Bad stack element.");
        return;
    }

    if (jvar < 0) {
        strcpy(name, "<foreign>");
    } else if (jvar < 0 || jvar > Rtgnvar(ndb)) {
        Glbpknam(ndb, pk);
        sprintf(msg,
                "Rtgvname: variable number %ld out of range for database %ld (%s).",
                (long) jvar, (long) ndb, pk);
        BasError(msg, strlen(msg));
        strcpy(name, "Bad stack element.");
    } else {
        StrgGet(database[ndb].var1[jvar - 1].varnum, name);
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGINNAME - used to find the real name of a function which is a
 * formal argument of a user function.
 */

void Rtginame(Integer *jvar, Integer *ndb, char *name)
{
    while (*ndb == LOCALDB)
        Rtcrack(database[*ndb].var1[*jvar - 1].dimen, jvar, ndb);

/* now have real jvar and ndb, get the name */
    StrgGet(database[*ndb].var1[*jvar - 1].varnum, name);

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGGNAME - return the name of a group.
 */

void Rtggname(Integer jgrp, Integer ndb, char *name)
{
    StrgGet(database[ndb].group1[jgrp - 1].grpnam, name);

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGNVAR1 - return the number of the first variable in the group.
 */

Integer Rtgnvar1(Integer jgrp, Integer ndb)
{
/* Add 1 because this result goes to FORTRAN */
    return database[ndb].group1[jgrp - 1].gvmap + 1;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGNASTR - get the number of the action string in the string table.
 */

Integer Rtgnastr(Integer jvar, Integer ndb)
{
    return database[ndb].var1[jvar - 1].actstr;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGFTEXT -
 *      Programmed by: Zane C. Motteler
 *      Date         : 9 September 1993
 *      Purpose      : return a pointer to the function's text, and its size.
 *                     the pointer is returned as the Address value of the
 *                     function; the size is returned as an output parameter.
 */

Address Rtgftext(Integer jvar, Integer ndb, Integer *size)
{
    Variable *pv;

    pv = &database[ndb].var1[jvar - 1];

    *size = pv->dimstr[5];

    return pv->setval;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

Integer Rtgloc()
{
    return locdepth;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGTVAR - return total number of variables in user and local db's.
 */

Integer Rtgtvar()
{
    return dblocal.nvars + dbuser.nvars;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGTC - Return the type code of a database entry.
 */

Integer Rtgtc(Integer jvar, Integer ndb)
{

    if (jvar < 0)
        return FILE_VAR;

    else
        return database[ndb].var1[jvar - 1].types;

}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGNDIMST - return the slot number of the dimension string
 *           - in the string table
 */

Integer Rtgndimst(Integer jvar, Integer ndb)
{
    return database[ndb].var1[jvar - 1].dimsto;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGNDIM -
 *        Programmed by Zane C. Motteler
 *                August 1990
 *        Purpose: find the number of dimensions of a variable.(Used to
 *                prevent lrsetdo broadcasting initial value to an array)
 *                Return the index into funparn for compiled functions
 */

Integer Rtgndim(Integer jvar, Integer ndb)
{
    if (jvar > 0)
        return database[ndb].var1[jvar - 1].dimen;

    else
        return 0;
}

/*--------------------------------------------------------------------------*/

/*                            COMMENT HANDLING                              */

/*--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/

/*                       ALLOCATE COMMENT SPACE                             */

/*--------------------------------------------------------------------------*/

/* RTCOMSP - allocates/changes comment space.
 */

static void Rtcomsp()
{
    int i;
    Integer size, oldsize;

    if (maxcomms == 0) {
        size = sizeof(Commentstruct) * (maxcomms + 300) / sizeof(Integer);
        Osallotn((Address *) (&dcomm), size, "PERM|RTCOMSP:dcomm");
    } else {
        oldsize = sizeof(Commentstruct) * maxcomms / sizeof(Integer);
        size = sizeof(Commentstruct) * (maxcomms + 300) / sizeof(Integer);
        Oschangn((Address *) (&dcomm), size, oldsize, "comment space");
    }

    freecom = maxcomms;
    for (i = maxcomms; i < maxcomms + 299; i++)
        dcomm[i].commn = i + 1;
    maxcomms += 300;
    dcomm[maxcomms - 1].commn = -1;

    return;
}

/*--------------------------------------------------------------------------*/

/*                            ENTER A COMMENT                               */

/*--------------------------------------------------------------------------*/

/* RTCOMEV - Enter a comment for a variable or a function.
 * Use this routine when the comment is not dynamic.
 * commn is used in two ways:
 * for a variable a chain begins with its comm entry which points into 
 * a dcomm record's commn entry,
 * giving a linked list of comments, which ends when commn is - 1.
 * if dcomm[n].commn is free, we maintain a linked list of free entries with
 * the chain beginning at freecom.  When more free storage is needed we
 * increase the storage
 */

void Rtcomev(Integer jvar, Integer ndb, char *comment)
{
    int lcomment, isave, k;
    Variable *pv;

/* Add more space if necessary */
    if (freecom == -1)
        Rtcomsp();
    lcomment = strlen(comment);

/* Store the comment */
    if (lcomment >= COMMENTSIZE) {
        strncpy(dcomm[freecom].comms, comment, COMMENTSIZE - 1);
        dcomm[freecom].comms[COMMENTSIZE - 1] = '\0';

    } else {
        strcpy(dcomm[freecom].comms, comment);
    }

/* Find the end of the chain of comments and add this one */
    isave = freecom;
    freecom = dcomm[freecom].commn;
    dcomm[isave].commn = -1;

    pv = &database[ndb].var1[jvar - 1];

    if (pv->comm == -1) {
        pv->comm = isave;
    } else {
        k = pv->comm;
        while (dcomm[k].commn > -1)
            k = dcomm[k].commn;
        dcomm[k].commn = isave;
    }

    return;
}

/*--------------------------------------------------------------------------*/

/*                      CHANGE A VARIABLE'S FIRST COMMENT                   */

/*--------------------------------------------------------------------------*/

/* RTCOMDCHV - Change the first comment for a variable or a function
 * If one doesn't exist, create it
 * This routine used by display routine to set comment of $, for example
 */

void Rtcomchv(Integer jvar, Integer ndb, char *comment)
{
    int lcomm;
    Variable *pv;

    pv = &database[ndb].var1[jvar - 1];

    if (pv->comm == -1) {
        Rtcomev(jvar, ndb, comment);
    } else {
        lcomm = strlen(comment);
        if (lcomm >= COMMENTSIZE) {
            strncpy(dcomm[pv->comm].comms, comment, COMMENTSIZE - 1);
            dcomm[pv->comm].comms[COMMENTSIZE - 1] = '\0';
        } else {
            strcpy(dcomm[pv->comm].comms, comment);
        }
    }

    return;
}

/*--------------------------------------------------------------------------*/

/*                      GET A VARIABLE's dcomm ENTRY                        */

/*--------------------------------------------------------------------------*/

/* RTGDCOMM - Return the dcomm entry for variable(jvar, ndb)
 *          - Note: add 1 to indices returned to FORTRAN
 */

Integer Rtgdcomm(Integer jvar, Integer ndb)
{
    return database[ndb].var1[jvar - 1].comm + 1;
}

/*--------------------------------------------------------------------------*/

/*                      RETURN THE k-th commn ENTRY                         */

/*--------------------------------------------------------------------------*/

/* RTGDCOMN - Return the kth commn entry(A FORTRAN code is
 *          - following the chain)
 */

Integer Rtgdcomn(Integer k)
{
    return dcomm[k - 1].commn + 1;
}

/*--------------------------------------------------------------------------*/

/*                   EXTRACT A VARIABLE'S icom-th COMMENT                   */

/*--------------------------------------------------------------------------*/

/* RTXCOMD - extract the icom'th comment for variable(jvar, ndb)
 *         - from the database
 */

Integer Rtxcomd(Integer jvar, Integer ndb, Integer icom, char *comment)
{
    int icount, k;

    if (jvar > 0) {
        icount = 0;
        k = database[ndb].var1[jvar - 1].comm;
        while (k > -1) {
            if (++icount == icom) {
                strcpy(comment, dcomm[k].comms);
                return (OK);
            }
            k = dcomm[k].commn;
        }
    }

    return (ERR);
}

/*--------------------------------------------------------------------------*/

/*                     EXTRACT A GROUP'S icom-th COMMENT                    */

/*--------------------------------------------------------------------------*/

/* RTXCOMG - extract the icom'th comment for Group(jgrp, ndb)
 *         - from the database
 */

Integer Rtxcomg(Integer jgrp, Integer ndb, Integer icom, char *comment)
{
    int icount, k;

    if (jgrp > 0) {
        icount = 0;
        k = database[ndb].group1[jgrp - 1].remark;
        while (k > -1) {
            if (++icount == icom) {
                strcpy(comment, dcomm[k].comms);
                return (OK);
            }
            k = dcomm[k].commn;
        }
    }

    return (ERR);
}

/*--------------------------------------------------------------------------*/

/*                        ENTER A COMMENT FOR A GROUP                       */

/*--------------------------------------------------------------------------*/

/* RTCOMEG - Enter a comment for a Group
 * Use this when comment is not dynamic
 * commn is used in two ways:
 * for a Group a chain begins its remark entry which points into commn
 * giving a linked list of comments, which ends when commn is -1
 * if dcomm[n].commn is free, we maintain a linked list of free entries with
 * the chain beginning at freecom.  When more free storage is needed we
 * increase the storage
 */

void Rtcomeg(Integer jvar, Integer ndb, char *comment)
{
    int lcomm, isave, k;
    Group *pg;

/* Add more space if necessary */
    if (freecom == -1)
        Rtcomsp();

    lcomm = strlen(comment);
    if (lcomm > COMMENTSIZE - 1) {
        strncpy(dcomm[freecom].comms, comment, COMMENTSIZE - 1);
        dcomm[freecom].comms[COMMENTSIZE - 1] = '\0';

    } else {
        strcpy(dcomm[freecom].comms, comment);
    }

/* Find the end of the chain of comments and add this one */
    isave = freecom;
    freecom = dcomm[freecom].commn;
    dcomm[isave].commn = -1;

    pg = &database[ndb].group1[jvar - 1];

    if (pg->remark == -1) {
        pg->remark = isave;
    } else {
        k = pg->remark;
        while (dcomm[k].commn > -1)
            k = dcomm[k].commn;
        dcomm[k].commn = isave;
    }

    return;
}

/*--------------------------------------------------------------------------*/

/*                        HANDLE INFO ABOUT PACKAGES                        */

/*--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/

/*                          SET UP PACKAGE INFORMATION                      */

/*--------------------------------------------------------------------------*/

/* PACKSET - this routine called by config-written routine setpkgs.
 */

void Packset(char *pkgnam, char *longnam, Integer maxits, Integer frn)
{
    int i;

    /* We need to initialize some information */
    if (numpak == 0) {
        for (i = 0; i <= MAXPAK; i++) {
            pakstat[i] = SLEEPING;
            pakmaxi[i] = 10000;
        }
        pakstat[LOCALDB] = UP;
        pakstat[USERDB] = UP;
    }

    numpak += 1;

    rdb_string_copy(paknam[numpak], pkgnam, PNSIZE);
    rdb_string_copy(lngnam[numpak], longnam, LNGNAMSIZE);

    pakmaxi[numpak] = maxits;
    if (frn > 0) {
        numfrgn += 1;
        foreign[numfrgn] = numpak;
    }

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* GLBMAX - return max number of iterations(default) for a package.
 */

Integer Glbmax(Integer n)
{
    return pakmaxi[n];
}

/*--------------------------------------------------------------------------*/

/*                             SET PACKAGE STATUS                           */

/*--------------------------------------------------------------------------*/

void Glbsstat(Integer n, Integer status)
{
    pakstat[n] = status;

    return;
}

/*--------------------------------------------------------------------------*/

/*                             GET PACKAGE STATUS                           */

/*--------------------------------------------------------------------------*/

Integer Glbstat(Integer n)
{
    static char msg[] = "glbstat: package number illegal";

    if ((n < 1) || ((n > numpak) && (n != USERDB) && (n != LOCALDB))) {
        BasError(msg, strlen(msg));
        Kaboom(1);
    }

    return pakstat[n];
}

/*--------------------------------------------------------------------------*/

/*                   GET PACKAGE STATUS AS CHARACTER STRING                 */

/*--------------------------------------------------------------------------*/

void Glbistat(Integer n, char *status)
{
    Integer k;

    k = Glbstat(n);
    if (k == SLEEPING)
        strcpy(status, "sleeping");

    else if (k == GENERATE)
        strcpy(status, "generate");

    else if (k == UP)
        strcpy(status, "-- up --");

    else
        strcpy(status, "????????");

    return;
}

/*--------------------------------------------------------------------------*/

/*                   FIND INDEX OF PACKAGE GIVEN ITS NAME                   */

/*--------------------------------------------------------------------------*/

/* GLBPKNUM - find index of package whose name is pn
 *          - glbpknux is same but won't complain
 */

Integer Glbpknum(char *pn)
{
    char msg[80];
    int i;

    if (!strcmp(pn, "global"))
        return USERDB;

    if (!strcmp(pn, "local"))
        return LOCALDB;

    for (i = 1; i <= numpak; i++)
        if (!strcmp(pn, paknam[i]))
            return i;

    sprintf(msg, "glbpknum: package not found -- %s.", pn);
    BasError(msg, strlen(msg));
    Kaboom(0);

    return ERR;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* GLBPKNUX - find index of package whose name is pn, but don't complain.
 */

Integer Glbpknux(char *pn)
{
    int i;

    if (!strcmp(pn, "global"))
        return USERDB;

    if (!strcmp(pn, "local"))
        return LOCALDB;

    for (i = 1; i <= numpak; i++)
        if (!strcmp(pn, paknam[i]))
            return i;

    return 0;
}

/*--------------------------------------------------------------------------*/

/*                      INITIALIZE A PACKAGE                                */

/*--------------------------------------------------------------------------*/

void Glbpinit(char *pn)
{
    Integer n;

/* Get the package number  */
    n = Glbpknum(pn);

/* Initialize that package */
    Glbninit(&n);

    return;
}

/*--------------------------------------------------------------------------*/

/*                      GET THE LONG NAME OF A PACKAGE                      */

/*--------------------------------------------------------------------------*/

void Glblngn(Integer n, char *longname)
{
    strcpy(longname, lngnam[n]);

    return;
}

/*--------------------------------------------------------------------------*/

/*                      GET THE SHORT NAME OF A PACKAGE                      */

/*--------------------------------------------------------------------------*/

/* GLBPKNAM - find name of package whose index is i.
 */

void Glbpknam(Integer i, char *pknam)
{

    if (i >= 1 && i <= numpak)
        strcpy(pknam, paknam[i]);

    else if (i == USERDB)
        strcpy(pknam, "global");

    else if (i == LOCALDB)
        strcpy(pknam, "local");

    else
        strcpy(pknam, "??");

    return;
}

/*--------------------------------------------------------------------------*/

/*                  LIST AVAILABLE PACKAGES AND THEIR STATUS                */

/*--------------------------------------------------------------------------*/

void Glblpackn(Integer ndb, Integer flag)
{
    Integer ngr, nvr;
    char msg[80], status[STATUSSIZE + 1];
    char longnam[LNGNAMSIZE + 1], pknam[PNSIZE + 1];

    Glbistat(ndb, status);
    rdb_string_set(status, STATUSSIZE);

    Glbpknam(ndb, pknam);
    rdb_string_set(pknam, PNSIZE);

    Glblngn(ndb, longnam);
    rdb_string_set(longnam, LNGNAMSIZE);

    if (edtfcnmod == 0) {
        if (flag) {
            sprintf(msg,
                    "Priority Name             Long Name                  Status");
            Remark(msg, strlen(msg));
        }

        sprintf(msg, "%8ld %s   %s   %s",
                (long) Rtgpri(ndb), pknam, longnam, status);
    } else {
        if (flag) {
            sprintf(msg,
                    "Name             Long Name                        ");
            sprintf(msg + strlen(msg), "#Groups #Variables  Status");
            Remark(msg, strlen(msg));
        }

        ngr = Rtgngrp(ndb);
        nvr = Rtgnvar(ndb);

        sprintf(msg, " %s   %s   %8ld  %8ld    %s",
                pknam, longnam, (long) ngr, (long) nvr, status);
    }

    Remark(msg, strlen(msg));

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void Glblpack()
{
    Integer i;
    char msg[80];

    sprintf(msg,
            "Priority Name             Long Name                  Status");
    Remark(msg, strlen(msg));

    for (i = 1; i <= numpak; i++)
        Glblpackn(i, 0);

    return;
}

/*--------------------------------------------------------------------------*/

/*                  CHECK WHETHER A PACKAGE IS GENERATED                    */

/*--------------------------------------------------------------------------*/

Integer Glbisgen(Integer n)
{
    char msg[80], pknam[PNSIZE + 1];

    if ((n >= 1) && (n <= numpak)) {
        if (Glbstat(n) == GENERATE) {
            return 1;
        } else {
            Glbpknam(n, pknam);
            sprintf(msg, "%s has never been generated.", pknam);
            Remark(msg, strlen(msg));
            return 0;
        }
    }

    return 0;
}

/*--------------------------------------------------------------------------*/

/*                     CHECK WHETHER A PACKAGE IS UP                        */

/*--------------------------------------------------------------------------*/

Integer Glbisup(Integer n)
{
    if ((n >= 1) && (n <= numpak)) {
        if (Glbstat(n) == UP) {
            return 1;
        }
    }

    return 0;
}

/*--------------------------------------------------------------------------*/

/*                      SET FIRST PACKAGE                                    */

/*--------------------------------------------------------------------------*/

void Frstpset(char *firstpk)
{
    nfirstps += 1;

    strcpy(firstpkg[nfirstps], firstpk);

    return;
}

/*--------------------------------------------------------------------------*/

/*                      GET NUMBER OF FIRSTS                                 */

/*--------------------------------------------------------------------------*/

Integer Getnfirst()
{
    return nfirstps;
}

/*--------------------------------------------------------------------------*/

/*                      GET NUMBER OF PACKAGES INSTALLED                     */

/*--------------------------------------------------------------------------*/

/* RTNUMPAK - the number of packages installed.
 */

Integer Rtnumpak()
{
    return numpak;
}

/*--------------------------------------------------------------------------*/

/*                      GET NUMBER OF FOREIGN PACKAGES                       */

/*--------------------------------------------------------------------------*/

/* GETNUMF - return the number of foreign packages */

Integer Getnumf()
{
    return numfrgn;
}

/*--------------------------------------------------------------------------*/

/*                 GET THE PKG NUMBER OF A FOREIGN PACKAGE                  */

/*--------------------------------------------------------------------------*/

/* GETFRGN - return the number of a foreign package.
 */

Integer Getfrgn(Integer k)
{
    return foreign[k];
}

/*--------------------------------------------------------------------------*/

/*                 GET THE SPECIFIED PACKAGE NAME                           */

/*--------------------------------------------------------------------------*/

void Getfirstp(Integer ifp, char *pknam)
{
    strcpy(pknam, firstpkg[ifp]);

    return;
}

/*--------------------------------------------------------------------------*/

/*                 PRINT MESSAGE FOR UNKNOWN VARIABLE                       */

/*--------------------------------------------------------------------------*/

void Unknownv(Integer *si)
{
    char name[256];
    char pkg[PNSIZE + 1];
    char errmsg[80];
    char cbell;
    Integer n1, nc, ni;

    nc = -si[NAML - 1];
    ni = si[N - 1];

/* Blank out the name prior to calling Parname */
    INIT_BLANK_STRING(name, nc + 1);

    n1 = Parname(si, name);
    if (bell == 1)
        cbell = (char) 7;
    else
        cbell = ' ';

    if ((ni > 0) && (ni != UNKNOWN)) {
        Glbpknam(ni, pkg);

        if (n1 > 37)
            name[37] = '\0';

        sprintf(errmsg, "%cThere is no variable named %s in package %s.",
                cbell, name, pkg);
    } else {
        if (n1 > 61)
            name[61] = '\0';
        sprintf(errmsg, "%cUnknown variable: %s", cbell, name);
    }

    BasError(errmsg, strlen(errmsg));

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* CBASNAME - try to find the name associated with the address fwa.
 */

void Cbasname(Address fwa, char *name)
{
    int i, j, nv, df;
    char msg[80];
    Address ad;
    Variable *pv;

    for (j = 0; j < dblocal.nvars; j++) {
        pv = &dblocal.var1[j];
        df = pv->dynflg;
        ad = (Address) (*((Address *) pv->addr));
        if ((df && (ad == fwa)) || (!df && (pv->addr == fwa))) {
            StrgGet(pv->varnum, name);
            return;
        }
    }

    for (j = 0; j < dbuser.nvars; j++) {
        pv = &dbuser.var1[j];
        df = pv->dynflg;
        ad = (Address) (*((Address *) pv->addr));
        if ((df && (ad == fwa)) || (!df && pv->addr == fwa)) {
            StrgGet(pv->varnum, name);
            return;
        }
    }

    for (i = 1; i <= prihigh - 2; i++) {
        nv = database[i].nvars;
        for (j = 0; j < nv; j++) {
            pv = &database[i].var1[j];
            df = pv->dynflg;
            ad = (Address) (*((Address *) pv->addr));
            if ((df && (ad == fwa)) || (!df && pv->addr == fwa)) {
                StrgGet(pv->varnum, name);
                return;
            }
        }
    }

    sprintf(msg, "basname: unable to find variable with address %p(0x%p).",
            fwa, fwa);
    BasError(msg, strlen(msg));

    Kaboom(0);

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* GLBGETCP - get the current package number.
 */

Integer Glbgetcp(void)
{
    return currpkg;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* RTGETVERB - get the verbose flag.
 */

Integer Rtgetverb(void)
{
    return (verbose);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
