/* 
 * STRG.C
 *
 *  String storage modules: facility for putting strings into storage and
 *  then referring to them with a Fortran integer.  This is accomplished 
 *  via indexing an array of char *s.
 *    Fortran entries are:
 *    call strginit(isize) -- initialize string table with approx # strings
 *    ientry =  strgfind(s) -- look up string s in table itable
 *                          -- inserts s in table if not already there
 *    ilength = strglen(ientry) -- return length of entry
 *    call strgget(ientry,t) -- fill t with entry, blank padded or chopped
 *    call strgbug -- print out diagnostics about itable
 *  Notes:   
 *    a. Uses Fcc protocols for Fortran-C communication
 *    b. Uses blocked string storage to reduce malloc load
 *    c. NStrg should be a prime since it is used in the hash function. 
 *    d. s="" iff strfind(s)=0
 *    Paul F. Dubois, July 1994
 */

#include "rt.h"

static unsigned int primes[] =
    { 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53,
    57, 59, 67, 71, 73, 79, 83, 89, 91, 97, 101
};
static int np = -1;

static char *blanks = NULL;

typedef struct node {
    char *string;
    int next;
} Node;

typedef struct {
    int h[NStrg];
    int capacity;
    int size;
    Node *nodes;
} StrgTable;

static StrgTable stringTable;

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

unsigned int StrgHash(char *s)
{
    register unsigned int result, i, c;

    result = 0;
    if (np < 0)
        np = sizeof(primes) / sizeof(int);

    for (i = 0; i < np; i++) {
        c = s[i];
        if (c == 0)
            break;
        result = ((result + c) * primes[i]) % NStrg;
    }
    return (result);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* STRGINIT - Initialize the string table holding around asize elements */

void StrgInit(integer asize)
{
    int i;
    integer iblank;

    assert(asize > 0);
    for (i = 0; i < NStrg; i++) {
        stringTable.h[i] = -1;
    }
    stringTable.capacity = asize;

#ifdef PCK_SCORE
    stringTable.nodes =
        FMAKE_N(Node, stringTable.capacity, "PERM|STRGINIT:nodes");
#else
    stringTable.nodes =
        (Node *) malloc(sizeof(Node) * (stringTable.capacity));
#endif

    stringTable.size = 0;

    iblank = StrgFind("");
    assert(iblank == 0);

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* save the strings in big blocks to avoid annoying malloc    */
#define STRINGBLOCKSIZE 8000
static char initialStringBlock[STRINGBLOCKSIZE];
static char *currentStringBlock = initialStringBlock;
static int currentStringIndex = 0;
static int stringBlockCount = 1;        /* diagnostic only */

static char *StrgSave(char *s)
{
    char *result;
    unsigned int n;

    n = strlen(s);
    assert(n < STRINGBLOCKSIZE);

    if (n + 1 + currentStringIndex >= STRINGBLOCKSIZE) {

#ifdef PCK_SCORE
        currentStringBlock = FMAKE_N(char, sizeof(initialStringBlock),
                                     "PERM|STRGSAVE:csb");
#else
        currentStringBlock = (char *) malloc(sizeof(initialStringBlock));
#endif

        currentStringIndex = 0;
        stringBlockCount++;
    }

    result = currentStringBlock + currentStringIndex;
    currentStringIndex += n + 1;

/* On the Crays, ensure that every string starts on a word boundary */
#if ( defined XMP | defined YMP | defined C90 | defined J90 | defined UNICOS | defined CRAY2 )
    {
        int pad;

        pad = 8 - (currentStringIndex & 7);
        currentStringIndex += pad;
    }
#endif

    memcpy(result, s, n + 1);

    return (result);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

static int StrgNewNode(char *s)
{
    Node *nd;
    int ns, nx;

    ns = stringTable.size++;
    nx = stringTable.capacity;
    nd = stringTable.nodes;

    if (ns >= nx) {
        nx *= 2;

        if (nd == NULL) {

#ifdef PCK_SCORE
            nd = FMAKE_N(Node, nx, "STRGNEWNODE:nodes");
#else
            nd = (Node *) malloc(sizeof(Node) * nx);
#endif

        } else {

#ifdef PCK_SCORE
            REMAKE_N(nd, Node, nx);
#else
            nd = (Node *) realloc(nd, sizeof(Node) * nx);
#endif
        }

        stringTable.capacity = nx;
        stringTable.nodes = nd;
    }

    nd[ns].string = StrgSave(s);
    nd[ns].next = -1;

    return (ns);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void StrgDone(void)
{
    Node *nd;

    nd = stringTable.nodes;

#ifdef PCK_SCORE
    SFREE(nd);
#else
    free(nd);
#endif

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

integer StrgFind(s)
char *s;
{
    int hvalue, kentry, kprevious;
    Node current;

    hvalue = StrgHash(s);
    kentry = stringTable.h[hvalue];
    if (kentry == -1) {
        kentry = StrgNewNode(s);
        stringTable.h[hvalue] = kentry;
        goto found;
    }

    current = stringTable.nodes[kentry];
    if (strcmp(s, current.string) != 0) {
        kprevious = kentry;
        kentry = current.next;
        while (kentry != -1) {
            current = stringTable.nodes[kentry];
            if (strcmp(s, current.string) == 0) {
                goto found;
            }
            kprevious = kentry;
            kentry = current.next;
        }

        kentry = StrgNewNode(s);
        stringTable.nodes[kprevious].next = kentry;
    }

  found:
    return ((integer) kentry);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

integer StrgFinz(char *s, integer len)
{

/* We want to include trailing blanks if there are any. */
    integer i;

    for (i = strlen(s); i < len; i++)
        s[i] = ' ';

    return (StrgFind(s));
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Comes in with the raw address of a FORTRAN string and its length.
 * Need to copy it to a C string for StrgFind, then free it.
 */

integer StrgFina(char *s, integer len)
{
    integer i;
    char *cpy;

#ifdef PCK_SCORE
    cpy = FMAKE_N(char, len + 1, "STRGFINA:cpy");
#else
    cpy = (char *) malloc(len + 1);
#endif

    strncpy(cpy, s, len);
    cpy[len] = '\0';
    i = StrgFind(cpy);

#ifdef PCK_SCORE
    SFREE(cpy)
#else
    free(cpy);
#endif

    return i;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* STRGADDR - return the address of a string. It is necessary
 *          - in order for compiled function calls to be able to pass the
 *          - address of a string
 */

void *StrgAddr(integer jentry)
{
    char *entry;

    assert(jentry >= 0);
    assert(jentry < stringTable.size);
    if (jentry == 0) {          /* blanks */
        if (!blanks) {

#ifdef PCK_SCORE
            blanks = FMAKE_N(char, 501, "STRGADDR:blanks");
#else
            blanks = (char *) malloc(501);
#endif

            INIT_BLANK_STRING(blanks, 501);
        }

        return ((void *) blanks);
    }

    entry = stringTable.nodes[jentry].string;
    return ((void *) entry);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* CSTRGGET - return a pointer to the string. */

char *CStrgGet(int jentry)
{

    assert(jentry >= 0);
    assert(jentry < stringTable.size);

    return stringTable.nodes[jentry].string;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* STRGGET - assumes s blank-filled, null terminated on entry */

void StrgGet(integer jentry, char *s)
{
    char *entry;
    int ls;
    void FccCopy(char *, int, char *);

    assert(jentry >= 0);
    assert(jentry < stringTable.size);

    entry = stringTable.nodes[jentry].string;
    ls = strlen(s);

    FccCopy(s, ls, entry);

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* in deference to Fortran, indx and newindx used to be 1-based.
 * However, all offsets sent to strg.c are now 0-based(ZCM 3/30/1999)
 * copies the string referred to by jentry, starting at position indx,
 * until either 1) n characters have been copied, or 2) the character
 * stopchr has been encountered, or 3) we run out of string. if 1)
 * happens, then newindx is advanced until it points at the next char
 * after the next stopchr, if any; in the other cases, it points at the
 * next available char(if any). in case 2, it points at the char(if
 * any) right after the stopchr. This routine replaces mcchrdln.
 * In deference to people who like to define empty macros, a call with
 * n==0 does nothing.
 *
 * jentry   - this is the entry no. for BODY of macro
 * s        - buffer where moved chars will be put
 * n        - size of string buffer s
 * indx     - 1-based spec of where in the string to start
 * newindx  - will point to next char to be moved
 * stopchr  - character(e. g., peol) to stop at
 */

void StrgGetn(integer jentry, char *s, integer n, integer indx,
              integer *newindx, integer stopchr)
{
    char *entry;
    int ls;
    void FccCopy(char *, int, char *);
    int ind,                    /* index of next stopchr after indx( if any ) */
        cm;                     /* number of characters actually moved */

    assert(n >= 0);
    assert(jentry >= 0);
    assert(jentry < stringTable.size);
    if (n == 0)
        return;
    entry = stringTable.nodes[jentry].string;
    ls = strlen(entry);
    if (indx >= ls)
        return;
    n = n < strlen(s) ? n : strlen(s);
    ind = strchr(entry + indx, (char) stopchr) - (entry + indx);
    if (ind < 0)
        ind = n;
    cm = ind < n ? ind + 1 : n;
    if (indx + cm > ls) {
        cm = ls - indx + 1;
        if (cm == 0) {
            *newindx = indx + 1;
            return;
        }
    }
    FccCopy(s, cm, entry + indx);
    *newindx = indx + cm;
    /* step through string till its end or until stopchr has been passed */
    for (; *newindx < ls && entry[*newindx - 1] != (char) stopchr;
         (*newindx)++) ;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* s is a Fortran character*1 array slen characters long.
 * clear it to blanks and null-terminate it for StrgGet.
 */

void StrgGetz(integer jentry, char *s, integer slen)
{
    int i;

    assert(slen >= 2);
    for (i = 0; i < slen - 1; i++)
        s[i] = ' ';
    s[slen - 1] = '\0';
    StrgGet(jentry, s);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Similar to the above but s comes in as a raw Address. */

void StrgGeta(integer jentry, char *s, integer slen)
{
    char *entry;
    void FccCopy(char *, int, char *);

    assert(jentry >= 0);
    assert(jentry < stringTable.size);
    assert(slen >= 1);
    entry = stringTable.nodes[jentry].string;
    FccCopy(s, slen, entry);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Same as StrgGetn except s ia an Address parameter
 * The difference is that we trust n to be the length of s.
 *
 * jentry  - this is the entry no. for BODY of macro
 * s       - buffer where moved chars will be put
 * n       - size of string buffer s
 * indx    - 1-based spec of where in the string to start
 * newindx - will point to next char to be moved
 * stopchr - character(e. g., peol) to stop at
 */

void StrgGetna(integer jentry, char *s, integer n, integer indx,
               integer *newindx, integer stopchr)
{
    char *entry;
    int ls;
    void FccCopy(char *, int, char *);
    int ind,                    /* index of next stopchr after indx( if any ) */
        cm;                     /* number of characters actually moved */

    assert(n >= 0);
    assert(jentry >= 0);
    assert(jentry < stringTable.size);
    if (n == 0)
        return;
    entry = stringTable.nodes[jentry].string;
    ls = strlen(entry);
    if (indx >= ls)
        return;
    ind = strchr(entry + indx, (char) stopchr) - (entry + indx);
    if (ind < 0)
        ind = n;
    cm = ind < n ? ind + 1 : n;
    if (indx + cm > ls) {
        cm = ls - indx + 1;
        if (cm == 0) {
            *newindx = indx + 1;
            return;
        }
    }
    FccCopy(s, cm, entry + indx);
    *newindx = indx + cm;
    /* step through string till its end or until stopchr has been passed */
    for (; *newindx < ls && entry[*newindx - 1] != (char) stopchr;
         (*newindx)++) ;
    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Print out an Address-style string. Used for debugging. */
void StrgPrt(char *strg, integer slen)
{
    int i;

    for (i = 0; i < slen; i++)
        fprintf(stderr, "%c", strg[i]);
    fprintf(stderr, "\n");
    fflush(stderr);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

integer StrgLen(integer jentry)
{
    char *entry;

    assert(jentry >= 0);
    assert(jentry < stringTable.size);
    entry = stringTable.nodes[jentry].string;
    return ((integer) strlen(entry));
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Returns the number of strings stored so far so edit routines
 * can cycle over them.
 */

integer NStrgs()
{
    return ((integer) stringTable.size);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* print out diagnostics */

void StrgBug()
{
    int i, ifull, ichain, j;
    float density, chainl;
    char errmsg[80];

    ifull = ichain = 0;
    for (i = 0; i < NStrg; i++) {
        if (stringTable.h[i] >= 0) {
/*
         sprintf(errmsg,"Occupied cell %d %d",i,stringTable.h[i]);
         BasError(errmsg,strlen(errmsg));
*/
            ifull++;
            ichain++;
            j = stringTable.nodes[stringTable.h[i]].next;
            while (j != -1) {
                ichain++;
                j = stringTable.nodes[j].next;
            }
        }
    }
    density = ((float) ifull) / ((float) NStrg);
    chainl = ((float) ichain) / ((float) (ifull == 0 ? 1 : ifull));
    sprintf(errmsg, "StrgBug Report");
    BasError(errmsg, strlen(errmsg));
    sprintf(errmsg, "Number of String Blocks %d", stringBlockCount);
    BasError(errmsg, strlen(errmsg));
    sprintf(errmsg, "Node Capacity %d", stringTable.capacity);
    BasError(errmsg, strlen(errmsg));
    sprintf(errmsg, "Number of Nodes %d", stringTable.size);
    BasError(errmsg, strlen(errmsg));
    sprintf(errmsg, "Load %e", density);
    BasError(errmsg, strlen(errmsg));
    sprintf(errmsg, "Chain Ave %e", chainl);
    BasError(errmsg, strlen(errmsg));
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* The purpose of the following routine is to move 'length' bytes from
 * source + soffset to target + toffset. The raw 'Addres'es are passed
 * from FORTRAN; this avoids dereferencing raw string pointers in
 * FORTRAN, and any worry about platform-dependent character descriptors.
 */

void SmovAtoA(char *target, integer toffset, char *source, integer soffset,
              integer length)
{
    memcpy(target + toffset, source + soffset, length);
    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Given a raw string pointer source, this copies length characters
 * from the source into the FORTRAN string target. target better
 * be long enough!
 */

void SmovAtoF(char *target, integer toffset, char *source, integer soffset,
              integer length)
{
    memcpy(target + toffset, source + soffset, length);
    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void SBlkSet(char *target, integer offset, integer length)
{
    if (length <= 0)
        return;
    memset(target + offset, ' ', length);
    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Given a raw string pointer target, this copies length characters
 * from the FORTRAN string source into the target. target better
 * be long enough!
 * This looks the same as the above, but the difference is that
 * in the above, source is an 'Address' and target is a '&string'
 * while below source is a 'string' and target is an Address.
 */

void SmovFtoA(char *target, integer toffset, char *source, integer soffset,
              integer length)
{
    integer nchar;

    if (length == 1 && *source == '\0') {
        target[toffset] = ' ';
        return;
    }

/* %$#@ FccMakeString truncates blanks off the end. */
    nchar = (length > strlen(source)) ? strlen(source) : length;

    memcpy(target + toffset, source + soffset, nchar);

    SBlkSet(target, toffset + nchar, length - nchar);

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* SMOVFTOF - this is for completeness and really does replace
 *          - ZMOVECHR which doesn't get optimized properly under
 *          - some FORTRAN compilers
 *          - memcpy will be correct
 */

void SmovFtoF(char *target, integer toffset, char *source, integer soffset,
              integer length)
{

    memcpy(target + toffset, source + soffset, length);

    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Intended to copy 'length' anonymous bytes from source to target.
 */

void SmovBtoA(char *target, integer toffset, void *source, integer soffset,
              integer length)
{
    memcpy(target + toffset, (char *) source + soffset, length);
    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Intended to copy 'length' anonymous bytes from source to target.
 * Now, however, the Address argument is the second.
 */

void SmovAtoB(void *target, integer toffset, void *source, integer soffset,
              integer length)
{
    memcpy((char *) target + toffset, (char *) source + soffset, length);
    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void StrChSet(char *target, integer offset, integer length, char *ch)
{
    if (length <= 0)
        return;

    memset(target + offset, *ch, length);
    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Convert strings to upper or lower case */

void StrTolc(char *oldstr, char *newstr, integer length)
{
    int i;

    for (i = 0; i < length; i++)
        newstr[i] = (char) tolower((int) oldstr[i]);
    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void StrTouc(char *oldstr, char *newstr, integer length)
{
    int i;

    for (i = 0; i < length; i++)
        newstr[i] = (char) toupper((int) oldstr[i]);
    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

integer StrIndex(char *str1, integer offset1, integer len1,
                 char *str2, integer offset2, integer len2)
{
    char *estr1, *estr2;
    int i;

    estr1 = str1 + offset1;
    estr2 = str2 + offset2;
    for (i = 0; i < len1 - len2 + 1; i++)
        if (!strncmp(estr1 + i, estr2, len2))
            return (i + 1);
    return 0;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Looks the same as the above, but the second argument is a non-dynamic string.
 */

integer StrIndexF(char *str1, integer offset1, integer len1,
                  char *str2, integer offset2, integer len2)
{
    char *estr1, *estr2;
    int i;

    estr1 = str1 + offset1;
    estr2 = str2 + offset2;
    for (i = 0; i < len1 - len2 + 1; i++)
        if (!strncmp(estr1 + i, estr2, len2))
            return (i + 1);
    return 0;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* return the index of the earliest one of the characters in str2 in str1,
 * or 0 if none of these occurs.
 */

integer StrIndex1(char *str1, integer offset1, integer len1, char *str2,
                  integer len2)
{
    char *estr1;
    int i, j;

    estr1 = str1 + offset1;
    for (j = 1; j <= len1; j++)
        for (i = 0; i < len2; i++)
            if (estr1[j - 1] == str2[i])
                return j;
    return 0;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void StrSkipBl(char *instring, integer length, integer *start)
{
    int i;

    for (i = 0; i < length; i++)
        if (instring[i] != ' ') {
            *start = i;
            return;
        }
    *start = length;
    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

integer StrGetCl(char *instring, integer offset, integer length)
{
    int i;

    for (i = offset + length - 1; i > offset; i--)
        if (instring[i] != ' ')
            return i - offset + 1;
    return 1;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

integer StrEQ(char *string1, integer offset1, integer length1,
              char *string2, integer offset2, integer length2)
{
    int i, minlen, maxlen;

    minlen = min(length1, length2);
    maxlen = max(length1, length2);

    if (strncmp(string1 + offset1, string2 + offset2, minlen))
        return 0;

    if (minlen == maxlen)
        return 1;

    if (minlen == length1)
        for (i = offset2 + minlen; i < offset2 + maxlen; i++) {
            if (string2[i] != ' ')
                return 0;
    } else
        for (i = offset1 + minlen; i < offset1 + maxlen; i++) {
            if (string1[i] != ' ')
                return 0;
        }
    return 1;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* Same as above, but second argument is a non-dynamic string
 */

integer StrEQ1F(char *string1, integer offset1, integer length1,
                char *string2, integer offset2, integer length2)
{

/* If string2 is blanks, it will have been reduced to the null string. */

    if (*(string2 + offset2) == '\0') {
        for (; length1 > 0; length1--, offset1++)
            if (*(string1 + offset1) != ' ')
                return 0;
        return 1;
    }

    return StrEQ(string1, offset1, length1, string2, offset2, length2);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

integer StrEQCh(char *c1, int offset1, char *c2, int offset2)
{
    return c1[offset1] == c2[offset2];
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

integer StrEQChF(char *c1, int offset1, char *c2, int size2)
{
    return c1[offset1] == c2[0];
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

integer StrGetCh(char *str, int offset)
{
    return (integer) str[offset];
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

void StrPutCh(char *str, int offset, char *ch, int size2)
{
    str[offset] = *ch != '\0' ? *ch : ' ';
    return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* UTGETCL - get character length of a fortran string.
 *         - name must be NULL terminated.
 */

Integer Utgetcl(char *name)
{
    int i;

    for (i = strlen(name); i > 1; i--) {
        if (name[i - 1] != ' ') {
            break;
        }
    }

    return i ? i : 1;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/* UTSTRLEN - Remove trailing spaces from a string, return the length.
 */

int utstrlen(char *s, int nx)
{
    int n;

    for (n = nx - 1; n >= 0; n--) {
        if (s[n] != ' ') {
            break;
        }
    }

    n++;
    s[n] = '\0';

    return n;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
