#
#arglist.py
#
#Ben Wolfson <rumjuggler@cryptarchy.org>
#
#A utility for retrieving arguments passed via the command line.
#
#Thanks to Trent Mick for many suggestions.
#
# Modified Oct 13, 2000 to support arguments of the form
#  -arg value
#
"""
Usage:

Instantiate an Arguments object with a series of Option objects.  An
Option object must be instantiated with at least a short form; it can
also be passed arguments indicating that the option has a long form,
takes a value and has a default value, in addition to a callback to be
performed on any passed-in value and a description string.

For instance, a standard "help" option could look like this:
    Option("h", "help", description="Displays this help message")



Options taking values:


An option taking a value should set has_value to a nonzero value.  If
an option has a value but does not have a default value, then it is an
error for that option to be present on the command line without a
value.

For instance, if the Arguments object is instantiated with an Option like this,
    Option("s", "startat", has_value=1)

then if the command line args looks like this:

    somescript.py -s

arglist.Error will be raised with the message "Option -s requires a value"

An Option can be instantiated with val_cb=somefunction.  In that case,
the result of calling that function on the value (which is normally a
string) will be stored in the value.  If an exception occurs,
arglist.CallbackError is raised with a message corresponding to the
description string of the Option.

For instance,
    Option('p', 'port', has_value=1, default_value=5555, val_cb=int)

will call builtin function "int" on a value, so that when you access
the value (below), instead of getting a string, you get an integer.



Processing the command line:


After the Arguments object is instantiated,

args = Arguments(...Option(...)...),

simply call the instance's process() method with the list of
command-line options--normally this will be sys.argv[1:].


Accessing the values:

The Arguments variable uses __getattr__ to allow you to refer to
options by either long or short names.  If the Arguments instance is
called "args" and it was instantiated like this,

args = Arguments(
    Option('h', 'help'),
    Option('p', 'port', has_value=1, default_value=5555, val_cb=int),
    Option('s', 'startat', has_value=1)
    )

and the command line has been processed, options can be accessed thusly:

if -h or --help was among the options processed,
    args.h and args.help will have value 1
if not, they will have value None

if -p or --port was among the options processed,
    args.p and args.port will have the value 5555, or whatever value
    corresponds to what the user passed in.
if not, they will have the value 5555.

if -s or --startat was among the options processed, and a value was provided,
    args.s and args.startat will both have whatever the argument to the option
    was.  It will be a string.
if not, they will have the value None.

"""

class Error(Exception):
    pass

class CallbackError(Error):
    pass

class Option:
    def __init__(self, short_f, long_f=None, has_value=None,
                 default_value=None,
                 val_cb=None, description=None):
        self.short_form, self.long_form, self.has_val, self.val_cb = \
                    short_f, long_f, has_value, val_cb
        self.value = default_value # or None
        self.description = description
        
    def get_keys(self):
        return self.short_form, self.long_form
    
    def handle(self, val):
        if self.has_val:
            if val:
                self.value = val
            elif self.value is None:
                    raise Error, 'Option -%s requires a value' % self.short_form
        else:
            self.value = 1
        if self.val_cb:
            try:
                self.value = self.val_cb(self.value)
            except:
                raise CallbackError, str(self)

    def __str__(self):
        tabs = (2, 20)
        return '%s-%s, --%s%s%s' % (' '*tabs[0], self.short_form,
                                    self.long_form or '', ' '*tabs[1],
                                    self.description or '')
    
        
class Arguments:
    def __init__(self, *opts):
        self.nonopts= []
        self.badopts= []
        self.opt_handler = {}
        for opt in opts:
            short_form, long_form = opt.get_keys()
            if short_form and long_form:
                self.opt_handler[short_form] = self.opt_handler[long_form] = opt
            elif short_form:
                self.opt_handler[short_form] = opt
            else:
                self.opt_handler[long_form] = opt

    def process(self, options):
        self._options = options[:]
        opt = self.nextarg()
        while opt:
            if opt[0] == '-':
                key = opt[:]
                while key[0] == '-':
                    key = key[1:]
                if self.opt_handler.has_key(key):
                    opt_obj = self.opt_handler[key]
                    if opt_obj.has_val:
                        val = self.nextarg()
                    else:
                        val = None
                    opt_obj.handle(val)
                else:
                    self.badopts.append(opt)
#                    raise RuntimeError, 'no such option %s' % opt
            else:
                self.nonopts.append(opt)
            opt = self.nextarg()

    def nextarg(self):
        if self._options:
            next = self._options[0]
            self._options.remove(next)
        else:
            next = None
        return next

    def __getattr__(self, key):
        if not self.opt_handler.has_key(key):
            raise Error, 'No option %s' % key
        return self.opt_handler[key].value

    def get_nonopts(self):
        return self.nonopts
            

def splitstrip(opt):
    return __optsplit(__optstrip(opt))

def __optstrip(opt):
    while opt[0] == '-':
        opt = opt[1:]
    return opt

def __optsplit(opt):
    if ':' in opt:
        return opt.split(':',1)
    elif '=' in opt:
        return opt.split('=',1)
    return opt, None
