#!/bin/env python

import string
from types import StringType

import rdp.parsetools as parsetools
AST = parsetools.AST

from rdp.visitor import walk

########################################
#
#
class Blank(AST):
    def __init__(self):
	AST.__init__(self, 'blank')

########################################
#
#
class Comment(AST):
    def __init__(self, *args):
	AST.__init__(self, 'comment')
        self.comments = args

########################################
#
#
class DeclarationSpecifiers(AST):
    """
    declaration_specifier ::= storage_class_specifier [declaration_specifier]
    declaration_specifier ::= type_specifier [declaration_specifier]
    declaration_specifier ::= type_qualifier [declaration_specifier]
    """

    def __init__(self, storage_class=None, type_specifiers=None,
                 type_qualifiers=None, struct=None):
	AST.__init__(self, 'declaration_specifiers')
        self.storage_class = storage_class or []
        self.type_specifiers = type_specifiers or []
        self.type_qualifiers = type_qualifiers or []
        self.spec_str = ' '.join(self.storage_class + self.type_qualifiers +
                                  self.type_specifiers)
        self.struct = struct

    def c_decl(self):
        # This has been pre-computed
        return self.spec_str

    # XXX - does this have any kids besides struct?
    def getChildren(self):
        if self.struct:
            return self._kids + [self.struct]
        else:
            return self._kids
    
########################################
#
#
class InitDeclarator(AST):
    """init declarator
    
    init_declarator ::= declarator
    init_declarator ::= declarator = initializer
    """

    # XXX - what is name vs declarator?
    def __init__(self, name, declarator=None, text=None, initializer=None,
                 type_spec=None):
        """
        name - name of variable
        declarator - full declaration (with pointer and such)
        """
	AST.__init__(self, 'init_declarator')
        self.name = name
        if declarator is None:
            self.declarator = name
        else:
            self.declarator = declarator
        self.text = text
        self.initializer = initializer
        self.type_spec = type_spec

        # if this is a function declartion...
        self.arg_names = []
        self.arg_decl  = []

    def c_decl(self):
        if self.initializer is None:
            return self.declarator
        else:
            return '%s = %s' % (self.declarator, self.initializer.c_decl())

    def c_decl_parameter(self):
        return '%s %s' % (self.type_spec.c_decl(), self.declarator)

########################################
#
#
class Initializer(AST):
    """
    initializer ::= assignment-expression
    initializer ::= { initializer_list }
    initializer ::= { initializer_list , }

    The args is a list of strings.
    """
    
    def __init__(self, *args):
	AST.__init__(self, 'initializer')
        self.args = list(args)

    def append(self, node):
        self.args.append(node)
        
    def c_decl(self):
        # If more the length is more than 1,
        # then an array is assumed and it is surrounded by curley braces
        # XXX - need a better system for inserting line breaks and indenting
        if len(self.args) > 1:
            return '\n{%s}' % string.join(self.args, ',\n ')
        else:
            return self.args[0]
        

########################################
#
#
class Declaration(AST):
    """declaration ::= declaration_specifiers init_declarator_list;

    specifier = <string> of "int" or ... or a DeclarationSpecifiers
    args      = <string> of name or a InitDeclarator

    The first child is a type_spec. The remaining children are InitDeclarators
    and have references to the type_spec.

    If given a string type_spec or init_declaration, then an object of the
    correct class is created.
    """
    
    def __init__(self, type_spec, *args):
	AST.__init__(self, 'declaration')

        if type(type_spec) == StringType:
            self.type_spec = DeclarationSpecifiers(type_specifiers=[type_spec])
        else:
            self.type_spec = type_spec

        self.inits = []
        self._kids = self.inits  # XXX- 
        for arg in args:
            if type(arg) == StringType:
                self.append(InitDeclarator(arg))
            else:
                self.append(arg)

    def append(self, node):
        # when appending children, add a pointer to the type_spec
        node.type_spec = self.type_spec
        self.inits.append(node)

    def getChildren(self):
        return [self.type_spec] + self.inits


########################################
#
#
class StructOrUnion(AST):
    def __init__(self, name='', struct='struct'):
	AST.__init__(self, 'struct_or_union')
        self.name = name
        self.struct = struct
        self.members = self._kids # XXXX - 

    def append(self, node):
        self.members.append(node)

    def getChildren(self):
        return self.members
        
########################################
#
#
class Stmt(AST):
    def __init__(self, code):
	AST.__init__(self, 'stmt')
        self.code = code

########################################
#
#
class StmtBlock(AST):
    def __init__ (self, *args):
	AST.__init__(self, 'stmt_block')
        self.add_kids(args)

########################################
#
#
class Return(AST):
    # jump_statement ::= return [expression]
    #
    def __init__(self, value):
	AST.__init__(self, 'return')
        self.value = value

########################################
#
#
class Function(AST):
    def __init__(self, decl, args=None, stmts=None):
	AST.__init__(self, 'function')
        self.decl = decl
        self.args = args or []
        self.stmts = stmts or []

    def getChildren(self):
        return [self.decl] + self.args + self.stmts

########################################
#
#
class File(AST):
    def __init__(self, name):
	AST.__init__(self, 'file')
        self.name = name

########################################
#
#
class Print_C_Code(parsetools.Dump_Mixin):
    def __init__(self, fp=None):
        parsetools.Dump_Mixin.__init__(self)
        self.max_line = 79
        self.set_indent(0)
        import sys
        self.fp = fp or sys.stdout

        self.comma = ''
        self.in_declaration = 0

    def visitBlank(self, node):
        self.writeto("\n")

    def visitComment(self, node):
        comments = node.comments
        if len(comments) == 1:
            self.writeto("%s\n" % str(comments[0]))
        else:
            self.writeto("/* %s\n" % str(comments[0]))
            for comment in comments[1:]:
                self.writeto(" * %s\n" % str(comment))
            self.writeto(" */")
        
    def visitDeclarationSpecifiers(self, node):
        self.writeto('%s ' % node.c_decl())

    def visitDeclaration(self, node):
        self.comma = ''
        self.in_declaration += 1
        self.visit(node.type_spec)
        for init in node.inits:
            self.visit(init)
        self.append_to_line(";\n")
        self.in_declaration -= 1

    def visitInitDeclarator(self, node):
        self.append_to_line('%s%s' % (self.comma, node.c_decl()))
        self.comma = ', '
        
    def visitStructOrUnion(self, node):
        if node.name:
            self.append_to_line('%s {\n' % node.struct)
        else:
            self.append_to_line('%s %s {\n' % (node.struct, node.name))
        self.doindent()
        for member in node.members:
            self.visit(member)
        self.deindent()
        if self.in_declaration == 0:
            self.writeto("};\n")
        else:
            self.writeto("} ")
        self.comma = ''

    def visitStmt(self, node):
        self.writeto("%s;\n" % node.code)

    def visitReturn(self, node):
        self.writeto("return %s;\n" % node.value)
        
    def visitFunction(self, node):
        self.writeto('%s\n' % node.decl.c_decl_parameter())
        for arg in node.args:
            self.visit(arg)
        self.writeto("{\n")
        self.doindent()
        for stmt in node.stmts:
            self.visit(stmt)
        self.deindent()
        self.writeto("}\n")

    def visitFile(self, node):
        self.writeto("/* file:%s */\n\n" % node.name)
        for child in node.getChildren():
            self.visit(child)

    def default(self, node):
        print "0000:missing node:", node

def print_C_code(ast, fp=None):
    walk(ast, Print_C_Code(fp))

########################################
#
#
class visit_C_Nodes(parsetools.Dump_Mixin):
    def __init__(self, fp=None):
        parsetools.Dump_Mixin.__init__(self)
        self.set_indent(0)
        import sys
        self.fp = fp or sys.stdout

    def visitBlank(self, node):
        self.writeto('Blank\n')

    def visitComment(self, node):
        self.writeto('Comments\n')
        
    def visitDeclarationSpecifiers(self, node):
        self.writeto('DeclarationSpecifiers: %s\n' % node.spec_str)
        if node.struct:
            self.doindent()
            self.visit(node.struct)
            self.deindent()

    def visitDeclaration(self, node):
        self.writeto('Declaration\n')
        self.doindent()
        self.visit(node.type_spec)
        for init in node.inits:
            self.visit(init)
        self.deindent()

    def visitInitDeclarator(self, node):
        self.writeto('InitDeclarator: %s\n' % node.declarator)
#        self.writeto('   %s %s\n' % (node.text, node.type_spec.spec_str))
        
    def visitStructOrUnion(self, node):
        self.writeto('StructOrUnion\n')
        self.doindent()
        for field in node.getChildren():
            self.visit(field)
        self.deindent()

    def visitStmt(self, node):
        self.writeto('Stmt\n')
        self.writeto('%s;\n' % node.code)

    def visitReturn(self, node):
        self.writeto('Return\n')
        self.writeto('return %s;\n' % node.value)
        
    def visitFunction(self, node):
        self.writeto('Function\n')
        self.doindent()
        for child in node.getChildren():
            self.visit(child)
        self.deindent()

    def visitFile(self, node):
        self.writeto('File\n')
        self.doindent()
        for child in node.getChildren():
            self.visit(child)
        self.deindent()

def print_C_nodes(ast, fp=None):
    walk(ast, visit_C_Nodes(fp))

########################################
#
#
if __name__ == '__main__':
    pass
        
