#!/bin/env python
#
# Module to support f90
#
from string import join
import types

import rdp.parsetools as parsetools
AST = parsetools.AST

from rdp.visitor import walk

########################################
#
#
class Blank(AST):
    def __init__(self):
	AST.__init__(self, 'blank')

########################################
#
#
class Comment(AST):
    def __init__(self, *args):
	AST.__init__(self, 'comment')
        self.comments = args

########################################
#
#
class FortranArrayBound(AST):
    def __init__(self, number=None, name=None, star=None):
	AST.__init__(self, 'array_bound')
	self.number = number
	self.name   = name
	self.start  = star

    def f90_decl(self):
	if self.number:
	    s = self.number
	elif self.name:
	    s = self.name
	elif self.star:
	    s = '*'
	else:
	    s = ''
	return s

########################################
#
#
class FortranArrayShape(AST):
    def __init__(self, lower=None, upper=None):
	AST.__init__(self, 'array_shape')
        self.lower = lower
	self.upper = upper
        if lower:
            self.append(lower)
        if upper:
            self.append(upper)

    def f90_decl(self):
	if self.lower is not None:
	    s = '%s:%s' % (
                self.lower.f90_decl(), self.upper.f90_decl())
	elif self.upper is not None:
	    s = self.upper.f90_decl()
	else:
	    s = ':'
	return s

    def first_element(self):
	if self.lower:
	    s = self.lower.f90_decl()
	else:
	    s = 1
	return s

    def get_size(self):
        # return number of elements
        if self.lower:
            return '(%s-(%s)+1)' % (self.upper.f90_decl(),
                                    self.lower.f90_decl())
        else:
            return self.upper.f90_decl()
        
########################################
#
#
class FortranArraySpec(AST):
    def __init__(self):
	AST.__init__(self, 'array_spec')

    def rank(self):
        return len(self._kids)

    def first_element(self):
	'''return the indexing for first element of the array'''
	s1 = []
	for item in self:
	    s1.append(str(item.first_element()))
        return '(%s)' % join(s1, ',')

    def assumed_shape(self):
	return '(%s)' % join(self.rank() * ':', ',')

    def get_shape(self):
        # return a list of dimensions
        shape = []
        for item in self:
	    shape.append(item.f90_decl())
        return shape
    
    def get_size(self, size=None):
        # size will be the total linear size of the array

        # XXX - if a complex, is dimension always given
        size = size or []
        for item in self:
            size.append(item.get_size())
        return size

    def get_dim_list(self, current=None):
        # dimlist will be of the form lower1,upper1,...
        current = current or []
        for rank in self:
            if rank.lower:
                current.append(rank.lower.f90_decl())
                current.append(rank.upper.f90_decl())
            else:
                current.append('1')
                current.append(rank.upper.f90_decl())
        return current

    def explicit_shape(self):
	return '(%s)' % join(self.get_shape(), ',')

    f90_decl = explicit_shape
	
########################################
#
#
class FortranType(AST):
    "Description of a Fortran type."
    def __init__ (self, name, kind=None, length=None, derived=0, decl=None):
        assert type(name) == types.StringType
#        assert kind is None or type(kind) == types.StringType
	AST.__init__(self, 'type')
        self.name = name
        self.kind = kind
        self.length = length
	self.derived = derived

        if decl:
            self.decl = decl
        else:
            self.default_decl()

    def default_decl(self):
        '''set the declaration based on the name, length, and kind.'''
        
        if self.length:
            try:
                self.decl = '%s*%d' % (self.name, int(self.length))
            except ValueError:
                self.decl = '%s*(%s)' % (self.name, self.length)
        elif self.kind:
            self.decl = '%s(%s)' % (self.name, self.kind)
        elif self.derived:
            self.decl = 'type(%s)' % self.name
        else:
            self.decl = self.name
            
    def f90_decl(self):
        return self.decl

########################################
#
#
class DerivedType(AST):
    def __init__ (self, name):
        self.name = name

    def f90_decl(self):
        return 'type(%s)' % self.name

########################################
#
#
class FortranTypeSpec(AST):

    def type_handler(self, attr, value):
        return value.f90_decl()

    def dimension_handler(self, attr, value):
        if self.dynamic():
            return 'dimension' + value.assumed_shape()
        else:
            return 'dimension' + value.explicit_shape()

    def intent_handler(self, attr, value):
        return 'intent(%s)' % value

#    def text_handler(self, attr, value):
#        return str(value)

    # list of all attributes
    # actual attributes must be passed by keyword. They are gathered into kw,
    # then installed into attr_dict.
    f90_handlers = {'var_type':type_handler,
                    'allocatable':None,
                    'dimension':dimension_handler,
                    'external':None, 
                    'intent':intent_handler,
                    'intrinsic':None,
                    'optional':None,
                    'parameter':None,
                    'pointer':None,
                    'private':None,
                    'public':None,
                    'save':None,
                    'target':None} # , 'init']

    # order of attributes in f90 declaration
    f90_order = ['var_type', 'allocatable', 'dimension', 'external',
                 'intent', 'intrinsic', 'optional',
                 'parameter', 'pointer',
                 'private', 'public', 'save', 'target'] # , 'init']

    attr_handlers = {}
    attr_handlers.update(f90_handlers)
    attribute_order = f90_order

    def __init__(self, attrs={}, **kw):
	AST.__init__(self, 'type_spec')
        self.attrs = parsetools.AttrDict(attrs)
        self.attrs.update(kw)
        
    def dynamic(self):
	# flag if variable is dynamic
        return 	self.attrs.pointer or self.attrs.allocatable

    def f90_decl(self):
	"""format the attributes for a fortran declaration list"""
        return self.attrs.decl(self, self.f90_handlers, self.f90_order)

########################################
#
#
class FortranVariable(AST):
    default_integer = FortranTypeSpec(var_type=FortranType('integer'))
    default_real = FortranTypeSpec(var_type=FortranType('real'))
    
    def __init__(self, name, dim_spec=None, init=None, type_spec=None):
	AST.__init__(self, 'variable_decl')
	self.name = name
	self.dim_spec = dim_spec
        self.init = init

        if type_spec is not None:
            self.type_spec = type_spec
        elif name[0] in 'ijklmn':
            self.type_spec = self.default_integer
        else:
            self.type_spec = self.default_real
	# used to identify if this declaration is within a derived type
	self.derived_component = 0
        if dim_spec:
            self.append(dim_spec)

    def f90_decl(self, assumed_shape=1):
        return '%s :: %s' % (
            self.type_spec.f90_decl(), self.f90_variable_decl(assumed_shape))
        
    def f90_variable_decl(self, assumed_shape=1):
        if self.type_spec.dynamic() and assumed_shape:
            dim = self.assumed_shape()
        else:
            dim = self.explicit_shape()

        if self.init is None:
            s = '%s%s' % (self.name, dim)
        else:
            s = '%s%s = %s' % \
                (self.name, dim, self.init)
	return s

    def cray_pointer(self):
        '''Return the the declaration of a variable using cray pointer
        for dynamic variables instead of f90 pointers.

        type
        name for common block
        name for type declaration
        pointer statement
        '''

        if self.dim_spec:
            dims = self.dim_spec.explicit_shape()
        else:
            dims = ''
        if self.type_spec.dynamic():
            return ('p' + self.name,
                    self.name,
                    'pointer (p%s, %s%s)'  % (self.name, self.name, dims))
        else:
            return (self.name,
                    '%s%s' % (self.name, dims),
                    None)
            
    def first_element(self):
        if self.dim_spec:
            return self.dim_spec.first_element()
        else:
            return ''

    def assumed_shape(self):
        if self.dim_spec:
            return self.dim_spec.assumed_shape()
        else:
            return ''

    def explicit_shape(self):
        if self.dim_spec:
            return self.dim_spec.explicit_shape()
        else:
            return ''

########################################
#
#
class FortranDerived(AST):
    def __init__(self, name, field_list):
	AST.__init__(self, 'derived_type')
	self.name = name
	self.field_list = field_list
	for field in field_list:
	    field.derived_component = 1

########################################
#
#
class TypeDeclaration(AST):
    def __init__(self, type_spec, *args):
	AST.__init__(self, 'type_decl')
        self.type_spec = type_spec
        for var in args:
            self.append(var)

    def append(self, node):
        # when appending children, add a pointer to the type_spec
        node.type_spec = self.type_spec
        AST.append(self, node)

    def f90_decl(self):
        attrs = self.type_spec.f90_decl()
        list = []
        for node in self:
            list.append(node.f90_variable_decl())
        s = '%s :: %s' % (attrs, join(list, ', '))
        return s

########################################
#
#
class External(AST):
    def __init__(self, *args):
	AST.__init__(self, 'external')
        self.names = []
        for arg in args:
            if type(arg) is type([]):
                for name in arg:
                    self.names.apend(name)
            else:
                self.names.append(arg)
        
########################################
#
#
class Implicit(AST):
    def __init__(self):
	AST.__init__(self, 'implicit')

########################################
#
#
class Specification(AST):
    def __init__(self):
	AST.__init__(self, 'specification')
    
########################################
#
#
class Common(AST):
    def __init__(self, name):
	AST.__init__(self, 'common')
        self.name = name
    
########################################
#
#
class If(AST):
    def __init__(self, condition, stmt):
	AST.__init__(self, 'if')
        self.condition = condition
        self.stmt = stmt

########################################
#
#
class IfBlock(AST):
    '''usage:  IfBlock( (condition1, stmt1), (condition2, stmt2) ...
                          (None, stmtn))'''
    def __init__(self, *args):
	AST.__init__(self, 'if_block')
        self.clauses = args

########################################
#
#
class Stmt(AST):
    def __init__(self, code):
	AST.__init__(self, 'stmt')
        self.code = code

########################################
#
#
class StmtBlock(AST):
    def __init__ (self, *args):
	AST.__init__(self, 'stmt_block')
        self.add_kids(args)

########################################
#
#
class Subroutine(AST):
    def __init__(self, name, args=None):
	AST.__init__(self, 'subroutine')
        self.name = name
        self.args = args or []
        self.ext_decl = External(name)
    
    def f90_decl(self):
        list = []
        for arg in self.args:
            list.append(arg.name)
        if list:
            args = '(%s)' % join(list, ', ')
        else:
            args = ''
        return "subroutine %s%s" % (self.name, args)

    def append_arg_decl(self, new=None):
        '''Append all of the argument, to new.
        If new is None, then append to self.'''

        if not new:
            new = self
        for arg in self.args:
            new.append(arg)

    def call_subroutine(self):
        '''return string to call argument with default argument names.'''
        list = []
        for arg in self.args:
            list.append(arg.name)
        if list:
            args = join(list, ', ')
        else:
            args = ''
        return Stmt('call %s(%s)' % (self.name, args))

########################################
#
#
class Function(AST):
    def __init__(self, name, args=None, result=None):
	AST.__init__(self, 'function')
        self.name = name
        self.args = args or []
        self.result = result

        if isinstance(result, FortranType):
            self.result_type = result
            self.result_clause = 0
            self.result_decl = FortranVariable(
                name, type_spec=FortranTypeSpec(var_type=result))
        elif isinstance(result, FortranTypeSpec):
            self.result_type = result.attrs.var_type
            self.result_clause = 0
            self.result_decl = FortranVariable(name, type_spec=result)
#        elif isinstance(result, TypeDeclaration):
# cannot be TypeDeclaration because it can contain several variables
#            self.result_type = result.type_spec.var_type
#            self.result_clause = 0
#            self.result_decl = FortranVariable(name,
#            type_spec=result.type_spec)
        elif isinstance(self.result, FortranVariable):
            self.result_type = result.type_spec.attrs.var_type
            self.result_clause = 1
            self.result_decl = self.result
        else:
            raise RuntimeError, "wrong type for result " + str(result)

        self.ext_decl = FortranVariable(
            name, type_spec=FortranTypeSpec(var_type=self.result_type,
                                          external=1))
        self.arg_decl = StmtBlock()
        self.arg_decl.add_kids(args)

    def f90_decl(self):
        if self.result_clause:
            result = ' result (%s)' % self.result.name
        else:
            result = ''

        list = []
        for arg in self.args:
            list.append(arg.name)
        if list:
            args = '(%s)' % join(list, ', ')
        else:
            args = ''
        return "function %s%s%s" % (self.name, args, result)

    def call_function(self):
        '''return string to call argument with default argument names.'''
        list = []
        for arg in self.args:
            list.append(arg.name)
        if list:
            args = join(list, ', ')
        else:
            args = ''
        return '%s(%s)' % (self.name, args)

    def append_arg_decl(self, new=None):
        '''Append all of the argument, to new.

        This is used to added the argument declaration to a function.
        The declaration must come after any implicit or use statement.

        if new is None, then add the argument to itself. Thus add the result
        declaration.  If new is not None, we are adding to another function
        so add the function declaration instead.
        '''

        if new is None:
            self.append(self.arg_decl)
        else:
            new.append(self.arg_decl)
            new.append(self.ext_decl)

    def append_result_decl(self, new=None):
        if new is None:
            new = self
        new.append(self.result_decl)

########################################
#
#
class Contains(AST):
    def __init__(self):
	AST.__init__(self, 'contains')

########################################
#
#
#class UseRename(AST):
#    def __init__(self, local, name=None):
#	AST.__init__(self, 'use_rename')
#	self.local = local
#	self.name = name
#
#    def f90_decl(self):
#        return '%s => %s' % (self.local, self.name)

########################################
#
#
class Use(AST):
    def __init__(self, name, rename=None, only=0):
	AST.__init__(self, 'use')
	self.name = name
        self.rename = rename or []
	self.only = only
        self.module_node = None

    def f90_decl_rename(self):
        rename = []
        for name in self.rename:
            if name[0] is name[1]:
                rename.append(name[0])
            else:
                rename.append('%s=>%s' % name)
        return join(rename, ', ')

    def f90_decl(self):
        rename = self.f90_decl_rename()
        if self.only:
            s = 'use %s, only: %s' % (self.name, rename)
        elif rename:
            s = 'use %s, %s' % (self.name, rename)
        else:
            s = 'use %s' % self.name
        return s

    def lookup_symbol(self, name):
        return self.module_node.lookup_symbol(name)
    # XXX - add all of the only and rename stuff
#        if self.only:
#            for sym in self.rename:
#if name == sym[0]: return name


########################################
#
#
class Module(AST):
    def __init__(self, name="undefined"):
	AST.__init__(self, 'module')
	self.name = name
	self.use_list = []
        self.symtab = {}

    def lookup_symbol(self, name):
        where = []
        what = self.symtab.get(name)
        if what is not None:
            where.append(self.name)
        for use in self.use_list:
            what1 = use.lookup_symbol(name)
            if what1 is not None:
                where.append(use.name)
                what = what1
        if len(where) == 0:
            raise RuntimeError, "module %s: name %s not found" % (
                self.name, name)
        if len(where) > 1:
            raise RuntimeError, "module %s: multiple symbols found for %s in %s" % (
                self.name, name, join(where, ", "))
        return what

########################################
#
#
class File(AST):
    def __init__(self, name=None):
	AST.__init__(self, 'file')
        self.name  = name
        
########################################
#
#
class Create_Module_Dict_Visitor:
    '''scan thru the tree.
    For each module, enter its name in the module_dict.
    After all modules are found, update the use statment
    to point to the module.
    '''

    def __init__(self, module_dict):
        self.module_dict = module_dict

    def visitUse(self, node):
        self.current_module.use_list.append(node)

    def visitModule(self, node):
        self.module_dict[node.name] = node
        self.current_module = node
        for child in node.getChildren():
            self.visit(child)
        
def Create_Module_Dict(ast, module_dict):
    walk(ast, Create_Module_Dict_Visitor(module_dict))
    for mod in module_dict.values():
        # XXX - by only using once, we lose rename and only information
        tmp_mod = []
        tmp_use = []
        for node in mod.use_list:
            node.module_node = module_dict[node.name]
            if node.module_node not in tmp_mod:
                tmp_use.append(node)
                tmp_mod.append(node.module_node)
        mod.use_list = tmp_use

########################################
#
#
# XXX - does this belong in frame.py since dictclass is there?
class Create_Module_Symtab_Visitor:
    '''scan thru the tree.
    For each module, create a new dictionary.
    '''

# XXX - useful?
#    def __init__(self):
#        self.symtab = None

    def visitModule(self, node):
        self.symtab = node.symtab
        for child in node.getChildren():
            self.visit(child)
        self.symtab = None

    def visitDictClass(self, node):
        self.symtab[node.name] = node

def Create_Module_Symtab(ast):
    walk(ast, Create_Module_Symtab_Visitor())
    
########################################
#
#
class Print_Module_Symtab_Visitor:

    def visitModule(self, node):
        print "Module %s:" % node.name
        print "%s : %s" % (node.name, node.symtab.keys())

def Print_Module_Symtab(ast):
    print "Symbol Table"
    walk(ast, Print_Module_Symtab_Visitor())

########################################
#
#
class Visit_F90_Node(parsetools.Dump_Mixin):
    def __init__(self, fp=None):
        parsetools.Dump_Mixin.__init__(self)
        self.max_line = 72
        self.set_indent(0)
        import sys
        self.fp = fp or sys.stdout
    
    def visitFortranTypeSpec(self, node):
        self.writeto(node.f90_decl() + '\n')

def dump_f90_nodes(ast, fp=None):
    walk(ast, Visit_F90_Node(fp))

########################################
#
#
# XXX - add code to continue long lines
class Print_Fortran_Code_Visitor(parsetools.Dump_Mixin):
    def __init__(self, fp=None):
        parsetools.Dump_Mixin.__init__(self)
        self.max_line = 72
        self.set_indent(0)
        import sys
        self.fp = fp or sys.stdout

    def visitFortranType(self, node):
        self.writeto(node.f90_decl())
        
    def visitComment(self, node):
        for comment in node.comments:
            self.writeto("!%s\n" % str(comment))

    def visitImplicit(self,node):
        self.writeto("implicit none\n")

    def visitCommon(self,node):
        self.writeto("common /%s/" % node.name)

    def visitUse(self, node):
        self.writeto("%s" % node.f90_decl())
        
    def visitModule(self, node):
        self.writeto("module %s" % node.name)
        self.doindent()
        for child in node.getChildren():
            self.visit(child)
        self.deindent()
        self.writeto("end module %s" % node.name)
        self.new_line()

    def visitDerivedType(self, node):
        self.writeto("type %s\n" % node.name)
        self.doindent()
        for child in node.getChildren():
            self.visit(child)
        self.deindent()
        self.writeto("end type %s\n" % node.name)
        self.new_line()

    def visitFortranVariable(self, node):
        self.writeto('%s\n' % node.f90_decl())

    def visitTypeDeclaration(self, node):
        self.writeto('%s\n' % node.f90_decl())

    def visitExternal(self, node):
        self.writeto("external %s\n" % join(node.names, ','))
        
    def visitSubroutine(self, node):
        self.writeto("%s\n" % node.f90_decl())
        self.doindent()
        for child in node.getChildren():
            self.visit(child)
        self.deindent()
        self.writeto("end subroutine %s\n" % node.name)
        self.new_line()

    def visitFunction(self, node):
        self.writeto("%s\n" % node.f90_decl())
        self.doindent()
        for child in node.getChildren():
            self.visit(child)
        self.deindent()
        self.writeto("end function %s\n" % node.name)
        self.new_line()

    def visitContains(self, node):
        self.new_line()
        self.writeto("contains\n")
        self.new_line()
        self.doindent()
        for child in node.getChildren():
            self.visit(child)
        self.deindent()

    def visitStmt(self, node):
        self.writeto('%s\n' % node.code)

    def visitIf(self, node):
        self.writeto("if (%s) %s\n" % (node.condition, node.stmt.code))

    def visitIfBlock(self, node):
        ifcmd = 'if'
        for (condition, stmt) in node.clauses:
            if condition is None:
                self.writeto('else\n')
            else:
                self.writeto('%s (%s) then\n' % (ifcmd, condition))
            self.doindent()
            self.visit(stmt)
            self.deindent()
            ifcmd = 'else if'
        self.writeto('endif\n')
        
    def visitBlank(self, node):
        self.new_line()
        
def print_fortran_code(ast, fp=None):
    walk(ast, Print_Fortran_Code_Visitor(fp))

########################################
#
#
if __name__ == '__main__':
    pass


