#!/bin/env python
#
#  Convert the mac input into fortran.
#
import os.path as path

import rdp.parsetools as parsetools
import f90
import vdf
from rdp.visitor import walk

def f90ize_types():
    '''Take the table from vdf.py and convert it to use f90 kinds.'''

    vdf.type_table['real'].decl = 'real(Kind8)'

########################################
#
#
class Print_Parameters:
    '''Print the parameters and defines as f90 parameters'''
    
    def visitBasisPackage(self, node):
        self.pkg_name = node.name

    def visitParameter(self, node):
        expr_type = node.expr_type
        if expr_type == vdf.type_int:
            self.writeto('integer, parameter :: %s = %s\n' %
                         (node.name, node.expr_str))
        elif expr_type == vdf.type_real:
            self.writeto('real, parameter :: %s = %s\n' %
                         (node.name, node.expr_str))
        else:
            self.writeto('character*(*), parameter :: %s = %s\n' %
                         (node.name, node.expr_str))

    def visitDefine(self, node):
        if type(node.value) == type(0):
            self.writeto('integer, parameter :: %s=%s\n' %
                         (node.name, node.expr_str))
        elif type(node.value) == type(0.0):
            self.writeto('real, parameter :: %s=%s\n' %
                         (node.name, node.expr_str))
        else:
            self.writeto('character*(*), parameter :: %s="%s"\n' %
                         (node.name, node.expr_str))
    
########################################
#
#
class VDF_using_Cray_Pointers_Visitor(f90.Print_Fortran_Code_Visitor,
                                      Print_Parameters):
    '''Take the VDF tree and create include files using cray pointers
    for dynamic variables.
    Make a file for each group, and add common blocks to it.
    '''

    def __init__(self, opts):
        f90.Print_Fortran_Code_Visitor.__init__(self)
        self.opts = opts
        self.incdir = opts.incdir

    def continue_line(self, out):
        '''Use F90 continuations.

        Start at 72 and scan back for a comma.'''
	# todo: add indent into header
        # remove exception by looking for other tokens

#        self.max_line = 40
        j = 0
        len_out = len(out)
        max_line = self.max_line
        header = '       '
        len_header = 0
        lines = []

#        print "0001:%s:%s:" % (len_out, out)

        while 1:
            j1 = min(j + max_line, len_out)
            if j1 == len_out:
                break;
#            print "0002:%s:%s:%s:" % (j, j1, out[j:j1])
            j1 = out.rfind(',', j, j1)
            if j1 < 0:
                raise RuntimeError, "comma not found:" + str(j1)

            j1 += 1
            if len_header == 0:
#                print "0003:%s:%s:%s" % (j,j1, out[j:j1])
                lines.append(out[j:j1] + ' &')
                len_header = len(header)
                len_line = max_line - len_header
            else:
#                print "0004:%s:%s:%s%s" % (j,j1,header, out[j:j1])
                lines.append(header + out[j:j1] + ' &')

            j = j1
        lines.append(header + out[j:j1])
#        print "0005:%s:%s:%s%s" % (j,j1,header, out[j:j1])
#        self.max_line = 72
        return '\n'.join(lines)

    def visitBasisPackage(self, node):
        filename = path.join(self.incdir, node.name + '.inc')
        self.fp = open(filename, 'w')
        self.pkg_name = node.name
        self.writeto('! Package %s\n' % node.name)
        self.writeto('!\n')
        for comment in node.comments:
            self.writeto('!%s\n' % comment)
        self.set_indent(6)
        for child in node.parameters:
            self.visit(child)
        self.writeto('\n')
        self.fp.close()

        for child in node.groups:
            self.visit(child)

    def visitGroupNode(self, group):
        if group.nv == 0:
            return
        self.set_indent(0)
        filename = path.join(self.incdir,
                             '%s_%s.inc' % (self.pkg_name, group.name))
        self.fp = open(filename, 'w')
        self.writeto('! --- Group %s\n' % group.name)
        self.set_indent(6)
        self.commons = {}
        self.declarations = {}

        for child in group.getChildren():
            self.visit(child)

        # add all of the common blocks in sorted order
        declarations = self.declarations
        keys = declarations.keys()
        keys.sort()
        for name in keys:
            self.writeto('%s %s\n' %
                         (name, ', '.join(declarations[name])))

        if not group.local:
            commons = self.commons
            keys = commons.keys()
            keys.sort()
            for name in keys:
                self.writeto('common /%s/ %s\n' %
                             (name, ', '.join(commons[name])))

        self.fp.close()
        self.fp = None

    def visitVDF_Variable(self, variable):
        type_decl = variable.type_spec.attrs.var_type.decl

        if variable.dim_spec:
            dims = variable.dim_spec.explicit_shape()
        else:
            dims = ''
        if variable.type_spec.dynamic():
            if self.opts.std or variable.type_spec.attrs.var_type.character:
                name = variable.name
                decl_name = name
                type_decl = 'type(address)'
            else:
                name = self.opts.ptrprefix + variable.name
                decl_name = variable.name
                self.writeto('pointer (%s, %s%s)' %
                             (name, variable.name, dims))
                self.writeto('\n')
        else:
            name = variable.name
            decl_name = '%s%s' % (variable.name, dims)

        self.commons.setdefault(variable.common, []).append(name)
        self.declarations.setdefault(type_decl, []).append(decl_name)

def vdf_using_cray_pointers(ast, opts):
    walk(ast, VDF_using_Cray_Pointers_Visitor(opts))

########################################
#
#
class VDF_to_F90_Visitor(f90.Print_Fortran_Code_Visitor, Print_Parameters):
    '''Write the VDF tree as a f90 module.
    Use f90 style pointers.
    Create a module for each group.
    '''

    def visitBasisPackage(self, node):
        self.pkg_name = node.name
        self.writeto('! Package %s\n' % node.name)
        self.writeto('!\n')
        for comment in node.comments:
            self.writeto('!%s\n' % comment)
        self.writeto('\n')
        self.set_indent(6)
        for child in node.parameters:
            self.visit(child)
        # XXX-XXX close file

        for child in node.groups:
            self.visit(child)


    # XXX-XXX this is now package.parameters
    def visitParameters(self, node):
        self.name = ('%s_params' % self.pkg_name).lower()
        self.writeto('module %s\n' % self.name)
        self.doindent()
        for child in node.getChildren():
            self.visit(child)
        self.deindent()
        self.writeto('end module %s\n' % self.name)
        self.writeto('\n')
        # close the file
    
    def visitGroupNode(self, group):
        if group.no_names == 0:
            return
        self.name = ('%s_%s' % (self.pkg_name, group.name)).lower()
        self.writeto('module %s\n' % self.name)
        self.doindent()

        for child in group.getChildren():
            self.visit(child)

        self.deindent()
        self.writeto('end module %s\n' % self.name)
        self.writeto('\n')
        
def vdf_to_f90(ast, fp=None):
    walk(ast, VDF_to_F90_Visitor(fp))

########################################
#
#
if __name__ == '__main__':
    import mac
    import mppl

    opts = mac.MACoptions()

    filename = opts.vdf

    ast = vdf.read_vdf(filename)

#   debug prints
#    import rdp.parsedebug as parsedebug
#    parsedebug.print_tree(ast)

    mppl.PkgDump_Cname(ast, opts)
    mppl.PkgDump_PreProcess(ast, opts)

    de_usertype()
    vdf_using_cray_pointers(ast)

    f90ize_types()
#    vdf_to_f90(ast)



